/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

let startTime = Date.now();

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');

const {FC} = require(REL_PATH + 'FUNCTIONS-FILES/functions-computation');
const {FS} = require(REL_PATH + 'FUNCTIONS-FILES/functions-substitutions');
const {FS_DATA} = require(REL_PATH + 'INIT-FILES/options-substitutions');

let latexSubstitution = null;
let latexToSave = '';

OPT.SHOW.showItemNumber = false;



FH.showTitle('BEGINNING: ' + path.basename(__filename));


FH.showSubtitle('SUBSTITUTIONS OF g');

FH.showSubsubtitle('g with repeated applications of hnabla');


FC.showSubstitution(SUBST.gNC);

let pg1 = PROTO.partial.makeCopy()
	.applyTo(PROTO.g.makeCopy());

let pg2 = PROTO.partial.makeCopy()
	.applyTo(pg1.makeCopy());

let pg3 = PROTO.partial.makeCopy()
	.applyTo(pg2.makeCopy());

let pg4 = PROTO.partial.makeCopy()
	.applyTo(pg3.makeCopy());

// let Tg1 = new MathElement();
// Tg1.setID('Tg1')
// 	.setLetter('T')
// 	.setLaTeXLetter('T')
// 	.addIndiceUp()
// 	.addIndiceUp()
// 	.addIndiceUp()
// 	.setProductKindToNonCommutative()
// 	.setSumWeight(1)
// 	.setProductWeight(1);
// let Tg1g = MathProduct.newFrom(Tg1)
// 	.multiplyOnRightWith(pg1)
// 	.setContraction([0,0], [1,-1,0])
// 	.setContraction([0,1], [1,-2,0])
// 	.setContraction([0,2], [1,-3,0]);
// let Tg1gS = MathSum.newFrom(Tg1g);
//
// Tg1gS.showConsole(OPT.SHOW, 'Tg1gS = ');
//
// latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tg1gS, SUBST.gNC);

let Tg2 = new MathElement();
Tg2.setID('Tg2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let Tg2g = MathProduct.newFrom(Tg2)
	.multiplyOnRightWith(pg2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-3,1]);
let Tg2gS = MathSum.newFrom(Tg2g);

Tg2gS.showConsole(OPT.SHOW, 'Tg2gS = ');

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tg2gS, SUBST.gNC);

let Tg3 = new MathElement();
Tg3.setID('Tg3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let Tg3g = MathProduct.newFrom(Tg3)
	.multiplyOnRightWith(pg3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0])
	.setContraction([0,4], [1,-4,1]);
let Tg3gS = MathSum.newFrom(Tg3g);

Tg3gS.showConsole(OPT.SHOW, 'Tg3gS = ');

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tg3gS, SUBST.gNC);

let Tg4 = new MathElement();
Tg4.setID('Tg4')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let Tg4g = MathProduct.newFrom(Tg4)
	.multiplyOnRightWith(pg4)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0])
	.setContraction([0,4], [1,-5,0])
	.setContraction([0,5], [1,-5,1]);
let Tg4gS = MathSum.newFrom(Tg4g);

Tg4gS.showConsole(OPT.SHOW, 'Tg4gS = ');

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tg4gS, SUBST.gNC);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-g-to-repeated-hnabla.tex', latexToSave);



FH.showSubsubtitle('g with powered applications of hnabla');


FC.showSubstitution(SUBST.g_powNC);

latexToSave = '';

let pg_pow1 = PROTO.partial.makeCopy()
	.applyTo(PROTO.g.makeCopy());

let pg_pow2 = PROTO.partial.makeCopy()
	.applyTo(pg_pow1.makeCopy());

let pg_pow3 = PROTO.partial.makeCopy()
	.applyTo(pg_pow2.makeCopy());

let pg_pow4 = PROTO.partial.makeCopy()
	.applyTo(pg_pow3.makeCopy());

let Tg_pow2 = new MathElement();
Tg_pow2.setID('Tg_pow2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let Tg_pow2g = MathProduct.newFrom(Tg_pow2)
	.multiplyOnRightWith(pg_pow2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-3,1]);
let Tg_pow2gS = MathSum.newFrom(Tg_pow2g);

Tg_pow2gS.showConsole(OPT.SHOW, 'Tg_pow2gS = ');

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tg_pow2gS, SUBST.g_powNC);

let Tg_pow3 = new MathElement();
Tg_pow3.setID('Tg_pow3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let Tg_pow3g = MathProduct.newFrom(Tg_pow3)
	.multiplyOnRightWith(pg_pow3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0])
	.setContraction([0,4], [1,-4,1]);
let Tg_pow3gS = MathSum.newFrom(Tg_pow3g);

Tg_pow3gS.showConsole(OPT.SHOW, 'Tg_pow3gS = ');

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tg_pow3gS, SUBST.g_powNC);

let Tg_pow4 = new MathElement();
Tg_pow4.setID('Tg_pow4')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let Tg_pow4g = MathProduct.newFrom(Tg_pow4)
	.multiplyOnRightWith(pg_pow4)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0])
	.setContraction([0,4], [1,-5,0])
	.setContraction([0,5], [1,-5,1]);
let Tg_pow4gS = MathSum.newFrom(Tg_pow4g);

Tg_pow4gS.showConsole(OPT.SHOW, 'Tg_pow4gS = ');

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tg_pow4gS, SUBST.g_powNC);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-g-to-powered-hnabla.tex', latexToSave);



FH.showSubsubtitle('g inverse');

FC.showSubstitution(SUBST.gi);

latexToSave = '';

let pgi1 = PROTO.partial.makeCopy()
	.applyTo(PROTO.gi.makeCopy());

let Tgi1 = new MathElement();
Tgi1.setID('Tgi1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceDown()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let Tgi1gi = MathProduct.newFrom(Tgi1)
	.multiplyOnRightWith(pgi1)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-2,1]);
let Tgi1giS = MathSum.newFrom(Tgi1gi);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tgi1giS, SUBST.gi);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-g-inverse.tex', latexToSave);


FH.showSubtitle('GAUGE');

FH.showSubsubtitle('nabla in post element');

latexToSave = '';

latexSubstitution = FC.showSubstitution(SUBST.NABLA_POST_ELEMENT);
latexToSave += latexSubstitution + CONST.CR;

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-nabla-postelement.tex', latexToSave);



FH.showSubsubtitle('nabla^3 on 1 in post element');

latexToSave = '';

latexSubstitution = FC.showSubstitution(SUBST.NABLA3_ONE_TO_NABLA_F_POST_ELEMENT);
latexToSave += latexSubstitution + CONST.CR;

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-nabla3-postelement.tex', latexToSave);



FH.showSubtitle('SUBSTITUTIONS OF H');

FH.showSubsubtitle('Load computed substitutions of H');

let filePathH = REL_PATH + FILES.SUBST.H_to_nc;
let substitutionH = FC.getSubstitutionsFromFile(filePathH);

substitutionH.orgItem.showConsole(OPT.SHOW, 'orgItem = ');



FH.showSubsubtitle('Display substitutions of H');

latexSubstitution = FC.showSubstitution(substitutionH);
// latexToSave += latexSubstitution + CONST.CR;
latexToSave = '';

let TH0 = new MathElement();
TH0.setID('TH0')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceDown()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TH0H = MathProduct.newFrom(TH0)
	.multiplyOnRightWith(PROTO.H)
	.setContraction([0,0], [1,0])
	.setContraction([0,1], [1,1]);
// TH0H = MathArgument.newFrom(TH0H);
let TH0HS = MathSum.newFrom(TH0H);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TH0HS, substitutionH);

let TH1 = new MathElement();
TH1.setID('TH1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceDown()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TH1H = MathProduct.newFrom(TH1)
	.multiplyOnRightWith(PROTO.H1)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-2,1]);
// TH1H.showConsole(OPT.SHOW, 'TH1H = ');
// TH1H = MathArgument.newFrom(TH1H);
let TH1HS = MathSum.newFrom(TH1H);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TH1HS, substitutionH);

let TH2 = new MathElement();
TH2.setID('TH2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TH2H = MathProduct.newFrom(TH2)
	.multiplyOnRightWith(PROTO.H2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-3,1]);
// TH2H.showConsole(OPT.SHOW, 'TH2H = ');
// TH2H = MathArgument.newFrom(TH2H);
let TH2HS = MathSum.newFrom(TH2H);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TH2HS, substitutionH);

let TH3 = new MathElement();
TH3.setID('TH3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TH3H = MathProduct.newFrom(TH3)
	.multiplyOnRightWith(PROTO.H3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0])
	.setContraction([0,4], [1,-4,1]);
let TH3HS = MathSum.newFrom(TH3H);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TH3HS, substitutionH);

let TH4 = new MathElement();
TH4.setID('TH4')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TH4H = MathProduct.newFrom(TH4)
	.multiplyOnRightWith(PROTO.H4)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0])
	.setContraction([0,4], [1,-5,0])
	.setContraction([0,5], [1,-5,1]);
let TH4HS = MathSum.newFrom(TH4H);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TH4HS, substitutionH);


FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-H.tex', latexToSave);




FH.showSubtitle('SUBSTITUTIONS OF L');

FH.showSubsubtitle('Load computed substitutions of L');

let filePathL = REL_PATH + FILES.SUBST.L_to_nc;
let substitutionL = FC.getSubstitutionsFromFile(filePathL);

substitutionL.orgItem.showConsole(OPT.SHOW, 'orgItem = ');



FH.showSubsubtitle('Display substitutions of L');


latexSubstitution = FC.showSubstitution(substitutionL);
// latexToSave += latexSubstitution + CONST.CR;
latexToSave = '';

let TL0 = new MathElement();
TL0.setID('TL0')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TL0L = MathProduct.newFrom(TL0)
	.multiplyOnRightWith(PROTO.L)
	.setContraction([0,0], [1,0]);
// TL0L = MathArgument.newFrom(TL0L);
let TL0LS = MathSum.newFrom(TL0L);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TL0LS, substitutionL);

let TL1 = new MathElement();
TL1.setID('TL1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TL1L = MathProduct.newFrom(TL1)
	.multiplyOnRightWith(PROTO.L1)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0]);
// TL1L = MathArgument.newFrom(TL1L);
let TL1LS = MathSum.newFrom(TL1L);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TL1LS, substitutionL);

let TL2 = new MathElement();
TL2.setID('TL2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TL2L = MathProduct.newFrom(TL2)
	.multiplyOnRightWith(PROTO.L2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0]);
let TL2LS = MathSum.newFrom(TL2L);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TL2LS, substitutionL);


let TL3 = new MathElement();
TL3.setID('TL3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TL3L = MathProduct.newFrom(TL3)
	.multiplyOnRightWith(PROTO.L3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0]);
let TL3LS = MathSum.newFrom(TL3L);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TL3LS, substitutionL);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-L.tex', latexToSave);


FH.showSubtitle('SUBSTITUTIONS OF Gamma');

FH.showSubsubtitle('Load computed substitutions of Gamma');

let filePathGamma = REL_PATH + FILES.SUBST.Gamma_to_nc;
let substitutionGamma = FC.getSubstitutionsFromFile(filePathGamma);

substitutionGamma.orgItem.showConsole(OPT.SHOW, 'orgItem = ');


FH.showSubsubtitle('Display substitutions of Gamma');

let Gamma = substitutionGamma.orgItem.makeCopy();
Gamma.setSumWeight(5)
	.setProductWeight(5);

let partial1Gamma = PROTO.partial.makeCopy().applyTo(Gamma.makeCopy());

let partial2Gamma = PROTO.partial.makeCopy().applyTo(partial1Gamma.makeCopy());

let partial3Gamma = PROTO.partial.makeCopy().applyTo(partial2Gamma.makeCopy());


latexSubstitution = FC.showSubstitution(substitutionGamma);
// latexToSave += latexSubstitution + CONST.CR;
latexToSave = '';

let TGamma0 = new MathElement();
TGamma0.setID('TGamma0')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceDown()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TGamma0Gamma = MathProduct.newFrom(TGamma0)
	.multiplyOnRightWith(Gamma)
	.setContraction([0,0], [1,0])
	.setContraction([0,1], [1,1])
	.setContraction([0,2], [1,2]);
// TGamma0Gamma.showConsole(OPT.SHOW, 'prod = ');
let TGamma0GammaS = MathSum.newFrom(TGamma0Gamma);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TGamma0GammaS, substitutionGamma);

let TGamma1 = new MathElement();
TGamma1.setID('TGamma1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceDown()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TGamma1Gamma = MathProduct.newFrom(TGamma1)
	.multiplyOnRightWith(partial1Gamma)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-2,1])
	.setContraction([0,3], [1,-2,2]);
// TGamma1Gamma.showConsole(OPT.SHOW, 'prod = ');
let TGamma1GammaS = MathSum.newFrom(TGamma1Gamma);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TGamma1GammaS, substitutionGamma);

let TGamma2 = new MathElement();
TGamma2.setID('TGamma2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TGamma2Gamma = MathProduct.newFrom(TGamma2)
	.multiplyOnRightWith(partial2Gamma)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-3,1])
	.setContraction([0,4], [1,-3,2]);
// TGamma2Gamma.showConsole(OPT.SHOW, 'prod = ');
let TGamma2GammaS = MathSum.newFrom(TGamma2Gamma);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TGamma2GammaS, substitutionGamma);


let TGamma3 = new MathElement();
TGamma3.setID('TGamma3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TGamma3Gamma = MathProduct.newFrom(TGamma3)
	.multiplyOnRightWith(partial3Gamma)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0])
	.setContraction([0,4], [1,-4,1])
	.setContraction([0,5], [1,-4,2]);
// TGamma3Gamma.showConsole(OPT.SHOW, 'prod = ');
let TGamma3GammaS = MathSum.newFrom(TGamma3Gamma);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TGamma3GammaS, substitutionGamma);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-Gamma.tex', latexToSave);



FH.showSubtitle('SUBSTITUTIONS OF HNABLA ON u');

FH.showSubsubtitle('Load computed substitutions of hnabla on u to Gamma');

{
	// local context for variables

let filePathHNu = REL_PATH + FILES.SUBST.hnabla_u_to_Gamma;
let substitutionHNu = FC.getSubstitutionsFromFile(filePathHNu);

substitutionHNu.orgItem.showConsole(OPT.SHOW, 'orgItem = ');



FH.showSubsubtitle('Display substitutions of hnabla on u to Gamma');


latexSubstitution = FC.showSubstitution(substitutionHNu);
// latexToSave += latexSubstitution + CONST.CR;
latexToSave = '';

// let THNu0 = new MathElement();
// THNu0.setID('THNu0')
// 	.setLetter('T')
// 	.setLaTeXLetter('T')
// 	.setProductKindToNonCommutative()
// 	.setSumWeight(1)
// 	.setProductWeight(1);
// let THNu0u = MathProduct.newFrom(THNu0)
// 	.multiplyOnRightWith(PROTO.u);
let THNu0u = MathProduct.newFrom(PROTO.u);
let THNu0uS = MathSum.newFrom(THNu0u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNu0uS, substitutionHNu);

let THNu1 = new MathElement();
THNu1.setID('THNu1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNu1u = MathProduct.newFrom(THNu1)
	.multiplyOnRightWith(PROTO.u1)
	.setContraction([0,0], [1,-1,0]);
let THNu1uS = MathSum.newFrom(THNu1u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNu1uS, substitutionHNu);

let THNu2 = new MathElement();
THNu2.setID('THNu2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNu2u = MathProduct.newFrom(THNu2)
	.multiplyOnRightWith(PROTO.u2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0]);
let THNu2uS = MathSum.newFrom(THNu2u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNu2uS, substitutionHNu);


let THNu3 = new MathElement();
THNu3.setID('THNu3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNu3u = MathProduct.newFrom(THNu3)
	.multiplyOnRightWith(PROTO.u3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0]);
let THNu3uS = MathSum.newFrom(THNu3u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNu3uS, substitutionHNu);


let THNu4 = new MathElement();
THNu4.setID('THNu4')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNu4u = MathProduct.newFrom(THNu4)
	.multiplyOnRightWith(PROTO.u4)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0]);
let THNu4uS = MathSum.newFrom(THNu4u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNu4uS, substitutionHNu);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-u-Gamma.tex', latexToSave);

}
// local context for variables


FH.showSubsubtitle('Load computed substitutions of hnabla on u to NC');

{
	// local context for variables

let filePathHNu = REL_PATH + FILES.SUBST.hnabla_u_to_nc;
let substitutionHNu = FC.getSubstitutionsFromFile(filePathHNu);

substitutionHNu.orgItem.showConsole(OPT.SHOW, 'orgItem = ');


FH.showSubsubtitle('Display substitutions of hnabla on u to NC');


latexSubstitution = FC.showSubstitution(substitutionHNu);
// latexToSave += latexSubstitution + CONST.CR;
latexToSave = '';

// let THNu0 = new MathElement();
// THNu0.setID('THNu0')
// 	.setLetter('T')
// 	.setLaTeXLetter('T')
// 	.setProductKindToNonCommutative()
// 	.setSumWeight(1)
// 	.setProductWeight(1);
// let THNu0u = MathProduct.newFrom(THNu0)
// 	.multiplyOnRightWith(PROTO.u);
let THNu0u = MathProduct.newFrom(PROTO.u);
let THNu0uS = MathSum.newFrom(THNu0u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNu0uS, substitutionHNu);

let THNu1 = new MathElement();
THNu1.setID('THNu1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNu1u = MathProduct.newFrom(THNu1)
	.multiplyOnRightWith(PROTO.u1)
	.setContraction([0,0], [1,-1,0]);
let THNu1uS = MathSum.newFrom(THNu1u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNu1uS, substitutionHNu);

let THNu2 = new MathElement();
THNu2.setID('THNu2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNu2u = MathProduct.newFrom(THNu2)
	.multiplyOnRightWith(PROTO.u2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0]);
let THNu2uS = MathSum.newFrom(THNu2u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNu2uS, substitutionHNu);


let THNu3 = new MathElement();
THNu3.setID('THNu3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNu3u = MathProduct.newFrom(THNu3)
	.multiplyOnRightWith(PROTO.u3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0]);
let THNu3uS = MathSum.newFrom(THNu3u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNu3uS, substitutionHNu);


let THNu4 = new MathElement();
THNu4.setID('THNu4')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNu4u = MathProduct.newFrom(THNu4)
	.multiplyOnRightWith(PROTO.u4)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0]);
let THNu4uS = MathSum.newFrom(THNu4u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNu4uS, substitutionHNu);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-u-NC.tex', latexToSave);

}
// local context for variables


FH.showSubtitle('SUBSTITUTIONS OF HNABLA ON q');

FH.showSubsubtitle('Load computed substitutions of hnabla on q to Gamma');

{
	// local context for variables

let filePathHNq = REL_PATH + FILES.SUBST.hnabla_q_to_Gamma;
let substitutionHNq = FC.getSubstitutionsFromFile(filePathHNq);

substitutionHNq.orgItem.showConsole(OPT.SHOW, 'orgItem = ');



FH.showSubsubtitle('Display substitutions of hnabla on q to Gamma');


latexSubstitution = FC.showSubstitution(substitutionHNq);
// latexToSave += latexSubstitution + CONST.CR;
latexToSave = '';

// let THNq0 = new MathElement();
// THNq0.setID('THNq0')
// 	.setLetter('T')
// 	.setLaTeXLetter('T')
// 	.setProductKindToNonCommutative()
// 	.setSumWeight(1)
// 	.setProductWeight(1);
// let THNq0q = MathProduct.newFrom(THNq0)
// 	.multiplyOnRightWith(PROTO.q);
let THNq0q = MathProduct.newFrom(PROTO.q);
let THNq0qS = MathSum.newFrom(THNq0q);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNq0qS, substitutionHNq);

let THNq1 = new MathElement();
THNq1.setID('THNq1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNq1q = MathProduct.newFrom(THNq1)
	.multiplyOnRightWith(PROTO.q1)
	.setContraction([0,0], [1,-1,0]);
let THNq1qS = MathSum.newFrom(THNq1q);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNq1qS, substitutionHNq);

let THNq2 = new MathElement();
THNq2.setID('THNq2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNq2q = MathProduct.newFrom(THNq2)
	.multiplyOnRightWith(PROTO.q2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0]);
let THNq2qS = MathSum.newFrom(THNq2q);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNq2qS, substitutionHNq);


let THNq3 = new MathElement();
THNq3.setID('THNq3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNq3q = MathProduct.newFrom(THNq3)
	.multiplyOnRightWith(PROTO.q3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0]);
let THNq3qS = MathSum.newFrom(THNq3q);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNq3qS, substitutionHNq);


let THNq4 = new MathElement();
THNq4.setID('THNq4')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNq4q = MathProduct.newFrom(THNq4)
	.multiplyOnRightWith(PROTO.q4)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0]);
let THNq4qS = MathSum.newFrom(THNq4q);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNq4qS, substitutionHNq);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-q-Gamma.tex', latexToSave);

}
// local context for variables


FH.showSubsubtitle('Load computed substitutions of hnabla on q to NC');

{
	// local context for variables

let filePathHNq = REL_PATH + FILES.SUBST.hnabla_q_to_nc;
let substitutionHNq = FC.getSubstitutionsFromFile(filePathHNq);

substitutionHNq.orgItem.showConsole(OPT.SHOW, 'orgItem = ');


FH.showSubsubtitle('Display substitutions of hnabla on q to NC');


latexSubstitution = FC.showSubstitution(substitutionHNq);
// latexToSave += latexSubstitution + CONST.CR;
latexToSave = '';

// let THNq0 = new MathElement();
// THNq0.setID('THNq0')
// 	.setLetter('T')
// 	.setLaTeXLetter('T')
// 	.setProductKindToNonCommutative()
// 	.setSumWeight(1)
// 	.setProductWeight(1);
// let THNq0q = MathProduct.newFrom(THNq0)
// 	.multiplyOnRightWith(PROTO.q);
let THNq0q = MathProduct.newFrom(PROTO.q);
let THNq0qS = MathSum.newFrom(THNq0q);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNq0qS, substitutionHNq);

let THNq1 = new MathElement();
THNq1.setID('THNq1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNq1q = MathProduct.newFrom(THNq1)
	.multiplyOnRightWith(PROTO.q1)
	.setContraction([0,0], [1,-1,0]);
let THNq1qS = MathSum.newFrom(THNq1q);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNq1qS, substitutionHNq);

let THNq2 = new MathElement();
THNq2.setID('THNq2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNq2q = MathProduct.newFrom(THNq2)
	.multiplyOnRightWith(PROTO.q2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0]);
let THNq2qS = MathSum.newFrom(THNq2q);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNq2qS, substitutionHNq);


let THNq3 = new MathElement();
THNq3.setID('THNq3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNq3q = MathProduct.newFrom(THNq3)
	.multiplyOnRightWith(PROTO.q3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0]);
let THNq3qS = MathSum.newFrom(THNq3q);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNq3qS, substitutionHNq);


let THNq4 = new MathElement();
THNq4.setID('THNq4')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNq4q = MathProduct.newFrom(THNq4)
	.multiplyOnRightWith(PROTO.q4)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0]);
let THNq4qS = MathSum.newFrom(THNq4q);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNq4qS, substitutionHNq);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-q-NC.tex', latexToSave);

}
// local context for variables



FH.showSubtitle('SUBSTITUTIONS OF HNABLA ON N');

FH.showSubsubtitle('Load computed substitutions of hnabla on N to Gamma');

{
	// local context for variables

let filePathHNN = REL_PATH + FILES.SUBST.hnabla_N_to_Gamma;
let substitutionHNN = FC.getSubstitutionsFromFile(filePathHNN);

substitutionHNN.orgItem.showConsole(OPT.SHOW, 'orgItem = ');



FH.showSubsubtitle('Display substitutions of hnabla on N to Gamma');


latexSubstitution = FC.showSubstitution(substitutionHNN);
// latexToSave += latexSubstitution + CONST.CR;
latexToSave = '';

let THNN0 = new MathElement();
THNN0.setID('THNN0')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNN0N = MathProduct.newFrom(THNN0)
	.multiplyOnRightWith(PROTO.N)
	.setContraction([0,0], [1,0]);
// THNN0N = MathArgument.newFrom(THNN0N);
let THNN0NS = MathSum.newFrom(THNN0N);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNN0NS, substitutionHNN);

let THNN1 = new MathElement();
THNN1.setID('THNN1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNN1N = MathProduct.newFrom(THNN1)
	.multiplyOnRightWith(PROTO.N1)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0]);
// THNN1N = MathArgument.newFrom(THNN1N);
let THNN1NS = MathSum.newFrom(THNN1N);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNN1NS, substitutionHNN);

let THNN2 = new MathElement();
THNN2.setID('THNN2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNN2N = MathProduct.newFrom(THNN2)
	.multiplyOnRightWith(PROTO.N2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0]);
let THNN2NS = MathSum.newFrom(THNN2N);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNN2NS, substitutionHNN);


let THNN3 = new MathElement();
THNN3.setID('THNN3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNN3N = MathProduct.newFrom(THNN3)
	.multiplyOnRightWith(PROTO.N3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0]);
let THNN3NS = MathSum.newFrom(THNN3N);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNN3NS, substitutionHNN);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-N-Gamma.tex', latexToSave);

}
// local context for variables



FH.showSubsubtitle('Load computed substitutions of hnabla on N to NC');

{
	// local context for variables

let filePathHNN = REL_PATH + FILES.SUBST.hnabla_N_to_nc;
let substitutionHNN = FC.getSubstitutionsFromFile(filePathHNN);

substitutionHNN.orgItem.showConsole(OPT.SHOW, 'orgItem = ');



FH.showSubsubtitle('Display substitutions of hnabla on N to NC');


latexSubstitution = FC.showSubstitution(substitutionHNN);
// latexToSave += latexSubstitution + CONST.CR;
latexToSave = '';

let THNN0 = new MathElement();
THNN0.setID('THNN0')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNN0N = MathProduct.newFrom(THNN0)
	.multiplyOnRightWith(PROTO.N)
	.setContraction([0,0], [1,0]);
// THNN0N = MathArgument.newFrom(THNN0N);
let THNN0NS = MathSum.newFrom(THNN0N);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNN0NS, substitutionHNN);

let THNN1 = new MathElement();
THNN1.setID('THNN1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNN1N = MathProduct.newFrom(THNN1)
	.multiplyOnRightWith(PROTO.N1)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0]);
// THNN1N = MathArgument.newFrom(THNN1N);
let THNN1NS = MathSum.newFrom(THNN1N);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNN1NS, substitutionHNN);

let THNN2 = new MathElement();
THNN2.setID('THNN2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNN2N = MathProduct.newFrom(THNN2)
	.multiplyOnRightWith(PROTO.N2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0]);
let THNN2NS = MathSum.newFrom(THNN2N);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNN2NS, substitutionHNN);


let THNN3 = new MathElement();
THNN3.setID('THNN3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let THNN3N = MathProduct.newFrom(THNN3)
	.multiplyOnRightWith(PROTO.N3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0]);
let THNN3NS = MathSum.newFrom(THNN3N);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, THNN3NS, substitutionHNN);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-N-NC.tex', latexToSave);

}
// local context for variables




console.log('');
let endTime = Date.now();
durationTime = endTime - startTime;
console.log('Duration (ms): ' + durationTime);


FH.showTitle('END: ' + path.basename(__filename));
