/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');

const {FC} = require(REL_PATH + 'FUNCTIONS-FILES/functions-computation');
const {FI} = require(REL_PATH + 'FUNCTIONS-FILES/functions-input');
const {FR} = require(REL_PATH + 'FUNCTIONS-FILES/functions-rules');

OPT.TITLE.widthTitle = 120;
OPT.TITLE.widthSubtitle = 120;
OPT.TITLE.widthSubsubtitle = 120;



// get all the fileKeys to check if all split expressions are taken into account
let fileKeysArrJSON = fs.readFileSync(REL_PATH + 'RESULTS/LISTS/SPLIT-R4-list.json', 'utf8');
let fileKeysArr = JSON.parse(fileKeysArrJSON);

removeFileKeysInArray = function(fileKey) {
	let index = fileKeysArr.indexOf(fileKey);
	if (index > -1) {
	  fileKeysArr.splice(index, 1);
	} else {
		console.log('fileKey not found: ' + fileKey);
		throw "not in array";
	};
};



// some functions to test some Simplifications

compareUsinghnabla2Substitution = function(simplificationFileRoot, sumResultFileRoots, showSums=false) {

	let simplifiedSum = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationFileRoot + '.json');
	simplifiedSum.fillNormalizedIndicesArray();

	console.log(simplificationFileRoot);

	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	// copy to be transformed to be saved in latex
	let simplifiedSumCopy = simplifiedSum.makeCopy();
	for (let i=0; i<simplifiedSumCopy.list.length; i++) {
		FH.printProgressMessage('term number ' + i + '/' + simplifiedSumCopy.list.length);
		let term = simplifiedSumCopy.list[i];
		FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_3, term, RULE_DATA);
	};
	FH.printProgressMessage('');

	// necessary to clean zero terms
	simplifiedSumCopy.updateExpression();
	simplifiedSumCopy.sortBySumWeight();
	simplifiedSumCopy.fillNormalizedIndicesArray();
	simplifiedSumCopy.simplifyExpression();

	let latex = simplifiedSumCopy.getLaTeXEquation(OPT.LTX, 'R = ');
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationFileRoot + '.tex', latex);
	inputLtx += '\\input{SIMPLIFICATIONS/' + simplificationFileRoot + '.tex' + '}\n\n';

	// transform simplified sum to compare to result
	simplifiedSum.substituteItemAndDerivativesWithData(SUBST.hnabla2u);
	simplifiedSum.collectCommutativeElementsInPreElement();
	simplifiedSum.sortPreElementsByProductWeight();
	simplifiedSum.sortBySumWeight();
	simplifiedSum.fillNormalizedIndicesArray();
	simplifiedSum.simplifyExpression();

	for (let i=0; i<simplifiedSum.list.length; i++) {
		FH.printProgressMessage('term number ' + i + '/' + simplifiedSum.list.length);
		let term = simplifiedSum.list[i];
		FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_3, term, RULE_DATA);
	};
	FH.printProgressMessage('');

	// necessary to clean zero terms
	simplifiedSum.updateExpression();
	simplifiedSum.sortBySumWeight();
	simplifiedSum.fillNormalizedIndicesArray();
	simplifiedSum.simplifyExpression();

	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let sumResult = new MathSum();
	for (let fileRoot of sumResultFileRoots) {
		sumResult.addTo(FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json'));
		removeFileKeysInArray(fileRoot);
	};

	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
		console.log('');
	};

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSum, sumResult, opt);

	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

	return;
};



compareUsingNablaFSubstitution = function(simplificationFileRoot, sumResultFileRoots, showSums=false) {

	let simplifiedSum = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationFileRoot + '.json');
	simplifiedSum.fillNormalizedIndicesArray();

	console.log(simplificationFileRoot);


	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let latex = simplifiedSum.getLaTeXEquation(OPT.LTX, 'R = ');
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationFileRoot + '.tex', latex);
	inputLtx += '\\input{SIMPLIFICATIONS/' + simplificationFileRoot + '.tex' + '}\n\n';

	// substitute hnabla F by sum of nabla3 on 1
	simplifiedSum.substituteItemAndDerivativesWithData(SUBST.HNABLA_F_TO_NABLA3_ONE_POST_ELEMENT);
	simplifiedSum.collectCommutativeElementsInPreElement();
	simplifiedSum.sortPreElementsByProductWeight();
	simplifiedSum.sortBySumWeight();
	simplifiedSum.fillNormalizedIndicesArray();
	simplifiedSum.simplifyExpression();

	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let sumResult = new MathSum();
	for (let fileRoot of sumResultFileRoots) {
		sumResult.addTo(FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json'));
		removeFileKeysInArray(fileRoot);
	};

	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
		console.log('');
	};

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSum, sumResult, opt)
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

	return;
};








FH.showTitle('BEGINNING: ' + path.basename(__filename));

let arraySimplifiedFileRoots = [
	// R4_0
	// 1.1
	'SPLIT-0/RTS+u',
	'SPLIT-0/RiccS+u',
	'SPLIT-0/hlaplacianSC+u',
	'SPLIT-0/SCS+u',
	// 1.2
	'SPLIT-0/gi(nu1nu2)*hnabla(nu2)SC+hnabla1(nu1)u',
	// 1.3
	'SPLIT-0/gi(nu1nu2)*SC+hnabla1(nu1)u*hnabla1(nu2)u',
	'SPLIT-0/RicU(nu1nu2)+hnabla1(nu1)u*hnabla1(nu2)u',
	// 1.4
	// the first 20 terms of the files are the same through subsitution of indices on the metric
	// the last 5 terms of the first file have to be added
	// so this reduces to 25 terms
	// REPLACED BY SPECIFIC TESTS
	// 'SPLIT-0/gi(nu1nu2)*gi(nu3nu4)+hnabla1(nu1)u*hnabla1(nu2)u*hnabla1(nu3)u*hnabla1(nu4)u',
	// 'SPLIT-0/gi(nu1nu3)*gi(nu2nu4)+hnabla1(nu1)u*hnabla1(nu2)u*hnabla1(nu3)u*hnabla1(nu4)u',
	// 'SPLIT-0/gi(nu1nu4)*gi(nu2nu3)+hnabla1(nu1)u*hnabla1(nu2)u*hnabla1(nu3)u*hnabla1(nu4)u',
	// 1.5
	'SPLIT-0/SC+hlaplacianu',
	// 1.5.1
	'SPLIT-0-DOUBLE-DERIVATIONS-1-5-1/RicU(nu1nu2)+Shnabla2(nu1nu2)u',
	// 1.6
	'SPLIT-0/gi(nu1nu2)+hlaplacianu*hnabla1(nu1)u*hnabla1(nu2)u',
	'SPLIT-0/gi(nu1nu2)+hnabla1(nu1)u*hlaplacianu*hnabla1(nu2)u',
	'SPLIT-0/gi(nu1nu2)+hnabla1(nu1)u*hnabla1(nu2)u*hlaplacianu',
	// 1.6.1
	'SPLIT-0-DOUBLE-DERIVATIONS-1-6-1/one+Shnabla2U(nu1nu2)u*hnabla1(nu1)u*hnabla1(nu2)u',
	// 1.6.2
	'SPLIT-0-DOUBLE-DERIVATIONS-1-6-2/one+hnabla1(nu1)u*Shnabla2U(nu1nu2)u*hnabla1(nu2)u',
	'SPLIT-0-DOUBLE-DERIVATIONS-1-6-2/one+hnabla1(nu1)u*comFUu(nu1nu2)*hnabla1(nu2)u',
	// 1.6.3
	'SPLIT-0-DOUBLE-DERIVATIONS-1-6-3/one+hnabla1(nu1)u*hnabla1(nu2)u*Shnabla2U(nu1nu2)u',
	'SPLIT-0-DOUBLE-DERIVATIONS-1-6-3/one+hnabla1(nu1)u*hnabla1(nu2)u*comFUu(nu1nu2)',
	// 1.7
	'SPLIT-0/one+hlaplacianu*hlaplacianu',
	// 1.7.1
	// REPLACED BY SPECIFIC TESTS
	// 'SPLIT-0-DOUBLE-DERIVATIONS-1-7/one+Shnabla2U(nu1nu2)u*Shnabla2(nu1nu2)u',
	// 'SPLIT-0-DOUBLE-DERIVATIONS-1-7/one+Shnabla2(nu1nu2)u*comFUu(nu1nu2)',
	// 'SPLIT-0-DOUBLE-DERIVATIONS-1-7/one+comFUu(nu1nu2)*Shnabla2(nu1nu2)u',
	// 'SPLIT-0-DOUBLE-DERIVATIONS-1-7/one+comFUu(nu1nu2)*comFu(nu1nu2)',
	// 1.8
	// REPLACED BY SPECIFIC TESTS
	// 'SPLIT-0/gi(nu1nu2)+hlaplacianhnabla(nu1)u*hnabla1(nu2)u',
	// 'SPLIT-0/gi(nu1nu2)+hnabla(nu1)hlaplacianu*hnabla1(nu2)u',
	'SPLIT-0/gi(nu1nu2)+hnabla1(nu1)u*hlaplacianhnabla(nu2)u',
	'SPLIT-0/gi(nu1nu2)+hnabla1(nu1)u*hnabla(nu2)hlaplacianu',
	'SPLIT-0/gi(nu1nu3)*gi(nu2nu4)+hnabla1(nu1)u*hnabla3(nu2nu3nu4)u',
	'SPLIT-0/gi(nu1nu3)*gi(nu2nu4)+hnabla3(nu1nu2nu3)u*hnabla1(nu4)u',
	// 1.9
	'SPLIT-0/gi(nu1nu3)*gi(nu2nu4)+hnabla4(nu1nu2nu3nu4)u',
	'SPLIT-0/gi(nu1nu2)+hnabla(nu1)hlaplacianhnabla(nu2)u',
	'SPLIT-0/one+hlaplacianhlaplacianu',
	// 1.10
	'SPLIT-0/SC+q',
	'SPLIT-0/one+q*q',
	// 1.11
	'SPLIT-0/gi(nu1nu2)+hnabla1(nu1)u*hnabla1(nu2)u*q',
	'SPLIT-0/gi(nu1nu2)+hnabla1(nu1)u*q*hnabla1(nu2)u',
	'SPLIT-0/gi(nu1nu2)+q*hnabla1(nu1)u*hnabla1(nu2)u',
	// 1.12
	// REPLACED BY SPECIFIC TESTS
	// 'SPLIT-0/one+hlaplacianu*q',
	// 'SPLIT-0/one+q*hlaplacianu',
	// 1.13
	'SPLIT-0/gi(nu1nu2)+hnabla1(nu1)q*hnabla1(nu2)u',
	'SPLIT-0/gi(nu1nu2)+hnabla1(nu1)u*hnabla1(nu2)q',
	// 1.14
	'SPLIT-0/one+hlaplacianq',
	// R4_2
	'SPLIT-2/gi(nu2nu3)+hnabla1(nu3)N(nu1)',
	'SPLIT-2/gi(nu2nu3)+hnabla1(nu3)u*N(nu1)',
	'SPLIT-2/gi(nu2nu3)+N(nu1)*hnabla1(nu3)u',
	'SPLIT-2/one+hnabla1(nu1)u*hnabla1(nu2)u',
	'SPLIT-2/one+N(nu1)*N(nu2)',
	'SPLIT-2-DOUBLE-DERIVATIONS/one+comFu(nu1nu2)',
	// R4_3
	// REPLACED BY SPECIFIC TESTS
	// 'SPLIT-3/gi(nu1nu2)*gi(nu3nu4)+hnabla1(nu4)u',
	// 'SPLIT-3/gi(nu1nu3)*gi(nu2nu4)+hnabla1(nu4)u',
	// 'SPLIT-3/gi(nu1nu4)*gi(nu2nu3)+hnabla1(nu4)u',
	// REPLACED BY SPECIFIC TESTS
	// 'SPLIT-3/gi(nu1nu2)+N(nu3)',
	// 'SPLIT-3/gi(nu1nu3)+N(nu2)',
	// 'SPLIT-3/gi(nu2nu3)+N(nu1)',
	// R4_4
	'SPLIT-4/one+u',
];

let inputLtx = '';
let showSums = false;


FH.showSubtitle('TESTS OF SPECIFIC SIMPLIFIED EXPRESSIONS');


FH.showSubsubtitle('batch of terms with laplacian and covariant derivative (section 1.8)');


{ // local group for variables
	let showSums = false;
	let simplificationFileRoot = 'SPLIT-0/gi(nu1nu2)+anticomHLHN(nu1)u*hnabla1(nu2)u';
	let sumResultFileRoots = [
		'SPLIT-0/gi(nu1nu2)+hnabla(nu1)hlaplacianu*hnabla1(nu2)u',
		'SPLIT-0/gi(nu1nu2)+hlaplacianhnabla(nu1)u*hnabla1(nu2)u'
	];

	let simplifiedSum = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationFileRoot + '.json');
	simplifiedSum.fillNormalizedIndicesArray();

	console.log(simplificationFileRoot);


	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let latex = simplifiedSum.getLaTeXEquation(OPT.LTX, 'R = ');
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationFileRoot + '.tex', latex);
	inputLtx += '\\input{SIMPLIFICATIONS/' + simplificationFileRoot + '.tex' + '}\n\n';

// substitute anticommutator {laplacian, nabla} on u to corresponding sum
simplifiedSum.substituteItemAndDerivativesWithData(SUBST.anticomHLHNu);
simplifiedSum.collectCommutativeElementsInPreElement();
simplifiedSum.sortPreElementsByProductWeight();
simplifiedSum.sortBySumWeight();
simplifiedSum.fillNormalizedIndicesArray();
simplifiedSum.simplifyExpression();

if (showSums) {
	simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
	console.log('');
};


	let sumResult = new MathSum();
	for (let fileRoot of sumResultFileRoots) {
		sumResult.addTo(FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json'));
		removeFileKeysInArray(fileRoot);
	};

	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
		console.log('');
	};

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSum, sumResult, opt)
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

} // end of local group for variables





// process.exit(0);




FH.showSubsubtitle('1st batch of terms in symmetric and antisymmetric second derivatives (section 1.7)');

{ // local group for variables
	let simplificationFileRoot = 'SPLIT-0-DOUBLE-DERIVATIONS-1-7/gi(nu1nu3)*gi(nu2nu4)+hnabla2(nu1nu2)u*Shnabla2(nu3nu4)u';
	let sumResultFileRoots = [
		'SPLIT-0-DOUBLE-DERIVATIONS-1-7/one+Shnabla2U(nu1nu2)u*Shnabla2(nu1nu2)u',
		'SPLIT-0-DOUBLE-DERIVATIONS-1-7/one+comFUu(nu1nu2)*Shnabla2(nu1nu2)u'
	];

	compareUsinghnabla2Substitution(simplificationFileRoot, sumResultFileRoots, false);

} // end of local group for variables

FH.showSubsubtitle('2nd batch of terms in symmetric and antisymmetric second derivatives (section 1.7)');

{ // local group for variables
	let simplificationFileRoot = 'SPLIT-0-DOUBLE-DERIVATIONS-1-7/gi(nu1nu3)*gi(nu2nu4)+hnabla2(nu1nu2)u*comFu(nu3nu4)';
	let sumResultFileRoots = [
		'SPLIT-0-DOUBLE-DERIVATIONS-1-7/one+Shnabla2(nu1nu2)u*comFUu(nu1nu2)',
		'SPLIT-0-DOUBLE-DERIVATIONS-1-7/one+comFUu(nu1nu2)*comFu(nu1nu2)'
	];

	compareUsinghnabla2Substitution(simplificationFileRoot, sumResultFileRoots, false);

} // end of local group for variables

// process.exit(0);


FH.showSubsubtitle('propagation of q (section 1.12)');

{ // local group for variables
	let showSums = false;
	let simplificationFileRoot = 'SPLIT-0/one+hlaplacianu-propagate-q';
	let sumResultFileRoots = [
		'SPLIT-0/one+hlaplacianu*q',
		'SPLIT-0/one+q*hlaplacianu'
	];

	let simplifiedSum = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationFileRoot + '.json');
	simplifiedSum.fillNormalizedIndicesArray();

	console.log(simplificationFileRoot);


	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};


	let obj = FC.propagateItemInSum(PROTO.q, simplifiedSum, OPT.LTX, 'R = ');
	let latex = obj.ltx;
	// console.log('latex = ' + latex);
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationFileRoot + '.tex', latex);
	inputLtx += '\\input{SIMPLIFICATIONS/' + simplificationFileRoot + '.tex' + '}\n\n';

	simplifiedSum = obj.sum;
	simplifiedSum.collectCommutativeElementsInPreElement();
	simplifiedSum.sortPreElementsByProductWeight();
	simplifiedSum.sortBySumWeight();
	simplifiedSum.fillNormalizedIndicesArray();
	simplifiedSum.simplifyExpression();

	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let sumResult = new MathSum();
	for (let fileRoot of sumResultFileRoots) {
		sumResult.addTo(FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json'));
		removeFileKeysInArray(fileRoot);
	};

	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
		console.log('');
	};

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSum, sumResult, opt);

	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

} // end of local group for variables


// process.exit(0);


FH.showSubsubtitle('4 first order derivatives of u (section 1.4)');

{ // local group for variables
	let showSums = false;
	let simplificationFileRoot = 'SPLIT-0/gi(nu1nu2)*gi(nu3nu4)+hnabla1(nu1)u*hnabla1(nu2)u*hnabla1(nu3)u*hnabla1(nu4)u-part1';
	let sumResultFileRoots = [
		'SPLIT-0/gi(nu1nu2)*gi(nu3nu4)+hnabla1(nu1)u*hnabla1(nu2)u*hnabla1(nu3)u*hnabla1(nu4)u',
		'SPLIT-0/gi(nu1nu3)*gi(nu2nu4)+hnabla1(nu1)u*hnabla1(nu2)u*hnabla1(nu3)u*hnabla1(nu4)u',
		'SPLIT-0/gi(nu1nu4)*gi(nu2nu3)+hnabla1(nu1)u*hnabla1(nu2)u*hnabla1(nu3)u*hnabla1(nu4)u'
	];
	let simplificationRemainderFileRoot = 'SPLIT-0/gi(nu1nu2)*gi(nu3nu4)+hnabla1(nu1)u*hnabla1(nu2)u*hnabla1(nu3)u*hnabla1(nu4)u-part2';

	let simplifiedSum = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationFileRoot + '.json');
	simplifiedSum.fillNormalizedIndicesArray();

	let simplifiedRemainderSum = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationRemainderFileRoot + '.json');
	simplifiedRemainderSum.fillNormalizedIndicesArray();

	console.log(simplificationFileRoot);
	console.log(simplificationRemainderFileRoot);


	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	if (showSums) {
		simplifiedRemainderSum.showNormalized(OPT.SHOW, 'Simplified Remainder Sum = ');
		console.log('');
	};

	let latex = simplifiedSum.getLaTeXEquation(OPT.LTX, 'R = ');
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationFileRoot + '.tex', latex);
	inputLtx += '\\input{SIMPLIFICATIONS/' + simplificationFileRoot + '.tex' + '}\n\n';

	let latexRemainder = simplifiedRemainderSum.getLaTeXEquation(OPT.LTX, 'R = ');
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationRemainderFileRoot + '.tex', latexRemainder);
	inputLtx += '\\input{SIMPLIFICATIONS/' + simplificationRemainderFileRoot + '.tex' + '}\n\n';

	// substitute tensor with 4 indices
	simplifiedSum.substituteItemAndDerivativesWithData(SUBST.sumgigi);
	simplifiedSum.collectCommutativeElementsInPreElement();
	simplifiedSum.sortPreElementsByProductWeight();
	simplifiedSum.sortBySumWeight();
	simplifiedSum.fillNormalizedIndicesArray();
	simplifiedSum.simplifyExpression();

	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let sumResult = new MathSum();
	for (let fileRoot of sumResultFileRoots) {
		sumResult.addTo(FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json'));
		removeFileKeysInArray(fileRoot);
	};

	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
		console.log('');
	};

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSum, sumResult, opt);
	sumResult = FC.removeSimplifiedSumFrom(simplifiedRemainderSum, sumResult, opt);

	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

} // end of local group for variables




FH.showSubsubtitle('1st batch of terms in R4_3');
// substitute hnabla on F with sum of nabla3 on 1
// test against 3 contributions

{ // local group for variables
	let simplificationFileRoot = 'SPLIT-3/gi(nu1nu2)*gi(nu3nu4)+hnabla1(nu4)u+hnabla(nu1)F(nu2nu3)';
	let sumResultFileRoots = [
		'SPLIT-3/gi(nu1nu2)*gi(nu3nu4)+hnabla1(nu4)u',
		'SPLIT-3/gi(nu1nu3)*gi(nu2nu4)+hnabla1(nu4)u',
		'SPLIT-3/gi(nu1nu4)*gi(nu2nu3)+hnabla1(nu4)u'
	];
	compareUsingNablaFSubstitution(simplificationFileRoot, sumResultFileRoots);
} // end of local group for variables




FH.showSubsubtitle('2nd batch of terms in R4_3');
// substitute hnabla on F with sum of nabla3 on 1
// test against 3 contributions

{ // local group for variables
	let simplificationFileRoot = 'SPLIT-3/gi(nu1nu2)+N(nu3)+hnabla(nu1)F(nu2nu3)';
	let sumResultFileRoots = [
		'SPLIT-3/gi(nu1nu2)+N(nu3)',
		'SPLIT-3/gi(nu1nu3)+N(nu2)',
		'SPLIT-3/gi(nu2nu3)+N(nu1)'
	];
	compareUsingNablaFSubstitution(simplificationFileRoot, sumResultFileRoots);
} // end of local group for variables




FH.showSubtitle('TESTS OF GENERIC SIMPLIFIED EXPRESSIONS');

for (let fileRoot of arraySimplifiedFileRoots) {

	// FH.showSubtitle(fileRoot);
	console.log(fileRoot);

	// FH.showSubsubtitle('Loading simplified expression');

	let simplifiedSum = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + fileRoot + '.json');
	simplifiedSum.fillNormalizedIndicesArray();

	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
	};


	// FH.showSubsubtitle('Save LaTeX of simplified expression');

	let latex = simplifiedSum.getLaTeXEquation(OPT.LTX, 'R = ');
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + fileRoot + '.tex', latex);
	inputLtx += '\\input{SIMPLIFICATIONS/' + fileRoot + '.tex' + '}\n\n';


	// FH.showSubsubtitle('Loading computed expression');

	let sumResult = FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json');
	removeFileKeysInArray(fileRoot);
	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
	};

	// FH.showSubsubtitle('Computing the difference');

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSum, sumResult, opt)
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

	console.log('');
};

// save file with latex inputs
FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/input-split-simplifications.tex', inputLtx);

console.log('');
console.log('remainding file keys: ' + JSON.stringify(fileKeysArr));


FH.showTitle('END: ' + path.basename(__filename));
