/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

let startTime = Date.now();

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');

const {FC} = require(REL_PATH + 'FUNCTIONS-FILES/functions-computation');
const {FR} = require(REL_PATH + 'FUNCTIONS-FILES/functions-rules');

let selection = null;
let selectionSum = null;
let partialStartTime = null;
let partialEndTime = null;
let durationTime = null;
let mathString = null;

FH.showTitle('BEGINNING: ' + path.basename(__filename));


FH.showSubtitle('INITIALIZATION');

FH.showSubsubtitle('Create G tensor');
// tensors G symmetric
let pmax = 3;
const dataGTensor = FC.prepareGtensor(PROTO.g, pmax, PROTO.dim);
const {arrG, arrGIndices, arrRedPerm, arrGPrefactors} = dataGTensor;


FH.showSubsubtitle('Load substitutions');

// get computed substitutions
let substitutionH = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.H_to_nc);
let substitutionL = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.L_to_nc);
let substitutionHNu = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.hnabla_u_to_nc);
let substitutionHNq = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.hnabla_q_to_nc);
let substitutionHNN = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.hnabla_N_to_nc);



FH.showSubsubtitle('Term - P');
let T1 = MathSum.newFrom(PROTO.P)
				.multiplyByInteger(-1)
				.normalizeToArguments();


FH.showSubsubtitle('Term K otimes K = - iK otimes iK');
let T2 = MathSum.newFrom(PROTO.iK)
				.tensorOnRightWith(PROTO.iK)
				.multiplyByInteger(-1)
				.normalizeToArguments();


FH.showSubsubtitle('The original expression to consider');
let T = new MathSum()
				.addTo(T1)
				.addTo(T2);

T.sortBySumWeight()
	.fillNormalizedIndicesArray();

T.showNormalized(OPT.SHOW, 'Original Sum = ');



FH.showSubtitle('COMPUTATION (STEP 1): PROPAGATION NABLA');

let sumResult = FC.computePropagationInArguments(T, PROTO.H);

sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')



FH.showSubtitle('STATISTICS');

let c1 = 0;
let cN1 = 0;
let cNN1 = 0;
let sum1 = new MathSum();
let sumN1 = new MathSum();
let sumNN1 = new MathSum();
let regExpr = null;

regExpr = FH.constructRegExpFromSelection(SELECTION.PENabla0,
	 											GLOBAL.sumWeightData,
												GLOBAL.sumWeightOperatorData);
let fullRE0 = new RegExp(regExpr.postStr + '-' + regExpr.mainStr, "g");

regExpr = FH.constructRegExpFromSelection(SELECTION.PENabla1,
	 											GLOBAL.sumWeightData,
												GLOBAL.sumWeightOperatorData);
let fullRE1 = new RegExp(regExpr.postStr + '-' + regExpr.mainStr, "g");

regExpr = FH.constructRegExpFromSelection(SELECTION.PENabla2,
	 											GLOBAL.sumWeightData,
												GLOBAL.sumWeightOperatorData);
let fullRE2 = new RegExp(regExpr.postStr + '-' + regExpr.mainStr, "g");

let outerIndicesStat = [];

for (let i=0; i<sumResult.list.length; i++) {
	let n = sumResult.list[i].getOuterIndices().length;
	let p = n/2;
	if (outerIndicesStat[p] === undefined) {
		outerIndicesStat[p] = 1;
	} else {
		outerIndicesStat[p] = outerIndicesStat[p] + 1;
	};

	if (sumResult.list[i].sumStringWeight.match(fullRE0)) {
		c1++;
		sum1.addTo(sumResult.list[i], false);
	} else if (sumResult.list[i].sumStringWeight.match(fullRE1)) {
		cN1++;
		sumN1.addTo(sumResult.list[i], false);
	} else if (sumResult.list[i].sumStringWeight.match(fullRE2)) {
		cNN1++;
		sumNN1.addTo(sumResult.list[i], false);
	};
};
console.log("Termes 1: " + c1);
console.log("Termes ∇1: " + cN1);
console.log("Termes ∇∇1: " + cNN1);

console.log('');
for (let p=0; p<outerIndicesStat.length; p++) {
	console.log('outer indices p = ' + p + ' => # = ' + outerIndicesStat[p]);
};


FH.showSubtitle('SUBSTITUTIONS');


FH.showSubsubtitle('Substitutions for H');

sumResult.substituteItemAndDerivativesWithData(substitutionH);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');



FH.showSubsubtitle('Substitutions for L');

sumResult.substituteItemAndDerivativesWithData(substitutionL);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');


FH.showSubsubtitle('Substitutions in post element');

sumResult.substituteItemAndDerivativesWithData(SUBST.NABLA_POST_ELEMENT);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');


FH.showSubsubtitle('Substitutions gauge to Riemann on u');

sumResult.substituteItemAndDerivativesWithData(substitutionHNu);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');


FH.showSubsubtitle('Substitutions gauge to Riemann on q');

sumResult.substituteItemAndDerivativesWithData(substitutionHNq);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');


FH.showSubsubtitle('Substitutions gauge to Riemann on N');

sumResult.substituteItemAndDerivativesWithData(substitutionHNN);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');



FH.showSubtitle('COMPUTATIONS (STEP 2): G, RULES…');


FH.showSubsubtitle('Contractions with G tensor');

sumResult = FC.contractWithGTensor(sumResult, dataGTensor, PROTO.gi.getID());
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')



FH.showSubsubtitle('Apply primary and secondary rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')



FH.showSubsubtitle('Apply operator rules');

let somethingFound = true;
let workingSum = sumResult;
// let loopC = 0;
while(somethingFound) {
	// loopC++;
	// console.log('loopC: ' + loopC);
	let {flag, sum}
					= FC.applyOperatorSimplificationTopDown(workingSum, PROTO.u);
	// let {flag, sum}
	// 				= FC.applyOperatorSimplificationIgnoreDegree(workingSum, PROTO.u);
	workingSum = sum;
	somethingFound = flag;
};

sumResult = workingSum;

sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
sumResult.showNormalized(OPT.SHOW, 'R = ');
console.log('');
console.log('The expression contains ' + sumResult.getListLength() + ' terms')



FH.showSubsubtitle('Save result');
// save result
FH.saveItemToFile(sumResult, REL_PATH + FILES.JSON.R2);


FH.showSubsubtitle('Mathematica expression');
mathString = sumResult.getMathematica(OPT.MATH);

// console.log(mathString);
// save result
FH.saveToFileWithSureFolder(REL_PATH + FILES.MATH.R2, mathString);



FH.showSubtitle('COMPUTATIONS (STEP 2): N=0');

FH.showSubsubtitle('Substitutions of N to 0');

sumResult.substituteItemAndDerivativesWithData(SUBST.N_TO_ZERO);
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
sumResult.showNormalized(OPT.SHOW, 'R(N=0) = ');
console.log('');
console.log('The expression contains ' + sumResult.getListLength() + ' terms')


FH.showSubsubtitle('Save result');
// save result
FH.saveItemToFile(sumResult, REL_PATH + FILES.JSON.R2_N0);


FH.showSubsubtitle('Mathematica expression');
mathString = sumResult.getMathematica(OPT.MATH);

// console.log(mathString);
// save result
FH.saveToFileWithSureFolder(REL_PATH + FILES.MATH.R2_N0, mathString);


FH.showSubtitle('SAVE LATEX RESULTS');

let latex = '';

FH.showSubsubtitle('Full expression');

// non simplified expression with N non zero
sumResult = FC.restoreFromFile(REL_PATH + FILES.JSON.R2);

// substitute cumulative derivations on 'u' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();


latex = sumResult.getLaTeXEquation(OPT.LTX, 'R_2 = ');
// console.log(latex);
FH.saveToFileWithSureFolder(REL_PATH + FILES.TEX.R2, latex);



FH.showSubsubtitle('Expression for N=0');

// simplified expression with N = zero
sumResult = FC.restoreFromFile(REL_PATH + FILES.JSON.R2_N0);

// substitute cumulative derivations on 'u' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();


latex = sumResult.getLaTeXEquation(OPT.LTX, 'R_2(N=0) = ');
// console.log(latexN);
FH.saveToFileWithSureFolder(REL_PATH + FILES.TEX.R2_N0, latex);



let endTime = Date.now();
durationTime = endTime - startTime;
console.log('Duration (ms): ' + durationTime);


FH.showTitle('END: ' + path.basename(__filename));
