/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

let startTime = Date.now();

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');

const {FC} = require(REL_PATH + 'FUNCTIONS-FILES/functions-computation');

let selection = null;
let selectionSum = null;
let partialStartTime = null;
let partialEndTime = null;
let durationTime = null;


FH.showTitle('BEGINNING: ' + path.basename(__filename));


partialStartTime = Date.now();

FH.showSubtitle('SAVE LATEX RESULTS');

let sumResult = null;
let latex = null;


FH.showSubsubtitle('Non simplified expression');

// non simplified expression with N non zero
sumResult = FC.restoreFromFile(REL_PATH + FILES.JSON.R4_ns[0]);

console.log('');
console.log('The expression contains ' + sumResult.getListLength() + ' terms')

// substitute cumulative derivations on 'u' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

// substitute cumulative derivations on 'q' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

// substitute cumulative derivations on 'N' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R_{4,0}(\\text{n.s.}) = ');
// console.log(latex);
FH.saveToFileWithSureFolder(REL_PATH + FILES.TEX.R4_ns[0], latex);



FH.showSubsubtitle('Simplified expression');

// simplified expression with N non zero
sumResult = FC.restoreFromFile(REL_PATH + FILES.JSON.R4[0]);

sumResult.fillNormalizedIndicesArray();

console.log('');
console.log('The expression contains ' + sumResult.getListLength() + ' terms')

// substitute cumulative derivations on 'u' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

// substitute cumulative derivations on 'q' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

// substitute cumulative derivations on 'N' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R_{4,0} = ');
// console.log(latex);
FH.saveToFileWithSureFolder(REL_PATH + FILES.TEX.R4[0], latex);



FH.showSubsubtitle('Non simplified expression, N=0');

// non simplified expression with N = 0
sumResult = FC.restoreFromFile(REL_PATH + FILES.JSON.R4_ns_N0[0]);

sumResult.fillNormalizedIndicesArray();

console.log('');
console.log('The expression contains ' + sumResult.getListLength() + ' terms')

// substitute cumulative derivations on 'u' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

// substitute cumulative derivations on 'q' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R_{4,0}(\\text{n.s.}, N=0) = ');
// console.log(latex);
FH.saveToFileWithSureFolder(REL_PATH + FILES.TEX.R4_ns_N0[0], latex);



FH.showSubsubtitle('Simplified expression, N=0');

// non simplified expression with N = 0
sumResult = FC.restoreFromFile(REL_PATH + FILES.JSON.R4_N0[0]);

sumResult.fillNormalizedIndicesArray();

console.log('');
console.log('The expression contains ' + sumResult.getListLength() + ' terms')

// substitute cumulative derivations on 'u' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

// substitute cumulative derivations on 'q' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R_{4,0}(N=0) = ');
// console.log(latex);
FH.saveToFileWithSureFolder(REL_PATH + FILES.TEX.R4_N0[0], latex);


let endTime = Date.now();
durationTime = endTime - startTime;
console.log('Duration (ms): ' + durationTime);


FH.showTitle('END: ' + path.basename(__filename));
