/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

let startTime = Date.now();

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');

const {FC} = require(REL_PATH + 'FUNCTIONS-FILES/functions-computation');
const {FR} = require(REL_PATH + 'FUNCTIONS-FILES/functions-rules');

let selection = null;
let selectionSum = null;
let partialStartTime = null;
let partialEndTime = null;
let durationTime = null;

FH.showTitle('BEGINNING: ' + path.basename(__filename));


FH.showSubtitle('INITIALIZATION');

FH.showSubsubtitle('Create G tensor');
// tensors G symmetric
let pmax = 6;
const dataGTensor = FC.prepareGtensor(PROTO.g, pmax, PROTO.dim);
const {arrG, arrGIndices, arrRedPerm, arrGPrefactors} = dataGTensor;

FH.showSubsubtitle('Load substitutions');

// get computed substitutions
let substitutionH = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.H_to_nc);
let substitutionL = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.L_to_nc);
let substitutionHNu = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.hnabla_u_to_nc);
let substitutionHNq = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.hnabla_q_to_nc);
let substitutionHNN = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.hnabla_N_to_nc);


FH.showSubsubtitle('Restore from file');

partialStartTime = Date.now();

let sumResult = FC.restoreFromFile(REL_PATH + FILES.JSON.R4_split[1]);
sumResult.fillNormalizedIndicesArray();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');

partialEndTime = Date.now();
durationTime = partialEndTime - partialStartTime;
console.log('Duration for restauration (ms): ' + durationTime);



FH.showSubtitle('SUBSTITUTIONS');

FH.showSubsubtitle('Substitutions for H');

sumResult.substituteItemAndDerivativesWithData(substitutionH);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');





FH.showSubtitle('COMPUTATIONS: G, RULES…');


FH.showSubsubtitle('Contractions with G tensor');

sumResult = FC.contractWithGTensor(sumResult, dataGTensor, PROTO.gi.getID());
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');



FH.showSubsubtitle('Apply primary and secondary rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')





FH.showSubsubtitle('Apply operator rules');

let somethingFound = true;
let workingSum = sumResult;
// let loopC = 0;
while(somethingFound) {
	// loopC++;
	// console.log('loopC: ' + loopC);
	// let {flag, sum}
	// 				= FC.applyOperatorSimplificationTopDown(workingSum, PROTO.u);
	let {flag, sum}
					= FC.applyOperatorSimplificationIgnoreDegree(workingSum, PROTO.u);
	workingSum = sum;
	somethingFound = flag;
};

sumResult = workingSum;

sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
sumResult.showNormalized(OPT.SHOW, 'R = ');
console.log('');
console.log('The expression contains ' + sumResult.getListLength() + ' terms')


FH.showSubsubtitle('Save simplified result');
// save result
FH.saveItemToFile(sumResult, REL_PATH + FILES.JSON.R4[1]);


FH.showSubtitle('SAVE LATEX RESULTS');

let latex = '';

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R_{4,1} = ');
// console.log(latex);
FH.saveToFileWithSureFolder(REL_PATH + FILES.TEX.R4[1], latex);


let endTime = Date.now();
durationTime = endTime - startTime;
console.log('Duration (ms): ' + durationTime);


FH.showTitle('END: ' + path.basename(__filename));
