/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');

const {FC} = require(REL_PATH + 'FUNCTIONS-FILES/functions-computation');
const {FI} = require(REL_PATH + 'FUNCTIONS-FILES/functions-input');
const {FR} = require(REL_PATH + 'FUNCTIONS-FILES/functions-rules');




FH.showTitle('BEGINNING');

let fileRoot = null;
let fileRoot_1 = null;
let fileRoot_2 = null;
let sumResult = null;
let sumResult_1 = null;
let sumResult_2 = null;
let latex = null;
let preLatex = null;


// get list of split files
let fileKeysArr = [];
let listFilesArr = [
	'RESULTS/LISTS/SPLIT-0-list.json',
	'RESULTS/LISTS/SPLIT-2-list.json',
	'RESULTS/LISTS/SPLIT-3-list.json',
	'RESULTS/LISTS/SPLIT-4-list.json'
];
for (let listFile of listFilesArr) {
	let fileKeysArrJSON = fs.readFileSync(REL_PATH + listFile, 'utf8');
	fileKeysArr = fileKeysArr.concat(JSON.parse(fileKeysArrJSON));
};

// console.log(JSON.stringify(fileKeysArr, null, "\t"));
// process.exit(0);

replaceFileKeysInArray = function(oldFileKeysArr, newFileKeysArr) {
	for(let fileKey of oldFileKeysArr) {
		let index = fileKeysArr.indexOf(fileKey);
		if (index > -1) {
		  fileKeysArr.splice(index, 1);
		} else {
			console.log('fileKey not found: ' + fileKey);
			throw "not in array";
		};
	};
	fileKeysArr = fileKeysArr.concat(newFileKeysArr);
};

getNewFileKeysArrFromSplit = function(postStr, rel_path) {
	// return array saved using:
	// // save file with list of file keys
	// FH.saveToFileWithSureFolder(rel_path + 'RESULTS/LISTS/SPLIT' + postStr + '-list.json', JSON.stringify(fileKeysArr, null, "\t"));
	let dataJSON = fs.readFileSync(rel_path + 'RESULTS/LISTS/SPLIT' + postStr + '-list.json', 'utf8');
	return JSON.parse(dataJSON);
};




FH.showSubtitle('PACKET OF TERMS IN 1.5')

FH.showSubsubtitle('Load results and add them');


fileRoot = 'SPLIT-0/RicU(nu1nu2)+hnabla2(nu1nu2)u';
sumResult = FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json');

preLatex = 'Simplifications of terms in:' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += '\\fileName{' + fileRoot + '}' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += 'produce:'
preLatex += '\n\n';
preLatex += '\\medskip\n\n';


// substitute cumulative derivations on 'u' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);

sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');


FH.showSubsubtitle('Substitution');

sumResult.substituteItemAndDerivativesWithData(SUBST.hnabla2u);

sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');


FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_3, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')




FH.showSubsubtitle('Save LaTeX and splitted results');

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R = ');
FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SPLIT-0-DOUBLE-DERIVATIONS/R4-1-5-1.tex', latex);

FC.saveSplitSum(sumResult, PROTO.u, '-0-DOUBLE-DERIVATIONS-1-5-1', REL_PATH, preLatex);


replaceFileKeysInArray([fileRoot], getNewFileKeysArrFromSplit('-0-DOUBLE-DERIVATIONS-1-5-1', REL_PATH));
// console.log(JSON.stringify(fileKeysArr, null, "\t"));







FH.showSubtitle('1ST PACKET OF 2 TERMS IN 1.6')

FH.showSubsubtitle('Load results and add them');

fileRoot_1 = 'SPLIT-0/gi(nu1nu3)*gi(nu2nu4)+hnabla2(nu1nu2)u*hnabla1(nu3)u*hnabla1(nu4)u';
sumResult_1 = FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot_1 + '.json');

fileRoot_2 = 'SPLIT-0/gi(nu1nu4)*gi(nu2nu3)+hnabla2(nu1nu2)u*hnabla1(nu3)u*hnabla1(nu4)u';
sumResult_2 = FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot_2 + '.json');

preLatex = 'Simplifications of terms in:' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += '\\fileName{' + fileRoot_1 + '}' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += '\\fileName{' + fileRoot_2 + '}' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += 'produce:'
preLatex += '\n\n';
preLatex += '\\medskip\n\n';


// add two sums
sumResult = new MathSum();
sumResult.addTo(sumResult_1, false).addTo(sumResult_2, false);

// substitute cumulative derivations on 'u' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);

sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');

FH.showSubsubtitle('Substitution');

sumResult.substituteItemAndDerivativesWithData(SUBST.hnabla2u);

sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');


FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_3, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')




FH.showSubsubtitle('Save LaTeX and splitted results');

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R = ');
FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SPLIT-0-DOUBLE-DERIVATIONS/R4-1-6-1.tex', latex);

FC.saveSplitSum(sumResult, PROTO.u, '-0-DOUBLE-DERIVATIONS-1-6-1', REL_PATH, preLatex);

replaceFileKeysInArray([fileRoot_1, fileRoot_2], getNewFileKeysArrFromSplit('-0-DOUBLE-DERIVATIONS-1-6-1', REL_PATH));
// console.log(JSON.stringify(fileKeysArr, null, "\t"));





FH.showSubtitle('2ND PACKET OF 2 TERMS IN 1.6')

FH.showSubsubtitle('Load results and add them');

fileRoot_1 = 'SPLIT-0/gi(nu1nu3)*gi(nu2nu4)+hnabla1(nu1)u*hnabla2(nu2nu3)u*hnabla1(nu4)u';
sumResult_1 = FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot_1 + '.json');

fileRoot_2 = 'SPLIT-0/gi(nu1nu2)*gi(nu3nu4)+hnabla1(nu1)u*hnabla2(nu2nu3)u*hnabla1(nu4)u';
sumResult_2 = FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot_2 + '.json');

preLatex = 'Simplifications of terms in:' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += '\\fileName{' + fileRoot_1 + '}' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += '\\fileName{' + fileRoot_2 + '}' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += 'produce:'
preLatex += '\n\n';
preLatex += '\\medskip\n\n';

// add two sums
sumResult = new MathSum();
sumResult.addTo(sumResult_1, false).addTo(sumResult_2, false);

// substitute cumulative derivations on 'u' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);

sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');

FH.showSubsubtitle('Substitution');

sumResult.substituteItemAndDerivativesWithData(SUBST.hnabla2u);

sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');


FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_3, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')



FH.showSubsubtitle('Save LaTeX and splitted results');

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R = ');
FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SPLIT-0-DOUBLE-DERIVATIONS/R4-1-6-2.tex', latex);

FC.saveSplitSum(sumResult, PROTO.u, '-0-DOUBLE-DERIVATIONS-1-6-2', REL_PATH, preLatex);

replaceFileKeysInArray([fileRoot_1, fileRoot_2], getNewFileKeysArrFromSplit('-0-DOUBLE-DERIVATIONS-1-6-2', REL_PATH));
// console.log(JSON.stringify(fileKeysArr, null, "\t"));




FH.showSubtitle('3RD PACKET OF 2 TERMS IN 1.6')

FH.showSubsubtitle('Load results and add them');

fileRoot_1 = 'SPLIT-0/gi(nu1nu3)*gi(nu2nu4)+hnabla1(nu1)u*hnabla1(nu2)u*hnabla2(nu3nu4)u';
sumResult_1 = FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot_1 + '.json');

fileRoot_2 = 'SPLIT-0/gi(nu1nu4)*gi(nu2nu3)+hnabla1(nu1)u*hnabla1(nu2)u*hnabla2(nu3nu4)u';
sumResult_2 = FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot_2 + '.json');

preLatex = 'Simplifications of terms in:' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += '\\fileName{' + fileRoot_1 + '}' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += '\\fileName{' + fileRoot_2 + '}' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += 'produce:'
preLatex += '\n\n';
preLatex += '\\medskip\n\n';

// add two sums
sumResult = new MathSum();
sumResult.addTo(sumResult_1, false).addTo(sumResult_2, false);

// substitute cumulative derivations on 'u' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);

sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');

FH.showSubsubtitle('Substitution');

sumResult.substituteItemAndDerivativesWithData(SUBST.hnabla2u);

sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');


FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_3, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')



FH.showSubsubtitle('Save LaTeX and splitted results');

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R = ');
FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SPLIT-0-DOUBLE-DERIVATIONS/R4-1-6-3.tex', latex);

FC.saveSplitSum(sumResult, PROTO.u, '-0-DOUBLE-DERIVATIONS-1-6-3', REL_PATH, preLatex);

replaceFileKeysInArray([fileRoot_1, fileRoot_2], getNewFileKeysArrFromSplit('-0-DOUBLE-DERIVATIONS-1-6-3', REL_PATH));
// console.log(JSON.stringify(fileKeysArr, null, "\t"));






FH.showSubtitle('PACKET OF 2 TERMS IN 1.7')

FH.showSubsubtitle('Load results and add them');



fileRoot_1 = 'SPLIT-0/gi(nu1nu3)*gi(nu2nu4)+hnabla2(nu1nu2)u*hnabla2(nu3nu4)u';
sumResult_1 = FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot_1 + '.json');

fileRoot_2 = 'SPLIT-0/gi(nu1nu4)*gi(nu2nu3)+hnabla2(nu1nu2)u*hnabla2(nu3nu4)u';
sumResult_2 = FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot_2 + '.json');

preLatex = 'Simplifications of terms in:' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += '\\fileName{' + fileRoot_1 + '}' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += '\\fileName{' + fileRoot_2 + '}' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += 'produce:'
preLatex += '\n\n';
preLatex += '\\medskip\n\n';

// add two sums
sumResult = new MathSum();
sumResult.addTo(sumResult_1, false).addTo(sumResult_2, false);

// substitute cumulative derivations on 'u' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);

sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');

FH.showSubsubtitle('Substitution');

sumResult.substituteItemAndDerivativesWithData(SUBST.hnabla2u);

sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');


FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_3, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')



FH.showSubsubtitle('Save LaTeX and splitted results');

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R = ');
FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SPLIT-0-DOUBLE-DERIVATIONS/R4-1-7.tex', latex);

FC.saveSplitSum(sumResult, PROTO.u, '-0-DOUBLE-DERIVATIONS-1-7', REL_PATH, preLatex);


replaceFileKeysInArray([fileRoot_1, fileRoot_2], getNewFileKeysArrFromSplit('-0-DOUBLE-DERIVATIONS-1-7', REL_PATH));
// console.log(JSON.stringify(fileKeysArr, null, "\t"));







FH.showSubtitle('SOME TERMS IN R_{4,2}')

FH.showSubsubtitle('Load results and add them');

fileRoot = 'SPLIT-2/one+hnabla2(nu1nu2)u';
sumResult = FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json');

preLatex = 'Simplifications of terms in:' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += '\\fileName{' + fileRoot + '}' + CONST.LATEX_SUM_NEWLINE + '\n';
preLatex += 'produce:'
preLatex += '\n\n';
preLatex += '\\medskip\n\n';

// substitute cumulative derivations on 'u' with "powered" derivations
sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);

sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');

FH.showSubsubtitle('Substitution');

sumResult.substituteItemAndDerivativesWithData(SUBST.hnabla2u);

sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');



FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_3, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')




FH.showSubsubtitle('Save LaTeX and splitted results');

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R = ');
FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SPLIT-2-DOUBLE-DERIVATIONS/R4-1.tex', latex);

FC.saveSplitSum(sumResult, PROTO.u, '-2-DOUBLE-DERIVATIONS', REL_PATH, preLatex);

replaceFileKeysInArray([fileRoot], getNewFileKeysArrFromSplit('-2-DOUBLE-DERIVATIONS', REL_PATH));
// console.log(JSON.stringify(fileKeysArr, null, "\t"));





FH.showSubtitle('SAVE LIST OF FILES')

// save file with list of file keys
FH.saveToFileWithSureFolder(REL_PATH + 'RESULTS/LISTS/SPLIT-R4-list.json', JSON.stringify(fileKeysArr, null, "\t"));


FH.showTitle('END');
