/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

let startTime = Date.now();

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');

const {FC} = require(REL_PATH + 'FUNCTIONS-FILES/functions-computation');

let selection = null;
let selectionSum = null;
let partialStartTime = null;
let partialEndTime = null;
let durationTime = null;

FH.showTitle('BEGINNING: ' + path.basename(__filename));



FH.showSubtitle('INITIALIZATION');


FH.showSubsubtitle('Create G tensor');
// tensors G symmetric
let pmax = 6;
const dataGTensor = FC.prepareGtensor(PROTO.g, pmax, PROTO.dim);
const {arrG, arrGIndices, arrRedPerm, arrGPrefactors} = dataGTensor;


FH.showSubsubtitle('Term T1 = P otimes P');

let T1 = MathSum.newFrom(PROTO.P)
				.tensorOnRightWith(PROTO.P)
				.normalizeToArguments();

// T1.showConsole(OPT.SHOW, 'T1 = ');


FH.showSubsubtitle('Term T2 = iK otimes iK otimes P');

let T2 = MathSum.newFrom(PROTO.iK)
					.tensorOnRightWith(PROTO.iK)
					.tensorOnRightWith(PROTO.P)
					.normalizeToArguments();

// T2.showConsole(OPT.SHOW, 'T2 = ');

FH.showSubsubtitle('Term T3 = iK otimes P otimes iK');

let T3 = MathSum.newFrom(PROTO.iK)
					.tensorOnRightWith(PROTO.P)
					.tensorOnRightWith(PROTO.iK)
					.normalizeToArguments();

// T3.showConsole(OPT.SHOW, 'T3 = ');

FH.showSubsubtitle('Term T4 = P otimes iK otimes iK');

let T4 = MathSum.newFrom(PROTO.P)
					.tensorOnRightWith(PROTO.iK)
					.tensorOnRightWith(PROTO.iK)
					.normalizeToArguments();

// T4.showConsole(OPT.SHOW, 'T4 = ');

FH.showSubsubtitle('Term T5 = iK otimes iK otimes iK otimes iK');

let T5 = MathSum.newFrom(PROTO.iK)
					.tensorOnRightWith(PROTO.iK)
					.tensorOnRightWith(PROTO.iK)
					.tensorOnRightWith(PROTO.iK)
					.normalizeToArguments();

// T5.showConsole(OPT.SHOW, 'T5 = ');

FH.showSubsubtitle('The original expression to consider');
let T = new MathSum()
				.addTo(T1)
				.addTo(T2)
				.addTo(T3)
				.addTo(T4)
				.addTo(T5);

T.sortBySumWeight()
	.fillNormalizedIndicesArray();

T.showNormalized(OPT.SHOW, 'Original Sum = ');



FH.showSubtitle('COMPUTATION');

partialStartTime = Date.now();

FH.showSubsubtitle('Propagations of nabla operators');

let sumResult = FC.computePropagationInArguments(T, PROTO.H);

console.log('');
console.log('The expression contains ' + sumResult.getListLength() + ' terms')


FH.showSubsubtitle('Simplifications');

sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')

partialEndTime = Date.now();
durationTime = partialEndTime - partialStartTime;
console.log('Duration for propagation and simplification (ms): ' + durationTime);



// FH.showSubsubtitle('Save full result');
// file is big and useless for the moment
// FH.saveItemToFile(sumResult, REL_PATH + FILES.JSON.R4_save);




FH.showSubtitle('SPLIT RESULT AND STATISTICS');

partialStartTime = Date.now();

let c1 = 0;
let cN1 = 0;
let cNN1 = 0;
let cNNN1 = 0;
let cNNNN1 = 0;

let sumByNabla = [];
let regExpr = null;
for (let nab=0; nab<5; nab++) {
	sumByNabla[nab] = new MathSum();
};

regExpr = FH.constructRegExpFromSelection(SELECTION.PENabla0,
												GLOBAL.sumWeightData,
												GLOBAL.sumWeightOperatorData);
let fullRE0 = new RegExp(regExpr.postStr + '-' + regExpr.mainStr, "g");

regExpr = FH.constructRegExpFromSelection(SELECTION.PENabla1,
												GLOBAL.sumWeightData,
												GLOBAL.sumWeightOperatorData);
let fullRE1 = new RegExp(regExpr.postStr + '-' + regExpr.mainStr, "g");

regExpr = FH.constructRegExpFromSelection(SELECTION.PENabla2,
												GLOBAL.sumWeightData,
												GLOBAL.sumWeightOperatorData);
let fullRE2 = new RegExp(regExpr.postStr + '-' + regExpr.mainStr, "g");

regExpr = FH.constructRegExpFromSelection(SELECTION.PENabla3,
												GLOBAL.sumWeightData,
												GLOBAL.sumWeightOperatorData);
let fullRE3 = new RegExp(regExpr.postStr + '-' + regExpr.mainStr, "g");

regExpr = FH.constructRegExpFromSelection(SELECTION.PENabla4,
												GLOBAL.sumWeightData,
												GLOBAL.sumWeightOperatorData);
let fullRE4 = new RegExp(regExpr.postStr + '-' + regExpr.mainStr, "g");

let outerIndicesStat = [];
let outerIndicesSum = [];

let N = sumResult.list.length;

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('item ' + i + '/' + N);

	let n = sumResult.list[i].getOuterIndices().length;
	let p = n/2;
	if (outerIndicesStat[p] === undefined) {
		outerIndicesStat[p] = 1;
		outerIndicesSum[p] = new MathSum();
		outerIndicesSum[p].addTo(sumResult.list[i], false);
	} else {
		outerIndicesStat[p] = outerIndicesStat[p] + 1;
		outerIndicesSum[p].addTo(sumResult.list[i], false);
	};

	if (sumResult.list[i].sumStringWeight.match(fullRE0)) {
		c1++;
		sumByNabla[0].addTo(sumResult.list[i], false);
	} else if (sumResult.list[i].sumStringWeight.match(fullRE1)) {
		cN1++;
		sumByNabla[1].addTo(sumResult.list[i], false);
	} else if (sumResult.list[i].sumStringWeight.match(fullRE2)) {
		cNN1++;
		sumByNabla[2].addTo(sumResult.list[i], false);
	} else if (sumResult.list[i].sumStringWeight.match(fullRE3)) {
		cNNN1++;
		sumByNabla[3].addTo(sumResult.list[i], false);
	} else if (sumResult.list[i].sumStringWeight.match(fullRE4)) {
		cNNNN1++;
		sumByNabla[4].addTo(sumResult.list[i], false);
	};
};
FH.printProgressMessage('');

console.log("Termes 1: " + c1);
console.log("Termes ∇1: " + cN1);
console.log("Termes ∇∇1: " + cNN1);
console.log("Termes ∇∇∇1: " + cNNN1);
console.log("Termes ∇∇∇∇1: " + cNNNN1);

console.log('');
for (let p=0; p<outerIndicesStat.length; p++) {
	console.log('outer indices p = ' + p + ' => # = ' + outerIndicesStat[p]);
};

partialEndTime = Date.now();
durationTime = partialEndTime - partialStartTime;
console.log('Duration for statistiques (ms): ' + durationTime);




FH.showSubtitle('SAVE SEPARATED FILES');

for (let nab=0; nab<5; nab++) {
	console.log('save #nabla = ' + nab);
	FH.saveItemToFile(sumByNabla[nab], REL_PATH + FILES.JSON.R4_split[nab]);
};



let endTime = Date.now();
durationTime = endTime - startTime;
console.log('Duration (ms): ' + durationTime);


FH.showTitle('END: ' + path.basename(__filename));
