/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');

const {FC} = require(REL_PATH + 'FUNCTIONS-FILES/functions-computation');
const {FI} = require(REL_PATH + 'FUNCTIONS-FILES/functions-input');
const {FR} = require(REL_PATH + 'FUNCTIONS-FILES/functions-rules');

OPT.TITLE.widthTitle = 120;
OPT.TITLE.widthSubtitle = 120;
OPT.TITLE.widthSubsubtitle = 120;


// some functions to test some Simplifications




compareUsingNablaFSubstitution = function(simplificationFileRoot, sumResultFileRoots, showSums=false) {

	let simplifiedSum = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationFileRoot + '.json');
	simplifiedSum.fillNormalizedIndicesArray();

	console.log(simplificationFileRoot);


	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let latex = simplifiedSum.getLaTeXEquation(OPT.LTX, 'R = ');
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationFileRoot + '.tex', latex);

	// substitute hnabla F by sum of nabla3 on 1
	simplifiedSum.substituteItemAndDerivativesWithData(SUBST.HNABLA_F_TO_NABLA3_ONE_POST_ELEMENT);
	simplifiedSum.collectCommutativeElementsInPreElement();
	simplifiedSum.sortPreElementsByProductWeight();
	simplifiedSum.sortBySumWeight();
	simplifiedSum.fillNormalizedIndicesArray();
	simplifiedSum.simplifyExpression();

	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let sumResult = new MathSum();
	for (let fileRoot of sumResultFileRoots) {
		sumResult.addTo(FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json'))
	};

	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
		console.log('');
	};

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSum, sumResult, opt)
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

	return;
};








FH.showTitle('BEGINNING: ' + path.basename(__filename));

let arraySimplifiedFileRoots = [
	'SPLIT-u-parallel/RTS+u',
	'SPLIT-u-parallel/RiccS+u',
	'SPLIT-u-parallel/hlaplacianSC+u',
	'SPLIT-u-parallel/SCS+u',
	'SPLIT-u-parallel/SC+q',
	'SPLIT-u-parallel/one+q*q',
	'SPLIT-u-parallel/one+hlaplacianq',
	'SPLIT-u-parallel/hnabla(nu1)SC+N(nu1)',
	'SPLIT-u-parallel/one+hnabla1(nu1)q*N(nu1)',
	'SPLIT-u-parallel/one+N(nu1)*hnabla1(nu1)q',
	'SPLIT-u-parallel/Ric(nu1nu2)+N(nu1)*N(nu2)',
	'SPLIT-u-parallel/g(nu1nu2)*SC+N(nu1)*N(nu2)',
	// SPECIFIC TESTS
	// 'SPLIT-u-parallel/g(nu1nu2)+q*N(nu1)*N(nu2)',
	// 'SPLIT-u-parallel/g(nu1nu2)+N(nu1)*q*N(nu2)',
	// 'SPLIT-u-parallel/g(nu1nu2)+N(nu1)*N(nu2)*q',
	// SPECIFIC TESTS
	// 'SPLIT-u-parallel/g(nu1nu2)*g(nu3nu4)+N(nu1)*N(nu2)*N(nu3)*N(nu4)',
	// 'SPLIT-u-parallel/g(nu1nu3)*g(nu2nu4)+N(nu1)*N(nu2)*N(nu3)*N(nu4)',
	// 'SPLIT-u-parallel/g(nu1nu4)*g(nu2nu3)+N(nu1)*N(nu2)*N(nu3)*N(nu4)',
	'SPLIT-u-parallel/gi(nu1nu3)*Ric(nu2nu3)+hnabla1(nu1)N(nu2)',
	'SPLIT-u-parallel/SC+hnabla1(nu1)N(nu1)',
	// SPECIFIC TESTS
	// 'SPLIT-u-parallel/one+q*hnabla1(nu1)N(nu1)',
	// 'SPLIT-u-parallel/one+hnabla1(nu1)N(nu1)*q',
	'SPLIT-u-parallel/g(nu1nu3)+N(nu1)*hnabla1(nu2)N(nu2)*N(nu3)',
	'SPLIT-u-parallel/g(nu2nu3)+N(nu1)*hnabla1(nu1)N(nu2)*N(nu3)',
	'SPLIT-u-parallel/g(nu1nu3)+N(nu1)*N(nu2)*hnabla1(nu2)N(nu3)',
	'SPLIT-u-parallel/g(nu1nu2)+N(nu1)*N(nu2)*hnabla1(nu3)N(nu3)',
	'SPLIT-u-parallel/g(nu1nu3)+N(nu1)*hnabla1(nu2)N(nu3)*N(nu2)',
	'SPLIT-u-parallel/g(nu2nu3)+N(nu1)*N(nu2)*hnabla1(nu1)N(nu3)',
	'SPLIT-u-parallel/g(nu2nu3)+hnabla1(nu1)N(nu2)*N(nu3)*N(nu1)',
	'SPLIT-u-parallel/g(nu2nu3)+hnabla1(nu1)N(nu2)*N(nu1)*N(nu3)',
	'SPLIT-u-parallel/g(nu2nu3)+hnabla1(nu1)N(nu1)*N(nu2)*N(nu3)',
	'SPLIT-u-parallel/g(nu2nu4)*gi(nu1nu3)+hnabla1(nu1)N(nu2)*hnabla1(nu3)N(nu4)',
	'SPLIT-u-parallel/one+hnabla1(nu1)N(nu1)*hnabla1(nu2)N(nu2)',
	'SPLIT-u-parallel/one+hnabla1(nu1)N(nu2)*hnabla1(nu2)N(nu1)',
	// SPECIFIC TESTS
	// 'SPLIT-u-parallel/one+N(nu1)*hnabla2(nu1nu2)N(nu2)',
	// 'SPLIT-u-parallel/one+N(nu1)*hnabla2(nu2nu1)N(nu2)',
	// SPECIFIC TESTS
	// 'SPLIT-u-parallel/one+hnabla2(nu1nu2)N(nu2)*N(nu1)',
	// 'SPLIT-u-parallel/one+hnabla2(nu1nu2)N(nu1)*N(nu2)',
	'SPLIT-u-parallel/g(nu1nu2)+hlaplacianN(nu1)*N(nu2)',
	'SPLIT-u-parallel/g(nu1nu2)+N(nu1)*hlaplacianN(nu2)',
	'SPLIT-u-parallel/gi(nu1nu3)+hnabla3(nu1nu2nu3)N(nu2)',
	// SPECIFIC TESTS
	// 'SPLIT-u-parallel/one+hnabla(nu1)hlaplacianN(nu1)',
	// 'SPLIT-u-parallel/one+hlaplacianhnabla(nu1)N(nu1)',
	// R4_2
	'SPLIT-u-parallel/one+N(nu1)*N(nu2)',
	'SPLIT-u-parallel/gi(nu2nu3)+hnabla1(nu3)N(nu1)',
	// R4_3
	// SPECIFIC TESTS
	// 'SPLIT-u-parallel/gi(nu2nu3)+N(nu1)',
	// 'SPLIT-u-parallel/gi(nu1nu3)+N(nu2)',
	// 'SPLIT-u-parallel/gi(nu1nu2)+N(nu3)',
	// R4_4
	'SPLIT-u-parallel/one+u'
];


let inputLtx = '';
let showSums = false;


FH.showSubtitle('TESTS OF SPECIFIC SIMPLIFIED EXPRESSIONS');


FH.showSubsubtitle('batch of terms with laplacian and covariant derivative on N');

{ // local group for variables
	let showSums = false;
	let simplificationFileRoot = 'SPLIT-u-parallel/one+anticomHLHN(nu1)N(nu1)';
	let sumResultFileRoots = [
		'SPLIT-u-parallel/one+hnabla(nu1)hlaplacianN(nu1)',
		'SPLIT-u-parallel/one+hlaplacianhnabla(nu1)N(nu1)'
	];

	let simplifiedSum = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationFileRoot + '.json');
	simplifiedSum.fillNormalizedIndicesArray();

	console.log(simplificationFileRoot);


	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let latex = simplifiedSum.getLaTeXEquation(OPT.LTX, 'R = ');
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationFileRoot + '.tex', latex);

// substitute anticommutator {laplacian, nabla} on u to corresponding sum
simplifiedSum.substituteItemAndDerivativesWithData(SUBST.anticomHLHNN);
simplifiedSum.collectCommutativeElementsInPreElement();
simplifiedSum.sortPreElementsByProductWeight();
simplifiedSum.sortBySumWeight();
simplifiedSum.fillNormalizedIndicesArray();
simplifiedSum.simplifyExpression();

if (showSums) {
	simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
	console.log('');
};


	let sumResult = new MathSum();
	for (let fileRoot of sumResultFileRoots) {
		sumResult.addTo(FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json'))
	};

	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
		console.log('');
	};

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSum, sumResult, opt)
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

} // end of local group for variables


// process.exit(0);



FH.showSubsubtitle('1st batch of second derivatives of N');

{ // local group for variables
	let showSums = false;
	let simplificationFileRootShnablaN = 'SPLIT-u-parallel/one+Shnabla2(nu1nu2)N(nu2)*N(nu1)';
	let sumResultFileRoots = [
		'SPLIT-u-parallel/one+hnabla2(nu1nu2)N(nu2)*N(nu1)',
		'SPLIT-u-parallel/one+hnabla2(nu1nu2)N(nu1)*N(nu2)'
	];

	let simplifiedSumShnablaN = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationFileRootShnablaN + '.json');
	simplifiedSumShnablaN.fillNormalizedIndicesArray();

	console.log(simplificationFileRootShnablaN);


	if (showSums) {
		simplifiedSumShnablaN.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let latexShnablaN = simplifiedSumShnablaN.getLaTeXEquation(OPT.LTX, 'R = ');
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationFileRootShnablaN + '.tex', latexShnablaN);


	let sumResult = new MathSum();
	for (let fileRoot of sumResultFileRoots) {
		sumResult.addTo(FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json'))
	};

	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	// substitute double derivation with sum symmetric and antisymmetric parts
	sumResult.substituteItemAndDerivativesWithData(SUBST.hnabla2N);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
		console.log('');
	};

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSumShnablaN, sumResult, opt);

	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

} // end of local group for variables

// process.exit(0);


FH.showSubsubtitle('2nd batch of second derivatives of N');

{ // local group for variables
	let showSums = false;
	let simplificationFileRootFN = 'SPLIT-u-parallel/one+N(nu1)*comFN(nu1nu2nu2)';
	let simplificationFileRootShnablaN = 'SPLIT-u-parallel/one+N(nu1)*Shnabla2(nu1nu2)N(nu2)';
	let sumResultFileRoots = [
		'SPLIT-u-parallel/one+N(nu1)*hnabla2(nu1nu2)N(nu2)',
		'SPLIT-u-parallel/one+N(nu1)*hnabla2(nu2nu1)N(nu2)'
	];

	let simplifiedSumFN = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationFileRootFN + '.json');
	simplifiedSumFN.fillNormalizedIndicesArray();
	let simplifiedSumShnablaN = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationFileRootShnablaN + '.json');
	simplifiedSumShnablaN.fillNormalizedIndicesArray();

	console.log(simplificationFileRootFN);
	console.log(simplificationFileRootShnablaN);


	if (showSums) {
		simplifiedSumFN.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
		simplifiedSumShnablaN.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let latexFN = simplifiedSumFN.getLaTeXEquation(OPT.LTX, 'R = ');
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationFileRootFN + '.tex', latexFN);
	let latexShnablaN = simplifiedSumShnablaN.getLaTeXEquation(OPT.LTX, 'R = ');
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationFileRootShnablaN + '.tex', latexShnablaN);


	let sumResult = new MathSum();
	for (let fileRoot of sumResultFileRoots) {
		sumResult.addTo(FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json'))
	};

	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	// substitute double derivation with sum symmetric and antisymmetric parts
	sumResult.substituteItemAndDerivativesWithData(SUBST.hnabla2N);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
		console.log('');
	};

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSumFN, sumResult, opt);
	sumResult = FC.removeSimplifiedSumFrom(simplifiedSumShnablaN, sumResult, opt);

	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

} // end of local group for variables

// process.exit(0);





FH.showSubsubtitle('propagation of q in hnabla N terms');

{ // local group for variables
	let showSums = false;
	let simplificationFileRoot = 'SPLIT-u-parallel/one+hnabla1(nu1)N(nu1)-propagate-q';
	let sumResultFileRoots = [
		'SPLIT-u-parallel/one+q*hnabla1(nu1)N(nu1)',
		'SPLIT-u-parallel/one+hnabla1(nu1)N(nu1)*q'
	];

	let simplifiedSum = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationFileRoot + '.json');
	simplifiedSum.fillNormalizedIndicesArray();

	console.log(simplificationFileRoot);


	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};


	let obj = FC.propagateItemInSum(PROTO.q, simplifiedSum, OPT.LTX, 'R = ');
	let latex = obj.ltx;
	// console.log('latex = ' + latex);
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationFileRoot + '.tex', latex);

	simplifiedSum = obj.sum;
	simplifiedSum.collectCommutativeElementsInPreElement();
	simplifiedSum.sortPreElementsByProductWeight();
	simplifiedSum.sortBySumWeight();
	simplifiedSum.fillNormalizedIndicesArray();
	simplifiedSum.simplifyExpression();

	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let sumResult = new MathSum();
	for (let fileRoot of sumResultFileRoots) {
		sumResult.addTo(FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json'))
	};

	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
		console.log('');
	};

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSum, sumResult, opt);

	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

} // end of local group for variables


// process.exit(0);


FH.showSubsubtitle('4 N with tensor in front');

{ // local group for variables
	let showSums = false;
	let simplificationFileRoot = 'SPLIT-u-parallel/gi(nu1nu2)*gi(nu3nu4)+N(nu1)*N(nu2)*N(nu3)*N(nu4)-tensor';
	let sumResultFileRoots = [
		'SPLIT-u-parallel/g(nu1nu2)*g(nu3nu4)+N(nu1)*N(nu2)*N(nu3)*N(nu4)',
		'SPLIT-u-parallel/g(nu1nu3)*g(nu2nu4)+N(nu1)*N(nu2)*N(nu3)*N(nu4)',
		'SPLIT-u-parallel/g(nu1nu4)*g(nu2nu3)+N(nu1)*N(nu2)*N(nu3)*N(nu4)'
	];

	let simplifiedSum = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationFileRoot + '.json');
	simplifiedSum.fillNormalizedIndicesArray();

	console.log(simplificationFileRoot);


	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let latex = simplifiedSum.getLaTeXEquation(OPT.LTX, 'R = ');
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationFileRoot + '.tex', latex);

	// substitute tensor with 4 indices
	simplifiedSum.substituteItemAndDerivativesWithData(SUBST.sumgg);
	simplifiedSum.collectCommutativeElementsInPreElement();
	simplifiedSum.sortPreElementsByProductWeight();
	simplifiedSum.sortBySumWeight();
	simplifiedSum.fillNormalizedIndicesArray();
	simplifiedSum.simplifyExpression();

	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let sumResult = new MathSum();
	for (let fileRoot of sumResultFileRoots) {
		sumResult.addTo(FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json'))
	};

	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
		console.log('');
	};

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSum, sumResult, opt);
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

} // end of local group for variables


// process.exit(0);



FH.showSubsubtitle('propagation of q in N N terms');

{ // local group for variables
	let showSums = false;
	let simplificationFileRoot = 'SPLIT-u-parallel/g(nu1nu2)+N(nu1)*N(nu2)-propagate-q';
	let sumResultFileRoots = [
		'SPLIT-u-parallel/g(nu1nu2)+q*N(nu1)*N(nu2)',
		'SPLIT-u-parallel/g(nu1nu2)+N(nu1)*q*N(nu2)',
		'SPLIT-u-parallel/g(nu1nu2)+N(nu1)*N(nu2)*q'
	];

	let simplifiedSum = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + simplificationFileRoot + '.json');
	simplifiedSum.fillNormalizedIndicesArray();

	console.log(simplificationFileRoot);


	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};


	let obj = FC.propagateItemInSum(PROTO.q, simplifiedSum, OPT.LTX, 'R = ');
	let latex = obj.ltx;
	// console.log('latex = ' + latex);
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + simplificationFileRoot + '.tex', latex);

	simplifiedSum = obj.sum;
	simplifiedSum.collectCommutativeElementsInPreElement();
	simplifiedSum.sortPreElementsByProductWeight();
	simplifiedSum.sortBySumWeight();
	simplifiedSum.fillNormalizedIndicesArray();
	simplifiedSum.simplifyExpression();

	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
		console.log('');
	};

	let sumResult = new MathSum();
	for (let fileRoot of sumResultFileRoots) {
		sumResult.addTo(FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json'))
	};

	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
		console.log('');
	};

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSum, sumResult, opt);

	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

} // end of local group for variables


// process.exit(0);

FH.showSubsubtitle('batch of terms with hnabla F');
// substitute hnabla on F with sum of nabla3 on 1
// test against 3 contributions

{ // local group for variables
	let simplificationFileRoot = 'SPLIT-u-parallel/gi(nu1nu2)+N(nu3)+hnabla(nu1)F(nu2nu3)';
	let sumResultFileRoots = [
		'SPLIT-u-parallel/gi(nu1nu2)+N(nu3)',
		'SPLIT-u-parallel/gi(nu1nu3)+N(nu2)',
		'SPLIT-u-parallel/gi(nu2nu3)+N(nu1)'
	];
	compareUsingNablaFSubstitution(simplificationFileRoot, sumResultFileRoots);
} // end of local group for variables




FH.showSubtitle('TESTS OF GENERIC SIMPLIFIED EXPRESSIONS');

for (let fileRoot of arraySimplifiedFileRoots) {

	// FH.showSubtitle(fileRoot);
	console.log(fileRoot);

	// FH.showSubsubtitle('Loading simplified expression');

	let simplifiedSum = FC.parseSumFromFile(REL_PATH + 'SIMPLIFICATIONS/' + fileRoot + '.json');
	simplifiedSum.fillNormalizedIndicesArray();

	if (showSums) {
		simplifiedSum.showNormalized(OPT.SHOW, 'Simplified Sum = ');
	};


	// FH.showSubsubtitle('Save LaTeX of simplified expression');

	let latex = simplifiedSum.getLaTeXEquation(OPT.LTX, 'R = ');
	FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/' + fileRoot + '.tex', latex);
	inputLtx += '\\input{SIMPLIFICATIONS/' + fileRoot + '.tex' + '}\n\n';


	// FH.showSubsubtitle('Loading computed expression');

	let sumResult = FC.restoreFromFile(REL_PATH + 'RESULTS/' + fileRoot + '.json');
	// substitute cumulative derivations on 'u' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAu);
	// substitute cumulative derivations on 'q' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAq);
	// substitute cumulative derivations on 'N' with "powered" derivations
	sumResult.substituteItemAndDerivativesWithData(SUBST.HNABLAN);
	sumResult.collectCommutativeElementsInPreElement();
	sumResult.sortPreElementsByProductWeight();
	sumResult.sortBySumWeight();
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();

	if (showSums) {
		sumResult.showNormalized(OPT.SHOW, 'Computed Sum = ');
	};

	// FH.showSubsubtitle('Computing the difference');

	let opt = {
		useOperatorRules: true,
		u: PROTO.u
	};

	sumResult = FC.removeSimplifiedSumFrom(simplifiedSum, sumResult, opt)
	sumResult.fillNormalizedIndicesArray();
	sumResult.simplifyExpression();
	sumResult.showNormalized(OPT.SHOW, 'Difference = ');

	console.log('');
};

// save file with latex inputs
FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SIMPLIFICATIONS/input-split-simplifications-u-parallel.tex', inputLtx);



FH.showTitle('END: ' + path.basename(__filename));
