/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

let startTime = Date.now();

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Rational,
				Integer
			} = require(REL_PATH + 'MATH-OBJECTS/MathRational');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');
const {PROTO_NCT, SUBST_NCT} = require(REL_PATH + 'INIT-FILES/initialization-data-NCT');

const {FC} = require(REL_PATH + 'FUNCTIONS-FILES/functions-computation');
const {FR} = require(REL_PATH + 'FUNCTIONS-FILES/functions-rules');
const {FR_NCT} = require(REL_PATH + 'FUNCTIONS-FILES/functions-rules-NCT');


let selection = null;
let selectionSum = null;
let partialStartTime = null;
let partialEndTime = null;
let durationTime = null;
let mathString = null;
let somethingFound = true;
let workingSum = null;
let spectralMathematicaTxt = null;

FH.showTitle('BEGINNING: ' + path.basename(__filename));

OPT.SPEC.applyDimensionValue = true;
OPT.SPEC.dimension = 2;
// shift to add to 'k' (number of arguments)
// to construct the '+' after the letter 'half dimension'
// here: k=1 => m,, k=2 => m+1, k=3 => m+3, etc, so shift = -1
OPT.SPEC.shiftHalfDimension = -1;

let spectralData = RULE_DATA.spectralData;
let opt_spec = OPT.SPEC;


FH.showSubtitle('INITIALIZATION');

FH.showSubsubtitle('Create G tensor');
// tensors G symmetric
let pmax = 3;
const dataGTensor = FC.prepareGtensor(PROTO.g, pmax, PROTO.dim);
const {arrG, arrGIndices, arrRedPerm, arrGPrefactors} = dataGTensor;




FH.showSubsubtitle('Term - P');
let T1 = MathSum.newFrom(PROTO.P)
				.multiplyByInteger(-1)
				.normalizeToArguments();


FH.showSubsubtitle('Term K otimes K = - iK otimes iK');
let T2 = MathSum.newFrom(PROTO.iK)
				.tensorOnRightWith(PROTO.iK)
				.multiplyByInteger(-1)
				.normalizeToArguments();


FH.showSubsubtitle('The original expression to consider');
let T = new MathSum()
				.addTo(T1)
				.addTo(T2);

T.sortBySumWeight()
	.fillNormalizedIndicesArray();

T.showNormalized(OPT.SHOW, 'Original Sum = ');


// extra rules to apply
let arrayOfRules = [
	FR_NCT.contractionsMetricWithTwofTensors,
	FR_NCT.ruleContractionsfTensorWithDoubleDerivation
];



FH.showSubtitle('COMPUTATION (STEP 1): PROPAGATION NABLA');

let sumResult = FC.computePropagationInArguments(T, PROTO.H);

sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')



FH.showSubtitle('STATISTICS');

let c1 = 0;
let cN1 = 0;
let cNN1 = 0;
let sum1 = new MathSum();
let sumN1 = new MathSum();
let sumNN1 = new MathSum();
let regExpr = null;

regExpr = FH.constructRegExpFromSelection(SELECTION.PENabla0,
	 											GLOBAL.sumWeightData,
												GLOBAL.sumWeightOperatorData);
let fullRE0 = new RegExp(regExpr.postStr + '-' + regExpr.mainStr, "g");

regExpr = FH.constructRegExpFromSelection(SELECTION.PENabla1,
	 											GLOBAL.sumWeightData,
												GLOBAL.sumWeightOperatorData);
let fullRE1 = new RegExp(regExpr.postStr + '-' + regExpr.mainStr, "g");

regExpr = FH.constructRegExpFromSelection(SELECTION.PENabla2,
	 											GLOBAL.sumWeightData,
												GLOBAL.sumWeightOperatorData);
let fullRE2 = new RegExp(regExpr.postStr + '-' + regExpr.mainStr, "g");

let outerIndicesStat = [];

for (let i=0; i<sumResult.list.length; i++) {
	let n = sumResult.list[i].getOuterIndices().length;
	let p = n/2;
	if (outerIndicesStat[p] === undefined) {
		outerIndicesStat[p] = 1;
	} else {
		outerIndicesStat[p] = outerIndicesStat[p] + 1;
	};

	if (sumResult.list[i].sumStringWeight.match(fullRE0)) {
		c1++;
		sum1.addTo(sumResult.list[i], false);
	} else if (sumResult.list[i].sumStringWeight.match(fullRE1)) {
		cN1++;
		sumN1.addTo(sumResult.list[i], false);
	} else if (sumResult.list[i].sumStringWeight.match(fullRE2)) {
		cNN1++;
		sumNN1.addTo(sumResult.list[i], false);
	};
};
console.log("Termes 1: " + c1);
console.log("Termes ∇1: " + cN1);
console.log("Termes ∇∇1: " + cNN1);

console.log('');
for (let p=0; p<outerIndicesStat.length; p++) {
	console.log('outer indices p = ' + p + ' => # = ' + outerIndicesStat[p]);
};


FH.showSubtitle('SUBSTITUTIONS FOR P1');

FH.showSubsubtitle('Load substitutions');

// get computed substitutions for NCT
let substitutionH = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.nabla_H_to_delta);
let substitutionL = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.nabla_L_to_delta);
let substitutionq = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.nabla_q_to_delta);


// only sum1 is used since no gauge curvature
// make a copy since we will substitute for P1
sumResult = sum1.makeCopy();


FH.showSubsubtitle('Substitutions for H to g times u');

sumResult.substituteItemAndDerivativesWithData(SUBST_NCT.H_to_gi_u);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');


FH.showSubsubtitle('Substitutions for H');

sumResult.substituteItemAndDerivativesWithData(substitutionH);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');



FH.showSubsubtitle('Substitutions for L');

sumResult.substituteItemAndDerivativesWithData(substitutionL);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');


FH.showSubsubtitle('Substitutions for q');

sumResult.substituteItemAndDerivativesWithData(substitutionq);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');



FH.showSubsubtitle('Contractions with G tensor');

sumResult = FC.contractWithGTensor(sumResult, dataGTensor, PROTO.gi.getID());
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')



FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	// TODO: use FR functions, with FR.RULES_ARRAY_1 and FR.RULES_ARRAY_2, or FR.RULES_ARRAY_1_2
	// FC.applyRulesOnTerm(term, RULE_DATA);
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')



FH.showSubsubtitle('Save result');

// save result
FH.saveItemToFile(sumResult, REL_PATH + FILES.JSON.R2_NCT_P1_ns);

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R_2(P_1) = ');
// // console.log(latexN);
FH.saveToFileWithSureFolder(REL_PATH + FILES.TEX.R2_NCT_P1_ns, latex);




FH.showSubsubtitle('Apply operator rules');

somethingFound = true;
workingSum = sumResult;
// let loopC = 0;
while(somethingFound) {
	// loopC++;
	// console.log('loopC: ' + loopC);
	// let {flag, sum}
	// 				= FC.applyOperatorSimplificationTopDown(workingSum, PROTO.u);
	let {flag, sum}
					= FC.applyOperatorSimplificationIgnoreDegree(workingSum, PROTO.u);
	workingSum = sum;
	somethingFound = flag;
};

sumResult = workingSum;

sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
sumResult.showNormalized(OPT.SHOW, 'R = ');
console.log('');
console.log('The expression contains ' + sumResult.getListLength() + ' terms')


// save result
FH.saveItemToFile(sumResult, REL_PATH + FILES.JSON.R2_NCT_P1);

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R_2 = ');
// // console.log(latexN);
FH.saveToFileWithSureFolder(REL_PATH + FILES.TEX.R2_NCT_P1, latex);



FH.showSubsubtitle('Split according to patterns');


FC.saveSplitSum(sumResult, PROTO.u, '-NCT-R2-P1', REL_PATH);


FH.showSubsubtitle('Save spectral function to Mathematica');

spectralMathematicaTxt = FC.getSpectralFunction(sumResult, spectralData, opt_spec)
FH.saveToFileWithSureFolder(REL_PATH + 'MATHEMATICA/SPECTRAL-FUNCTIONS/R2-NCT-P1.txt', spectralMathematicaTxt);





FH.showSubtitle('SUBSTITUTIONS FOR P2');


FH.showSubsubtitle('Load substitutions');

// get computed substitutions for NCT
substitutionH = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.nabla_H_to_delta);
substitutionL = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.nabla_L_P2_to_delta);
substitutionq = SUBST_NCT.qTo0;


// only sum1 is used since no gauge curvature
// make a copy since we will substitute for P2
sumResult = sum1.makeCopy();

FH.showSubsubtitle('Substitutions for H to g times u');

sumResult.substituteItemAndDerivativesWithData(SUBST_NCT.H_to_gi_u);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');



FH.showSubsubtitle('Substitutions for H');

sumResult.substituteItemAndDerivativesWithData(substitutionH);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');



FH.showSubsubtitle('Substitutions for L');

sumResult.substituteItemAndDerivativesWithData(substitutionL);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');


FH.showSubsubtitle('Substitutions for q');

sumResult.substituteItemAndDerivativesWithData(substitutionq);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');



FH.showSubsubtitle('Contractions with G tensor');

sumResult = FC.contractWithGTensor(sumResult, dataGTensor, PROTO.gi.getID());
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')


FH.showSubsubtitle('Substitutions for hi');

sumResult.substituteItemAndDerivativesWithData(SUBST_NCT.hi_to_gi_fi);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');



FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	// TODO: use FR functions, with FR.RULES_ARRAY_1 and FR.RULES_ARRAY_2, or FR.RULES_ARRAY_1_2
	// FC.applyRulesOnTerm(term, RULE_DATA);
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')


FH.showSubsubtitle('Apply additional rules');


for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	FR.applyArrayOfRulesOnTerm(arrayOfRules, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')





FH.showSubsubtitle('Save result');

// save result
FH.saveItemToFile(sumResult, REL_PATH + FILES.JSON.R2_NCT_P2_ns);

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R_2(P_2) = ');
// // console.log(latexN);
FH.saveToFileWithSureFolder(REL_PATH + FILES.TEX.R2_NCT_P2_ns, latex);



FH.showSubsubtitle('Apply operator rules');

somethingFound = true;
workingSum = sumResult;
// let loopC = 0;
while(somethingFound) {
	// loopC++;
	// console.log('loopC: ' + loopC);
	// let {flag, sum}
	// 				= FC.applyOperatorSimplificationTopDown(workingSum, PROTO.u);
	let {flag, sum}
					= FC.applyOperatorSimplificationIgnoreDegree(workingSum, PROTO.u);
	workingSum = sum;
	somethingFound = flag;
};

sumResult = workingSum;

sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
sumResult.showNormalized(OPT.SHOW, 'R = ');
console.log('');
console.log('The expression contains ' + sumResult.getListLength() + ' terms')


// save result
FH.saveItemToFile(sumResult, REL_PATH + FILES.JSON.R2_NCT_P2);

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R_2 = ');
// // console.log(latexN);
FH.saveToFileWithSureFolder(REL_PATH + FILES.TEX.R2_NCT_P2, latex);


FH.showSubsubtitle('Save spectral function to Mathematica');

spectralMathematicaTxt = FC.getSpectralFunction(sumResult, spectralData, opt_spec)
FH.saveToFileWithSureFolder(REL_PATH + 'MATHEMATICA/SPECTRAL-FUNCTIONS/R2-NCT-P2.txt', spectralMathematicaTxt);





FH.showSubtitle('ADD BOTH SUMS');

FH.showSubsubtitle('Restore from file');

let sumResultP1 = FC.restoreFromFile(REL_PATH + FILES.JSON.R2_NCT_P1_ns);
sumResultP1.fillNormalizedIndicesArray();

let sumResultP2 = FC.restoreFromFile(REL_PATH + FILES.JSON.R2_NCT_P2_ns);
sumResultP2.fillNormalizedIndicesArray();

sumResult = new MathSum();
sumResult.addTo(sumResultP1);
sumResult.addTo(sumResultP2);

sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();

console.log('The expression contains ' + sumResult.getListLength() + ' terms')




FH.showSubsubtitle('Substitutions for hi');
//
// let latexSubstitution = FC.showSubstitution(SUBST_NCT.hi_to_gi_fi);
//
// let test = MathProduct.newFrom(PROTO_NCT.fi);
// test.showConsole(OPT.SHOW, 'test = ');
//
sumResult.substituteItemAndDerivativesWithData(SUBST_NCT.hi_to_gi_fi);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');





FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	// TODO: use FR functions, with FR.RULES_ARRAY_1 and FR.RULES_ARRAY_2, or FR.RULES_ARRAY_1_2
	// FC.applyRulesOnTerm(term, RULE_DATA);
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')



FH.showSubsubtitle('Apply additional rules');


for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	FR.applyArrayOfRulesOnTerm(arrayOfRules, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')




FH.showSubsubtitle('Apply operator rules');

somethingFound = true;
workingSum = sumResult;
// let loopC = 0;
while(somethingFound) {
	// loopC++;
	// console.log('loopC: ' + loopC);
	// let {flag, sum}
	// 				= FC.applyOperatorSimplificationTopDown(workingSum, PROTO.u);
	let {flag, sum}
					= FC.applyOperatorSimplificationIgnoreDegree(workingSum, PROTO.u);
	workingSum = sum;
	somethingFound = flag;
};

sumResult = workingSum;

sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
sumResult.showNormalized(OPT.SHOW, 'R = ');
console.log('');
console.log('The expression contains ' + sumResult.getListLength() + ' terms')


// save result
FH.saveItemToFile(sumResult, REL_PATH + FILES.JSON.R2_NCT);

latex = sumResult.getLaTeXEquation(OPT.LTX, 'R_2 = ');
// // console.log(latexN);
FH.saveToFileWithSureFolder(REL_PATH + FILES.TEX.R2_NCT, latex);



FH.showSubsubtitle('Split according to patterns');


FC.saveSplitSum(sumResult, PROTO.u, '-NCT-R2', REL_PATH);



FH.showSubsubtitle('Save spectral function to Mathematica');

spectralMathematicaTxt = FC.getSpectralFunction(sumResult, spectralData, opt_spec)
FH.saveToFileWithSureFolder(REL_PATH + 'MATHEMATICA/SPECTRAL-FUNCTIONS/R2-NCT.txt', spectralMathematicaTxt);


FH.showSubsubtitle('Save latex for d=2');

let d = 2;
let newSum = new MathSum();
for (let term of sumResult.list) {
	// work with copy to change part of it
	let termCopy = term.makeCopy();
	// evaluate at given dimension (construct option for that)
	let preElement = termCopy.preElement;
	// preElement.showConsole(OPT.SHOW, 'pre El. = ');
	preElement.prefactor = preElement.prefactor.evaluateAtInteger(d);
	termCopy.updateExpression();
	if (!termCopy.isZero()) {
		newSum.addTo(termCopy, false);
	};
};

newSum.updateExpression();
newSum.sortBySumWeight();
newSum.fillNormalizedIndicesArray();
newSum.simplifyExpression();
newSum.showNormalized(OPT.SHOW, 'R = ');
console.log('');
console.log('The expression contains ' + newSum.getListLength() + ' terms')

latex = newSum.getLaTeXEquation(OPT.LTX, 'R_2(d=2) = ');
// // console.log(latexN);
FH.saveToFileWithSureFolder(REL_PATH + FILES.TEX.R2_NCT_d2, latex);


let endTime = Date.now();
durationTime = endTime - startTime;
console.log('Duration (ms): ' + durationTime);


FH.showTitle('END: ' + path.basename(__filename));
