/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

let startTime = Date.now();

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');
const {PROTO_NCT, SUBST_NCT} = require(REL_PATH + 'INIT-FILES/initialization-data-NCT');

const {FC} = require(REL_PATH + 'FUNCTIONS-FILES/functions-computation');
const {FR} = require(REL_PATH + 'FUNCTIONS-FILES/functions-rules');

let selection = null;
let selectionSum = null;
let partialStartTime = null;
let partialEndTime = null;
let durationTime = null;

FH.showTitle('BEGINNING: ' + path.basename(__filename));


// { // local variables
// 	let sumResult = FC.restoreFromFile(REL_PATH + FILES.JSON.R4_ns[0]);
// 	sumResult.fillNormalizedIndicesArray();
// 	sumResult.showNormalized(OPT.SHOW, 'Result = ');
//
// } // end of local variables
//
// process.exit(0);




FH.showSubtitle('INITIALIZATION');

FH.showSubsubtitle('Create G tensor');
// tensors G symmetric
let pmax = 6;
const dataGTensor = FC.prepareGtensor(PROTO.g, pmax, PROTO.dim);
const {arrG, arrGIndices, arrRedPerm, arrGPrefactors} = dataGTensor;

FH.showSubsubtitle('Load substitutions');

// get computed substitutions for NCT
let substitutionH = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.nabla_H_to_delta);
let substitutionL = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.nabla_L_to_delta);
let substitutionq = FC.getSubstitutionsFromFile(REL_PATH + FILES.SUBST.nabla_q_to_delta);


FH.showSubsubtitle('Restore from file');

partialStartTime = Date.now();

let sumResult = FC.restoreFromFile(REL_PATH + FILES.JSON.R4_split[0]);
sumResult.fillNormalizedIndicesArray();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');

partialEndTime = Date.now();
durationTime = partialEndTime - partialStartTime;
console.log('Duration for restauration (ms): ' + durationTime);



FH.showSubtitle('SUBSTITUTIONS');


FH.showSubsubtitle('Substitutions for H to g times u');

sumResult.substituteItemAndDerivativesWithData(SUBST_NCT.H_to_gi_u);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');




FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	// TODO: use FR functions, with FR.RULES_ARRAY_1 and FR.RULES_ARRAY_2, or FR.RULES_ARRAY_1_2
	// FC.applyRulesOnTerm(term, RULE_DATA);
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')


FH.showSubsubtitle('Substitutions for H up to 1');

substitutionH.maxLevel = 1;
sumResult.substituteItemAndDerivativesWithData(substitutionH);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');



FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	// TODO: use FR functions, with FR.RULES_ARRAY_1 and FR.RULES_ARRAY_2, or FR.RULES_ARRAY_1_2
	// FC.applyRulesOnTerm(term, RULE_DATA);
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')




FH.showSubsubtitle('Substitutions for H up to 2');

substitutionH.maxLevel = 2;
sumResult.substituteItemAndDerivativesWithData(substitutionH);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');



FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	// TODO: use FR functions, with FR.RULES_ARRAY_1 and FR.RULES_ARRAY_2, or FR.RULES_ARRAY_1_2
	// FC.applyRulesOnTerm(term, RULE_DATA);
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')



process.exit(0);



FH.showSubsubtitle('Substitutions for L');

sumResult.substituteItemAndDerivativesWithData(substitutionL);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');



FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	// TODO: use FR functions, with FR.RULES_ARRAY_1 and FR.RULES_ARRAY_2, or FR.RULES_ARRAY_1_2
	// FC.applyRulesOnTerm(term, RULE_DATA);
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')






FH.showSubsubtitle('Substitutions for q');

sumResult.substituteItemAndDerivativesWithData(substitutionq);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');


FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	// TODO: use FR functions, with FR.RULES_ARRAY_1 and FR.RULES_ARRAY_2, or FR.RULES_ARRAY_1_2
	// FC.applyRulesOnTerm(term, RULE_DATA);
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')


FH.showSubsubtitle('Contractions with G tensor');

sumResult = FC.contractWithGTensor(sumResult, dataGTensor, PROTO.gi.getID());
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')



FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	// TODO: use FR functions, with FR.RULES_ARRAY_1 and FR.RULES_ARRAY_2, or FR.RULES_ARRAY_1_2
	// FC.applyRulesOnTerm(term, RULE_DATA);
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')




FH.showSubsubtitle('Substitutions for H up to 4');

substitutionH.maxLevel = 4;
sumResult.substituteItemAndDerivativesWithData(substitutionH);
sumResult.collectCommutativeElementsInPreElement();
sumResult.sortPreElementsByProductWeight();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms');


FH.showSubsubtitle('Apply rules');

for (let i=0; i<sumResult.list.length; i++) {
	FH.printProgressMessage('term number ' + i + '/' + sumResult.list.length);
	let term = sumResult.list[i];
	// TODO: use FR functions, with FR.RULES_ARRAY_1 and FR.RULES_ARRAY_2, or FR.RULES_ARRAY_1_2
	// FC.applyRulesOnTerm(term, RULE_DATA);
	FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
};
FH.printProgressMessage('');

// necessary to clean zero terms
sumResult.updateExpression();
sumResult.sortBySumWeight();
sumResult.fillNormalizedIndicesArray();
sumResult.simplifyExpression();
// sumResult.showNormalized(OPT.SHOW, 'R = ');

console.log('The expression contains ' + sumResult.getListLength() + ' terms')



sumResult.showNormalized(OPT.SHOW, 'R = ');


FH.showSubtitle('SAVE RESULT');
// save result
FH.saveItemToFile(sumResult, REL_PATH + FILES.JSON.R4_NCT_ns);


FH.showTitle('END: ' + path.basename(__filename));
