/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

let startTime = Date.now();

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');
const {PROTO_NCT, SUBST_NCT} = require(REL_PATH + 'INIT-FILES/initialization-data-NCT');

const {FC} = require(REL_PATH + 'FUNCTIONS-FILES/functions-computation');
const {FS} = require(REL_PATH + 'FUNCTIONS-FILES/functions-substitutions');
const {FS_NCT} = require(REL_PATH + 'FUNCTIONS-FILES/functions-substitutions-NCT');
const {FS_DATA} = require(REL_PATH + 'INIT-FILES/options-substitutions');



FH.showTitle('BEGINNING: ' + path.basename(__filename));



// compute substitution of H for NonCommutative Torus
// in that case:
// all derivative of metric are zero
// no gauge field strength

// PROTO_NCT.H_NCT_0.showConsole(OPT.SHOW, 'H = ');
//
// PROTO_NCT.delta1k.showConsole(OPT.SHOW, '1 = ');
// PROTO_NCT.delta2k.showConsole(OPT.SHOW, '2 = ');
// PROTO_NCT.delta3k.showConsole(OPT.SHOW, '3 = ');
// PROTO_NCT.delta4k.showConsole(OPT.SHOW, '4 = ');


FH.showSubtitle('COMPUTATION OF SUBSTITUTIONS OF H');

PROTO_NCT.H_NCT_0.showConsole(OPT.SHOW, 'H => ');


// from these data, the code should generate all the substitutions in normal coordinates
// data given at level 0 of derivation…
let initialSubstitutionH = {
		optLaTeX: OPT.LTX,
		termElement: CONST.MAIN_ELEMENT,
		filePath: FILES.SUBST.nabla_H_to_delta, // file in which substitution data are saved
		sumFileRoot: './RESULTS/SUBSTITUTIONS-NCT/Substitutions-NCT-H-sum', // file root to save sums
		latexLetter: 'H',
		orgItem: PROTO.H,
		derivationOperator: PROTO.nabla,
		minLevel: 0, // in final substitution data
		maxLevel: 4, // in final substitution data
		newItems: // list all new items
			[
				{
					item: PROTO_NCT.H_NCT_0, // MathProduct
					// all mapIndices should have same "from" in same order
					mapIndices: [
						{from:[0], to:[0,0]},
						{from:[1], to:[0,1]}
					],
				}
			]
	};

let latexToSaveH = FS_NCT.createSubstitutionsDataNCT(initialSubstitutionH, FS_DATA.operatorData, FS_DATA.placeHolderTensorData, [SUBST.giToConstant], REL_PATH);


// save latex of computations
// FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-NCT-H.tex', latexToSaveH);

FH.showSubtitle('COMPUTATION OF SUBSTITUTIONS OF u');

PROTO_NCT.u_NCT_0.showConsole(OPT.SHOW, 'u => ');


// from these data, the code should generate all the substitutions in normal coordinates
// data given at level 0 of derivation…
let initialSubstitutionu = {
		optLaTeX: OPT.LTX,
		termElement: CONST.MAIN_ELEMENT,
		filePath: FILES.SUBST.hnabla_u_to_delta, // file in which substitution data are saved
		sumFileRoot: './RESULTS/SUBSTITUTIONS-NCT/Substitutions-NCT-u-sum', // file root to save sums
		latexLetter: 'u',
		orgItem: PROTO.u,
		derivationOperator: PROTO.hnabla,
		minLevel: 0, // in final substitution data
		maxLevel: 4, // in final substitution data
		newItems: // list all new items
			[
				{
					item: PROTO_NCT.u_NCT_0, // MathProduct
					// all mapIndices should have same "from" in same order
					mapIndices: [],
				}
			]
	};

let latexToSaveu = FS_NCT.createSubstitutionsDataNCT(initialSubstitutionu, FS_DATA.operatorData, FS_DATA.placeHolderTensorData, [SUBST.giToConstant], REL_PATH);


// save latex of computations
// FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-NCT-H.tex', latexToSaveH);


FH.showSubtitle('COMPUTATION OF SUBSTITUTIONS OF L (P1 CASE)');


PROTO_NCT.L_NCT_0.showConsole(OPT.SHOW, 'L => ');

let initialSubstitutionL = {
		optLaTeX: OPT.LTX,
		termElement: CONST.MAIN_ELEMENT,
		filePath: FILES.SUBST.nabla_L_to_delta, // file in which substitution data are saved
		sumFileRoot: './RESULTS/SUBSTITUTIONS-NCT/Substitutions-NCT-L-sum', // file root to save sums
		latexLetter: 'L',
		orgItem: PROTO.L,
		derivationOperator: PROTO.nabla,
		minLevel: 0, // in final substitution data
		maxLevel: 3, // in final substitution data
		newItems: // list all new items
			[
				{
					item: PROTO_NCT.L_NCT_0, // MathProduct
					// all mapIndices should have same "from" in same order
					mapIndices: [
						{from:[0], to:[0,0]}
					],
				}
			]
	};

let latexToSaveL = FS_NCT.createSubstitutionsDataNCT(initialSubstitutionL, FS_DATA.operatorData, FS_DATA.placeHolderTensorData, [SUBST.giToConstant], REL_PATH);


// save latex of computations
// FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-NCT-L.tex', latexToSaveL);



FH.showSubtitle('COMPUTATION OF SUBSTITUTIONS OF q (P1 CASE)');


PROTO_NCT.q_NCT_0.showConsole(OPT.SHOW, 'q => ');


let initialSubstitutionq = {
		optLaTeX: OPT.LTX,
		termElement: CONST.MAIN_ELEMENT,
		filePath: FILES.SUBST.nabla_q_to_delta, // file in which substitution data are saved
		sumFileRoot: './RESULTS/SUBSTITUTIONS-NCT/Substitutions-NCT-q-sum', // file root to save sums
		latexLetter: 'q',
		orgItem: PROTO.q,
		derivationOperator: PROTO.nabla,
		minLevel: 0, // in final substitution data
		maxLevel: 2, // in final substitution data
		newItems: // list all new items
			[
				{
					item: PROTO_NCT.q_NCT_0, // MathProduct
					// all mapIndices should have same "from" in same order
					mapIndices: [
					],
				}
			]
	};

let latexToSaveq = FS_NCT.createSubstitutionsDataNCT(initialSubstitutionq, FS_DATA.operatorData, FS_DATA.placeHolderTensorData, [SUBST.giToConstant], REL_PATH);


// save latex of computations
// FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-NCT-q.tex', latexToSaveq);

FH.showSubtitle('COMPUTATION OF SUBSTITUTIONS OF L (P2 CASE)');


PROTO_NCT.L_NCT_0.showConsole(OPT.SHOW, 'L => ');

let initialSubstitutionL_P2 = {
		optLaTeX: OPT.LTX,
		termElement: CONST.MAIN_ELEMENT,
		filePath: FILES.SUBST.nabla_L_P2_to_delta, // file in which substitution data are saved
		sumFileRoot: './RESULTS/SUBSTITUTIONS-NCT/Substitutions-NCT-L-P2-sum', // file root to save sums
		latexLetter: 'L',
		orgItem: PROTO.L,
		derivationOperator: PROTO.nabla,
		minLevel: 0, // in final substitution data
		maxLevel: 3, // in final substitution data
		newItems: // list all new items
			[
				{
					item: PROTO_NCT.L_P2_NCT_0, // MathProduct
					// all mapIndices should have same "from" in same order
					mapIndices: [
						{from:[0], to:[0,0]}
					],
				},
				{
					item: PROTO_NCT.L_P2_NCT_1, // MathProduct
					// all mapIndices should have same "from" in same order
					mapIndices: [
						{from:[0], to:[0,0]}
					],
				}
			]
	};

let latexToSaveL_P2 = FS_NCT.createSubstitutionsDataNCT(initialSubstitutionL_P2, FS_DATA.operatorData, FS_DATA.placeHolderTensorData, [SUBST.giToConstant, SUBST_NCT.hiToConstant], REL_PATH);


// save latex of computations
// FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-NCT-L.tex', latexToSaveL);



// TODO: create substitutions for P2 operator
// use
// SUBST_NCT.hiToConstant


FH.showSubtitle('EXPORT SUBSTITUTIONS TO LATEX');

let latexToSave = '';


let filePathH = REL_PATH + FILES.SUBST.nabla_H_to_delta;
let substitutionH = FC.getSubstitutionsFromFile(filePathH);
latexToSave = '';

let TH0 = new MathElement();
TH0.setID('TH0')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceDown()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TH0H = MathProduct.newFrom(TH0)
	.multiplyOnRightWith(PROTO.H)
	.setContraction([0,0], [1,0])
	.setContraction([0,1], [1,1]);
// TH0H = MathArgument.newFrom(TH0H);
let TH0HS = MathSum.newFrom(TH0H);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TH0HS, substitutionH);

let TH1 = new MathElement();
TH1.setID('TH1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceDown()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TH1H = MathProduct.newFrom(TH1)
	.multiplyOnRightWith(PROTO.H1)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-2,1]);
// TH1H.showConsole(OPT.SHOW, 'TH1H = ');
// TH1H = MathArgument.newFrom(TH1H);
let TH1HS = MathSum.newFrom(TH1H);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TH1HS, substitutionH);

let TH2 = new MathElement();
TH2.setID('TH2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TH2H = MathProduct.newFrom(TH2)
	.multiplyOnRightWith(PROTO.H2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-3,1]);
// TH2H.showConsole(OPT.SHOW, 'TH2H = ');
// TH2H = MathArgument.newFrom(TH2H);
let TH2HS = MathSum.newFrom(TH2H);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TH2HS, substitutionH);

let TH3 = new MathElement();
TH3.setID('TH3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TH3H = MathProduct.newFrom(TH3)
	.multiplyOnRightWith(PROTO.H3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0])
	.setContraction([0,4], [1,-4,1]);
let TH3HS = MathSum.newFrom(TH3H);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TH3HS, substitutionH);

let TH4 = new MathElement();
TH4.setID('TH4')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TH4H = MathProduct.newFrom(TH4)
	.multiplyOnRightWith(PROTO.H4)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0])
	.setContraction([0,4], [1,-5,0])
	.setContraction([0,5], [1,-5,1]);
let TH4HS = MathSum.newFrom(TH4H);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TH4HS, substitutionH);


FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-NCT-H.tex', latexToSave);



let filePathu = REL_PATH + FILES.SUBST.hnabla_u_to_delta;
let substitutionu = FC.getSubstitutionsFromFile(filePathu);
latexToSave = '';

// let Tu0 = new MathElement();
// Tu0.setID('Tu0')
// 	.setLetter('T')
// 	.setLaTeXLetter('T')
// 	.setProductKindToNonCommutative()
// 	.setSumWeight(1)
// 	.setProductWeight(1);
// let Tu0u = MathProduct.newFrom(Tu0)
// 	.multiplyOnRightWith(PROTO.u);
let Tu0u = MathProduct.newFrom(PROTO.uh0);
let Tu0uS = MathSum.newFrom(Tu0u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tu0uS, substitutionu);

let Tu1 = new MathElement();
Tu1.setID('Tu1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let Tu1u = MathProduct.newFrom(Tu1)
	.multiplyOnRightWith(PROTO.uh1)
	.setContraction([0,0], [1,-1,0]);
let Tu1uS = MathSum.newFrom(Tu1u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tu1uS, substitutionu);

let Tu2 = new MathElement();
Tu2.setID('Tu2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let Tu2u = MathProduct.newFrom(Tu2)
	.multiplyOnRightWith(PROTO.uh2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0]);
let Tu2uS = MathSum.newFrom(Tu2u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tu2uS, substitutionu);

let Tu3 = new MathElement();
Tu3.setID('Tu3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let Tu3u = MathProduct.newFrom(Tu3)
	.multiplyOnRightWith(PROTO.uh3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0]);
let Tu3uS = MathSum.newFrom(Tu3u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tu3uS, substitutionu);

let Tu4 = new MathElement();
Tu4.setID('Tu4')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let Tu4u = MathProduct.newFrom(Tu4)
	.multiplyOnRightWith(PROTO.uh4)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0]);
let Tu4uS = MathSum.newFrom(Tu4u);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tu4uS, substitutionu);


FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-NCT-u.tex', latexToSave);




let filePathL = REL_PATH + FILES.SUBST.nabla_L_to_delta;
let substitutionL = FC.getSubstitutionsFromFile(filePathL);
latexToSave = '';

let TL0 = new MathElement();
TL0.setID('TL0')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TL0L = MathProduct.newFrom(TL0)
	.multiplyOnRightWith(PROTO.L)
	.setContraction([0,0], [1,0]);
// TL0L = MathArgument.newFrom(TL0L);
let TL0LS = MathSum.newFrom(TL0L);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TL0LS, substitutionL);

let TL1 = new MathElement();
TL1.setID('TL1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TL1L = MathProduct.newFrom(TL1)
	.multiplyOnRightWith(PROTO.L1)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0]);
// TL1L = MathArgument.newFrom(TL1L);
let TL1LS = MathSum.newFrom(TL1L);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TL1LS, substitutionL);

let TL2 = new MathElement();
TL2.setID('TL2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TL2L = MathProduct.newFrom(TL2)
	.multiplyOnRightWith(PROTO.L2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0]);
let TL2LS = MathSum.newFrom(TL2L);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TL2LS, substitutionL);


let TL3 = new MathElement();
TL3.setID('TL3')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TL3L = MathProduct.newFrom(TL3)
	.multiplyOnRightWith(PROTO.L3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0]);
let TL3LS = MathSum.newFrom(TL3L);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TL3LS, substitutionL);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-NCT-L.tex', latexToSave);





let filePathq = REL_PATH + FILES.SUBST.nabla_q_to_delta;
let substitutionq = FC.getSubstitutionsFromFile(filePathq);
latexToSave = '';

// let latexSubstitutionq = FC.showSubstitution(substitutionq);
// latexToSave += latexSubstitutionq + CONST.CR;

// let Tq0 = new MathElement();
// Tq0.setID('Tq0')
// 	.setLetter('T')
// 	.setLaTeXLetter('T')
// 	.setProductKindToNonCommutative()
// 	.setSumWeight(1)
// 	.setProductWeight(1);
// let Tq0q = MathProduct.newFrom(Tq0)
// 	.multiplyOnRightWith(PROTO.q);
let Tq0q = MathProduct.newFrom(PROTO.q);
let Tq0qS = MathSum.newFrom(Tq0q);


latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tq0qS, substitutionq);

let Tq1 = new MathElement();
Tq1.setID('Tq1')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let Tq1q = MathProduct.newFrom(Tq1)
	.multiplyOnRightWith(PROTO.q1)
	.setContraction([0,0], [1,-1,0]);
let Tq1qS = MathSum.newFrom(Tq1q);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tq1qS, substitutionq);

let Tq2 = new MathElement();
Tq2.setID('Tq2')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let Tq2q = MathProduct.newFrom(Tq2)
	.multiplyOnRightWith(PROTO.q2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0]);
let Tq2qS = MathSum.newFrom(Tq2q);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, Tq2qS, substitutionq);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-NCT-q.tex', latexToSave);



let filePathL_P2 = REL_PATH + FILES.SUBST.nabla_L_P2_to_delta;
let substitutionL_P2 = FC.getSubstitutionsFromFile(filePathL_P2);
latexToSave = '';

let TLP20 = new MathElement();
TLP20.setID('TLP20')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TLP20L = MathProduct.newFrom(TLP20)
	.multiplyOnRightWith(PROTO.L)
	.setContraction([0,0], [1,0]);
// TLP20L = MathArgument.newFrom(TLP20L);
let TLP20LS = MathSum.newFrom(TLP20L);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TLP20LS, substitutionL_P2);

let TLP21 = new MathElement();
TLP21.setID('TLP21')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TLP21L = MathProduct.newFrom(TLP21)
	.multiplyOnRightWith(PROTO.L1)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0]);
// TLP21L = MathArgument.newFrom(TLP21L);
let TLP21LS = MathSum.newFrom(TLP21L);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TLP21LS, substitutionL_P2);

let TLP22 = new MathElement();
TLP22.setID('TLP22')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TLP22L = MathProduct.newFrom(TLP22)
	.multiplyOnRightWith(PROTO.L2)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0]);
let TLP22LS = MathSum.newFrom(TLP22L);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TLP22LS, substitutionL_P2);


let TLP23 = new MathElement();
TLP23.setID('TLP23')
	.setLetter('T')
	.setLaTeXLetter('T')
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(1)
	.setProductWeight(1);
let TLP23L = MathProduct.newFrom(TLP23)
	.multiplyOnRightWith(PROTO.L3)
	.setContraction([0,0], [1,-1,0])
	.setContraction([0,1], [1,-2,0])
	.setContraction([0,2], [1,-3,0])
	.setContraction([0,3], [1,-4,0]);
let TLP23LS = MathSum.newFrom(TLP23L);

latexToSave += FC.getLaTeXSubstitution(OPT.LTX, TLP23LS, substitutionL_P2);

FH.saveToFileWithSureFolder(REL_PATH + 'LATEX/SUBSTITUTIONS/substitutions-computed-NCT-L-P2.tex', latexToSave);







FH.showTitle('END: ' + path.basename(__filename));
