/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');
const {FI} = require(REL_PATH + 'FUNCTIONS-FILES/functions-input');
const {FR} = require(REL_PATH + 'FUNCTIONS-FILES/functions-rules');

const {
				Rational,
				Integer
			} = require(REL_PATH + 'MATH-OBJECTS/MathRational');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument,
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

// all the functions defined here are properties of the object FC
const FC = {};

// return the MathExpression saved in the given file
FC.restoreFromFile = function(file) {
	let dataJSON = FH.getDataFromFile(file);
	return FC.constructItemFromDataJSON(dataJSON);
};

FC.parseSumFromFile = function(file) {
	let dataJSON = fs.readFileSync(file, 'utf8');
	// console.log('dataJSON: ' + dataJSON);
	let dataArr = JSON.parse(dataJSON);
	// console.log('dataArr: ' + JSON.stringify(dataArr));
	return FI.parseMathSum(dataArr);
};

// return the MathExpression given by the JSON text
FC.constructItemFromDataJSON = function(dataJSON) {
	let data = JSON.parse(dataJSON);
	// save the current global indice ID
	// since it will be unecessarily increased in resconstruction
	let currentGlobalIndiceID = FH.copyData(GLOBAL.currentIndiceID);
	let {maxIndiceID, item} = FC.constructItemFromData(data);
	// set global indice ID to its saved value
	GLOBAL.currentIndiceID = FH.copyData(currentGlobalIndiceID);
	// console.log('(restoreFromFile) maxIndiceID: ' + maxIndiceID);
	FH.setCurrentIndiceIDFromString(maxIndiceID, GLOBAL.currentIndiceID);
	// console.log('(restoreFromFile) GLOBAL.currentIndiceID: ' + JSON.stringify(GLOBAL.currentIndiceID));
	return item;
}


// data: object representation of an item (.toObject())
FC.constructItemFromData = function(data) {
	let maxIndiceID = 'a';
	let item = null;
	let objConstructed = null;
	let currentClass = data.class;
	switch (currentClass){
		case 'MathSum':
		 	item = new MathSum();
			item.contractions = data.cont;
			objConstructed = FC.constructItemFromData(data.pf);
			item.prefactor = objConstructed.item;
			for (let subItemData of data.list) {
				let {maxIndiceID: maxSubIndiceID, item:subItem} = FC.constructItemFromData(subItemData);
				item.list.push(subItem);
				if (maxSubIndiceID > maxIndiceID) {maxIndiceID = maxSubIndiceID;}
			};
			break;

		case 'MathArgument':
			item = new MathArgument();
			item.contractions = data.cont;
			objConstructed = FC.constructItemFromData(data.pf);
			item.prefactor = objConstructed.item;
			for (let subItemData of data.list) {
				let {maxIndiceID: maxSubIndiceID, item:subItem} = FC.constructItemFromData(subItemData);
				item.list.push(subItem);
				if (maxSubIndiceID > maxIndiceID) {maxIndiceID = maxSubIndiceID;}
			};
			// preElement
			objConstructed = FC.constructItemFromData(data.pre);
			item.preElement = objConstructed.item;
			if (objConstructed.maxIndiceID > maxIndiceID) {maxIndiceID = objConstructed.maxIndiceID;}
			// postElement
			objConstructed = FC.constructItemFromData(data.post);
			item.postElement = objConstructed.item;
			if (objConstructed.maxIndiceID > maxIndiceID) {maxIndiceID = objConstructed.maxIndiceID;}
			// console.log('MathArgument maxIndiceID: ' + maxIndiceID);
			break;

		case 'MathTensorProduct':
			item = new MathTensorProduct();
			item.contractions = data.cont;
			objConstructed = FC.constructItemFromData(data.pf);
			item.prefactor = objConstructed.item;
			for (let subItemData of data.list) {
				let {maxIndiceID: maxSubIndiceID, item:subItem} = FC.constructItemFromData(subItemData);
				item.list.push(subItem);
				if (maxSubIndiceID > maxIndiceID) {maxIndiceID = maxSubIndiceID;}
			};
			break;

		case 'MathProduct':
			item = new MathProduct();
			item.contractions = data.cont;
			objConstructed = FC.constructItemFromData(data.pf);
			item.prefactor = objConstructed.item;
			for (let subItemData of data.list) {
				let {maxIndiceID: maxSubIndiceID, item:subItem} = FC.constructItemFromData(subItemData);
				item.list.push(subItem);
				if (maxSubIndiceID > maxIndiceID) {maxIndiceID = maxSubIndiceID;}
			};
			// console.log('MathProduct maxIndiceID: ' + maxIndiceID);
			break;

		case 'MathElement':
			item = new MathElement();
			item.flipID(item.getID(), data.id);
			item.productKind = data.pK;
			item.letter = data.l;
			item.math_letter = data.math;
			item.latex_letter = data.ltx;
			item.sum_weight = data.sW;
			item.product_weight = data.pW;
			item.indices = data.ind;
			item.contractions = data.cont;
			maxIndiceID = data.ind.id;
			// console.log('MathElement maxIndiceID: ' + maxIndiceID);
			break;

		case 'MathOperator':
			item = new MathOperator();
			item.flipID(item.getID(), data.id);
			item.productKind = data.pK;
			item.letter = data.l;
			item.math_letter = data.math;
			item.latex_letter = data.ltx;
			item.sum_weight = data.sW;
			item.product_weight = data.pW;
			item.indices = data.ind;
			item.contractions = data.cont;
			item.behavior = data.be;
			item.status = data.st;
			maxIndiceID = data.ind.id;
			// console.log('MathOperator maxIndiceID (avant): ' + maxIndiceID);
			if (data.st === CONST.OPERATOR_STATUS_APPLIED) {
				// construct argument
				let {maxIndiceID: maxSubIndiceID, item:arg} = FC.constructItemFromData(data.arg);
				// apply to argument
				item.argument = arg;
				// check for maxIndiceID
				if (maxSubIndiceID > maxIndiceID) {maxIndiceID = maxSubIndiceID;}
			};
			// console.log('MathOperator maxIndiceID (après): ' + maxIndiceID);
			break;

		case 'RationalExpression':
			let {item: numerator} = FC.constructItemFromData(data.num);
			item = new RationalExpression(numerator, data.den);
			break;

		case 'Polynomials':
			item = Polynomials.newFromLetterAndArrayOfCoefficient(data.l, data.mon);
			break;

		default:
			throw "Class unknown";
			break;
	}; // switch class

	// item.showConsole(OPT.SHOW, 'item: ');
	item.updateExpression();
	// by recursive application, goes up to first property of data
	return {maxIndiceID: maxIndiceID, item: item};
};


// format substitutions computed and previously saved in a file
// the result can be used directly in 'substituteItemAndDerivativesWithData'
FC.getSubstitutionsFromFile = function(filePath) {
	let savedsubstitutionObj = JSON.parse(FH.getDataFromFile(filePath));
	// console.log(JSON.stringify(savedsubstitutionObj));

	let substitutionObj = {};
	substitutionObj.termElement = savedsubstitutionObj.termElement;
	substitutionObj.minLevel = savedsubstitutionObj.minLevel;
	substitutionObj.maxLevel = savedsubstitutionObj.maxLevel;
	substitutionObj.orgItem = FC.constructItemFromDataJSON(savedsubstitutionObj.orgItemJSON);
	substitutionObj.derivationOperators = [ FC.constructItemFromDataJSON(savedsubstitutionObj.derivationOperatorJSON)];
	substitutionObj.newItemsByDerivatives = [];

	substitutionObj.newItemsByDerivatives = [];
	for (let i=0; i<=substitutionObj.maxLevel; i++) {
	// for (let i=substitutionObj.minLevel; i<=2; i++) {
		let newItemDerivativeData = savedsubstitutionObj.newItemsByDerivatives[i];
		// console.log('');
		// console.log('');
		let substitutionDerivativeData = [];
		for (let newItemData of newItemDerivativeData) {
			// console.log('');
			let substitutionNewItemData = {};
			if (newItemData.itemJSON === '0') {
				// new item is null
				substitutionNewItemData.item = null;
			} else {
				substitutionNewItemData.item = FC.constructItemFromDataJSON(newItemData.itemJSON);
				// substitutionNewItemData.item.showConsole(OPT.SHOW, 'item = ');
				// substitutionNewItemData.item.makeCopy().showConsole(OPT.SHOW, 'item = ');
				substitutionNewItemData.mapIndices = newItemData.mapIndices;
				substitutionNewItemData.permutationsSum = newItemData.permutationsSum;
				// console.log('substitutionNewItemData: ' + JSON.stringify(substitutionNewItemData));
				// collect this new item at this derivation level
				substitutionDerivativeData.push(substitutionNewItemData);
			};
		};
		// collect for this derivation level
		substitutionObj.newItemsByDerivatives.push(substitutionDerivativeData);
	};
	return substitutionObj;
}

// first function used in the computation,
// see the scientific paper to understand its meaning
FC.computePropagationInArguments = function(sumOfArguments, Hproto) {
	let currentSum = null;
	let currentTerm = null;
	let currArrIdx = null;
	let latentList = null;
	let lastLatentArrIdx = null;
	let currentOperator = null;
	let currentMathElement = null;
	let currentNewSum = null;
	let startIdx = null;
	let endIdx = null;
	let retInfo = {};

	currentSum = sumOfArguments;
	// latentList = currentSum.getLatentOperator();
	// let nLatentOperators = currentSum.getLatentOperator().length;
	lastLatentArrIdx = currentSum.getLastLatentOperator();

	// while (latentList.length > 0) {
	let counter = 0;
	while (lastLatentArrIdx !== null) {
		counter++;
		// progress message
		FH.printProgressMessage('Latent operator at '
									+ JSON.stringify(lastLatentArrIdx)
									+ ' (' + counter + ')');

		// arrIdx to the operator
		// start with the last => operator in then on the right in terms
		currArrIdx = lastLatentArrIdx;

		// test coherence…
		if (currArrIdx.length !== 3) {throw 'Error in length of currArrIdx'};
		// currArrIdx[0] => the term in the sum
		// currArrIdx[1] => the factor in the tensor product
		// currArrIdx[2] => the factor in the product (should always be 1 in our computation)

		// get the term to replace in the sum
		currentTerm = currentSum.getItemAtIndex(currArrIdx[0]);

		// get the idx range after the operator inside tensor product
		startIdx = currArrIdx[1] + 1;
		endIdx = currentTerm.list.length;

		 // currentTerm will be replaced by this new sum
		currentNewSum = new MathSum();

		// PROPAGATION ON ARGUMENTS
		// add new terms by applying the operator on tensor factors on its right
		for (let i=startIdx; i<endIdx; i++) {
			// make a copy of the term to modify
			let newCurrentTerm = currentTerm.makeCopy();

			// get reference to the operator to move
			currentOperator = newCurrentTerm.getItemAtIndex(currArrIdx[1]).getItemAtIndex(currArrIdx[2]);

			// remove the operator from the term
			// updateContractions is set to true, but this only affects
			// "newCurrentTerm.getItemAtIndex(currArrIdx[1])" (the MathProduct)
			// and does not propagate to newCurrentTerm (the MathArgument)
			// => newCurrentTerm contains the contractions
			retInfo = {}; // just an empty object for informations…
			newCurrentTerm.getItemAtIndex(currArrIdx[1]).removeItemAtIndex(currArrIdx[2], retInfo, true);

			// get the MathElement on which to apply the operator
			// in our computation, it is always the 0-th factor in the i-th tensor product
			currentMathElement = newCurrentTerm.getItemAtIndex(i).getItemAtIndex(0);

			// apply operator on the MathElement
			currentOperator.applyTo(currentMathElement);

			// replace by the new factor
			retInfo = {};
			newCurrentTerm.getItemAtIndex(i).replaceByItemAtIndex(currentOperator, 0, retInfo, true)

			// add this new term to the sum
			// no need to copy
			currentNewSum.addTo(newCurrentTerm, false);
		};

		// PROPAGATION ON EXP OF H
		// add new terms where the operator is applied to a copy of Hproto
		for (let i=startIdx; i<=endIdx; i++) {
			// make a copy of the term to modify
			let newCurrentTerm = currentTerm.makeCopy();

			// get reference to the operator to move
			currentOperator = newCurrentTerm.getItemAtIndex(currArrIdx[1]).getItemAtIndex(currArrIdx[2]);

			// remove the operator from the term
			retInfo = {}; // just an empty object for informations…
			newCurrentTerm.getItemAtIndex(currArrIdx[1]).removeItemAtIndex(currArrIdx[2], retInfo, true);

			// construct the new factor to insert: currentOperator applied to a copy of Hproto
			currentOperator.applyTo(Hproto.makeCopy());

			// insert the new factor with minus sign
			newCurrentTerm.insertItemAtIndex(currentOperator, i, false).multiplyByInteger(-1);

			// add this new term to the sum
			currentNewSum.addTo(newCurrentTerm, false);
		};

		// ACTION ON POST ELEMENT
		// add the operator applied to the post element
		// no loop here, but create a 'local block'
		{
			// make a copy of the term to modify
			let newCurrentTerm = currentTerm.makeCopy();

			// get reference to the operator to move
			currentOperator = newCurrentTerm.getItemAtIndex(currArrIdx[1]).getItemAtIndex(currArrIdx[2]);

			// remove the operator from the term
			retInfo = {};
			newCurrentTerm.getItemAtIndex(currArrIdx[1]).removeItemAtIndex(currArrIdx[2], retInfo, true);

			// get the MathElement inside newCurrentTerm.postElement => index 0 since only 1 factor
			currentMathElement = newCurrentTerm.postElement.getItemAtIndex(0);

			// apply operator on the MathElement
			currentOperator.applyTo(currentMathElement);

			// replace by the new factor
			retInfo = {};
			newCurrentTerm.postElement.replaceByItemAtIndex(currentOperator, 0, retInfo, true)

			// add this new term to the sum
			currentNewSum.addTo(newCurrentTerm, false);
		};

		// replace currentTerm by currentNewSum
		retInfo = {};
		currentSum.replaceByItemAtIndex(currentNewSum, currArrIdx[0], retInfo, true)

		// latentList = currentSum.getLatentOperator();
		lastLatentArrIdx = currentSum.getLastLatentOperator();
	};
	FH.printProgressMessage('');
	return currentSum;
};

// THIS FUNCTION IS NOT NEEDED ANYMORE WITH SUBSTITUTIONS IN TERMS OF RIEMANNIAN TENSOR
// sumOfArguments: sum obtained from computation
// nablaOperatorID: the ID of the nabla operator to replace
// arrPartialOperators: the array of partial operators, increasing number of indices; 0=>1 indice, 1=>2 indices…
// should contains sufficient elements to cover all the repeated applications of the nabla operator
FC.replaceNablaByPartialOnCommutativeElements = function(sumOfArguments, nablaOperatorID, arrPartialOperators) {
	let arrNablaComm = null;
	let currentTerm = null;
	let currentElement = null;
	let currentOperator = null;
	let replaceItemContainer = null;
	let currentIdx = null;
	let partialOp = null;
	let currArrIdx = null;
	let arrCont = null;
	let strCont = null;
	let newIndiceID = null;
	let arrIndicesID = null;
	let depth = null;
	let nIndices = null;
	let retInfo = {};

	// collect array of idx for commutative elements on which nabla acts
	arrNablaComm = sumOfArguments.getOperatorAppliedToCommutative(nablaOperatorID);

	// perform change on each found element
	for (let i=0; i<arrNablaComm.length; i++) {
		FH.printProgressMessage('item ' + i + '/' + arrNablaComm.length);
		// array to access the element
		currArrIdx = FH.copyData(arrNablaComm[i]);
		// check that the last idx is strictly superior to -1 since inside operator
		if (currArrIdx[currArrIdx.length -1] >= -1) {
			throw 'last index should be negative';
		};

		// get current term in sum
		currentTerm = sumOfArguments.getItemAtIndex(currArrIdx[0]);

		// get contractions array to update
		arrCont = currentTerm.contractions;
		strCont = JSON.stringify(arrCont);

		// get the element on which to apply the new operator
		currentElement = sumOfArguments.getItemAt(currArrIdx);

		// get number of applied operator => determine the new operator to use
		// remove the last idx of currArrIdx
		depth = currArrIdx.splice(-1,1);
		depth = depth[0];

		// get new operator to apply
		// get its new indiceID
		nIndices = -1-depth;
		partialOp = arrPartialOperators[nIndices-1].makeCopy();
		newIndiceID = partialOp.indices.id;

		// get current operator and array of indiceID
		currentOperator = sumOfArguments.getItemAt(currArrIdx);
		arrIndicesID = FH.flattenArray(currentOperator.getIndicesID());
		// remove last item which is associated to the element
		arrIndicesID.splice(-1,1);

		// get the idx at which to replace current operator in replaceItemContainer
		currentIdx = currArrIdx.splice(-1,1);
		currentIdx = currentIdx[0];
		replaceItemContainer = sumOfArguments.getItemAt(currArrIdx);

		// apply new operator to the element
		partialOp.applyTo(currentElement);
		retInfo = {};
		// replace in its container
		replaceItemContainer.replaceByItemAtIndex(partialOp, currentIdx, retInfo, false);
		// update strCont
		for (let j=0; j<nIndices; j++) {
			let re = new RegExp(arrIndicesID[j] + '-0', "g");
			strCont = strCont.replace(re, newIndiceID + '-' + j);
		}
		// update contractions
		currentTerm.contractions = JSON.parse(strCont);
	};
	FH.printProgressMessage('');
};


// fired to prepare data on the G tensor, with pmax depending on the computation
FC.prepareGtensor = function(metricElement, pmax, dim) {
	// tensors G completely symmetric
	// prepare data until p <= pmax
	// arrG[p] => element to multiply preElement with
	// arrGIndices[p] => list of arrIdx to get contraction indiceID
	// arrRedPerm[p] => list of permutations of indices for contractions
	let arrG = [];
	let arrGIndices = [];
	let arrGInitialPermutation = [];
	let arrRedPerm = [];
	let arrGPrefactors = [];

	// nothing for p=0
	for (let p=1; p<pmax+1; p++) {
		arrG[p] = new MathProduct();
		arrGIndices[p] = [];
		arrGInitialPermutation[p] = [];
		for (let i=0; i<p; i++) {
			arrG[p].multiplyOnRightWith(metricElement);
			arrGIndices[p].push([i,0]);
			arrGIndices[p].push([i,1]);
		};
		for (let i=0; i<2*p; i++) {
			arrGInitialPermutation[p].push(i);
		};
		arrRedPerm[p] = FH.reducedPermutations(arrGInitialPermutation[p]);
		arrG[p].multiplyByRational( 1, Math.pow(2, 2*p)*FH.factorial(p) );
		arrG[p].multiplyByRational( FH.factorial(2*p), arrRedPerm[p].length );
		let num = Polynomials.newFromLetterAndArrayOfCoefficient(dim.numerator.letter, [2*p-2, 1]);
		arrGPrefactors[p] = RationalExpression.newFromPolynomials(num);
		arrGPrefactors[p].multiplyByRational(1,2);
	};
	return {
					arrG: arrG,
					arrGIndices: arrGIndices,
					arrRedPerm: arrRedPerm,
					arrGPrefactors: arrGPrefactors
				};
};


// contract all terms in sumToContract with the tensor G
// multiply by the tensor and perform contractions of 'free' indices in terms
// return a new MathSum, don't modify the original MathSum
FC.contractWithGTensor = function(sumToContract, dataGTensor, giID) {
	let contractedSum = new MathSum();
	let ruleData = {};
	ruleData.applyRules = false;

	let c = 0;
	for (let term of sumToContract.list) {
		c++;
		FC.contractTermWithGTensor(term, contractedSum, dataGTensor, giID, ruleData, 'term ' + c);
	}; // loop on term

	FH.printProgressMessage('');
	return contractedSum;
};

// contract all terms in sumToContract with the tensor G
// multiply by the tensor and perform contractions of 'free' indices in terms
// apply simplification rules on terms immediately
// return a new MathSum which is simplified
// don't modify the original MathSum
FC.contractWithGTensorAndApplyRules = function(sumToContract, dataGTensor, giID, ruleData) {
	let contractedSum = new MathSum();
	ruleData.applyRules = true;

	let c = 0;
	let cS = 0;
	let msg = '';
	for (let term of sumToContract.list) {
		c++;
		cS = contractedSum.list.length;
		msg = 'contr. sum: ' + cS + '/ term ' + c;
		FC.contractTermWithGTensor(term, contractedSum, dataGTensor, giID, ruleData, msg);
		contractedSum.updateExpression();
		contractedSum.fillNormalizedIndicesArray();
		contractedSum.simplifyExpression();
	}; // loop on term

	FH.printProgressMessage('');
	return contractedSum;
};

// Main part of the computation
// contract a term with the G tensor at "free" indices (see paper)
// if an inverse metric fully contract with the G tensor
// there is a recursive formula: this makes appear polynomials in dim
// TODO: prevent 'term' to be modified by a making copy
FC.contractTermWithGTensor = function(term, contractedSum, dataGTensor, giID, ruleData, msg) {
	// array of free indices to contract with G
	let outerIndicesArrIdx = term.getOuterIndices();
	let n = outerIndicesArrIdx.length;
	// use to determine which tensor G to take
	// data at level p must have been prepared before
	let p = n/2;
	FH.printProgressMessage(msg + ' / p = ' + p);
	// console.log('');
	// term.showConsole(OPT.SHOW, 'before while loop term: ');
	// console.log('outerIndicesArrIdx: ' + JSON.stringify(outerIndicesArrIdx));

	// fist step: use contraction of inverse metric with tensor G
	// remove all inverse metrics (giID) with outer indices using rule p -> p-1
	let somethingFound = true;
	while(somethingFound && p>0) {
		somethingFound = false;
		// update normalized array of indices
		term.fillNormalizedIndicesArray();
		// console.log('');
		// term.showNormalized(OPT.SHOW, 'while loop term: ');
		// term.showConsole(OPT.SHOW, 'while loop term: ');

		let preElement = term.preElement;
		for (let i=0; i<preElement.list.length; i++) {
			if (preElement.list[i].getID() === giID) {
				let giFactor = preElement.list[i];
				// check if outer indices
				let id = giFactor.indices.id;
				if (giFactor.normalizedIndices[id+'-0'].kind === CONST.INDICE_OUTER
							&& giFactor.normalizedIndices[id+'-1'].kind === CONST.INDICE_OUTER) {
					// console.log('');
					// console.log('double outer indices in gi')
					somethingFound = true;
					// multiply prefactor by arrGPrefactors[p]
					// dataGTensor.arrGPrefactors[p].showConsole(OPT.SHOW, 'dataGTensor.arrGPrefactors[p]: ');
					preElement.prefactor = preElement.prefactor.multiply(dataGTensor.arrGPrefactors[p]);
					// remove giFactor
					preElement.list.splice(i,1);
					// decrease p for while loop
					p = p - 1;
					FH.printProgressMessage(msg + ' / p = ' + p);
					// return to while loop
					break;
				}; // test both indices are outer
			}; // if giID
		}; // for termPreElement.list
	}; // while somethingFound

	// update outer indices data
	outerIndicesArrIdx = term.getOuterIndices();
	n = outerIndicesArrIdx.length;
	p = n/2;
	FH.printProgressMessage(msg + ' / p = ' + p);

	// no remaining contraction to perform
	// just add a copy of the term
	if (p == 0) {
		contractedSum.addTo(term.makeCopy(), false);
		// new term
		return;
	};

	// needed contractions information
	// no error if missing: do nothing and alert
	if (dataGTensor.arrG[p] === undefined) {
		console.log('Missing data for G tensor at p=' + p);
		// next term in the for loop
		return;
	}
	// get all data at level p
	let currentG = dataGTensor.arrG[p];
	let currentArrIdxG = dataGTensor.arrGIndices[p];
	let currentPerm = dataGTensor.arrRedPerm[p];

	let cp = 0;
	for (let perm of currentPerm) {
		cp++;
		FH.printProgressMessage(msg + ' / p = ' + p + ' / perm: ' + cp);
		// console.log('term ' + c + ' / p = ' + p + ' / perm: ' + cp);
		let currentTerm = term.makeCopy();
		// currentTerm.propagateNormalizedIndicesArray([]);
		let arrCont = currentTerm.contractions;
		// collect the srtIdx of the target contractions
		let targetArrSrtIdx = [];
		for (arrIdx of outerIndicesArrIdx) {
			let srtIdx = currentTerm.getTextIndiceAt(arrIdx);
			targetArrSrtIdx.push(srtIdx);
		};

		// multiply currentG on preElement
		let currentPreElement = currentTerm.preElement;
		currentPreElement.multiplyOnLeftWith(currentG);

		// collect the srtIdx of the source contractions,
		// since multiplication makes a copy of currentG
		// the srtIdx of indices are (only) now well-defined,
		// since multiplication is on the left,
		// currentArrIdxG still indexes at the right place
		// this is fortunate…
		let sourceArrSrtIdx = [];
		for (let arrIdx of currentArrIdxG) {
			let srtIdx = currentPreElement.getTextIndiceAt(arrIdx);
			sourceArrSrtIdx.push(srtIdx);
		};
		for (let i=0; i<sourceArrSrtIdx.length; i++) {
			// the i-th element in sourceArrSrtIdx is contracted with
			// the perm[i]-th element in targetArrSrtIdx
			let tSrtIdx = targetArrSrtIdx[perm[i]];
			arrCont.push([sourceArrSrtIdx[i], tSrtIdx].sort(FH.alphaSort));
		};
		currentTerm.contractions = FH.copyData(arrCont);
		if (ruleData.applyRules) {
			// TODO: use FR functions, with FR.RULES_ARRAY_1 and FR.RULES_ARRAY_2, or FR.RULES_ARRAY_1_2
			// FC.applyRulesOnTerm(currentTerm, rulesData);
			FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, currentTerm, rulesData);
		};
		contractedSum.addTo(currentTerm, false);
	}; // loop on perm
	return;
};



// collect arrays of terms suspected to be related by Lemma about X_k and X_{k+1}
// criteria (work for R2):
// - same expression modulo presence of u in tensor product (pre, main, post)
// - same degree in d in the rational expression in front => good ?
FC.collectTermsForOperatorRelation = function(originalSum, elID) {
	let arrOperatorCollections = [];
	for (let i=0; i<originalSum.list.length; i++) {
		let term = originalSum.list[i];
		term.fillNormalizedIndicesArray();
		// term.showNormalized(OPT.SHOW, 'term: ');
		let {counter: counter, string: mainTxt, preString: preTxt, postString: postTxt}
							= term.getReducedStringRepresentationCountingElementID(elID);

		let preRE = term.preElement.prefactor;
		let deg = preRE.numerator.getDegree();

		let found = false;
		for (let obj of arrOperatorCollections) {
			if (obj.post === postTxt
						&& obj.main === mainTxt
						&& obj.pre === preTxt
						&& obj.deg === deg) {
				let newTermObj = {idx: i, counter: counter};
				obj.terms.push(newTermObj);
				obj.terms.sort( (o1,o2) => {return o1.counter - o2.counter;} );
				found = true;
				break;
			}; // if already created
		}; // for obj
		if (!found) {
			// create new entry
			let obj = {};
			obj.main = mainTxt;
			obj.post = postTxt;
			obj.pre = preTxt;
			obj.deg = deg;
			obj.terms = [{idx: i, counter: counter}];
			arrOperatorCollections.push(obj);
		};
	}; // for i
	return arrOperatorCollections;
};

// collect arrays of terms suspected to be related by Lemma about X_k and X_{k+1}
// criteria (work for R2):
// - same expression modulo presence of u in tensor product (pre, main, post)
// - ignore degree in d in the rational expression in front
FC.collectTermsForOperatorRelationIgnoreDegree = function(originalSum, elID) {
	let arrOperatorCollections = [];
	for (let i=0; i<originalSum.list.length; i++) {
		let term = originalSum.list[i];
		term.fillNormalizedIndicesArray();
		// term.showNormalized(OPT.SHOW, 'term: ');
		let {counter: counter, string: mainTxt, preString: preTxt, postString: postTxt}
							= term.getReducedStringRepresentationCountingElementID(elID);

		let preRE = term.preElement.prefactor;

		let found = false;
		for (let obj of arrOperatorCollections) {
			if (obj.post === postTxt
						&& obj.main === mainTxt
						&& obj.pre === preTxt) {
				let newTermObj = {idx: i, counter: counter};
				obj.terms.push(newTermObj);
				obj.terms.sort( (o1,o2) => {return o1.counter - o2.counter;} );
				found = true;
				break;
			}; // if already created
		}; // for obj
		if (!found) {
			// create new entry
			let obj = {};
			obj.main = mainTxt;
			obj.post = postTxt;
			obj.pre = preTxt;
			obj.terms = [{idx: i, counter: counter}];
			arrOperatorCollections.push(obj);
		};
	}; // for i
	return arrOperatorCollections;
};

// apply the relation on the operators X_k, at all occurences of (k, k+1)
// u: instance of uproto
FC.applyOperatorSimplificationAllAtOnce = function(originalSum, u) {
	let uID = u.getID();
	let arrOperatorCollections = FC.collectTermsForOperatorRelation(originalSum, uID);
	let operatorSimplifiedSum = new MathSum();

	for (let obj of arrOperatorCollections) {
		console.log('');
		console.log(JSON.stringify(obj));

		if (obj.terms.length > 1) {
			let maxTermData = obj.terms[obj.terms.length-1];
			// console.log(JSON.stringify(maxTermData));
			let maxCounter = maxTermData.counter;
			let maxIdx = maxTermData.idx;
			let maxTerm = originalSum.list[maxIdx];
			let maxTermRE = maxTerm.preElement.prefactor;
			let maxNum = maxTermRE.numerator.getMaxDegreeCoefficient();
			let maxDen = maxTermRE.denominator;
			let maxSign = Math.sign(maxNum);
			// console.log('maxTerm: ' + maxNum + '/' + maxDen);
			for (let termData of obj.terms) {
				let counter = termData.counter;
				let idx = termData.idx;
				let term = originalSum.list[idx];
				let termRE = term.preElement.prefactor;
				let num = termRE.numerator.getMaxDegreeCoefficient();
				let den = termRE.denominator;
				let sign = Math.sign(num);
				if (counter + 1 == maxCounter && sign == - maxSign) {
					// console.log('term: ' + num + '/' + den);
					// term.showConsole(OPT.SHOW, 'current term: ');
					// console.log('coef: ' + num*maxDen + '/' + den*maxNum);
					// add original term
					operatorSimplifiedSum.addTo(term, false);
					// construct terms to add to the new sum

					// first, add same term with another preElement.prefactor
					let newTerm = term.makeCopy();
					// newTerm.showConsole(OPT.SHOW, 'new term: ');
					// create new preElement.prefactor for this newTerm
					let newRE = maxTermRE.makeCopy();
					newRE.multiplyByRational(num*maxDen, den*maxNum);
					// newRE.showConsole(OPT.SHOW, 'new RE: ');
					let newTermRE = newRE.makeCopy();
					newTermRE.multiplyByInteger(-1);
					// newTerm.preElement.showConsole(OPT.SHOW, 'pre element: ');
					newTerm.preElement.prefactor = newTermRE;
					// newTerm.showConsole(OPT.SHOW, 'new term: ');
					operatorSimplifiedSum.addTo(newTerm, false);

					// second, add all terms with u at all places
					for (let i=0; i<term.list.length+1; i++) {
						let newTermPropagation = term.makeCopy();
						let newTermPropagationRE = newRE.makeCopy();
						// newTermPropagationRE.showConsole(OPT.SHOW, 'new prop. RE: ');
						newTermPropagation.preElement.prefactor = newTermPropagationRE;
						let newU = MathProduct.newFrom(u);
						newTermPropagation.list.splice(i,0, newU);
						// newTermPropagation.showConsole(OPT.SHOW, 'new prop. term: ');
						operatorSimplifiedSum.addTo(newTermPropagation, false);
					};

				} else {
					// not a good candidate => add it to the new sum
					operatorSimplifiedSum.addTo(term, false);
				};
			};
		} else {
			// only one term in obj.terms => add it to the new sum
			let idx = obj.terms[0].idx;
			let term = originalSum.list[idx];
			operatorSimplifiedSum.addTo(term, false);
		};
	};
	operatorSimplifiedSum.sortBySumWeight();
	operatorSimplifiedSum.fillNormalizedIndicesArray();
	operatorSimplifiedSum.simplifyExpression();

	return operatorSimplifiedSum;
};


// apply the relation on the operators X_k, at max #u
// for all the patterns found, take into account degrees of polynomials in prefactors
// u: instance of PROTO.u
FC.applyOperatorSimplificationTopDown = function(originalSum, u) {
	// used in a loop which fires this function => repeat until nothing found
	let somethingFound = false;

	let uID = u.getID();
	let arrOperatorCollections = FC.collectTermsForOperatorRelation(originalSum, uID);
	let operatorSimplifiedSum = new MathSum();

	for (let obj of arrOperatorCollections) {
		// console.log('');
		// console.log(JSON.stringify(obj));

		if (obj.terms.length > 1) {
			// corresponds to max #u since arrOperatorCollections is sorted with increasing #u
			let maxTermData = obj.terms[obj.terms.length-1];
			// console.log(JSON.stringify(maxTermData));
			let maxCounter = maxTermData.counter;
			let maxIdx = maxTermData.idx;
			let maxTerm = originalSum.list[maxIdx];
			let maxTermRE = maxTerm.preElement.prefactor;
			let maxNum = maxTermRE.numerator.getMaxDegreeCoefficient();
			let maxDen = maxTermRE.denominator;
			let maxSign = Math.sign(maxNum);
			// console.log('maxTerm: ' + maxNum + '/' + maxDen);
			for (let termData of obj.terms) {
				let counter = termData.counter;
				let idx = termData.idx;
				let term = originalSum.list[idx];
				let termRE = term.preElement.prefactor;
				let num = termRE.numerator.getMaxDegreeCoefficient();
				let den = termRE.denominator;
				let sign = Math.sign(num);
				// only retain #u-1
				if (counter + 1 == maxCounter && sign == - maxSign) {
					// console.log('perform term data: ' + JSON.stringify(termData));
					// console.log('');
					somethingFound = true;
					// console.log('term: ' + num + '/' + den);
					// term.showConsole(OPT.SHOW, 'current term: ');
					// console.log('coef: ' + num*maxDen + '/' + den*maxNum);
					// add original term
					operatorSimplifiedSum.addTo(term, false);
					// construct terms to add to the new sum

					// first, add same term with another preElement.prefactor
					let newTerm = term.makeCopy();
					// newTerm.showConsole(OPT.SHOW, 'new term: ');
					// create new preElement.prefactor for this newTerm
					let newRE = maxTermRE.makeCopy();
					newRE.multiplyByRational(num*maxDen, den*maxNum);
					// newRE.showConsole(OPT.SHOW, 'new RE: ');
					let newTermRE = newRE.makeCopy();
					newTermRE.multiplyByInteger(-1);
					// newTerm.preElement.showConsole(OPT.SHOW, 'pre element: ');
					newTerm.preElement.prefactor = newTermRE;
					// newTerm.showConsole(OPT.SHOW, 'new term: ');
					operatorSimplifiedSum.addTo(newTerm, false);

					// second, add all terms with u at all places
					for (let i=0; i<term.list.length+1; i++) {
						let newTermPropagation = term.makeCopy();
						let newTermPropagationRE = newRE.makeCopy();
						// newTermPropagationRE.showConsole(OPT.SHOW, 'new prop. RE: ');
						newTermPropagation.preElement.prefactor = newTermPropagationRE;
						let newU = MathProduct.newFrom(u);
						newTermPropagation.list.splice(i,0, newU);
						// newTermPropagation.showConsole(OPT.SHOW, 'new prop. term: ');
						operatorSimplifiedSum.addTo(newTermPropagation, false);
					};

				} else {
					// not a good candidate => add it to the new sum
					operatorSimplifiedSum.addTo(term, false);
				};
			};
		} else {
			// only one term in obj.terms => add it to the new sum
			let idx = obj.terms[0].idx;
			let term = originalSum.list[idx];
			operatorSimplifiedSum.addTo(term, false);
		};
	};
	operatorSimplifiedSum.sortBySumWeight();
	operatorSimplifiedSum.fillNormalizedIndicesArray();
	operatorSimplifiedSum.simplifyExpression();

	// flag is used in the loop over all rule functions => repeat
	return {flag: somethingFound, sum: operatorSimplifiedSum};

};

// apply the relation on the operators X_k, at max #u
// for all the patterns found, ignore degrees of polynomials in prefactors
// u: instance of PROTO.u
FC.applyOperatorSimplificationIgnoreDegree = function(originalSum, u) {
	// used in a loop which fires this function => repeat until nothing found
	let somethingFound = false;

	let uID = u.getID();
	let arrOperatorCollections = FC.collectTermsForOperatorRelationIgnoreDegree(originalSum, uID);
	let operatorSimplifiedSum = new MathSum();

	for (let obj of arrOperatorCollections) {
		// console.log('');
		// console.log(JSON.stringify(obj));

		if (obj.terms.length > 1) {
			// corresponds to max #u since arrOperatorCollections is sorted with increasing #u
			let maxTermData = obj.terms[obj.terms.length-1];
			let maxCounter = maxTermData.counter;
			// corresponds to min #u
			let minTermData = obj.terms[0];
			let minCounter = minTermData.counter;
			let minIdx = minTermData.idx;
			let minTerm = originalSum.list[minIdx];

			// add minTermData or a "propaged version" to the sum
			if (minCounter < maxCounter) {
				// console.log('perform term data: ' + JSON.stringify(minTermData));
				// console.log('');
				somethingFound = true;
				// construct terms to add to the new sum
				// add all terms with u at all places
				for (let i=0; i<minTerm.list.length+1; i++) {
					let newTermPropagation = minTerm.makeCopy();
					let newU = MathProduct.newFrom(u);
					newTermPropagation.list.splice(i,0, newU);
					// newTermPropagation.showConsole(OPT.SHOW, 'new prop. term: ');
					operatorSimplifiedSum.addTo(newTermPropagation, false);
				};
			} else {
					// not a good candidate => add it to the new sum
					operatorSimplifiedSum.addTo(minTerm, false);
			};

			// add all other terms in obj.terms for index > 0
			for (let i=1; i<obj.terms.length; i++) {
				let termData = obj.terms[i];
				let idx = termData.idx;
				operatorSimplifiedSum.addTo(originalSum.list[idx], false);
			};
		} else {
			// only one term in obj.terms => add it to the new sum
			let idx = obj.terms[0].idx;
			let term = originalSum.list[idx];
			operatorSimplifiedSum.addTo(term, false);
		};
	};
	operatorSimplifiedSum.sortBySumWeight();
	// operatorSimplifiedSum.showConsole(OPT.SHOW, 'operatorSimplifiedSum = ');
	operatorSimplifiedSum.fillNormalizedIndicesArray();
	operatorSimplifiedSum.simplifyExpression();

	// flag is used in the loop over all rule functions => repeat
	return {flag: somethingFound, sum: operatorSimplifiedSum};
};



// propagate given MathProduct or MathElement in terms of a sum
FC.propagateItemInSum = function(item, sum, opt, text = '') {
	getPropagationLaTeX = function(ltxItem, k) {
		let n = k+1;
		let propagationTxt = '\\sum_{\\ell=1}'
													+ '^{' + n + '}'
													+ '('
													+ 'i^{(\\ell)}_'
													+ '{' + ltxItem + '} '
													+ CONST.LATEX_SPECTRAL_OPERATOR_LETTER
													+ CONST.LATEX_SUB + '{' + n + '}'
													+ ')';
		return propagationTxt;
	};

	let newSum = new MathSum();
	let objItem = item.getLaTeX(opt);
	let ltxItem = objItem.ltx;
	let signItem = objItem.sign;
	if (signItem<0) {
		throw "sign of item should be 1";
	};

	// code very similar to getLaTeX(opt) for MathExpression (MathSum case)
	let latexJoinList = CONST.LATEX_SUM_NEWLINE
										+ CONST.CR
										+ CONST.LATEX_ALIGN_SEP;

	// console.log('ltxItem = ' + ltxItem);
	let latex = '';
	for (let i=0; i<sum.list.length; i++) {
		let term = sum.list[i];
		// term.showConsole(OPT.SHOW, 'term = ');
		let objTerm = term.getLaTeX(opt);

		// sign string
		let signTerm = objTerm.sign;
		let signTxt = '';
		if (signTerm<0) {
			signTxt = '-';
		} else if(i>0) {
			// no sign '+' for first item
			signTxt = '+';
		};

		let ltxTerm = objTerm.ltx;
		// console.log('ltxTerm = ' + ltxTerm);
		let k = term.list.length;
		let opTxt = CONST.LATEX_SPECTRAL_OPERATOR_LETTER
							+ CONST.LATEX_SUB + '{' + k + '}';
		let propTxt = getPropagationLaTeX(ltxItem, k);
		// console.log('opTxt = ' + opTxt);
		// console.log('propTxt = ' + propTxt);
		// replace opTxt by propTxt in ltxTerm
		ltxTerm = FH.replaceAll(ltxTerm, opTxt, propTxt);

		latex += signTxt + ltxTerm;
		if (opt.showLineNumbers) {
			let nLine = i+1;
			latex += CONST.LATEX_LINE_NUMBER + '{' + nLine + '} ';
		};
		if (i !== sum.list.length-1) {
			latex += latexJoinList;
		};

 		for (let i=0; i<term.list.length+1; i++) {
			let newTerm = term.makeCopy();
			let newItem = MathProduct.newFrom(item);
			newTerm.list.splice(i,0, newItem);
			newSum.addTo(newTerm, false);
		};
	}; // loop on terms
	latex = '{}' + CONST.LATEX_ALIGN_SEP + latex;
	let latexSum = CONST.LATEX_BEGIN_EQUATION
					+ CONST.CR
					+ text
					+ latex
					+ CONST.CR
					+ CONST.LATEX_END_EQUATION;

	// newSum.showConsole(OPT.SHOW, 'sum = ');
	return {sum: newSum, ltx: latexSum};
};


FC.removeSimplifiedSumFrom = function(removeSum, mainSum, opt) {

	let negRemoveSum = removeSum.makeCopy();
	negRemoveSum.multiplyByInteger(-1);
	let returnedSum = mainSum.makeCopy()
	mainSum.addTo(negRemoveSum);

	if (opt.useOperatorRules) {
		let somethingFound = true;
		let workingSum = mainSum;
		// let loopC = 0;
		while(somethingFound) {
			let {flag, sum}
							= FC.applyOperatorSimplificationIgnoreDegree(workingSum, opt.u);
			workingSum = sum;
			somethingFound = flag;
		};
		FH.printProgressMessage('');
		mainSum = workingSum;
	};

	mainSum.fillNormalizedIndicesArray();
	mainSum.simplifyExpression();

	return mainSum;
};

// return object containing data used to simplify expressions
FC.getPatternOfMathArgument = function(element) {
	let strArrRep = element.getStringArrayRepresentation();
	// let strPreElRep = element.preElement.getStringRepresentation();
	let strPreElArrRep = element.preElement.getStringArrayRepresentation();
	let poly = element.preElement.prefactor.numerator;
	let num = poly.getMaxDegreeCoefficient();
	let deg = poly.getDegree();
	let den = element.preElement.prefactor.denominator;
	let coeff = RationalExpression.newFromLetterAndRational(poly.getLetter(),num,den);
	let obj = {};
	obj.length = element.list.length;
	obj.mainPattern = strArrRep;
	// obj.preElStr = strPreElRep;
	obj.strPreElArrRep = strPreElArrRep;
	obj.coeff = coeff;
	obj.deg = deg;
	// console.log(JSON.stringify(obj));
	return obj;
};

// remove occurences of the u element in the array patternArr
FC.reducePatternArray = function(patternArr, u) {
	let ds = {useInternalDisplaySymbols: false};
	let stru = u.getStringRepresentation(ds);
	// console.log(stru);
	let returnedArr = [];
	for (let i=0;i<patternArr.length; i++) {
		if (patternArr[i] !== stru) {
			returnedArr.push(patternArr[i]);
		}
	};
	return returnedArr;
};

// create normalized pattern
// i.e. replace non 'u' by letters a,b,c,…
FC.createNormizedPattern = function(patternArr, u) {
	let ds = {useInternalDisplaySymbols: false};
	let stru = u.getStringRepresentation(ds);
	let returnedArr = [];
	let c = 0;
	for (let i=0;i<patternArr.length; i++) {
		if (patternArr[i] !== stru) {
			// add next letter in CONST.ARRAY_LETTERS_SHORT
			returnedArr.push(CONST.ARRAY_LETTERS_SHORT[c]);
			c++;
		} else {
			// add 'u'
			returnedArr.push(stru);
		}
	};
	return returnedArr;
};

// sort array of {term, termData, sortStr} using sortStr
FC.sortByPattern = function(a,b) {
	if(a.sortStr < b.sortStr) return -1;
	if(a.sortStr > b.sortStr) return 1;
	return 0;
};

FC.createSumSortedByPattern = function(sum, u) {
	let ds = {useInternalDisplaySymbols: false};
	let stru = u.getStringRepresentation(ds);
	let termsArr = [];
	for (let term of sum.list) {
		let termData = FC.getPatternOfMathArgument(term);
		let mainPattern = termData.mainPattern;
		let termPattern = FC.reducePatternArray(mainPattern, u);
		let termPatternStr = termPattern.join('*');
		if (termPatternStr === '') {
			termPatternStr = stru;
		};
		let strPreElArrRep = termData.strPreElArrRep;
		let termPreElStr = strPreElArrRep.join('*');
		let key = termPreElStr + '+' + termPatternStr;
		let mainPatternClean = FH.copyData(mainPattern);
		for (let i=0; i<mainPattern.length; i++) {
			if (mainPatternClean[i] === stru) {
				// first character in ascii table
				mainPatternClean[i] = '!';
			};
		};
		let sortStr = strPreElArrRep.join('*') + '+' + mainPatternClean.join('*');

		termsArr.push({term: term, termData: termData, sortStr: sortStr});
	};
	termsArr.sort(FC.sortByPattern);
	let sumReturned = new MathSum();
	for (let termObj of termsArr) {
		let term = termObj.term;
		sumReturned.addTo(term, false);
	};
	return sumReturned;
};


// collect in an array terms with same patterns for simplification
// the "model" is the first occurence in the sum
FC.collectTermsWithSamePattern = function(sum, u) {
	N = sum.list.length;
	// collect terms with same pattern, degree, length, preElement
	let collectedTerms = [];
	let currentLength = 0;
	let currentDegree = 0;
	let currentReducedPatternStr = '';
	let currentPreElementStr = '';
	for (let i=0; i<N; i++) {
		let obj = FC.getPatternOfMathArgument(sel13_4.list[i]);
		let {length, mainPattern, strPreElArrRep, coeff, deg} = obj;
		let initialPattern = FC.reducePatternArray(mainPattern, u);
		let initialPatternStr = initialPattern.join('+');
		let preElStr = strPreElArrRep.join('+');

		if (collectedTerms.length === 0) {
			collectedTerms.push(obj);
			currentLength = length;
			currentDegree = deg;
			currentReducedPatternStr = initialPatternStr;
			currentPreElementStr = preElStr;
		} else if (currentLength === length) {
			// collect highest degree
			if (deg > currentDegree){
				collectedTerms = [];
				collectedTerms.push(obj);
				currentLength = length;
				currentDegree = deg;
				currentReducedPatternStr = initialPatternStr;
				currentPreElementStr = preElStr;
			} else if (currentDegree === deg
									&& currentReducedPatternStr === initialPatternStr
									&& currentPreElementStr === preElStr) {
				collectedTerms.push(obj);
			};
		};

	};
	return collectedTerms;
};

FC.incrementArrayPattern = function(arrPattern, u) {
	let ds = {useInternalDisplaySymbols: false};
	let stru = u.getStringRepresentation(ds);

	let arrNewPatterns = [];
	// let arrNumber = [];
	// create list of new pattern adding 'u' at each places
	let cumulativeArray = [];
	for (let i=0;i<=arrPattern.length; i++) {
		let newPattern = FH.copyData(arrPattern);
		newPattern.splice(i, 0, stru);
		cumulativeArray.push(newPattern);
	};
	// collect number of occurences of same pattern
	// this is collected in a key=>value object
	// key: string representing the pattern
	// value: number of occurences
	let objNumber = {};
	// since same patterns are is successive order
	// need only compare to the "last found"
	// saved in currentPatternStr
	let currentPatternStr = '';
	for (let arr of cumulativeArray) {
		let testStr = arr.join('+');
		if (currentPatternStr === testStr) {
			// same as previous, so only increment number
			objNumber[testStr]++;
		} else {
			// new pattern
			// update currentPatternStr
			currentPatternStr = testStr;
			// number = 1
			objNumber[testStr] = 1;
			// collect the pattern in the final array
			arrNewPatterns.push(arr);
		};
	};
	// for (let arr of arrNewPatterns) {
	// 	arrNumber.push(objNumber[arr.join('+')]);
	// };
	return {arrPattern: arrNewPatterns, objNumber: objNumber};
};

// split sum into sub sums according to patterns
// save as LaTeX, JSON and Mathematica
// originalSum: original sum to split
// u: PROTO.u
// postStr: added to "SPLIT" folder
FC.saveSplitSum = function(originalSum, u, postStr, rel_path, preLatex = '') {
	let splitObj = {};
	let ds = {useInternalDisplaySymbols: false};
	let stru = u.getStringRepresentation(ds);

	for (let term of originalSum.list) {
		let termData = FC.getPatternOfMathArgument(term);
		let mainPattern = termData.mainPattern;
		let termPattern = FC.reducePatternArray(mainPattern, u);
		let termPatternStr = termPattern.join('*');
		if (termPatternStr === '') {
			termPatternStr = stru;
		};
		let strPreElArrRep = termData.strPreElArrRep;
		let termPreElStr = strPreElArrRep.join('*');
		let key = termPreElStr + '+' + termPatternStr;
		let mainPatternClean = FH.copyData(mainPattern);
		for (let i=0; i<mainPattern.length; i++) {
			if (mainPatternClean[i] === stru) {
				// first character in ascii table
				mainPatternClean[i] = '!';
			};
		};
		let sortStr = strPreElArrRep.join('*') + '+' + mainPatternClean.join('*');

		// console.log('pre element: ' + termPreElStr + ' / pattern: ' + termPatternStr);
		// console.log('string: ' + sortStr);

		// collect terms according to patterns
		if (splitObj[key] === undefined) {
			splitObj[key] = [{term: term, termData: termData, sortStr: sortStr}];
		} else {
			splitObj[key].push({term: term, termData: termData, sortStr: sortStr});
		};
		// console.log('');
	};

	let N = 0;
	let ltxFileNameCorr = [['^', ''],['_', ''],['·', '']];
	let inputLtx = preLatex;
	// format each group of pattern in various formats:
	// json to be reused latter
	// latex
	// mathematica
	// normalized/simplified mathematica
	let fileKeysArr = [];
	for (let k in splitObj) {
		N = N + splitObj[k].length;
		splitObj[k].sort(FC.sortByPattern);
		let sum = new MathSum();
		for (let termObj of splitObj[k]) {
			let term = termObj.term;
			sum.addTo(term, false);
			// console.log(termObj.sortStr);
		};
		console.log('key: ' + k + ' / #terms: ' + splitObj[k].length);
		sum.fillNormalizedIndicesArray();
		// sum.showNormalized(OPT.SHOW, 'Sum = ');
		let fileNameRoot = FH.replaceInStringUsingCorrespondance(k, ltxFileNameCorr);

		// collect this file name
		fileKeysArr.push('SPLIT' + postStr + '/' + fileNameRoot);

		console.log(fileNameRoot);
		let latex = sum.getLaTeXEquation(OPT.LTX, 'R = ');
		// console.log(latex);
		FH.saveToFileWithSureFolder(rel_path + 'LATEX/SPLIT' + postStr + '/' + fileNameRoot + '.tex', latex);
		// add "input" of this file to create master file
		inputLtx += '\\inputFile{SPLIT' + postStr + '/' + fileNameRoot + '.tex' + '}\n\n';

		let mathematica = sum.getMathematica(OPT.MATH);
		// save full mathematica expression
		FH.saveToFileWithSureFolder(rel_path + 'MATHEMATICA/SPLIT' + postStr + '/' + fileNameRoot + '.txt', mathematica);

		// create simplified mathematica expression
		// normalized pattern
		// polynomials
		let collectedTermsText = 'S = ';
		collectedTermsText += CONST.MATH_PLUS + CONST.MATH_BRACKET_OPEN + '\n';
		let termsArr = [];
		for (let i=0; i<sum.list.length; i++) {
			let term = sum.list[i];
			let obj = FC.getPatternOfMathArgument(term);
			let mainPattern = obj.mainPattern;
			let normalizedPattern = FC.createNormizedPattern(mainPattern, u);
			// console.log('normalizedPattern: ' + JSON.stringify(normalizedPattern));
			let mathRE = term.preElement.prefactor;

			let tensor = FH.convertArrayToMathematicaTensorProductString(normalizedPattern);
			let termTxt = CONST.MATH_TIMES + CONST.MATH_BRACKET_OPEN;
			termTxt += mathRE.getMathematica(OPT.MATH);
			termTxt += ', ';
			termTxt += tensor;
			termTxt += CONST.MATH_BRACKET_CLOSE;
			termsArr.push(termTxt);
		};
		collectedTermsText += termsArr.join(',\n');
		collectedTermsText += CONST.MATH_BRACKET_CLOSE + '\n\n';

		// save simplified/normalized mathematica expression
		FH.saveToFileWithSureFolder(rel_path + 'MATHEMATICA/SPLIT-NORMALIZED' + postStr + '/' + fileNameRoot + '.txt', collectedTermsText);

		// save expression in json file to manipulate it latter
		FH.saveItemToFile(sum, rel_path + 'RESULTS/SPLIT' + postStr + '/' + fileNameRoot + '.json');

		console.log('');
	};
	console.log('N = ' + N);

	// save file with list of file keys
	FH.saveToFileWithSureFolder(rel_path + 'RESULTS/LISTS/SPLIT' + postStr + '-list.json', JSON.stringify(fileKeysArr, null, "\t"));

	// save file with latex inputs
	FH.saveToFileWithSureFolder(rel_path + 'LATEX/INPUT-SPLIT/' + 'input-split' + postStr + '.tex', inputLtx);

};


// return the spectral functions of a MathSum
// in a format used by Mathematica
// for spectral functions "Ialpha" defined in the paper and its notebook
FC.getSpectralFunction = function(sum, spectralData, opt) {

	// collect data on spectral elements in 'factor' according to data in 'spectralItems'
	// 'spectralItems' contains elements in spectral calculus of 'u'
	getSpectralData = function(factor, spectralItems) {
		let inSpectralCalculus = false;
		let power = new Rational(0,1);
		for (let data of spectralItems) {
			// don't use 'getID()' since returns MathElement in MathOperator
			// here we want 'top' id
			// console.log(factor.id + ' / ' + data.itemID);
			if (factor.id === data.itemID) {
				// console.log('the same');
				inSpectralCalculus = true;
				power = data.power;
			};
		};
		return {inSpectralCalculus: inSpectralCalculus, power: power};
	};

	let applyDimensionValue = opt.applyDimensionValue;
	let d = 0;
	let halfDimension = 0;
	if (applyDimensionValue) {
		d = opt.dimension;
		halfDimension = opt.dimension/2;
	};
	let spectralFunctionString = opt.spectralFunctionString;
	let letterSpectralArgument = opt.letterSpectralArgument;

	let shiftHalfDimension = opt.shiftHalfDimension;

	let spectralTxtArr = [];
	for (let term of sum.list) {
		// console.log('');

		// work with copy to change part of it
		let termCopy = term.makeCopy();

		if (applyDimensionValue) {
			// evaluate at given dimension (construct option for that)
			let preElement = termCopy.preElement;
			// preElement.showConsole(OPT.SHOW, 'pre El. = ');
			preElement.prefactor = preElement.prefactor.evaluateAtInteger(d);
			termCopy.updateExpression();
			if (termCopy.isZero()) {
				continue;
			};
			// preElement.showConsole(OPT.SHOW, 'pre El. = ');
		};


		termCopy.fillNormalizedIndicesArray();
		// termCopy.showNormalized(OPT.SHOW, 'Term = ');
		// termCopy.showConsole(OPT.SHOW, 'term = ');
		// array to collect data on each tensor factor in main part
		// these data collect information on MathProduct in each tensor factor
		let factorDataArr = [];
		for (let tensorFactor of termCopy.list) {
			// tensorFactor.showConsole(OPT.SHOW, 'tensor factor = ');
			// initial data to collect
			let pl = new Rational(0,1);
			let pr = new Rational(0,1);
			// let kernel = tensorFactor;
			let factor = tensorFactor.list[0];
			// factor.showConsole(OPT.SHOW, 'factor : ');
			let obj = getSpectralData(factor, spectralData.spectralItems);
			let inSpectralCalculus = obj.inSpectralCalculus;
			let power = obj.power;
			let performRight = true;
			while (inSpectralCalculus) {
				// console.log('in spectral calculus');
				// add power
				// power.showConsole(OPT.SHOW, 'power = ');
				pl.addTo(power);
				// if last factor…
				if (tensorFactor.list.length === 1) {
					// console.log('last item');
					// kernel is 1
					tensorFactor = MathProduct.newFrom(MathElement_one);
					// no need to collect on the right
					performRight = false;
					// quit the while loop
					inSpectralCalculus = false;
				} else {
					// remove this factor (first one)
					tensorFactor.list.shift();
					// consider next factor
					factor = tensorFactor.list[0];
					// factor.showConsole(OPT.SHOW, 'factor : ');
					obj = getSpectralData(factor, spectralData.spectralItems);
					inSpectralCalculus = obj.nSpectralCalculus;
					power = obj.power;
				};
			};
			// get data from right factors
			// if (performRight) {
				factor = tensorFactor.list[tensorFactor.list.length-1];
				// factor.showConsole(OPT.SHOW, 'factor : ');
				obj = getSpectralData(factor, spectralData.spectralItems);
				inSpectralCalculus = obj.inSpectralCalculus;
				power = obj.power;
				while (inSpectralCalculus) {
					// console.log('in spectral calculus');
					// add power
					// power.showConsole(OPT.SHOW, 'power = ');
					pr.addTo(power);
					// if last factor…
					if (tensorFactor.list.length === 1) {
						// console.log('last item');
						// kernel is 1
						tensorFactor = MathProduct.newFrom(MathElement_one);
						// quit the while loop
						inSpectralCalculus = false;
					} else {
						// remove this factor (last one)
						tensorFactor.list.pop();
						// consider next factor
						factor = tensorFactor.list[tensorFactor.list.length-1];
						// factor.showConsole(OPT.SHOW, 'factor : ');
						obj = getSpectralData(factor, spectralData.spectralItems);
						inSpectralCalculus = obj.nSpectralCalculus;
						power = obj.power;
					};
				// };
			};
			factorDataArr.push({left: pl, right: pr, kernel: tensorFactor});
			// console.log('');
		};
		// display partial result
		// for (let factorData of factorDataArr) {
		// 	factorData.kernel.showConsole(OPT.SHOW, 'kernel = ');
		// 	factorData.left.showConsole(OPT.SHOW, 'left = ');
		// 	factorData.right.showConsole(OPT.SHOW, 'right = ');
		// };

		// analyse 'factorDataArr'
		let powerArr = [];
		let powerStringArr = [];
		let kernelArr = [];
		let p = new Rational(0,1);
		// indice of the spectral function
		let k = factorDataArr.length;
		// arguments of the spectral function
		let argumentArr = [];
		// current indice of argument of spectral function
		let indiceArg = 0;
		// add first argument, always 0 indice
		argumentArr.push(letterSpectralArgument + indiceArg);
		for (let i=0; i<factorDataArr.length; i++) {
			let factorData = factorDataArr[i];
			// factorData.kernel.showConsole(OPT.SHOW, 'kernel = ');
			// factorData.left.showConsole(OPT.SHOW, 'left = ');
			// factorData.right.showConsole(OPT.SHOW, 'right = ');
			if (factorData.kernel.isOne()) {
				p.addTo(factorData.left);
				p.addTo(factorData.right);
			} else {
				p.addTo(factorData.left);
				powerArr.push(p);
				powerStringArr.push(p.getStringRepresentation());
				p = Rational.newFrom(factorData.right);
				kernelArr.push(factorData.kernel);
				// increment indice of argument since new non trivial kernel element
				indiceArg++;
			};
			// add argument with current indice
			argumentArr.push(letterSpectralArgument + indiceArg);
		};
		// collect last power
		powerArr.push(p);
		powerStringArr.push(p.getStringRepresentation());
		// for (let p of powerArr) {
		// 	p.showConsole(OPT.SHOW, 'p = ');
		// };
		// for (let kernel of kernelArr) {
		// 	kernel.showConsole(OPT.SHOW, 'kernel = ');
		// };
		// console.log(JSON.stringify(powerStringArr));
		termCopy.list = kernelArr;
		termCopy.fillNormalizedIndicesArray();
		// termCopy.showNormalized(OPT.SHOW, 'Term simplified = ');
		// console.log(termCopy.getMathematica(OPT.MATH));

		// construct spectral function with its arguments
		// TODO: if applyDimensionValue is false, use letterHalfDimension
		let shift = halfDimension + shiftHalfDimension + k;
		let spectralFunctionTxt = spectralFunctionString
															+ k
															+ CONST.MATH_BRACKET_OPEN
															+ shift
															+ ', '
															+ argumentArr.join(', ')
															+ CONST.MATH_BRACKET_CLOSE;

		// construct the factor in front of the pectral function
		// powers of arguments
		let spectralPrefactor = '';
		// let argumentArr = [];
		for (let i=0; i<powerArr.length; i++) {
			let p = powerArr[i];
			if (!p.isZero()) {
				spectralPrefactor += CONST.MATH_POWER
														+ CONST.MATH_BRACKET_OPEN
														+ letterSpectralArgument
														+ i
														+ ', '
														+ p.getStringRepresentation()
														+ CONST.MATH_BRACKET_CLOSE
														+ ' ';
			};
			// argumentArr.push(letterSpectralArgument + i);
		};
		spectralFunctionTxt = spectralPrefactor + spectralFunctionTxt;
		let termMathematicaTxt = CONST.MATH_TIMES
														+ CONST.MATH_BRACKET_OPEN
														+ spectralFunctionTxt
														+ ', '
														+ termCopy.getMathematica(OPT.MATH)
														+ CONST.MATH_BRACKET_CLOSE;
		// console.log(termMathematicaTxt);

		spectralTxtArr.push(termMathematicaTxt);
	}; // term

	// construct global mathematica string
	let spectralMathematicaTxt = CONST.MATH_PLUS
															+ CONST.MATH_BRACKET_OPEN
															+ spectralTxtArr.join(', ')
															+ CONST.MATH_BRACKET_CLOSE;

	return spectralMathematicaTxt;
};



// create data for basis at different level
// starting from a pattern
FC.createBasisData = function(initialPattern, targetLevel, u) {
	// arrBasis will collect all the basis from
	// initialLevel to targetLevel (included)
	// returned by this function
	let arrBasis = null;
	// objBasis will collect all the basis transformation data
	// returned by this function
	// key: string of pattern (using join('+'))
	// value: object {arrPattern, objNumber}
	// (object returned by FC.incrementArrayPattern update at each level)
	let objBasis = null;

	let initialLevel = initialPattern.length;
	if (initialLevel < targetLevel) {
		// init with the initial pattern
		arrBasis = [initialPattern];
		objBasis = {};
		// initial data
		// objBasis[initialPatternStr] = {arrPattern:[initialPattern], arrNumber: [1]};
		// loop to increase basis size
		// at each level, collect more basis
		// last basis added => add data after loop
		let lastArrBasis = null;
		for (let i=initialLevel; i<targetLevel; i++) {
			// loop over basis collected so far
			for (let currentBasis of arrBasis) {
				// get already collected data for this basis
				let currentBasisStr = currentBasis.join('+');
				let currentBasisObj = objBasis[currentBasisStr];
				let nextObj = null;
				if (currentBasisObj !== undefined) {
					// this basis has been incremented before
					// compute next increment
					let arrPattern = currentBasisObj.arrPattern;
					let objNumber = currentBasisObj.objNumber;
					// to collect data for this current basis
					let nextBasisObj = {arrPattern: [], objNumber: {}};
					// loop over all pattern produced so far from this basis
					for (let pattern of arrPattern) {
						let patternStr = pattern.join('+');
						let n = objNumber[patternStr];
						nextObj = FC.incrementArrayPattern(pattern, u);
						// display data in console
						// console.log('pattern: ' + JSON.stringify(pattern));
						// for (let nextPattern of nextObj.arrPattern) {
						// 	console.log(' => ' + JSON.stringify(nextPattern));
						// };
						// console.log('');
						// update nextBasisObj with these data
						for (let newPattern of nextObj.arrPattern) {
							let newPatternStr = newPattern.join('+');
							let newObjNbr = nextBasisObj.objNumber;
							if (newObjNbr[newPatternStr] !== undefined) {
								// this new pattern has already been added
								// update only number
								newObjNbr[newPatternStr] = newObjNbr[newPatternStr] + nextObj.objNumber[newPatternStr]*n;
							} else {
								// this new pattern has not been added before
								// add the pattern
								nextBasisObj.arrPattern.push(newPattern);
								// add number value
								newObjNbr[newPatternStr] = nextObj.objNumber[newPatternStr]*n;
							};
						};
					};
					// replace the data for this basis
					objBasis[currentBasisStr] = nextBasisObj;
				} else {
					// this basis has not been incremented yet
					// this is a new basis added before
					nextObj = FC.incrementArrayPattern(currentBasis, u);
					objBasis[currentBasisStr] = nextObj;
				};
			};
			// the new array of basis at this level is inherited from "initialPattern"
			let initialPatternStr = initialPattern.join('+');
			let newArrBasis = FH.copyData(objBasis[initialPatternStr].arrPattern);
			lastArrBasis = newArrBasis;
			// add to array of basis
			arrBasis = arrBasis.concat(newArrBasis);
			// console.log('current arrBasis: ' + JSON.stringify(arrBasis));
		};
		// add data for the last array of basis
		// this basis is the "target one" => trivial data
		for (let basis of lastArrBasis) {
			let basisStr = basis.join('+');
			// the basis maps to itself with number 1
			let objNumber = {};
			objNumber[basisStr] = 1;
			let basisObj = {arrPattern: [basis], objNumber: objNumber};
			objBasis[basisStr] = basisObj;
		};
	} else {
		console.log("nothing to do…");
	};

		return {arrBasis: arrBasis, objBasis: objBasis};
};




// create the transition matrix from a basis at level initialLevel
// to a basis at level targetLevel for the (minimum) pattern initialPattern
// returns a double array of integers which can be converted to
// a Mathematica matrix using FH.convertDoubleArrayToMathematicaString
// returns the two associated basis (as arrays of patterns)
FC.getTransitionMatrixAndBasis = function(initialPattern, initialLevel, targetLevel, u) {
	let {arrBasis, objBasis} = FC.createBasisData(initialPattern, targetLevel, u);
	let initialPatternStr = initialPattern.join('+');
	let finalBasis = objBasis[initialPatternStr].arrPattern;
	// collect initial basis at initialLevel
	let initialBasis = [];
	for (let basis of arrBasis) {
		if (basis.length === initialLevel) {
			initialBasis.push(basis);
		};
	};
	let matrix = [];
	// console.log('finalBasis: ' + JSON.stringify(finalBasis));
	// console.log('initialBasis: ' + JSON.stringify(initialBasis));
	// fill O in matrix
	for (let fBasis of finalBasis) {
		let rowNumber = [];
		for (let iBasis of initialBasis) {
			rowNumber.push(0);
		};
		matrix.push(rowNumber);
	};
	// fill matrix with values
	let row = 0;
	let col = 0;
	for (let iBasis of initialBasis) {
		let iBasisStr = iBasis.join('+');
		let iBasisData = objBasis[iBasisStr];
		let iBasisArrPattern = iBasisData.arrPattern;
		let iBasisObjNumber = iBasisData.objNumber;
		row = 0;
		for (let fBasis of finalBasis) {
			let fBasisStr = fBasis.join('+');
			if (iBasisObjNumber[fBasisStr] !== undefined) {
				let n = iBasisObjNumber[fBasisStr];
				matrix[row][col] = n;
			};
			row++;
		};
		col++;
	};

	return {matrix:matrix, initialBasis:initialBasis, finalBasis:finalBasis};
};






FC.getFullTransitionMatrixAndBasis = function(initialPattern, targetLevel, u) {
	let {arrBasis, objBasis} = FC.createBasisData(initialPattern, targetLevel, u);
	// console.log('arrBasis: ' + JSON.stringify(arrBasis));
	// console.log('objBasis: ' + JSON.stringify(objBasis));
	let initialPatternStr = initialPattern.join('+');
	let finalBasis = objBasis[initialPatternStr].arrPattern;

	let matrix = [];
	// fill O in matrix
	for (let fBasis of finalBasis) {
		let rowNumber = [];
		for (let iBasis of arrBasis) {
			rowNumber.push(0);
		};
		matrix.push(rowNumber);
	};

	// fill matrix with values
	let row = 0;
	let col = 0;
	for (let iBasis of arrBasis) {
		let iBasisStr = iBasis.join('+');
		let iBasisData = objBasis[iBasisStr];
		let iBasisArrPattern = iBasisData.arrPattern;
		let iBasisObjNumber = iBasisData.objNumber;
		row = 0;
		for (let fBasis of finalBasis) {
			let fBasisStr = fBasis.join('+');
			if (iBasisObjNumber[fBasisStr] !== undefined) {
				let n = iBasisObjNumber[fBasisStr];
				matrix[row][col] = n;
			};
			row++;
		};
		col++;
	};

	return {matrix:matrix, finalBasis:finalBasis};
};





FC.showSubstitution = function(substitutionData) {
	let orgItem = substitutionData.orgItem;
	// take first derivation
	let derivationElement = substitutionData.derivationOperators[0];
	let newItemsByDerivatives = substitutionData.newItemsByDerivatives;
	let latexReturnedText = '';


	// used to display substitutions to 0
	let zero = new MathElement();
	zero
	.setID('0')
	.setLetter('0')
	.setLaTeXLetter('0')
	.setProductKindToCommutative()
	.setSumWeight(1)
	.setProductWeight(1);


	// determine in which part of MathArgument the substitution takes place
	let termElement = null;
	if (substitutionData.termElement === undefined) {
		termElement = CONST.MAIN_ELEMENT;
	} else {
		termElement = substitutionData.termElement;
	};

	// min level of derivations to consider
	// no value => 0
	let minLevel = null;
	if (substitutionData.minLevel === undefined) {
		minLevel = 0;
	} else {
		minLevel = substitutionData.minLevel;
	};
	// max level of derivations to consider
	let maxLevel = substitutionData.maxLevel;

	// general tests:
	// orgItem should be MathElement
	if (orgItem.constructor.name !== 'MathElement') {
		throw "MathElement for original item required";
	};
	// substitute items must be MathProduct or 'null'
	// length of perm in permutationsSum = length of mapIndices
	for (let newItemDerivativeData of newItemsByDerivatives) {
		for (let newItemData of newItemDerivativeData) {
			// check that each item is MathProduct if not 'null'
			if (newItemData.item !== null) {
				// newItemData.item.showConsole(OPT.SHOW, 'newItemData.item = ');
				if (newItemData.item.constructor.name !== 'MathProduct') {
					throw "MathProduct for new item required";
				};
				let l = newItemData.mapIndices.length;
				for (let perm of newItemData.permutationsSum) {
					if (perm.length !== l) {
						throw "Inconsistancy in permutationsSum";
					};
				};
			}; // test not 'null'
		};
	};

	// display original element with indices
	let orgFactor = MathProduct.newFrom(orgItem);
	orgFactor.fillNormalizedIndicesArray();
	orgFactor.showNormalized(OPT.SHOW, 'Original element: ');
	console.log('');

	let orgItemDerivatives = [];
	// loop from O to maxLevel
	for (let nDerivations = 0; nDerivations<maxLevel+1; nDerivations++) {
		let latexEquation = CONST.LATEX_BEGIN_EQUATION + CONST.CR;
		let latexObj = null;
		let latex = null;
		let sign = null;
		let signTxt = '';

		if (nDerivations === 0) {
			orgItemDerivatives.push(orgItem.makeCopy());
		} else {
			let currentDerivation = derivationElement.makeCopy();
			currentDerivation.applyTo(orgItemDerivatives[nDerivations-1]);
			orgItemDerivatives.push(currentDerivation);
		};

		// nothing unless we are >= minLevel
		// just fill orgItemDerivatives (above) to construct successives derivations
		if (nDerivations < minLevel) {
			continue;
		};

		console.log('level of derivations: ' + nDerivations);

		let currentFactor = MathProduct.newFrom(orgItemDerivatives[nDerivations]);
		currentFactor.fillNormalizedIndicesArray();
		currentFactor.showNormalized(OPT.SHOW, '');
		let currentIndiceIDtoIndiceObj = currentFactor.normalizedIndices;
		// console.log('currentIndiceIDtoIndiceObj: ' + JSON.stringify(currentIndiceIDtoIndiceObj));

		if (newItemsByDerivatives[nDerivations] === undefined) {
			console.log('No data for this level');
			// next derivation level
			continue;
		};

		latexObj = currentFactor.getLaTeX(OPT.LTX);
		latex = latexObj.ltx;
		sign = latexObj.sign;
		signTxt = '';
		if (sign<0) {
			signTxt = '-';
		};
		latexEquation += signTxt + latex + CONST.CR;
		latexEquation += CONST.LATEX_RIGHTARROW
													+ CONST.LATEX_ALIGN_SEP
													+ '{}'
													+ CONST.CR;

		console.log(' => ');

		let latexJoinList = CONST.LATEX_SUM_NEWLINE
											+ CONST.CR
											+ CONST.LATEX_ALIGN_SEP;
		let latexNewItemArray = [];
		for (let newItemData of newItemsByDerivatives[nDerivations]) {
			let newItem = newItemData.item;
			let newItemCopy = null;
			if (newItem !== null) {
				let mapIndices = newItemData.mapIndices;
				let permutationsSum = newItemData.permutationsSum;
				// console.log('mapIndices: ' + JSON.stringify(mapIndices));
				// console.log('permutationsSum: ' + JSON.stringify(permutationsSum));
				// loop over permutations
				for (let perm of permutationsSum) {
					// console.log('permutation: ' + JSON.stringify(perm));
					// local copy for this permutation
					newItemCopy = newItem.makeCopy();
					// create data "normalizedIndices"
					newItemCopy.fillNormalizedIndicesArray();
					// get the data "normalizedIndices"
					let newIndiceIDtoIndiceObj = newItemCopy.normalizedIndices;
					// console.log('newIndiceIDtoIndiceObj (before): ' + JSON.stringify(newIndiceIDtoIndiceObj));
					// adjust "normalizedIndices" to reproduce permutation
					for (let i=0; i<mapIndices.length; i++) {
						// add leading 0 since now in MathProduct
						let fromArrIdx = [0].concat(mapIndices[i].from);
						// console.log('fromArrIdx: ' + JSON.stringify(fromArrIdx));
						let fromIndiceID = currentFactor.getTextIndiceAt(fromArrIdx);
						let eFrom = currentIndiceIDtoIndiceObj[fromIndiceID];
						// console.log('eFrom: ' + JSON.stringify(eFrom));
						// use 'perm' to select 'to'
						let toArrIdx = mapIndices[perm[i]].to;
						// console.log('toArrIdx: ' + JSON.stringify(toArrIdx));
						let toIndiceID = newItemCopy.getTextIndiceAt(toArrIdx);
						let eTo = newIndiceIDtoIndiceObj[toIndiceID];
						// console.log('eTo (before): ' + JSON.stringify(eTo));
						eTo.number = eFrom.number;
						// console.log('eFrom.number: ' + eFrom.number);
						// console.log('eTo (after): ' + JSON.stringify(eTo));
						// console.log('map: ' + fromIndiceID + ' => ' + toIndiceID);
					};
					// console.log('newIndiceIDtoIndiceObj (after): ' + JSON.stringify(newIndiceIDtoIndiceObj));
					newItemCopy.showNormalized(OPT.SHOW, '     ');

					latexObj = newItemCopy.getLaTeX(OPT.LTX);
					latex = latexObj.ltx;
					sign = latexObj.sign;
					signTxt = '+';
					if (sign<0) {
						signTxt = '-';
					};
					let latexNewItemText = signTxt + latex;

					latexNewItemArray.push(latexNewItemText);
				};
			} else {
				newItemCopy = MathProduct.newFrom(zero);
				newItemCopy.showNormalized(OPT.SHOW, '     ');

				latexObj = newItemCopy.getLaTeX(OPT.LTX);
				latex = latexObj.ltx;
				let latexNewItemText = latex;

				latexNewItemArray.push(latexNewItemText);
			}; // if newItem null or 0
		}; // newItemData
		latexEquation += latexNewItemArray.join(latexJoinList);
		latexEquation += CONST.CR + CONST.LATEX_END_EQUATION;
		// console.log(latexEquation);

		console.log('');
		latexReturnedText += latexEquation + CONST.CR + CONST.CR;
	}; // nDerivations
	return latexReturnedText;
};

// used to get a LaTeX representation of substitutions rules
FC.getLaTeXSubstitution = function(opt, orgSum, substitutionData) {
	let localOrgSum = orgSum.makeCopy();
	let optNoLineNumber = FH.copyData(opt);
	optNoLineNumber.showLineNumbers = false;
	let latexTxt = '';
	latexTxt += CONST.CR + CONST.LATEX_BEGIN_EQUATION + CONST.CR;
	localOrgSum.sortBySumWeight();
	localOrgSum.fillNormalizedIndicesArray();
	latexTxt += localOrgSum.list[0].getLaTeXInEquation(optNoLineNumber, '');
	latexTxt += CONST.LATEX_RIGHTARROW + CONST.CR;
	localOrgSum.substituteItemAndDerivativesWithData(substitutionData);
	localOrgSum.fillNormalizedIndicesArray();
	latexTxt += localOrgSum.getLaTeXInEquation(opt, '');
	latexTxt += CONST.CR + CONST.LATEX_END_EQUATION + CONST.CR;
	return latexTxt;
};

module.exports = {
	FC: FC,
};
