/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Rational,
				Integer
			} = require(REL_PATH + 'MATH-OBJECTS/MathRational');


// all the functions defined here are properties of the object FF
let FF = {};

// remove extra zeros in arr after last non zero element
// if array contains only zeros, return [0]
FF.cleanZeros = function(arr) {
	let p = arr.length-1;
	let c = arr[p];
	while ( p>0 && (c === null || c == 0) ) {
		arr.splice(p,1);
		p--;
		// console.log('p='+p);
		c = arr[p];
	};
	return FF.fillArrayWithZeros(arr);
};

FF.fillArrayWithZeros = function(arr) {
	let filledArr = [];
	// console.log('arr: ' + JSON.stringify(arr));
	for (let i=0; i<arr.length; i++) {
		// console.log(arr[i]);
		if (arr[i] !== undefined && arr[i] !== null ) {
			filledArr.push(arr[i]);
		} else {
			filledArr.push(0);
		};
	};
	return filledArr;
};


// evaluate at Rational r "as if" 'arr' is a polynomials
// coefficients in array in increasing powers of monomials
// array is required to be filled with 0 for null monomials
// this not checked
FF.evaluateArrayAtRational = function(arr, r) {
	let a = arr[0];
	let result = new Rational(a,1);
	if (arr.length !== 1) {
		let arrNext = FH.copyData(arr);
		// remove first element
		arrNext.shift();
		let rToAdd = FF.evaluateArrayAtRational(arrNext, r).multiplyBy(r);
		result = result.addTo(rToAdd);
	};
	return result;
};

// multiply array by integer as if it was polynomials
// coefficients in array in increasing powers of monomials
// arrays are required to be filled with 0 for null monomials
FF.multiplyArrayByInteger = function(arr1, n) {
	if (!Number.isInteger(n)) {
		throw "multiply by integer";
	}
	let arr = [];
	let c = null;
	for (let i=0; i<arr1.length; i++) {
		c = arr1[i];
		if (c !== undefined && c !== null) {
			arr[i] = c*n;
		} else {
			arr[i] = 0;
		};
	};
	FF.cleanZeros(arr);
	return arr;
};


// add arrays as if they were polynomials
// coefficients in arrays in increasing powers of monomials
// arrays are required to be filled with 0 for null monomials
FF.addArrays = function(arr1, arr2) {
	let arr = [];
	let c = null;
	for (let i=0; i<arr1.length; i++) {
		c = arr1[i];
		if (c !== undefined && c !== null) {
			arr[i] = c;
		};
	};
	for (let i=0; i<arr2.length; i++) {
		c = arr2[i];
		if (c !== undefined && c !== null) {
			if (arr[i] === undefined) {
				arr[i] = c;
			} else {
				arr[i] += c;
			};
		};
	};
	FF.cleanZeros(arr);
	return arr;
};

// multiply arrays as if they were polynomials
// coefficients in arrays in increasing powers of monomials
// arrays are required to be filled with 0 for null monomials
FF.multiplyArrays = function(arr1, arr2) {
	let arr = [];
	let c1 = null;
	let c2 = null;
	for (let i=0; i<arr1.length; i++) {
		c1 = arr1[i];
		if (c1 !== undefined && c1 !== null) {
			for (let j=0; j<arr2.length; j++) {
				c2 = arr2[j];
				if (c2 !== undefined && c2 !== null) {
					if (arr[i+j] === undefined) {
						arr[i+j] = c1*c2;
					} else {
						arr[i+j] += c1*c2;
					};
				};
			};
		};
	};
	FF.cleanZeros(arr);
	return arr;
};


// Polynomials Long division
// divide arr1 by arr2 as if they were polynomials
// return arrQ for quotient and test if rest is zero
// https://math.stackexchange.com/questions/215734/pseudo-code-for-polynomial-long-division
FF.divideArrays = function(arr1, arr2) {
	let arrP = FF.cleanZeros(arr1);
	let arrD = FF.cleanZeros(arr2);
	// console.log('arrP: ' + JSON.stringify(arrP));
	// console.log('arrD: ' + JSON.stringify(arrD));
	let arrQ = [];
	let arrR = arrP;
	if (arrD.length>0 && arrD[arrD.length-1] !== 0 && arrP.length>=arrD.length) {
		// divisor non zero and degree of divisor less than of polynomials
		while (arrR[arrR.length-1] !== 0 && arrR.length>=arrD.length) {
			// console.log('arrQ: ' + JSON.stringify(arrQ));
			// console.log('arrR: ' + JSON.stringify(arrR));
			let l = arrR.length - arrD.length + 1;
			let arrT = FH.createNullArrayWithLength(l);
			let c = arrR[arrR.length-1]/arrD[arrD.length-1];
			if (!Number.isInteger(c)) {
				throw "arrays can't be divided";
			};
			arrT[arrT.length-1] = c;
			arrQ = FF.addArrays(arrQ, arrT);
			let arrS = FF.multiplyArrays(arrT, arrD);
			arrS = FF.multiplyArrayByInteger(arrS, -1);
			arrR = FF.addArrays(arrR, arrS);
		};
	} else {
		throw "divisor is not correct";
	};
	// return non empty rest array
	if (arrR.length === 0) {
		arrR = [0];
	};
	// test if arrR = [0]
	if (arrR.length === 1 && arrR[0] === 0) {
		return arrQ;
	} else {
		throw "remainder is non zero";
	};
};

// find factors of n, including 1
FF.getListOfFactors = function(n) {
	let arr = [];
	// let max = Math.ceil(n/2) + 1;
	let max = Math.abs(n);
	for (let i=1; i<=max; i++) {
		// console.log('i = ' + i);
		if (n%i === 0) {
			arr.push(i);
		};
	};
	return arr;
};

// test if 'Rational' r is in the array of 'Rational' arr
FF.inRationalArray = function(arr, r) {
	let found = false;
	for (let rTest of arr) {
		if (rTest.isEqual(r)) {
			found = true;
		};
	};
	return found;
};


// main function: return list of factors (as arrays) and their multiplicity
// all arrays are treated as polynomials:
// coefficients in array in increasing powers of monomials
FF.getFactorsFromArray = function(arr) {
	let globalFactor = Math.abs(FH.getGCDArray(arr));
	// put sign of max degree coeff to this global factor
	let sign = Math.sign(arr[arr.length-1]);
	globalFactor = globalFactor*sign;
	// console.log('globalFactor: ' + globalFactor);
	let arrReduced = FH.copyData(arr);
	for (let i=0; i<arrReduced.length; i++) {
		arrReduced[i] = arrReduced[i]/globalFactor;
	};
	// console.log('arrReduced: ' + JSON.stringify(arrReduced));

	let arrFactors = [];

	// collect 0 root factors
	let p=0;
	let addToFactors = false;
	while (arrReduced[0] === 0) {
		p++;
		addToFactors = true;
		arrReduced.shift();
	};
	// collect data
	if (addToFactors) {
		arrFactors.push({arr: [0,1], mult: p});
	};

	// construct possible numerators and denominator
	// see Rational Root Theorem (wikipedia for instance)
	let pArr = FF.getListOfFactors(arrReduced[0]);
	let cMax = arrReduced[arrReduced.length-1];
	let qArr = FF.getListOfFactors(cMax);
	// collect non equal rationals
	let rationalsArr = [];
	for (let q of qArr) {
		for (let p of pArr) {
			let r = new Rational(p,q);
			if (!FF.inRationalArray(rationalsArr, r)) {
				rationalsArr.push(r);
				let rn = r.multiply(-1);
				rationalsArr.push(rn);
			};
		};
	};

	// collect roots and construct factors with multiplicity
	let arrQ = arrReduced;
	// console.log('look for roots:')
	for (let r of rationalsArr) {
		// console.log('arrQ' + JSON.stringify(arrQ));
		// console.log('is integer: ' + Rational.isInteger(r));
		let n = FF.evaluateArrayAtRational(arrQ, r);
		// n.showConsole(OPT.SHOW, 'value: ');
		let isRoot = n.isZero() ? true : false;
		let addToFactors = isRoot;
		let arrD = [-r.numerator, r.denominator];
		// console.log('divisor: ' + JSON.stringify(arrD));
		// count multiplicity
		let p = 0;
		while (isRoot) {
			p++;
			arrQ = FF.divideArrays(arrQ, arrD);
			n = FF.evaluateArrayAtRational(arrQ, r);
			isRoot = n.isZero() ? true : false;
		};
		// collect data
		if (addToFactors) {
			arrFactors.push({arr: arrD, mult: p});
		};
		// console.log('');
	};
	// prepare final result
	if (arrQ.length === 1) {
		globalFactor = globalFactor*arrQ[0];
	} else {
		arrFactors.push({arr: arrQ, mult: 1});
	};
	// return results
	return {globalFactor: globalFactor, arrFactors: arrFactors};
};

// return the array corresponding to the factors data
// all arrays are treated as polynomials:
// coefficients in array in increasing powers of monomials
FF.getArrayFromFactors = function(factorsObj) {
	let {globalFactor: globalFactor, arrFactors: arrFactors} = factorsObj;
	// start with 'polynomials' 1
	let arr = [1];

	// loop over arrFactors
	// and multiply arrays as many times as multiplicity
	for (let factorData of arrFactors) {
		arrF = factorData.arr;
		multF = factorData.mult;
		for (let i=0; i<multF; i++) {
			arr = FF.multiplyArrays(arr, arrF);
		};
	};

	// multiply by global factor
	arr = FF.multiplyArrayByInteger(arr, globalFactor);

	return arr;
};

// test arr and its factor decomposition
FF.testFactorDecompositionOfArray = function(factorsObj, arr) {
	let arrProduct = FF.getArrayFromFactors(factorsObj);
	return JSON.stringify(arrProduct) === JSON.stringify(arr);
};


module.exports = {
	FF: FF,
};
