/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument,
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

const {PROTO} = require(REL_PATH + 'INIT-FILES/initialization-data');

// all the functions defined here are properties of the object FI
const FI = {};

// TODO: create combinaisons of derivations used in results
FI.inputDictionnary = {
	'H': PROTO.H,
	'L': PROTO.L,
	'N': PROTO.N,
	'q': PROTO.q,
	'u': PROTO.u,
	'F': PROTO.F,
	'FU': PROTO.FU,
	'g': PROTO.g,
	'gi': PROTO.gi,
	'R': PROTO.RT,
	'RU': PROTO.RTU,
	'R2': PROTO.RTSQUARE,
	'Ric': PROTO.Ric,
	'RicU': PROTO.RicU,
	'Ric2': PROTO.RicSQUARE,
	'S': PROTO.SC,
	'S2': PROTO.SCSQUARE,
	'hNu': PROTO.hnablau,
	'hN1u': PROTO.hnabla1u,
	'hN2u': PROTO.hnabla2u,
	'hN3u': PROTO.hnabla3u,
	'hN4u': PROTO.hnabla4u,
	'hL1u': PROTO.hlaplacian1u,
	'hL2u': PROTO.hlaplacian2u,
	'hSN2u': PROTO.Shnabla2u,
	'hSN2Uu': PROTO.Shnabla2Uu,
	'hNLu': PROTO.hnablahlaplacianu,
	'hLNu': PROTO.hlaplacianhnablau,
	'hNLNu': PROTO.hnablahlaplacianhnablau,
	'comFu': PROTO.comFu,
	'comFUu' : PROTO.comFUu,
	'anticomHLHNu': PROTO.anticomhlaplacianhnablau,
	'hN1N': PROTO.hnabla1N,
	'hN2N': PROTO.hnabla2N,
	'hN3N': PROTO.hnabla3N,
	'hL1N': PROTO.hlaplacian1N,
	'hNLN': PROTO.hnablahlaplacianN,
	'hLNN': PROTO.hlaplacianhnablaN,
	'hSN2N': PROTO.Shnabla2N,
	'comFN' : PROTO.comFN,
	'anticomHLHNN': PROTO.anticomhlaplacianhnablaN,
	'hN1q': PROTO.hnabla1q,
	'hN2q': PROTO.hnabla2q,
	'hL1q': PROTO.hlaplacian1q,
	'N31': PROTO.nabla31,
	'hN1F': PROTO.hnablaF,
	'hN1R': PROTO.hnabla1RT,
	'hN2R': PROTO.hnabla2RT,
	'hNSC': PROTO.hnablaSC,
	'hLSC': PROTO.hlaplacianSC,
	'sgi': PROTO.sumgigi,
	'sg': PROTO.sumgg,
};



// various helpful functions
FI.removeSpaces = function(txt) {
	return txt.replace(/\s/g, '');
};
FI.removeExtParenthesis = function(txt) {
	let newTxt = txt.replace(/[\s]*[(]/g, '');
	newTxt = newTxt.replace(/[)][\s]*/g, '');
	return newTxt;
};
FI.removeExtBraces = function(txt) {
	let newTxt = txt.replace(/[\s]*[{]/g, '');
	newTxt = newTxt.replace(/[}][\s]*/g, '');
	return newTxt;
};
FI.checkIndices = function(el, indicesArr) {
	let elIndicesArrIdx = el.getIndicesArrIdx();
	// console.log('elIndicesArrIdx: ' + JSON.stringify(elIndicesArrIdx));
	if (elIndicesArrIdx.length != indicesArr.length) {
		throw "Numbers of indices do not coincide";
	};
	for (let i=0; i<elIndicesArrIdx.length; i++) {
		// console.log('arrIdx: ' + JSON.stringify(elIndicesArrIdx[i]));
		let elIndicePos = el.getIndicePositionAt(elIndicesArrIdx[i]);
		if (indicesArr[i].pos != elIndicePos) {
			throw "Indices positions do not match";
		};
	};
	return true;
};

// parse a string to a RationalExpression
// ex: "(2 d^10 - d + 4) / 16"
// ex: "-2d^2 + 3d - 4"
FI.parseRationalExpression = function(string) {

	collectTerms = function(txt) {
		let regexp = /([+-]{0,1}[\d\w\^{}]+)/g;
		return txt.match(regexp);
	};

	getGlobalSign = function(txt) {
		let regexp = /[\s]*([+-]*)[\s]*[(]/g;
		let signArr = regexp.exec(txt);
		let sign = '+';
		if (signArr != null) {
			sign = signArr[1];
		};
		return sign;
	};

	removeGlobalSign = function(txt) {
		let regexp = /[\s]*[+-]*[\s]*[(]/g;
		let newTxt = txt.replace(regexp, '(');
		return newTxt;
	};

	decomposeMonomials = function(txt) {
		let regexpIntCoeff = /^[^\^+-\d]{0,1}([+-]*\d*)/g;
		let regexpLetter = /(['d'])/g;
		let regexpPower = /[\^]([{}\d]*)/g;
		let intCoeffArr = regexpIntCoeff.exec(txt);
		// let letterArr = regexpLetter.exec(txt);
		let letterIdx = txt.indexOf(CONST.DIMENSION_LETTER);
		let powerArr = regexpPower.exec(txt);
		// console.log("monomials: " + txt);
		// console.log("intCoeff: " + JSON.stringify(intCoeffArr));
		// console.log("power: " + JSON.stringify(powerArr));
		// analyse data
		let intCoeff = null;
		if (intCoeffArr != null) {
			let intCoeffTxt = intCoeffArr[1];
			if (intCoeffTxt === "+" || intCoeffTxt === "") {
				intCoeff = 1;
			} else if (intCoeffTxt === "-") {
				intCoeff = -1;
			} else {
				intCoeff = parseInt(intCoeffTxt);
				if (isNaN(intCoeff)) {
					throw "intCoeff is not an integer";
				};
			};
		} else {
			intCoeff = 1;
		};
		// console.log("intCoeff is the integer = " + intCoeff);

		let power = null;
		if (letterIdx == -1) {
			// constant term
			power = 0;
		} else {
			// non constant term
			if (powerArr != null) {
				// power > 1
				let powerTxt = powerArr[1];
				powerTxt = FI.removeExtBraces(powerTxt);
				power = parseInt(powerTxt);
				if (isNaN(power)) {
					throw "power is not an integer";
				};
			} else {
				// power = 1
				power = 1;
			};
		};

		// console.log("power is the integer = " + power);

		return {coeff: intCoeff, power: power};
	};

	// main part of the function
	// first find numerator and denominator
	// they are separated by "/"
	// console.log('parse RE, string = ' + string);
	let num = '';
	let den = '';
	let sign = '+';
	if (FI.removeSpaces(string) === "") {
		string = "1";
	};
	let idx = string.indexOf("/");
	// console.log('/ found at ' + idx);
	if (idx == -1) {
		// no denominator
		den = '1';
		num = string;
	} else {
		num = string.substring(0,idx);
		den = string.substring(idx+1);
	};
	// get global sign
	sign = getGlobalSign(num);
	// remove global sign
	num = removeGlobalSign(num);
	// remove spaces
	num = FI.removeSpaces(num);
	// remove external parenthesis
	num = FI.removeExtParenthesis(num);
	// remove spaces and external parenthesis
	den = FI.removeSpaces(den);
	den = FI.removeExtParenthesis(den);

	// console.log("num is |" + num + "|");
	// console.log("den is |" + den + "|");

	let denominator = parseInt(den);
	if (!isNaN(denominator)) {
		// console.log("denominator is the integer = " + denominator);
	} else {
		throw "denominator is not an integer";
	};

	let numArr = collectTerms(num);
	// console.log("terms of num: " + JSON.stringify(numArr));
	let numerator = null;
	if (numArr.length > 0) {
		numerator = Polynomials.newFromLetterAndCoefficient(CONST.DIMENSION_LETTER,0);
		for (let monomial of numArr) {
			let {coeff, power} = decomposeMonomials(monomial);
			let newTerm = Polynomials.newFromLetterCoefficientAndPower(CONST.DIMENSION_LETTER,coeff,power);
			numerator = numerator.add(newTerm);
		};
	};

	// numerator.showConsole(OPT.SHOW, 'num = ');
	let returnedRE = new RationalExpression(numerator, denominator);
	if (sign == '-') {
		returnedRE.multiplyByInteger(-1);
	};
	return new RationalExpression(numerator, denominator);
};


// parse a string representing a list of factors
// with possible indices contractions
// TODO: take into account MathProduct inside MathTensorProduct using '(...)' syntax
FI.parseMathExpression = function(expression, kind) {

	// collect in an array the factors in the initial expression
	// factors are separated by comas
	// clean factors by removing external spaces
	collectFactors = function(txt) {
		let factorsArr = txt.split(',');
		// console.log(JSON.stringify(factorsArr));
		for (let i=0; i<factorsArr.length; i++) {
			factorsArr[i] = FI.removeSpaces(factorsArr[i]);
		};
		return factorsArr;
	};

	// return the key of the factor and the data on the indices
	decomposeFactor = function(txt){
		let newTxt = FI.removeSpaces(txt);
		// states of the parser
		let STATE_KEY = 0;
		let STATE_INDICE_UP = 1;
		let STATE_INDICE_DOWN = 2;
		// initial state is reading the key
		let state = STATE_KEY;
		// collected data
		let key = '';
		let indicesArr = [];

		// current data
		let label = '';
		let pos = 0;

		// 'state machine' parser: look at characters and
		// change state according to situtation
		// collect data on the way
		for (let i=0; i<newTxt.length; i++) {
			let c = newTxt[i];
			// console.log(c);

			// change state if c is '^' or '_'
			if (c == '^') {
				if (state == STATE_INDICE_UP || state == STATE_INDICE_DOWN) {
					// save previous indice
					indicesArr.push({label: label, pos: pos});
				};
				state = STATE_INDICE_UP;
				label = '';
				pos = CONST.INDICE_UP;
				continue;
			};
			if (c == '_') {
				if (state == STATE_INDICE_UP || state == STATE_INDICE_DOWN) {
					// save previous indice
					indicesArr.push({label: label, pos: pos});
				};
				state = STATE_INDICE_DOWN;
				label = '';
				pos = CONST.INDICE_DOWN;
				continue;
			};

			switch(state) {
				case STATE_KEY:
					key += c;
					break;
				case STATE_INDICE_UP:
					label += c;
					break;
				case STATE_INDICE_DOWN:
					label += c;
					break;
			};
			// console.log('i='+ i + ' => ' + JSON.stringify(indicesArr));
			// console.log('');
		};
		if (state == STATE_INDICE_UP || state == STATE_INDICE_DOWN) {
			// save last indice
			indicesArr.push({label: label, pos: pos});
		};
		// console.log(key);
		// console.log(JSON.stringify(indicesArr));

		return {key: key, indices: indicesArr};
	};


	// main part of the function
	let initialTxt = FI.removeSpaces(expression);
	// console.log(expression);

	let returnedElement = null;
	let returnedIndicesByLabel = {};
	switch(kind) {
		case 'MathProduct':
			returnedElement = MathProduct.newFrom(MathElement_one);
			break;
		case 'MathTensorProduct':
			returnedElement = MathTensorProduct.newFrom(MathElement_one);
			break;
		default:
			throw "Only 'MathProduct' or 'MathTensorProduct' accepted";
			break;
	};
	// returnedElement.showConsole(OPT.SHOW, 'returned element: ');
	if (initialTxt != '') {
		let factorsArr = collectFactors(initialTxt);
		// console.log(JSON.stringify(factorsArr));
		for (let i=0; i<factorsArr.length; i++) {
			let factor = factorsArr[i];
			let decomposition = decomposeFactor(factor);
			// console.log(JSON.stringify(decomposition));
			if (FI.inputDictionnary[decomposition.key] === undefined) {
				let msg = "unknown input " + decomposition.key;
				throw msg;
			};
			let el = FI.inputDictionnary[decomposition.key].makeCopy();
			if (kind === 'MathTensorProduct') {
				// promote el to a MathProduct
				el = MathExpression.convertTo(el, 'MathProduct');
			};
			// el.showConsole(OPT.SHOW, 'element: ');
			// add element to main list returnedElement
			returnedElement.list[i] = el;
			// check that indices are in correspondance (numbers and positions)
			if (FI.checkIndices(el, decomposition.indices)) {
				// console.log('indices checked');
			};
			// data in returnedIndicesByLabel
			let elIndicesArrIdx = el.getIndicesArrIdx();
			for (let j=0; j<elIndicesArrIdx.length; j++) {
				let arrIdx = elIndicesArrIdx[j];
				let indIDn = el.getTextIndiceAt(arrIdx);
				let label = decomposition.indices[j].label;
				// console.log('label: ' + label);
				// console.log('indIDn: ' + indIDn);
				if (returnedIndicesByLabel[label] !== undefined) {
					returnedIndicesByLabel[label].push(indIDn);
				} else {
					returnedIndicesByLabel[label] = [indIDn];
				};
			};
		};
	} else {
		// default value of the returned element: '1'
	};

	return {element: returnedElement, indicesByLabel: returnedIndicesByLabel};
};


FI.parseMathArgument = function(arr) {
	// arr should contain 4 strings:
	// [0]: global rational fraction (in 'd') to be parsed as a RationalExpression
	// [1]: preElement to be parsed as a MathProduct
	// [2]: main element to be parsed as a MathTensorProduct
	// [3]: postElement to be parsed as a MathProduct

	combineIndicesByLabelArray = function(...arrs) {
		let newArr = {};
		for (let obj of arrs) {
			// console.log('obj: ' + JSON.stringify(obj));
			for (let label in obj) {
				// console.log('label: ' + label);
				let indIDnArr = obj[label];
				// console.log('indIDnArr: ' + JSON.stringify(indIDnArr));
				if (newArr[label] !== undefined) {
					for (let indIDn of indIDnArr) {
						// console.log('indIDn: ' + indIDn);
						newArr[label].push(indIDn);
					};
				} else {
					// console.log('add array')
					newArr[label] = indIDnArr;
				};
			};
		};
	return newArr;
	};


	// check length
	if (arr.length !== 4) {
		throw "array does not contains correct information";
	};
	// check strings
	for (let el of arr) {
		if (! typeof el === 'string') {
			throw "array does not contain only strings"
		};
	};
	let obj = null;
	// get RationalExpression
	let globalRE = FI.parseRationalExpression(arr[0]);
	// globalRE.showConsole(OPT.SHOW, 'globalRE: ');
	// get preElement
	obj = FI.parseMathExpression(arr[1], 'MathProduct');
	let preElement = obj.element;
	let preElementIndByLabel = obj.indicesByLabel;
	// preElement.showConsole(OPT.SHOW, 'preElement: ');
	// get main part
	obj = FI.parseMathExpression(arr[2], 'MathTensorProduct');
	let main = obj.element;
	let mainIndByLabel = obj.indicesByLabel;
	// console.log('mainIndicesByLabel: ' + JSON.stringify(mainIndByLabel));
	// main.showConsole(OPT.SHOW, 'main: ');
	// main.prefactor.showConsole(OPT.SHOW, 'main.prefactor: ');
	// get postElement
	obj = FI.parseMathExpression(arr[3], 'MathProduct');
	let postElement = obj.element;
	let postElementIndByLabel = obj.indicesByLabel;
	// postElement.showConsole(OPT.SHOW, 'postElement: ');

	// create MathArgument (avoid copies of 'main' and other parts)
	let returnedMathArgument = MathArgument.newFrom(MathElement_one);
	returnedMathArgument.list = main.list;
	returnedMathArgument.preElement = preElement;
	preElement.prefactor = globalRE;
	returnedMathArgument.postElement = postElement;

	returnedMathArgument.updateExpression();

	let indByLabelArr = combineIndicesByLabelArray(preElementIndByLabel, mainIndByLabel, postElementIndByLabel);
	// console.log('indByLabelArr: ' + JSON.stringify(indByLabelArr));

	// add contractions
	let contractions = [];
	for (let label in indByLabelArr) {
		let e = indByLabelArr[label];
		if (e.length === 2) {
			contractions.push(e.sort(FH.alphaSort));
		};
		if (e.length > 2) {
			throw "label " + label + " appears more than twice";
		};
	};
	returnedMathArgument.contractions = contractions;

	returnedMathArgument.preElement.sortByProductWeight();

	return returnedMathArgument;
};



FI.parseMathSum = function(arr) {
	// arr contains arrays of MathArguments to be parsed by FI.parseMathArgument
	let returnedMathSum = MathSum.newFrom(MathElement_one);
	returnedMathSum.list = [];
	returnedMathSum.updateExpression();
	for (let termData of arr) {
		let el = FI.parseMathArgument(termData);
		returnedMathSum.addTo(el, false);
	};
	returnedMathSum.collectCommutativeElementsInPreElement();
	returnedMathSum.sortPreElementsByProductWeight();
	returnedMathSum.sortBySumWeight();
	returnedMathSum.fillNormalizedIndicesArray();
	returnedMathSum.simplifyExpression();

	return returnedMathSum;
};



module.exports = {
	FI: FI,
};
