/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Polynomials,
				RationalExpression,
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument,
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');


const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');
const {PROTO_NCT, SUBST_NCT} = require(REL_PATH + 'INIT-FILES/initialization-data-NCT');
const {FR} = require(REL_PATH + 'FUNCTIONS-FILES/functions-rules');


let FR_NCT = {};


FR_NCT.contractionsMetricWithTwofTensors = function(term, rulesData) {
	let fiID = rulesData.id.fiID;
	let gID = rulesData.id.gID;
	let gi = rulesData.elements.gi;

	// change later if something is found
	// used in a loop which fires this function => repeat until nothing found
	let somethingFound = false;

	// the idea is to look at contractions:
	// find first contraction between gID and fiID
	let arrCont = term.contractions;
	let indiceIDtoIndiceObj = term.collectIndiceIDtoIndiceObj();

	for (let c of arrCont) {
		// source data to test
		let sObj = indiceIDtoIndiceObj[c[0]];
		let sID = sObj.elementID;
		// target data to test
		let tObj = indiceIDtoIndiceObj[c[1]];
		let tID = tObj.elementID;

		// test contraction: does it connect gID with fiID?
		// no test on derivation…
		if (
				(sID === fiID && tID === gID)
				||
				(sID === gID && tID === fiID)
			) {

			// normalize: source is metric, target is f tensor
			if (sID === fiID) {
				let nObj = sObj;
				sObj = tObj;
				tObj = nObj;
				sArrIdx = sObj.arrIdx;
			};

			// find associated indiceID of the metric
			let sIndiceIDRoot = sObj.id;
			let siblingIndiceObjects = FH.getSiblingIndiceObjects(sIndiceIDRoot, indiceIDtoIndiceObj);

			// collect all targets of metric which are f tensors
			let targetRootIDArray = [];
			let targetIndiceObjectsArray = [];
			let performContraction = true;
			for (let s of siblingIndiceObjects) {
				// get target t associated to source s
				let sIndiceID = s.id + '-' + s.n;
				let tIndiceID = FH.getTargetIndiceIDFromSourceIndiceID(sIndiceID, arrCont);
				let t = indiceIDtoIndiceObj[tIndiceID];
				let tID = t.elementID;
				if (tID === fiID
							&& !targetRootIDArray.includes(t.id) ) {
					targetRootIDArray.push(t.id);
					targetIndiceObjectsArray.push(t);
				} else {
					// not in expected situation => flag to leave current contraction
					performContraction = false;
					break;
				};
			};

			if (!performContraction) {
				// next contraction
				continue;
			};

			// we found something
			somethingFound = true;

			let giCopy = gi.makeCopy();
			let giIndiceIDRoot = giCopy.getIndicesID();

			// contractions indicesID to remove (metric tensor)
			let toSuppressIndiceIDArr = [];
			// couples of old indiceID => new indice ID to replace in arrCont
			let mapIndiceIDArr = [];
			for (let i=0; i<siblingIndiceObjects.length; i++) {
				let s = siblingIndiceObjects[i];
				let sIndiceID = s.id + '-' + s.n;
				// collect in to suppress
				toSuppressIndiceIDArr.push(sIndiceID);
				// target f tensor indice object
				let t = targetIndiceObjectsArray[i];
				// get indice number in f tensor
				let tn1 = t.n;
				let tn2 = tn1 === 0 ? 1 : 0;
				// indiceID of the other indice in f tensor
				let tIndiceID = t.id + '-' + tn2;
				// new indiceID in f tensor
				let giIndiceID = giIndiceIDRoot + '-' + i;
				// map tIndiceID => giIndiceID
				mapIndiceIDArr.push([ tIndiceID, giIndiceID ]);
			};

			// new contractions array: remove all contractions from metric
			let newArrCont = [];
			for (let cont of arrCont) {
				if (!toSuppressIndiceIDArr.includes(cont[0])
						&& !toSuppressIndiceIDArr.includes(cont[1])) {
					newArrCont.push(cont);
				};
			};

			// update contractions using mapIndiceIDArr
			let strNewArrCont = JSON.stringify(newArrCont);
			for (let map of mapIndiceIDArr) {
				let re = new RegExp(map[0], "g");
				strNewArrCont = strNewArrCont.replace(re, map[1]);
			};

			// install new contractions
			term.contractions = JSON.parse(strNewArrCont);

			let preElement = term.preElement;
			// preElement.showConsole(OPT.SHOW, 'pre el. before: ');
			// change source with inverse metric copy (in preElement)
			let sIdx = sArrIdx[sArrIdx.length-1]
			if (sArrIdx[0] === -2) {
				// in pre element
				preElement.list[sIdx] = giCopy;
			} else {
				throw "metric should be in pre element";
			};
			// preElement.showConsole(OPT.SHOW, 'pre el. after: ');

			// remove all f tensors: there are in pre element
			// collect idx in array
			let preArrIdx = [];
			for (let t of targetIndiceObjectsArray) {
				let arrIdx = t.arrIdx;
				if (arrIdx[0] === -2) {
					// add idx in preElement
					preArrIdx.push(arrIdx[arrIdx.length-1]);
				};
			};
			// remove from right to left to maintain idx
			preArrIdx.sort().reverse();
			// remove from pre element
			for (let idx of preArrIdx) {
				preElement.list.splice(idx,1);
			};

			// multiply by -1 (since contraction gives minus sign)
			term.multiplyByInteger(-1);
			term.updateExpression();
		}; // if source & target
		// if something is found, quit contraction loop
		if (somethingFound) {break;}
	}; // loop contractions
	// used in the loop over all rule functions
	return somethingFound;
};

// rule: full contraction of inverse metric and delta^2 => 0
// use FR.ruleContractionsSymmetricAntisymmetric2Tensors
FR_NCT.ruleContractionsfTensorWithDoubleDerivation = function(term, rulesData) {
	let fiID = rulesData.id.fiID;
	let delta2ID = rulesData.id.delta2ID;
	return FR.ruleContractionsSymmetricAntisymmetric2Tensors(term, delta2ID, fiID);
};

module.exports = {
	FR_NCT: FR_NCT,
};
