/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');
const {FI} = require(REL_PATH + 'FUNCTIONS-FILES/functions-input');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument,
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

// all the functions defined here are properties of the object FR
const FR = {};



// apply predefined rules on 'term'
// 'arrayOfRules' is an array of rule functions with two arguments:
// - 'term' is the term to consider
// - 'rulesData' contains informations about elements to use in the rule
// => rulesData collects all the items used, minimal one is 'RULE_DATA' in 'initialization-data'
// 'term' is modified directly, nothing is returned
// the order of application of the rules can be important,
// it is given by the order in 'arrayOfRules'
// rules can return 'zero' terms => use test with break to prevent next rules to be tested
// the rules are functions given below
// these functions look similar, but no universal pattern can be found to collect them in only one
// some have been "factorized" using "general rules" with preformatted data given as argument
// this permits to apply new rules without many code
FR.applyArrayOfRulesOnTerm = function(arrayOfRules, term, rulesData) {
	for (let ruleFunction of arrayOfRules) {
		let somethingFound = true;
		let termIsZero = false;
		while (somethingFound) {
			term.fillNormalizedIndicesArray();
			somethingFound = ruleFunction(term, rulesData);
			// if (somethingFound) {
			// 	console.log('somethingFound');
			// }
			termIsZero = term.isZero();
			// if the term is zero, quit the while loop
			if (termIsZero) {break;}
		}; // while nothingFound
		// if the term is zero, quit the for loop
		if (termIsZero) {break;}
	}; // loop on rules
	return;
};


// GENERAL RULES: used by specific rules
// general rule: full contraction of two 2-tensors: one symmetric the other antisymmetric => 0
FR.ruleContractionsSymmetricAntisymmetric2Tensors = function(term, tensorSymID, tensorAntiSymID) {
	// change later if something is found
	// used in a loop which fires this function => repeat until nothing found
	let somethingFound = false;

	// the idea is to look at contractions:
	// find first contraction between tensorSymID and tensorAntiSymID
	let arrCont = term.contractions;
	let indiceIDtoIndiceObj = term.collectIndiceIDtoIndiceObj();

	// console.log('');
	// term.showConsole(OPT.SHOW, 'term: ');
	// console.log('indiceIDtoIndiceObj: ' + JSON.stringify(indiceIDtoIndiceObj));
	// console.log('');

	for (let c of arrCont) {
		// source data to test
		let sObj = indiceIDtoIndiceObj[c[0]];
		let sID = sObj.elementID;
		// target data to test
		let tObj = indiceIDtoIndiceObj[c[1]];
		let tID = tObj.elementID;

		// test contraction: does it connect tensorSymID with tensorAntiSymID?
		if (	(sID === tensorSymID && tID === tensorAntiSymID)
						||
						(sID === tensorAntiSymID && tID === tensorSymID)  ) {

			// get all indices of source
			let sIndiceIDRoot = sObj.id;
			let siblingIndiceObjects = FH.getSiblingIndiceObjects(sIndiceIDRoot, indiceIDtoIndiceObj);
			// console.log('siblingIndiceObjects: ' + JSON.stringify(siblingIndiceObjects));
			// console.log('');

			let doubleContraction = true;
			let tIndiceIDRoot = tObj.id;
			for (let s of siblingIndiceObjects) {
				// get target t associated to source s
				let sIndiceID = s.id + '-' + s.n;
				let tIndiceID = FH.getTargetIndiceIDFromSourceIndiceID(sIndiceID, arrCont);
				if (!tIndiceID.includes(tIndiceIDRoot)) {
					doubleContraction = false;
				};
			};

			if (!doubleContraction) {
				continue;
			}

			// we found something
			somethingFound = true;

			// put term to 0
			term.multiplyByInteger(0);
			term.updateExpression();
		}; // if source & target
		// if something is found, quit contraction loop
		if (somethingFound) {break;}
	}; // loop contractions
	// used in the loop over all rule functions
	return somethingFound;
};


// general rule: perform full contractions of 2 given tensor with same number of indices
// the contraction must produce a MathProduct element (not a sum)
// TODO: accept MathSum as result of full contraction
// => new MathElement
FR.ruleFullContractionTwoTensors = function(term, tensorData) {
	// tensorData: {
	// 		tensorFirst: MathElement => instance of the first tensor
	// 		tensorSecond: MathElement => instance of the second tensor
	// 		contractionsList: [
	// 							{
	// 								indiceConfigurations: [ "0123", ... ] => list of configuration of contractions
	//								contractedItem: MathProduct => instance of the item to replace the 2 tensors
	// 							}, ...
	// 										]
	// 	}

	// indiceConfigurations is an array of string of numbers
	// which reproduce the full contraction configuration:
	// if the order of first tensor indices position is (0,1,2,…)
	// and the order of the second tensor indices position is (n_0, n_1, n_2,…)
	// then the string is the concatenation of n_0, n_1, n_2,…
	// contractedItem is an instance of the MathProduct which replaces in these configurations the full contraction
	// it can be 0 (the number) => then contraction is null
	// if no string in one of the indiceConfigurations is found, then contraction is considered as 0

	// change later if something is found
	// used in a loop which fires this function => repeat until nothing found
	let somethingFound = false;

	// the idea is to look at contractions:
	// find first contraction between giID and FID
	let arrCont = term.contractions;
	let indiceIDtoIndiceObj = term.collectIndiceIDtoIndiceObj();

	let tensorFirst = tensorData.tensorFirst;
	let tensorSecond = tensorData.tensorSecond;
	let tensorFirstID = tensorFirst.getID();
	let tensorSecondID = tensorSecond.getID();
	let nIndices = tensorFirst.indices.list.length;

	// perform some tests
	// same number of indices
	if (tensorSecond.indices.list.length !== nIndices) {
		throw "the two tensors should have same number of indices";
	};

	// TODO: test down and up indices?

	// console.log('');
	// term.showConsole(OPT.SHOW, 'term before: ');
	// term.showNormalized(OPT.SHOW, 'term before: ');

	for (let c of arrCont) {
		// source data to test
		let sObj = indiceIDtoIndiceObj[c[0]];
		let sArrIdx = sObj.arrIdx;
		let sID = sObj.elementID;
		// target data to test
		let tObj = indiceIDtoIndiceObj[c[1]];
		let tArrIdx = tObj.arrIdx;
		let tID = tObj.elementID;

		// test contraction: does it connect tensorFirstID with tensorSecondID?
		// restriction on derivation of tensor down:
		// no derivation
		if (	(sID === tensorFirstID && sArrIdx[sArrIdx.length-1] >= 0
								&& tID === tensorSecondID && tArrIdx[tArrIdx.length-1] >= 0)
						||
						(sID === tensorSecondID  && sArrIdx[sArrIdx.length-1] >= 0
								&& tID === tensorFirstID && tArrIdx[tArrIdx.length-1] >= 0)  ) {

			// normalize: source is tensor first, target is tensor second
			if (sID === tensorSecondID) {
				let nObj = sObj;
				sObj = tObj;
				tObj = nObj;
				sArrIdx = sObj.arrIdx;
				tArrIdx = tObj.arrIdx;
			};

			// loop over indices of source
			// check if always contracted with target
			// OK => collect in targetCollection
			// not OK => 'continue' the main contraction loop
			// find associated indiceID to construct contraction configuration
			let sIndiceIDRoot = sObj.id;
			let siblingIndiceObjects = FH.getSiblingIndiceObjects(sIndiceIDRoot, indiceIDtoIndiceObj);

			// check that targets ID are all tensorSecondID
			let targetIndiceObjectsArray = [];
			let toSuppressIndiceIDArr = [];
			let performContraction = true;
			for (let s of siblingIndiceObjects) {
				// get target t associated to source s
				let sIndiceID = s.id + '-' + s.n;
				toSuppressIndiceIDArr.push(sIndiceID);
				let tIndiceID = FH.getTargetIndiceIDFromSourceIndiceID(sIndiceID, arrCont);
				let t = indiceIDtoIndiceObj[tIndiceID];
				let tID = t.elementID;
				if (tID === tensorSecondID) {
					targetIndiceObjectsArray[s.n] = t;
				} else {
					// not in expected situation => flag to leave current contraction
					performContraction = false;
					break;
				};
			};

			if (!performContraction) {
				// next contraction
				continue;
			};

			// we found something
			somethingFound = true;

			// console.log('sArrIdx: ' + JSON.stringify(sArrIdx));
			// console.log('Something found');

			// find configuration
			let stringConfig = '';
			for (let t of targetIndiceObjectsArray) {
				stringConfig += t.n;
			};
			// console.log('stringConfig: '+ stringConfig);
			let contractedItem = null;
			for (let contList of tensorData.contractionsList) {
				if (contList.indiceConfigurations.includes(stringConfig)) {
					// console.log('config found');
					contractedItem = contList.contractedItem;
					// quit loop on contList
					break;
				};
			};

			if (contractedItem === null || contractedItem === 0) {
				// considered as 0
				// => multiply by 0
				term.preElement.prefactor = term.preElement.prefactor.multiplyByInteger(0);
			} else {
				// contractedItem.showConsole(OPT.SHOW, 'new factor: ');
				let contractedItemCopy = contractedItem.makeCopy();

				let preElement = term.preElement;
				let postElement = term.postElement;

				// change source with tensorUp
				let sIdx = sArrIdx[sArrIdx.length-1];
				let tIdx = tArrIdx[tArrIdx.length-1];
				// console.log('sIdx: ' + sIdx);
				if (sArrIdx[0] === -2) {
					// in pre element
					// console.log('in pre element');
					if (sIdx < tIdx) {
						preElement.list.splice(tIdx, 1);
						preElement.list.splice(sIdx, 1);
					} else {
						preElement.list.splice(sIdx, 1);
						preElement.list.splice(tIdx, 1);
					};
					// multiply by contractedItemCopy
					preElement.multiplyOnRightWith(contractedItemCopy, false);
					preElement.sortByProductWeight();
				} else if (sArrIdx[0] === -1) {
					// in post element
					// console.log('in post element');
					if (sIdx < tIdx) {
						postElement.list.splice(tIdx, 1);
						postElement.list.splice(sIdx, 1);
					} else {
						postElement.list.splice(sIdx, 1);
						postElement.list.splice(tIdx, 1);
					};
					// multiply by contractedItemCopy
					postElement.multiplyOnRightWith(contractedItemCopy, false);
					postElement.sortByProductWeight();
				} else {
					// in main element
					throw "tensors should be in pre or post elements";
				};

				// new contractions array: remove all contractions collected
				let newArrCont = [];
				for (let cont of arrCont) {
					if (!toSuppressIndiceIDArr.includes(cont[0])
							&& !toSuppressIndiceIDArr.includes(cont[1])) {
						newArrCont.push(cont);
					};
				};
				let strNewArrCont = JSON.stringify(newArrCont);
				// console.log('newArrCont (after suppress): ' + strNewArrCont);
				// install new contractions
				term.contractions = JSON.parse(strNewArrCont);

			};

			term.updateExpression();
			// term.showConsole(OPT.SHOW, 'term after: ');
		}; // if source & target
		// if something is found, quit contraction loop
		if (somethingFound) {break;}
	}; // loop contractions
	// used in the loop over all rule functions
	return somethingFound;
};


// general rule: raise 2 indices of a tensor with inverse metric
// tensorData = info about the tensor to contract and the result
// => new MathElement
FR.ruleRaiseIndicesWithInverseMetric = function(term, giID, tensorData) {
	// giID: ID of inverse metric
	// tensorData: {
	// 		tensorDown: MathElement => instance of tensor with down indices
	// 		tensorUp: MathElement => instance of tensor with up indices
	// 	}

	// change later if something is found
	// used in a loop which fires this function => repeat until nothing found
	let somethingFound = false;

	// the idea is to look at contractions:
	// find first contraction between giID and tensorDownID
	let arrCont = term.contractions;
	let indiceIDtoIndiceObj = term.collectIndiceIDtoIndiceObj();

	let tensorDown = tensorData.tensorDown;
	let tensorDownID = tensorDown.getID();
	let tensorUp = tensorData.tensorUp;
	let nIndices = tensorDown.indices.list.length;

	let tensorDerivationLevel = tensorData.tensorDerivationLevel; // NEW
	let tensorIsMathOperator = null;
	if (tensorDown.constructor.name === 'MathElement') {
		if (tensorDown.constructor.name !== 'MathElement') {
			throw "tensors should have same constructor";
		};
		tensorIsMathOperator = false;
	} else if (tensorDown.constructor.name === 'MathOperator') {
		if (tensorDown.constructor.name !== 'MathOperator') {
			throw "tensors should have same constructor";
		};
		tensorIsMathOperator = true;
	} else {
		throw "tensors should be MathElement or MathOperator";
	};

	// perform some tests
	// same number of indices
	if (tensorUp.indices.list.length !== nIndices) {
		throw "tensors should have same number of indices";
	};
	// same symmetries
	if (tensorUp.indices.symmetric.length !== tensorDown.indices.symmetric.length
				||
			tensorUp.indices.antisymmetric.length !== tensorDown.indices.antisymmetric.length) {
		throw "tensors should have same symmetries";
	};
	// TODO: test down and up indices?

	// console.log('');
	// term.showConsole(OPT.SHOW, 'term before: ');

	for (let c of arrCont) {
		// source data to test
		let sObj = indiceIDtoIndiceObj[c[0]];
		let sArrIdx = sObj.arrIdx;
		let sID = sObj.elementID;
		// target data to test
		let tObj = indiceIDtoIndiceObj[c[1]];
		let tArrIdx = tObj.arrIdx;
		let tID = tObj.elementID;

		// test contraction: does it connect giID with tensorDownID?
		// restriction on derivation of tensor down:
		// no derivation
		// TODO: add possibility to 1 derivation since in normal coordinate g commutes with nabla…
		if (	(sID === giID && sArrIdx[sArrIdx.length-1] >= 0
								&& tID === tensorDownID && tArrIdx[tArrIdx.length-1] >= tensorDerivationLevel) // NEW
						||
						(sID === tensorDownID  && sArrIdx[sArrIdx.length-1] >= tensorDerivationLevel // NEW
								&& tID === giID && tArrIdx[tArrIdx.length-1] >= 0)  ) {

			// normalize: source is tensor down, target is inverse metric
			if (sID === giID) {
				let nObj = sObj;
				sObj = tObj;
				tObj = nObj;
				sArrIdx = sObj.arrIdx;
				// sID = sObj.elementID;
				// tArrIdx = tObj.arrIdx;
				// tID = tObj.elementID;
			};

			// loop over indices of source
			// check if contracted with NEW inverse metric (check in targetCollection)
			// OK => collect in targetCollection
			// not OK => 'continue' the main contraction loop
			// find associated indiceID of the inverse metric
			let sIndiceIDRoot = sObj.id;
			let siblingIndiceObjects = FH.getSiblingIndiceObjects(sIndiceIDRoot, indiceIDtoIndiceObj);

			// collect all targets: inverse metric
			let targetRootIDArray = [];
			let targetIndiceObjectsArray = [];
			let performContraction = true;
			for (let s of siblingIndiceObjects) {
				// get target t associated to source s
				let sIndiceID = s.id + '-' + s.n;
				let tIndiceID = FH.getTargetIndiceIDFromSourceIndiceID(sIndiceID, arrCont);
				let t = indiceIDtoIndiceObj[tIndiceID];
				let tID = t.elementID;
				if (tID === giID
							&& !targetRootIDArray.includes(t.id) ) {
					targetRootIDArray.push(t.id);
					targetIndiceObjectsArray.push(t);
				} else {
					// not in expected situation => flag to leave current contraction
					performContraction = false;
					break;
				};
			};

			if (!performContraction) {
				// next contraction
				continue;
			};

			// we found something
			somethingFound = true;

			// console.log('sArrIdx: ' + JSON.stringify(sArrIdx));

			let tensorUpCopy = tensorUp.makeCopy();
			let tensorUpIndiceIDRoot = tensorUpCopy.getIndicesID();
			// tensorUpCopy.showConsole(OPT.SHOW, 'tensor up copy: ');

			// contractions indicesID to remove (tensor down)
			let toSuppressIndiceIDArr = [];
			// couples of old indiceID => new indice ID to replace in arrCont
			let mapIndiceIDArr = [];
			for (let i=0; i<siblingIndiceObjects.length; i++) {
				let s = siblingIndiceObjects[i];
				let sIndiceID = s.id + '-' + s.n;
				// collect in to suppress
				toSuppressIndiceIDArr.push(sIndiceID);
				// target inverse metric indice object
				let t = targetIndiceObjectsArray[i];
				// get indice number in inverse metric
				let tn1 = t.n;
				let tn2 = tn1 === 0 ? 1 : 0;
				// indiceID of the other indice in inverse metric
				let tIndiceID = t.id + '-' + tn2;
				// new indiceID in tensor up
				let tensorUpIndiceID = tensorUpIndiceIDRoot + '-' + i;
				// map tIndiceID => tensorUpIndiceID
				mapIndiceIDArr.push([ tIndiceID, tensorUpIndiceID ]);
			};

			// console.log('toSuppressIndiceIDArr: ' + JSON.stringify(toSuppressIndiceIDArr));
			// console.log('mapIndiceIDArr: ' + JSON.stringify(mapIndiceIDArr));

			// new contractions array: remove all contractions from tensor down
			let newArrCont = [];
			for (let cont of arrCont) {
				// console.log('cont: ' + JSON.stringify(cont));
				if (!toSuppressIndiceIDArr.includes(cont[0])
						&& !toSuppressIndiceIDArr.includes(cont[1])) {
					// console.log('cont added');
					newArrCont.push(cont);
				};
			};
			// console.log('newArrCont (after suppress): ' + JSON.stringify(newArrCont));

			// update contractions using mapIndiceIDArr
			let strNewArrCont = JSON.stringify(newArrCont);
			for (let map of mapIndiceIDArr) {
				let re = new RegExp(map[0], "g");
				strNewArrCont = strNewArrCont.replace(re, map[1]);
			};
			// console.log('newArrCont (after update): ' + strNewArrCont);

			// install new contractions
			term.contractions = JSON.parse(strNewArrCont);

			let preElement = term.preElement;
			let postElement = term.postElement;

			// TODO: CHANGE FOLLOWING CODE WITH 'container' APPROACH GIVEN BELOW?
			// // change source with tensorUp
			// let sIdx = sArrIdx[sArrIdx.length-1]
			// // console.log('sIdx: ' + sIdx);
			// if (sArrIdx[0] === -2) {
			// 	// in pre element
			// 	// console.log('in pre element');
			// 	preElement.list[sIdx] = tensorUpCopy;
			// } else if (sArrIdx[0] === -1) {
			// 	// in post element
			// 	// console.log('in post element');
			// 	postElement.list[sIdx] = tensorUpCopy;
			// } else {
			// 	// in main element
			// 	// console.log('in main element');
			// 	term.list[sIdx] = tensorUpCopy;
			// };

// 'container' APPROACH => to be tested
			// console.log('use new implementation');
			let containerArrIdx = FH.copyData(sArrIdx);
			// get position of tensorDown in container list
			let sIdx = sArrIdx[sArrIdx.length-1];
			if (sIdx >= 0) {
				// no derivations involved
				// remove last idx to get container
				containerArrIdx.splice(-1,1);
				let containerItem = term.getItemAt(containerArrIdx);
				// replace tensorDown by the copy of tensorUp
				containerItem.list[sIdx] = tensorUpCopy;
			} else {
				// derivation involved: two cases
				// 1- tensorDown is a MathOperator itself
				// 2- tensorDown is inside a MathOperator
				if (tensorIsMathOperator) {
					// console.log('tensor is MathOperator, use new code to be tested');
					// two possibilities:
					if (sIdx === -1) {
						// console.log('tensor is the first MathOperator, use new code to be tested');
						// 1- tensorDown is the first MathOperator, sIdx = -1
						// term.showConsole(OPT.SHOW, 'term = ');
						// console.log('containerArrIdx: ' + JSON.stringify(containerArrIdx));
						let operatorItem = term.getItemAt(containerArrIdx);
						// save its argument
						let argument = operatorItem.argument;
						// argument.showConsole(OPT.SHOW, 'argument = ');
						// remove last two idx to get container
						containerArrIdx.splice(-1,1);
						// but get position of tensorDown in list of containerItem
						let tensorIdx = containerArrIdx[containerArrIdx.length-1];
						containerArrIdx.splice(-1,1);
						// console.log('containerArrIdx: ' + JSON.stringify(containerArrIdx));
						let containerItem = term.getItemAt(containerArrIdx);
						// containerItem.showConsole(OPT.SHOW, 'containerItem = ');
						// put argument in tensorUp copy
						// tensorUpCopy.showConsole(OPT.SHOW, 'tensorUpCopy = ');
						tensorUpCopy.applyTo(argument);
						// replace tensorDown by the copy of tensorUp with its argument
						containerItem.list[tensorIdx] = tensorUpCopy;
						// term.showConsole(OPT.SHOW, 'term = ');
					} else {
						console.log('tensor is not the first MathOperator, use new code to be tested');
						// 2- tensorDown is inside a MathOperator, sIdx <= -2
						// the container is the MathOperator one level up
						containerArrIdx[sArrIdx.length-1] = sIdx + 1;
						let containerItem = term.getItemAt(containerArrIdx);
						// save its argument
						let argument = tensorDown.argument;
						// put argument in tensorUp copy
						tensorUpCopy.applyTo(argument);
						// apply the container to the copy of tensorUp with its argument
						containerItem.argument = tensorUpCopy;
					}
				} else {
					console.log('tensor is MathElement inside MathOperator, use new code to be tested');
					// tensorDown is a MathElement inside a MathOperator
					// the container is the MathOperator one level up
					containerArrIdx[sArrIdx.length-1] = sIdx + 1;
					let containerItem = term.getItemAt(containerArrIdx);
					// apply the container to the tensorUp copy
					containerItem.argument = tensorUpCopy;
				};
			};


			// remove all inverse metrics: there are in pre element
			// collect idx in array
			let preArrIdx = [];
			for (let t of targetIndiceObjectsArray) {
				let arrIdx = t.arrIdx;
				if (arrIdx[0] === -2) {
					// add idx in preElement
					preArrIdx.push(arrIdx[arrIdx.length-1]);
				};
			};
			// remove from right to left to maintain idx
			preArrIdx.sort().reverse();
			// console.log('preArrIdx (sorted+reverse): ' + JSON.stringify(preArrIdx));
			// remove from pre element
			let currentPreElement = term.preElement;
			for (let idx of preArrIdx) {
				preElement.list.splice(idx,1);
			};

			term.updateExpression();
			// term.showConsole(OPT.SHOW, 'term after: ');
		}; // if source & target
		// if something is found, quit contraction loop
		if (somethingFound) {break;}
	}; // loop contractions
	// used in the loop over all rule functions
	return somethingFound;
};


// general rule: check for occurence of same element in pre element
// TODO: search in other elements…
// replace these elements by unique element
FR.rulePoweredSameItem = function(term, powerData) {
	// powerData: {
	// 		item: MathElement to search for,
	// 		power: integer, the power to look at,
	// 		newItem: MathElement which replace these elements
	// }

	// change later if something is found
	// used in a loop which fires this function => repeat until nothing found
	let somethingFound = false;

	let preElement = term.preElement;
	let orgItemID = powerData.item.getID();
	let p = powerData.power;
	// collect all occurences of original item
	let arrArrIdx = preElement.getPositionsByID(orgItemID, 0, CONST.PRE_ELEMENT);
	if (arrArrIdx.length < p) {
		// not enough items
		return somethingFound;
	};
	// we found something
	somethingFound = true;

	// replace first occurence
	let newItem = powerData.newItem.makeCopy();
	// term.showConsole(OPT.SHOW, 'term = ');
	// console.log('arrArrIdx: ' + JSON.stringify(arrArrIdx));
	let returnedInfo = {};
	// since we are in pre element and no derivations are selected,
	// all arrArrIdx[i] are of length = 1
	preElement.replaceByItemAtIndex(newItem, arrArrIdx[0][0], returnedInfo, false);
	for (let i=1; i<arrArrIdx.length; i++) {
		preElement.removeItemAtIndex(arrArrIdx[i][0], returnedInfo, false);
	};
	// term.showConsole(OPT.SHOW, 'term = ');

	return somethingFound;
};


// SPECIFIC RULES
// rule: if main in MathArgument contains only k elements u
// => reduce to 1 element u and multiply by 1/k!
FR.ruleOperatorsAllOnlyu = function(term, rulesData) {
	let uID = rulesData.id.uID;
	// console.log('ruleOperatorsAllOnlyu');
	// console.log('uID = ' + uID);
	// term.showConsole(OPT.SHOW, 'term = ');

	// used in a loop which fires this function => repeat until nothing found
	let somethingFound = false;

	// update array of indices to fire getReducedStringRepresentationCountingElementID
	term.fillNormalizedIndicesArray();

	let {counter: counter, string: mainTxt, postString: postTxt} = term.getReducedStringRepresentationCountingElementID(uID);
	if (mainTxt === '' && counter > 1) {
		let somethingFound = true;
		term.list = [term.list[0]];
		term.preElement.multiplyByRational(1, FH.factorial(counter));
		term.updateExpression();
	};
	// used in the loop over all rule functions
	return somethingFound;
};


// rule: contractions of metric and its inverse
// no Kronecker symbol needed since change corresponding indices
// takes care of double contractions to speed up
// dim: MathRationalExpression which represents the dimension of space
FR.ruleContractionsMetricWithInverseMetric = function(term, rulesData) {
	let gID = rulesData.id.gID;
	let giID = rulesData.id.giID;
	let dim = rulesData.elements.dim;

	// console.log('ruleContractionsMetricWithInverseMetric');
	// console.log('gID = ' + gID);
	// console.log('giID = ' + giID);
	// term.showConsole(OPT.SHOW, 'term = ');

	// change later if something is found
	// used in a loop which fires this function => repeat until nothing found
	let somethingFound = false;

	// the idea is to look at contractions:
	// find first contraction between gID and giID
	let arrCont = term.contractions;
	let indiceIDtoIndiceObj = term.collectIndiceIDtoIndiceObj();
	// everything is concentred on preElement
	let currentPreElement = term.preElement;

	for (let c of arrCont) {
		// source data to test
		let sObj = indiceIDtoIndiceObj[c[0]];
		let sArrIdx = sObj.arrIdx;
		let sID = sObj.elementID;
		// target data to test
		let tObj = indiceIDtoIndiceObj[c[1]];
		let tArrIdx = tObj.arrIdx;
		let tID = tObj.elementID;

		// test contraction: does it connect gID with giID?
		if (
					sArrIdx[sArrIdx.length-1] >= 0 && tArrIdx[tArrIdx.length-1] >= 0
					&&
					(
						(sID === gID && tID === giID)
						||
						(sID === giID && tID === gID)
					)
				) {

			// console.log('➤ found simplification');
			// we found something
			somethingFound = true;

			// find associated indiceID to implement Kronecker contraction
			let sIndiceIDRoot = sObj.id;
			let sn1 = sObj.n;
			let tIndiceIDRoot = tObj.id;
			let tn1 = tObj.n;
			let sn2 = sn1 === 0 ? 1 : 0;
			let tn2 = tn1 === 0 ? 1 : 0;
			// console.log(sn1 + '/' + sn2 + ' -- ' + tn1 + '/' + tn2);
			let sIndiceID1 = sIndiceIDRoot + '-' + sn1;
			let sIndiceID2 = sIndiceIDRoot + '-' + sn2;
			let tIndiceID1 = tIndiceIDRoot + '-' + tn1;
			let tIndiceID2 = tIndiceIDRoot + '-' + tn2;
			// term.showConsole(OPT.SHOW, 'term (before) = ');
			// console.log(sIndiceID1 + '/' + sIndiceID2 + ' -- ' + tIndiceID1 + '/' + tIndiceID2);
			// idx of MathElement to remove from preElement
			let sIdx = sArrIdx[sArrIdx.length-1];
			let tIdx = tArrIdx[tArrIdx.length-1];

			// get target indiceID associated to sIndiceID2
			let newsIndiceID2 = FH.getTargetIndiceIDFromSourceIndiceID(sIndiceID2, arrCont);
			if (newsIndiceID2 === null) {
				// no indice found => no contraction
				continue;
			};
			// get target indiceID associated to tIndiceID2
			let newtIndiceID2 = FH.getTargetIndiceIDFromSourceIndiceID(tIndiceID2, arrCont);
			if (newtIndiceID2 === null) {
				// no indice found => no contraction
				continue;
			};

			// update arrCont
			let newArrCont = [];
			// detect self-contraction of g with gi => dim factor
			if (sIndiceID2 === newtIndiceID2) {
				currentPreElement.prefactor = currentPreElement.prefactor.multiply(dim);
				// console.log('');
				// console.log('  ➤ found self contraction');
				// currentPreElement.prefactor.showConsole(OPT.SHOW, 'prefactor: ');
			} else {
				// add new contraction, as if Kronecker symbol
				// console.log('  ➤ found Kronecker contraction');
				newArrCont.push([newsIndiceID2,newtIndiceID2].sort(FH.alphaSort));
			};
			// only keep couples without sIndiceID1, sIndiceID2, tIndiceID1, tIndiceID2
			for (let cont of arrCont) {
				if (!cont.includes(sIndiceID1) && !cont.includes(sIndiceID2)
						&& !cont.includes(tIndiceID1) && !cont.includes(tIndiceID2)	) {
					newArrCont.push(cont);
				};
			};
			term.contractions = FH.copyData(newArrCont);
			// remove g and gi from preElement
			if (sIdx < tIdx) {
				currentPreElement.list.splice(tIdx, 1);
				currentPreElement.list.splice(sIdx, 1);
			} else {
				currentPreElement.list.splice(sIdx, 1);
				currentPreElement.list.splice(tIdx, 1);
			};

			term.updateExpression();
		}; // if source & target
		// if something is found, quit contraction loop
		if (somethingFound) {break;}
	}; // loop contractions
	// used in the loop over all rule functions
	return somethingFound;
};


// rule: contractions of inverse metric and Riemann tensor
// Ric: Ricci MathElement
// if inverse metric contracts with (0,2), (0,3), (1,2) of (1,3) => Ric with a sign
// if inverse metric contracts with (0,1), (2,3) => 0
// otherwise, do nothing
FR.ruleContractionsInverseMetricWithRiemannTensor = function(term, rulesData) {
	let giID = rulesData.id.giID;
	let RTID = rulesData.id.RTID;
	let Ric = rulesData.elements.Ric;

	// change later if something is found
	// used in a loop which fires this function => repeat until nothing found
	let somethingFound = false;

	// the idea is to look at contractions:
	// find first contraction between giID and RTID
	let arrCont = term.contractions;
	let indiceIDtoIndiceObj = term.collectIndiceIDtoIndiceObj();
	// everything is concentred on preElement
	let currentPreElement = term.preElement;

	// console.log('');
	// term.showConsole(OPT.SHOW, 'term before: ');
	// console.log('giID: ' + giID + ' / RTID: ' + RTID);

	for (let c of arrCont) {
		// source data to test
		let sObj = indiceIDtoIndiceObj[c[0]];
		let sArrIdx = sObj.arrIdx;
		let sID = sObj.elementID;
		// target data to test
		let tObj = indiceIDtoIndiceObj[c[1]];
		// console.log('sObj: ' + JSON.stringify(sObj) + ' / tObj: ' + JSON.stringify(tObj));
		let tArrIdx = tObj.arrIdx;
		let tID = tObj.elementID;


		// test contraction: does it connect giID with RTID?
		// Riemann tensor can be in a derivation:
		// we suppose it is a Riemannian covariant derivative
		// => the metric commutes with it
		if (
					(sID === giID && tID === RTID)
					||
					(sID === RTID && tID === giID)
				) {

			// console.log('first test OK');

			// normalize: source is inverse metric, target is Riemann tensor
			if (sID === RTID) {
				// console.log('switch source and target')
				let nObj = sObj;
				sObj = tObj;
				tObj = nObj;
				sArrIdx = sObj.arrIdx;
				sID = sObj.elementID;
				tArrIdx = tObj.arrIdx;
				tID = tObj.elementID;
				// console.log('switch => sObj: ' + JSON.stringify(sObj) + ' / tObj: ' + JSON.stringify(tObj));
			};

			// check for double contraction of inverse metric with Riemann tensor
			// find associated indiceID of the inverse metric
			let sIndiceIDRoot = sObj.id;
			let sn1 = sObj.n;
			let sn2 = sn1 === 0 ? 1 : 0;
			let sIndiceID1 = sIndiceIDRoot + '-' + sn1;
			let sIndiceID2 = sIndiceIDRoot + '-' + sn2;

			// get target indiceID associated to sIndiceID2
			let newsIndiceID2 = FH.getTargetIndiceIDFromSourceIndiceID(sIndiceID2, arrCont);
			if (newsIndiceID2 === null) {
				// no indice found => no contraction
				continue;
			};

			let tIndiceIDRoot = tObj.id;
			let tn1 = tObj.n;
			let tIndiceID1 = tIndiceIDRoot + '-' + tn1;

			// console.log('indices:' + 'sIndiceID1: ' + sIndiceID1 + ' / sIndiceID2: ' + sIndiceID2);
			// console.log('newsIndiceID2: ' + newsIndiceID2 + ' / tIndiceID1: ' + tIndiceID1);

			// test if newsIndiceID2 belongs to Riemann tensor
			if (!newsIndiceID2.includes(tIndiceIDRoot)) {
				// console.log('second test failed');
				// not a double contraction
				// next contraction
				continue;
			};

			// we found something
			somethingFound = true;
			// console.log('something found');
			// console.log('sArrIdx: ' + JSON.stringify(sArrIdx));
			// console.log('tArrIdx: ' + JSON.stringify(tArrIdx));

			// find where newsIndiceID2 is connected in Riemann tensor
			let tn2 = null;
			for (tn2=0; tn2<4; tn2++) {
				if (newsIndiceID2 === tIndiceIDRoot + '-' + tn2) {
					// break this loop on tn2
					break;
				};
			};
			let tIndiceID2 = tIndiceIDRoot + '-' + tn2;
			// console.log('tIndiceID2: ' + tIndiceID2);

			// console.log(sIndiceID1 + '/' + sIndiceID2 + ' -- ' + tIndiceID1 + '/' + tIndiceID2);

			let contractedTargetIdx = [tn1,tn2].sort();
			let configIdx = contractedTargetIdx.join('');
			// console.log('configIdx: ' + configIdx);

			let sign = null;
			// if inverse metric contracts with (0,2), (0,3), (1,2) of (1,3) => Ric with a sign
			// if inverse metric contracts with (0,1), (2,3) => 0
			if (configIdx === '01' || configIdx === '23') {sign = 0;}
			if (configIdx === '02' || configIdx === '13') {sign = -1;}
			if (configIdx === '03' || configIdx === '12') {sign = 1;}

			currentPreElement.prefactor = currentPreElement.prefactor.multiplyByInteger(sign);

			// if sign is 0 no need to change replace…
			if (sign !== 0) {

				// update arrCont
				let newArrCont = [];
				// only keep couples without sIndiceID1, sIndiceID2, tIndiceID1, tIndiceID2
				for (let cont of arrCont) {
					if (!cont.includes(sIndiceID1) && !cont.includes(sIndiceID2)
							&& !cont.includes(tIndiceID1) && !cont.includes(tIndiceID2)	) {
						newArrCont.push(cont);
					};
				};

				// find the two remaining idx in Riemann tensor
				let remainingIdx = [];
				for (let i=0; i<4; i++) {
					if (i!== tn1 && i!== tn2) {
						remainingIdx.push(i);
					};
				};
				remainingIdx.sort();
				// console.log('remainingIdx: ' + JSON.stringify(remainingIdx));
				//
				let RicCopy = Ric.makeCopy();
				let RicIndiceID = RicCopy.getIndicesID();

				// update contractions: replace Riemann tensor by Ricci at the two remaining slots
				let strNewArrCont = JSON.stringify(newArrCont);
				// console.log('strNewArrCont (before): ' + strNewArrCont);
				let re0 = new RegExp(tIndiceIDRoot + '-' + remainingIdx[0], "g");
				let re1 = new RegExp(tIndiceIDRoot + '-' + remainingIdx[1], "g");
				strNewArrCont = strNewArrCont.replace(re0, RicIndiceID+ '-0');
				strNewArrCont = strNewArrCont.replace(re1, RicIndiceID+ '-1');
				// console.log('strNewArrCont (after): ' + strNewArrCont);
				term.contractions = JSON.parse(strNewArrCont);

				// idx of MathElement to remove from preElement
				let sIdx = sArrIdx[sArrIdx.length-1];
				let tIdx = tArrIdx[tArrIdx.length-1];

				if (tIdx >= 0) {
					// Riemann Tensor not in derivation
					// remove inverse metric and Riemann Tensor from preElement
					if (sIdx < tIdx) {
						currentPreElement.list.splice(tIdx, 1);
						currentPreElement.list.splice(sIdx, 1);
					} else {
						currentPreElement.list.splice(sIdx, 1);
						currentPreElement.list.splice(tIdx, 1);
					};
					// multiply by Ricci tensor
					currentPreElement.multiplyOnRightWith(RicCopy, false);
					currentPreElement.sortByProductWeight();
				} else {
					// Riemann Tensor in derivation
					// we suppose it is a Riemannian covariant derivative
					// => the metric commutes with it
					let arrIdxContainer = tArrIdx;
					// the container is the MathOperator one level up
					arrIdxContainer[tArrIdx.length-1] = tIdx + 1;
					let container = term.getItemAt(arrIdxContainer);

					// container.showConsole(OPT.SHOW, 'container = ');
					// remove inverse metric
					currentPreElement.list.splice(sIdx, 1);
					container.argument = RicCopy;
					currentPreElement.sortByProductWeight();
				};
			}; // if sign not 0

			term.updateExpression();
			// term.showConsole(OPT.SHOW, 'term after: ');
		}; // if source & target
		// if something is found, quit contraction loop
		if (somethingFound) {break;}
	}; // loop contractions
	// used in the loop over all rule functions
	return somethingFound;
};


// rule: contractions of inverse metric and Ricci tensor
FR.ruleContractionsInverseMetricWithRicciTensor = function(term, rulesData) {
	let giID = rulesData.id.giID;
	let RicID = rulesData.id.RicID;
	let SC = rulesData.elements.SC;

	// change later if something is found
	// used in a loop which fires this function => repeat until nothing found
	let somethingFound = false;

	// the idea is to look at contractions:
	// find first contraction between giID and RicID
	let arrCont = term.contractions;
	let indiceIDtoIndiceObj = term.collectIndiceIDtoIndiceObj();
	// everything is concentred on preElement
	let currentPreElement = term.preElement;

	for (let c of arrCont) {
		// source data to test
		let sObj = indiceIDtoIndiceObj[c[0]];
		let sArrIdx = sObj.arrIdx;
		let sID = sObj.elementID;
		// target data to test
		let tObj = indiceIDtoIndiceObj[c[1]];
		let tArrIdx = tObj.arrIdx;
		let tID = tObj.elementID;

		// test contraction: does it connect giID with RicID?
		if (
						(sID === giID && tID === RicID)
						||
						(sID === RicID && tID === giID)
				) {

			// normalize: source is inverse metric, target is Ricci tensor
			if (sID === RicID) {
				let nObj = sObj;
				sObj = tObj;
				tObj = nObj;
				sArrIdx = sObj.arrIdx;
				sID = sObj.elementID;
				tArrIdx = tObj.arrIdx;
				tID = tObj.elementID;
			};

			// check for double contraction of inverse metric with Ricci tensor
			// find associated indiceID of the inverse metric
			let sIndiceIDRoot = sObj.id;
			let sn1 = sObj.n;
			let sn2 = sn1 === 0 ? 1 : 0;
			let sIndiceID1 = sIndiceIDRoot + '-' + sn1;
			let sIndiceID2 = sIndiceIDRoot + '-' + sn2;

			// get target indiceID associated to sIndiceID2
			let newsIndiceID2 = FH.getTargetIndiceIDFromSourceIndiceID(sIndiceID2, arrCont);
			if (newsIndiceID2 === null) {
				// no indice found => no contraction
				continue;
			};

			let tIndiceIDRoot = tObj.id;
			let tn1 = tObj.n;
			let tIndiceID1 = tIndiceIDRoot + '-' + tn1;

			// test if newsIndiceID2 belongs to Ricci tensor
			if (!newsIndiceID2.includes(tIndiceIDRoot)) {
				// not a double contraction => return
				// used in the loop over all rule functions
				// return somethingFound;
				continue;
			}

			// we found something
			somethingFound = true;

			// the second contraction is with the second indice in Ricci tensor
			let tIndiceID2 = newsIndiceID2;

			// console.log(sIndiceID1 + '/' + sIndiceID2 + ' -- ' + tIndiceID1 + '/' + tIndiceID2);

			// update arrCont
			let newArrCont = [];
			// only keep couples without sIndiceID1, sIndiceID2, tIndiceID1, tIndiceID2
			for (let cont of arrCont) {
				if (!cont.includes(sIndiceID1) && !cont.includes(sIndiceID2)
						&& !cont.includes(tIndiceID1) && !cont.includes(tIndiceID2)	) {
					newArrCont.push(cont);
				};
			};

			term.contractions = FH.copyData(newArrCont);

			// idx of MathElement to remove from preElement
			let sIdx = sArrIdx[sArrIdx.length-1];
			let tIdx = tArrIdx[tArrIdx.length-1];

			if (tIdx >= 0) {
				// Ricci Tensor not in derivation
				// remove inverse metric and Ricci Tensor from preElement
				if (sIdx < tIdx) {
					currentPreElement.list.splice(tIdx, 1);
					currentPreElement.list.splice(sIdx, 1);
				} else {
					currentPreElement.list.splice(sIdx, 1);
					currentPreElement.list.splice(tIdx, 1);
				};
				// multiply by Scalar Curvature
				currentPreElement.multiplyOnRightWith(SC, true);
				currentPreElement.sortByProductWeight();
			} else {
				// Ricci Tensor in derivation
				// we suppose it is a Riemannian covariant derivative
				// => the metric commutes with it
				let arrIdxContainer = tArrIdx;
				// the container is the MathOperator one level up
				arrIdxContainer[tArrIdx.length-1] = tIdx + 1;
				let container = term.getItemAt(arrIdxContainer);

				// container.showConsole(OPT.SHOW, 'container = ');
				// remove inverse metric
				currentPreElement.list.splice(sIdx, 1);
				container.argument = SC.makeCopy();
				currentPreElement.sortByProductWeight();
			};

			term.updateExpression();
		}; // if source & target
		// if something is found, quit contraction loop
		if (somethingFound) {break;}
	}; // loop contractions
	// used in the loop over all rule functions
	return somethingFound;
};


// rule: g^{12} hnabla_1 Ric_{23} => 1/2 hnabla_3 SC
FR.ruleRiemannCovariantRicciToScalar = function(term, rulesData) {
	let giID = rulesData.id.giID;
	let RicID = rulesData.id.RicID;
	let hnablaID = rulesData.id.hnablaID;
	let SC = rulesData.elements.SC;

	// change later if something is found
	// used in a loop which fires this function => repeat until nothing found
	let somethingFound = false;

	// the idea is to look at contractions:
	// find first contraction between giID and RicID
	let arrCont = term.contractions;
	let indiceIDtoIndiceObj = term.collectIndiceIDtoIndiceObj();
	// everything is concentred on preElement
	let currentPreElement = term.preElement;

	for (let c of arrCont) {
		// source data to test
		let sObj = indiceIDtoIndiceObj[c[0]];
		let sArrIdx = sObj.arrIdx;
		let sID = sObj.elementID;
		// target data to test
		let tObj = indiceIDtoIndiceObj[c[1]];
		let tArrIdx = tObj.arrIdx;
		let tID = tObj.elementID;

		// test contraction: does it connect giID with RicID?
		if (
						(sID === giID && tID === RicID)
						||
						(sID === RicID && tID === giID)
				) {

			// normalize: source is inverse metric, target is Ricci tensor
			if (sID === RicID) {
				let nObj = sObj;
				sObj = tObj;
				tObj = nObj;
				sArrIdx = sObj.arrIdx;
				sID = sObj.elementID;
				tArrIdx = tObj.arrIdx;
				tID = tObj.elementID;
			};

			// check for contraction of the inverse metric with hnablaID
			// find associated indiceID of the inverse metric
			let sIndiceIDRoot = sObj.id;
			let sn1 = sObj.n;
			let sn2 = sn1 === 0 ? 1 : 0;
			// the two indiceIDn of the inverse metric
			let sIndiceID1 = sIndiceIDRoot + '-' + sn1;
			let sIndiceID2 = sIndiceIDRoot + '-' + sn2;

			// get target indiceID associated to sIndiceID2
			let newsIndiceID2 = FH.getTargetIndiceIDFromSourceIndiceID(sIndiceID2, arrCont);
			if (newsIndiceID2 === null) {
				// no indice found => no contraction
				continue;
			};

			// test if newsIndiceID2 belongs to hnablaID
			let opObj = indiceIDtoIndiceObj[newsIndiceID2];
			let opID = opObj.elementID;
			if (opID !== hnablaID) {
				// not a suitable contraction
				continue;
			};

			// now: sObj is inverse metric, tObj is Ricci, opObj is hnabla

			// check that Ricci tensor is argument of hnablaID
			// compare their ArrIdx:
			let opArrIdxNew = FH.copyData(opObj.arrIdx);
			let tArrIdxNew = FH.copyData(tObj.arrIdx);
			// last element should be 1 less for Ricci, as an direct argument of hnabla
			tArrIdxNew[tArrIdxNew.length-1] = tArrIdxNew[tArrIdxNew.length-1] + 1;
			// now these two arrays should be the same
			if (JSON.stringify(opArrIdxNew) !== JSON.stringify(tArrIdxNew)) {
				// Ricci is not direct argument of hnabla
				continue;
			};

			// we found something
			somethingFound = true;
			// console.log('found something');

			// now:
			// replace Ricci by 1/2 SC in hnabla
			// update contractions array
			// remove inverse metric

			// find associated indiceID of Ricci
			// since Ricci is symmetric, no matter what we start with
			let tIndiceIDRoot = tObj.id;
			let tn1 = tObj.n;
			let tn2 = tn1 === 0 ? 1 : 0;
			// this is the indiceIDn used in contraction with inverse metric
			let tIndiceID1 = tIndiceIDRoot + '-' + tn1;
			// this is the indiceIDn to be used to update contractions array
			let tIndiceID2 = tIndiceIDRoot + '-' + tn2;

			// indiceIDn of hnabla (just for human eyes and brain…)
			let opIndiceID1 = newsIndiceID2;

			// update arrCont
			let newArrCont = [];
			// only keep couples without sIndiceID1, sIndiceID2, tIndiceID1, opIndiceID1
			for (let cont of arrCont) {
				if (!cont.includes(sIndiceID1) && !cont.includes(sIndiceID2)
						&& !cont.includes(tIndiceID1) && !cont.includes(opIndiceID1)	) {
					newArrCont.push(cont);
				};
			};
			// replace tIndiceID2 by opIndiceID1:
			// the "free" indice in Ricci (tIndiceID2) becomes the indice in hnabla (opIndiceID1)
			// => any contraction with tIndiceID2 is now a contraction with opIndiceID1
			let strNewArrCont = JSON.stringify(newArrCont);
			let re = new RegExp(tIndiceID2, "g");
			strNewArrCont = strNewArrCont.replace(re, opIndiceID1);

			term.contractions = FH.copyData(JSON.parse(strNewArrCont));

			// replace Ricci element by SC elements
			let container = term.getItemAt(opObj.arrIdx);
			container.argument = SC.makeCopy();
			// remove inverse metric
			let sIdx = sArrIdx[sArrIdx.length-1];
			currentPreElement.list.splice(sIdx, 1);
			// multiply by 1/2
			term.multiplyByRational(1,2);
			// clean up order
			currentPreElement.sortByProductWeight();

			term.updateExpression();
		}; // if source & target
		// if something is found, quit contraction loop
		if (somethingFound) {break;}
	}; // loop contractions
	// used in the loop over all rule functions
	return somethingFound;
};


// rule: contraction of inverse metric with two consecutive nabla => hlaplacian
// hlaplacian: the hlaplacian MathOperator
// can work with nabla or hnabla according to hnablaID given and hlaplacian given
FR.ruleLaplacian = function(term, rulesData) {
	let giID = rulesData.id.giID;
	let hnablaID = rulesData.id.hnablaID;
	let hlaplacian = rulesData.elements.hlaplacian;

	// change later if something is found
	// used in a loop which fires this function => repeat until nothing found
	let somethingFound = false;

	// the idea is to look at contractions:
	// find first contraction between giID and RicID
	let arrCont = term.contractions;
	let indiceIDtoIndiceObj = term.collectIndiceIDtoIndiceObj();
	// everything is concentred on preElement
	let currentPreElement = term.preElement;

	for (let c of arrCont) {
		// source data to test
		let sObj = indiceIDtoIndiceObj[c[0]];
		let sArrIdx = sObj.arrIdx;
		let sID = sObj.elementID;
		// target data to test
		let tObj = indiceIDtoIndiceObj[c[1]];
		let tArrIdx = tObj.arrIdx;
		let tID = tObj.elementID;

		// test contraction: does it connect giID with hnablaID?
		if (
						(sID === giID && tID === hnablaID)
						||
						(sID === hnablaID && tID === giID)
				) {

			// normalize: source is inverse metric, target is nabla operator
			if (sID === hnablaID) {
				let nObj = sObj;
				sObj = tObj;
				tObj = nObj;
				sArrIdx = sObj.arrIdx;
				sID = sObj.elementID;
				tArrIdx = tObj.arrIdx;
				tID = tObj.elementID;
			};

			// check for second contraction of inverse metric with hnablaID
			// find associated indiceID of the inverse metric
			let sIndiceIDRoot = sObj.id;
			let sn1 = sObj.n;
			let sn2 = sn1 === 0 ? 1 : 0;
			// the two indiceIDn of the inverse metric
			let sIndiceID1 = sIndiceIDRoot + '-' + sn1;
			let sIndiceID2 = sIndiceIDRoot + '-' + sn2;

			// get target indiceID associated to sIndiceID2
			let newsIndiceID = FH.getTargetIndiceIDFromSourceIndiceID(sIndiceID2, arrCont);
			if (newsIndiceID === null) {
				// no indice found => no contraction
				continue;
			};

			// test if newsIndiceID belongs to hnablaID
			let opObj = indiceIDtoIndiceObj[newsIndiceID];
			let opID = opObj.elementID;
			if (opID !== hnablaID) {
				// not a suitable contraction
				continue;
			};

			// now: sObj is inverse metric, tObj operator is nabla, opObj operator is nabla

			// check that first nabla (tObj) is argument of second nabla (opObj)
			// note that if the situation is inverted, it will be found for another contraction
			// compare their ArrIdx:
			let opArrIdxNew = FH.copyData(opObj.arrIdx);
			let tArrIdxNew = FH.copyData(tObj.arrIdx);
			// last element should be 1 less for tObj operator,
			// as an direct argument of opObj operator
			tArrIdxNew[tArrIdxNew.length-1] = tArrIdxNew[tArrIdxNew.length-1] + 1;
			// now these two arrays should be the same
			if (JSON.stringify(opArrIdxNew) !== JSON.stringify(tArrIdxNew)) {
				// tObj operator is not direct argument of opObj operator
				continue;
			};

			// we found something
			somethingFound = true;
			// console.log('found something');
			// term.showConsole(OPT.SHOW, 'term = ');
			// now:
			// replace opObj operator by hlaplacian
			// remove tObj operator
			// take argument of tObj operator and put it as argument of hlaplacian
			// remove inverse metric and all indices in contractions

			// find associated indiceID of tObj operator
			let tIndiceIDRoot = tObj.id;
			let tn = tObj.n;
			// this is the indiceIDn used in contraction with inverse metric
			let tIndiceID = tIndiceIDRoot + '-' + tn;

			// indiceIDn of opObj operator (just for human eyes and brain…)
			let opIndiceID = newsIndiceID;

			// update arrCont
			let newArrCont = [];
			// only keep couples without sIndiceID1, sIndiceID2, tIndiceID, opIndiceID
			for (let cont of arrCont) {
				if (!cont.includes(sIndiceID1) && !cont.includes(sIndiceID2)
						&& !cont.includes(tIndiceID) && !cont.includes(opIndiceID)	) {
					newArrCont.push(cont);
				};
			};
			term.contractions = newArrCont;

			// capture argument to pass to hlaplacian
			// it is the argument of tObj operator
			let argument = term.getItemAt(tObj.arrIdx).argument;
			// argument.showConsole(OPT.SHOW, 'argument = ');
			let newOp = hlaplacian.makeCopy();
			newOp.applyTo(argument);
			// replace opObj operator by hlaplacian
			let opArrIdx = opObj.arrIdx;
			let opIdx = opArrIdx[opArrIdx.length-1];
			if (opIdx === -1) {
				// first operator in hierarchy
				// get idx in parent item
				let idx = opArrIdx[opArrIdx.length-2];
				// remove last two idx to get reference to parent item
				let parentArrIdx = opArrIdx.slice(0, -2);
				let parentItem = term.getItemAt(parentArrIdx);
				let retInfo = {};
				parentItem.replaceByItemAtIndex(newOp, idx, retInfo, false);
			} else {
				// find the operator to which opObj is argument
				// console.log('**** in derivation');
				let parentArrIdx = FH.copyData(opObj.arrIdx);
				parentArrIdx[parentArrIdx.length-1] = parentArrIdx[parentArrIdx.length-1] + 1;
				let parentItem = term.getItemAt(parentArrIdx);
				parentItem.argument = newOp;
			};

			// remove inverse metric
			let parentArrIdx = FH.copyData(sObj.arrIdx);
			// remove last idx to get parent
			parentArrIdx.splice(-1,1);
			let parentItem = term.getItemAt(parentArrIdx);
			let sIdx = sArrIdx[sArrIdx.length-1];
			parentItem.list.splice(sIdx, 1);

			term.updateExpression();
			// term.showConsole(OPT.SHOW, 'term = ');
			// console.log('');

		}; // if source & target
		// if something is found, quit contraction loop
		if (somethingFound) {break;}
	}; // loop contractions
	// used in the loop over all rule functions
	return somethingFound;
};


// rule: full contraction of inverse metric and gauge curvature => 0
// use FR.ruleContractionsSymmetricAntisymmetric2Tensors
FR.ruleContractionsInverseMetricWithGaugeCurvature = function(term, rulesData) {
	let giID = rulesData.id.giID;
	let FID = rulesData.id.FID;
	return FR.ruleContractionsSymmetricAntisymmetric2Tensors(term, giID, FID);
};


// rule: full contraction of Ricci tensor and gauge curvature => 0
// use FR.ruleContractionsSymmetricAntisymmetric2Tensors
FR.ruleContractionsRicciTensorWithGaugeCurvature = function(term, rulesData) {
	let RicID = rulesData.id.RicID;
	let FUID = rulesData.id.FUID;
	return FR.ruleContractionsSymmetricAntisymmetric2Tensors(term, RicID, FUID);
};

// rule: full contraction of Symmetric part of double derivations and gauge curvature => 0
// use FR.ruleContractionsSymmetricAntisymmetric2Tensors
FR.ruleContractionsSymmetricLaplacianWithGaugeCurvature = function(term, rulesData) {
	let Shnabla2ID = rulesData.id.Shnabla2ID;
	let FUID = rulesData.id.FUID;
	return FR.ruleContractionsSymmetricAntisymmetric2Tensors(term, Shnabla2ID, FUID);
};

// rule: full contraction of Ricci tensor and gauge curvature commutator with upper indices => 0
// use FR.ruleContractionsSymmetricAntisymmetric2Tensors
FR.ruleContractionsRicciTensorWithCommutatorFu = function(term, rulesData) {
	let RicUID = rulesData.id.RicUID;
	let comFuID = rulesData.id.comFuID;
	return FR.ruleContractionsSymmetricAntisymmetric2Tensors(term, RicUID, comFuID);
};


// rule: raise indices of gauge curvature
// use FR.ruleRaiseIndicesWithInverseMetric
FR.ruleRaiseIndicesOfGaugeCurvatureWithInverseMetric = function(term, rulesData) {
	let giID = rulesData.id.giID;
	let FTensorData = rulesData.tensorData.F;
	return FR.ruleRaiseIndicesWithInverseMetric(term, giID, FTensorData);
};


// rule: raise indices of Riemann tensor
// use FR.ruleRaiseIndicesWithInverseMetric
FR.ruleRaiseIndicesOfRiemannTensorWithInverseMetric = function(term, rulesData) {
	let giID = rulesData.id.giID;
	let RTTensorData = rulesData.tensorData.RT;
	return FR.ruleRaiseIndicesWithInverseMetric(term, giID, RTTensorData);
};


// rule: raise indices of Ricci tensor
// use FR.ruleRaiseIndicesWithInverseMetric
FR.ruleRaiseIndicesOfRicciTensorWithInverseMetric = function(term, rulesData) {
	let giID = rulesData.id.giID;
	let RicTensorData = rulesData.tensorData.Ric;
	return FR.ruleRaiseIndicesWithInverseMetric(term, giID, RicTensorData);
};

// rule: raise indices of comFu
// use FR.ruleRaiseIndicesWithInverseMetric
FR.ruleRaiseIndicesOfCommutatorFuWithInverseMetric = function(term, rulesData) {
	let giID = rulesData.id.giID;
	let comFuTensorData = rulesData.tensorData.comFu;
	return FR.ruleRaiseIndicesWithInverseMetric(term, giID, comFuTensorData);
};

// rule: raise indices of Shnabla2
// use FR.ruleRaiseIndicesWithInverseMetric
FR.ruleRaiseIndicesOfShnabla2WithInverseMetric = function(term, rulesData) {
	let giID = rulesData.id.giID;
	let Shnabla2TensorData = rulesData.tensorData.Shnabla2;
	return FR.ruleRaiseIndicesWithInverseMetric(term, giID, Shnabla2TensorData);
};


// rule: full contraction of Riemann tensor with itself
// use FR.ruleFullContractionTwoTensors
FR.ruleFullContractionRiemannTensorWithItself = function(term, rulesData) {
	let RTSTensorData = rulesData.tensorData.RTS;
	return FR.ruleFullContractionTwoTensors(term, RTSTensorData);
};


// rule: full contraction of Ricci tensor with itself
// use FR.ruleFullContractionTwoTensors
FR.ruleFullContractionRicciTensorWithItself = function(term, rulesData) {
	let RicSTensorData = rulesData.tensorData.RicS;
	return FR.ruleFullContractionTwoTensors(term, RicSTensorData);
};


// rule: two copies of scalar curvature
// use FR.rulePoweredSameItem
FR.ruleTwoCopiesOfScalarCurvature = function(term, rulesData) {
	let SCSPowerData = rulesData.powerData.SCS;
	return FR.rulePoweredSameItem(term, SCSPowerData);
};


// minimal rules
FR.RULES_ARRAY_1 = [
	FR.ruleOperatorsAllOnlyu,
	FR.ruleContractionsMetricWithInverseMetric,
	FR.ruleContractionsInverseMetricWithRiemannTensor,
	FR.ruleContractionsInverseMetricWithRicciTensor,
	FR.ruleContractionsInverseMetricWithGaugeCurvature,
	FR.ruleContractionsRicciTensorWithGaugeCurvature,
	FR.ruleRaiseIndicesOfGaugeCurvatureWithInverseMetric,
	FR.ruleRaiseIndicesOfRiemannTensorWithInverseMetric,
	FR.ruleRaiseIndicesOfRicciTensorWithInverseMetric,
	FR.ruleRiemannCovariantRicciToScalar
];

// second list of rules: combine elements with themselves
// and deals with nabla operators
FR.RULES_ARRAY_2 = [
	FR.ruleFullContractionRiemannTensorWithItself,
	FR.ruleFullContractionRicciTensorWithItself,
	FR.ruleTwoCopiesOfScalarCurvature,
	FR.ruleLaplacian
];

// combine 2 sets of rules
FR.RULES_ARRAY_1_2 = FR.RULES_ARRAY_1.concat(FR.RULES_ARRAY_2);

// third list of rules: symmetric and antisymmetric parts of double derivative
FR.RULES_ARRAY_3 = [
	FR.ruleRaiseIndicesOfCommutatorFuWithInverseMetric,
	FR.ruleRaiseIndicesOfShnabla2WithInverseMetric,
	FR.ruleContractionsSymmetricLaplacianWithGaugeCurvature,
	FR.ruleContractionsRicciTensorWithCommutatorFu
];


module.exports = {
	FR: FR,
};
