/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FC} = require(REL_PATH + 'FUNCTIONS-FILES/functions-computation');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');
const {FR} = require(REL_PATH + 'FUNCTIONS-FILES/functions-rules');
const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument,
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');


const {FS} = require(REL_PATH + 'FUNCTIONS-FILES/functions-substitutions');
const {PROTO_NCT, SUBST_NCT} = require(REL_PATH + 'INIT-FILES/initialization-data-NCT');


// all the functions defined here are properties of the object FS_NCT
const FS_NCT = {};



// function used to create subsitution for noncommutative torus
// create a file containing the data for the substitution
// in normal coordinates of an expression obtained
// as the repeated application of a derivation on a MathElement
// return a latex string containing equations
// => save to display result
// initial data are given in initialData
// file path is given in initialData
// operatorData = {AppliedOperator: MathOperator to apply,
// 								NonCommutative: MathOperator instance,
// 								Commutative: MathOperator instance}
// where
// AppliedOperator is the one applied: used for the rule of derivation
// and appears in display
// NonCommutative is the one applied on noncommutative elements
// Commutative is the one applied on commutative elements
// placeHolderTensorData = {
// 	tensorUp: instance of MathElement with one and only one indice up,
// 	tensorDown: instance of MathElement with one and only one indice down
// }
// => tensors used to maintain contractions
// => appear in returned latex expression
// substArr = [subst1, subst2,…]
// where substi is a substitution rule applied at each level
FS_NCT.createSubstitutionsDataNCT = function(initialData, operatorData, placeHolderTensorData, substArr, rel_path) {
	// initial parameters
	let latexToReturn = '';
	let latexLetter = initialData.latexLetter;
	let saveSumFileRoot = initialData.sumFileRoot;
	let optLaTeX = initialData.optLaTeX;

	// create initial arrays
	let indicePositionsArr = [];
	let initialFromArr = [];
	for (let map of initialData.newItems[0].mapIndices) {
		let pos = initialData.orgItem.indices.list[map.from[0]].pos;
		indicePositionsArr.push(pos);
		initialFromArr.push(map.from);
	};
	// console.log('indicePositionsArr: ' + JSON.stringify(indicePositionsArr));

	// initial data used in FS.applyOperatorToMathProduct
	let dataArr = [];
	for (let obj of initialData.newItems) {
		let mapIndicesIDn = [];
		for (let map of obj.mapIndices) {
			mapIndicesIDn.push(obj.item.getTextIndiceAt(map.to));
		};
		dataArr.push({term: obj.item, mapIndicesIDn: mapIndicesIDn});
	};

	// the final substitution data saved in a file
	// to be read at runtime and used in computation
	let finalSubstitution = {};
	finalSubstitution.termElement = initialData.termElement;
	finalSubstitution.minLevel = initialData.minLevel;
	finalSubstitution.maxLevel = initialData.maxLevel;
	finalSubstitution.orgItemJSON = initialData.orgItem.toJSON();
	finalSubstitution.derivationOperatorJSON = initialData.derivationOperator.toJSON();
	// items are stored in their JSON format
	// => construct them at runtime

	// array to collect substitution data
	finalSubstitution.newItemsByDerivatives = [];

	// loop to generate derivatives (i>0)
	for (let i=0; i<=finalSubstitution.maxLevel; i++) {
		console.log('');
		console.log('i=' + i);
		let newDataArr = [];
		if (i>0) {
			// apply derivation on all terms and collect results
			for (data of dataArr) {
				let appliedOperator = FS.applyOperatorToMathProduct(data.term, operatorData, data.mapIndicesIDn);
				for (let obj of appliedOperator) {
					newDataArr.push({term: obj.term, mapIndicesIDn: obj.mapIndicesIDn});
				};
			};
			indicePositionsArr = [CONST.INDICE_DOWN].concat(indicePositionsArr);
		} else {
			// just use the data, no need to apply operator
			newDataArr = dataArr;
		};
		// transform to sum
		let sum = FS.constructMathSum(newDataArr, indicePositionsArr, placeHolderTensorData);
		for (let subst of substArr) {
			// console.log('apply subst');
			sum.substituteItemAndDerivativesWithData(subst);
		};
		// // substitution of first derivation of metric inverse in "main"
		// sum.substituteItemAndDerivativesWithData(SUBST.giToConstant);
		// sum.showConsole(OPT.SHOW, 'sum = ');

		// create sum in final form
		let subSum = sum.makeCopy();

		// substitute derivations of k by powers of delta
		subSum.substituteItemAndDerivativesWithData(SUBST_NCT.NABLAk);
		// collect commutative objects in "pre"
		subSum.collectCommutativeElementsInPreElement();
		// update expression to suppress "1" in pre element
		subSum.updateExpression();
		// sort pre element
		subSum.sortPreElementsByProductWeight();
		// // simplify expression using rules
		// for (let j=0; j<subSum.list.length; j++) {
		// 	FH.printProgressMessage('term number ' + j + '/' + subSum.list.length);
		// 	let term = subSum.list[j];
		// // TODO: use FR functions, with FR.RULES_ARRAY_1 and FR.RULES_ARRAY_2, or FR.RULES_ARRAY_1_2
		// // FC.applyRulesOnTerm(term, RULE_DATA);
		// FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
		// };
		// FH.printProgressMessage('');
		subSum.updateExpression();
		subSum.sortBySumWeight();
		subSum.fillNormalizedIndicesArray();
		subSum.simplifyExpression();
		subSum.showNormalized(OPT.SHOW, 'subSum = ');

		// export as latex
		latexToReturn += subSum.getLaTeXEquation(optLaTeX, latexLetter + '('+ i +') = ') + CONST.CR + CONST.CR;

		// save sum as json
		let sumFilePath = saveSumFileRoot + '-' + i + '.json';
		FH.saveItemToFile(subSum, rel_path + sumFilePath);

		// collect data for the substitutions file
		let substitutionDataArr = FS.constructDataArrFromSum(subSum, placeHolderTensorData);
		// console.log('substitutionDataArr: ' + JSON.stringify(substitutionDataArr));
		// create substitution data for this number of derivations
		let substitutionData = FS.createItemsByDerivativesData(substitutionDataArr, i, indicePositionsArr, initialFromArr);
		// remove 'term' since we collect only JSON version
		for (let substitutionTerm of substitutionData) {
			substitutionTerm.item = undefined;
		};
		// console.log('');
		// console.log('substitutionData: ' + JSON.stringify(substitutionData));
		// collect these data
		finalSubstitution.newItemsByDerivatives.push(substitutionData);

		// prepare the sum for the next iteration on derivations
		console.log('');
		// collect commutative objects in "pre"
		sum.collectCommutativeElementsInPreElement();
		// update expression to suppress "1" in pre element
		sum.updateExpression();
		// sort pre element
		sum.sortPreElementsByProductWeight();
		sum.fillNormalizedIndicesArray();
		sum.simplifyExpression();
		sum.showNormalized(OPT.SHOW, 'sum = ');
		// data to be used at beginning of the loop
		dataArr = FS.constructDataArrFromSum(sum, placeHolderTensorData);
	}; // loop on derivations

	// save substitution data in JSON file
	FH.saveToFileWithSureFolder(rel_path + initialData.filePath, JSON.stringify(finalSubstitution, null, "\t"));

	return latexToReturn;
};

module.exports = {
	FS_NCT: FS_NCT,
};
