/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const fs = require('fs');
const path = require('path');
const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FC} = require(REL_PATH + 'FUNCTIONS-FILES/functions-computation');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');
const {FR} = require(REL_PATH + 'FUNCTIONS-FILES/functions-rules');
const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');

const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument,
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');

// all the functions defined here are properties of the object FS
const FS = {};



// function used to apply differential operator on a MathProduct
// SUPPOSE MATHOPERATORS HAVE ONE AND ONLY ONE INDICE
FS.applyOperatorToMathProduct = function(item, operatorData, initialMapIndicesIDn) {
	// item = MathProduct
	// operatorData = {AppliedOperator: MathOperator to apply,
	// 								NonCommutative: MathOperator instance,
	// 								Commutative: MathOperator instance,
	// 								Christoffel: MathElement Christoffel symbol}
	// initialMapIndicesIDn = [ indiceIDn, …] listed in predefined order
	// the function will return [{term: MathProduct computed, mapIndicesIDn: [indiceIDn, …]}, …]
	// where mapIndicesIDn is updated for each produced "term":
	// 1- add the indiceIDn of the differential operator in front
	// 2- other indicesIDn are updated from initialMapIndicesIDn

	// test if item is a MathProduct
	if (item.constructor.name !== 'MathProduct') {
		throw 'MathProduct required';
	};

	let AppliedOperator = operatorData.AppliedOperator;

	let NCOperator = operatorData.NonCommutative;
	let COperator = operatorData.Commutative;
	// test if operators have same behavior
	if (NCOperator.behavior !== COperator.behavior) {
		throw 'Operators should have same behavior';
	};
	// test if operators are already applied
	if (NCOperator.status !== CONST.OPERATOR_STATUS_LATENT ||
			COperator.behavior !== CONST.OPERATOR_STATUS_LATENT) {
		throw 'Operators should not be applied';
	};
	// test productKind of MathOperator
	if (NCOperator.productKind !== CONST.ELEMENT_NON_COMMUTATIVE ||
			COperator.productKind !== CONST.ELEMENT_COMMUTATIVE) {
		throw 'Operators should not be applied';
	};

	// the data to return: [{term: MathProduct, mapIndicesIDn: [indiceID-n,…]}, …]
	let returnedArr = [];

	// get indiceID of original item
	let itemIndicesIDArr = item.getIndicesID();

	switch(AppliedOperator.behavior) {
		case CONST.OPERATOR_LEIBNIZ_RULE:
			// apply differential operator on every MathElement in the MathProduct
			// loop over factors in MathProduct
			for (let i=0; i<item.list.length; i++) {
				let newItem = item.makeCopy();
				let copiedInfoMapIndicesID = newItem.copiedInfo.mapIndicesID;
				let newItemIndicesIDArr = newItem.getIndicesID();
				let currentItem = newItem.getItemAtIndex(i);
				let currentOperator = null;
				switch(currentItem.productKind) {
					case CONST.ELEMENT_COMMUTATIVE:
						currentOperator = COperator.makeCopy();
						break;
					case CONST.ELEMENT_NON_COMMUTATIVE:
						currentOperator = NCOperator.makeCopy();
						break;
				};
				// get indiceIDn of the applied MathOperator
				let currentOperatorIndicesIDn = currentOperator.indices.id + '-0';

				// replace "currentItem" by "currentOperator applied on currentItem"
				let retInfo = {}; // no need here
				newItem.removeItemAtIndex(i, retInfo, false);
				currentOperator.applyTo(currentItem);
				newItem.insertItemAtIndex(currentOperator,i);

				// returned data for this currentItem
				let currentItemReturnedData = {};
				// add "newItem"
				currentItemReturnedData.term = newItem;
				// add new mapIndicesIDn
				let mapIndicesIDnArr = [currentOperatorIndicesIDn].concat(initialMapIndicesIDn);
				let mapIndicesIDnStr = JSON.stringify(mapIndicesIDnArr);
				// console.log('mapIndicesIDnStr (before): ' + mapIndicesIDnStr);
				// console.log('copiedInfoMapIndicesID: ' +JSON.stringify(copiedInfoMapIndicesID));
				for (let j=0; j < copiedInfoMapIndicesID.length; j++) {
					let oldIndID = copiedInfoMapIndicesID[j][0];
					let newIndID = copiedInfoMapIndicesID[j][1];
					let re = new RegExp(oldIndID, "g");
					mapIndicesIDnStr = mapIndicesIDnStr.replace(re, newIndID);
				};
				// console.log('mapIndicesIDnStr (after): ' + mapIndicesIDnStr);
				// parse the string into an object
				currentItemReturnedData.mapIndicesIDn = JSON.parse(mapIndicesIDnStr);

				returnedArr.push(currentItemReturnedData);
			}; // loop on list in item
			break;

		case CONST.OPERATOR_RIEMANN_RULE:
		// step 1: apply differential operator on every MathElement in the MathProduct
		// loop over factors in MathProduct
			for (let i=0; i<item.list.length; i++) {
				// console.log('apply Leibniz rule');
				// first, act with MathOperator
				let newItem = item.makeCopy();
				let copiedInfoMapIndicesID = newItem.copiedInfo.mapIndicesID;
				let newItemIndicesIDArr = newItem.getIndicesID();
				let currentItem = newItem.getItemAtIndex(i);
				let currentOperator = null;
				switch(currentItem.productKind) {
					case CONST.ELEMENT_COMMUTATIVE:
						currentOperator = COperator.makeCopy();
						break;
					case CONST.ELEMENT_NON_COMMUTATIVE:
						currentOperator = NCOperator.makeCopy();
						break;
				};
				// get indiceIDn of the applied MathOperator
				let currentOperatorIndicesIDn = currentOperator.indices.id + '-0';

				// replace "currentItem" by "currentOperator applied on currentItem"
				let retInfo = {}; // no need here
				newItem.removeItemAtIndex(i, retInfo, false);
				currentOperator.applyTo(currentItem);
				newItem.insertItemAtIndex(currentOperator,i);

				// returned data for this currentItem
				let currentItemReturnedData = {};
				// add "newItem"
				currentItemReturnedData.term = newItem;
				// add new mapIndicesIDn
				let mapIndicesIDnArr = [currentOperatorIndicesIDn].concat(initialMapIndicesIDn);
				let mapIndicesIDnStr = JSON.stringify(mapIndicesIDnArr);
				// console.log('mapIndicesIDnStr (before): ' + mapIndicesIDnStr);
				// console.log('copiedInfoMapIndicesID: ' +JSON.stringify(copiedInfoMapIndicesID));
				for (let j=0; j < copiedInfoMapIndicesID.length; j++) {
					let oldIndID = copiedInfoMapIndicesID[j][0];
					let newIndID = copiedInfoMapIndicesID[j][1];
					let re = new RegExp(oldIndID, "g");
					mapIndicesIDnStr = mapIndicesIDnStr.replace(re, newIndID);
				};
				// console.log('mapIndicesIDnStr (after): ' + mapIndicesIDnStr);
				// parse the string into an object
				currentItemReturnedData.mapIndicesIDn = JSON.parse(mapIndicesIDnStr);

				returnedArr.push(currentItemReturnedData);
			}; // loop on list in item

			// step 2: add contraction with Christoffel for every outer indice,
			// with plus or minus sign
			// collect positions of outer indices
			let outerIndicesArrIdx = item.getOuterIndices();
			// console.log('');
			// console.log('');
			// console.log('outerIndicesArrIdx in item: ' + JSON.stringify(outerIndicesArrIdx));

			// loop over these outer indices
			for (let i=0; i<outerIndicesArrIdx.length; i++) {
				let arrIdx = outerIndicesArrIdx[i];
				// console.log('');
				// console.log('apply Christoffel contraction');
				// console.log('arrIdx: ' + JSON.stringify(arrIdx));
				let newItem = item.makeCopy();
				let copiedInfoMapIndicesID = newItem.copiedInfo.mapIndicesID;
				let currentIndicePosition = newItem.getIndicePositionAt(arrIdx);
				let currentIndiceIDn = newItem.getTextIndiceAt(arrIdx);
				// newItem.showConsole(OPT.SHOW, 'newItem = ');
				// console.log('currentIndiceIDn: ' + currentIndiceIDn)
				// console.log('currentIndicePosition: ' + currentIndicePosition);

				let newChristoffel = operatorData.Christoffel.makeCopy();
				// newChristoffel.showConsole(OPT.SHOW, 'Christoffel = ');
				let currentOperatorIndicesIDn = newChristoffel.getTextIndiceAt([1]);
				// console.log('currentOperatorIndicesIDn: ' + currentOperatorIndicesIDn);

				// multiply by Christoffel
				newItem.multiplyOnLeftWith(newChristoffel, false);
				// collect two remaining indiceIDn from Christoffel
				// depending on indice position => not the same role
				let contractedIndiceIDn = '';
				let replacementIndiceIDn = '';
				if (currentIndicePosition === CONST.INDICE_DOWN) {
					// change sign if down indice
					newItem.multiplyByInteger(-1);
					contractedIndiceIDn = newChristoffel.getTextIndiceAt([0]);
					replacementIndiceIDn = newChristoffel.getTextIndiceAt([2]);
				} else {
					contractedIndiceIDn = newChristoffel.getTextIndiceAt([2]);
					replacementIndiceIDn = newChristoffel.getTextIndiceAt([0]);
				};
				// add contraction: [currentIndiceIDn,contractedIndiceIDn];
				let newContraction = [currentIndiceIDn,contractedIndiceIDn].sort(FH.alphaSort);
				newItem.contractions.push(newContraction);

				// newItem.showConsole(OPT.SHOW, 'newItem (after) = ');
				// returned data for this outer indice
				let currentItemReturnedData = {};
				// add "newItem"
				currentItemReturnedData.term = newItem;
				// add new mapIndicesIDn
				let mapIndicesIDnArr = [currentOperatorIndicesIDn].concat(initialMapIndicesIDn);
				let mapIndicesIDnStr = JSON.stringify(mapIndicesIDnArr);
				for (let j=0; j < copiedInfoMapIndicesID.length; j++) {
					let oldIndID = copiedInfoMapIndicesID[j][0];
					let newIndID = copiedInfoMapIndicesID[j][1];
					let re = new RegExp(oldIndID, "g");
					mapIndicesIDnStr = mapIndicesIDnStr.replace(re, newIndID);
				};
				// replace currentIndiceIDn by replacementIndiceIDn
				let re = new RegExp(currentIndiceIDn, "g");
				mapIndicesIDnStr = mapIndicesIDnStr.replace(re, replacementIndiceIDn);
				// parse the string into an object
				currentItemReturnedData.mapIndicesIDn = JSON.parse(mapIndicesIDnStr);

				returnedArr.push(currentItemReturnedData);
			}; // loop over outer indices
			break;

		default:
			throw 'MathOperator behavior not implemented';
			break;
	};

	return returnedArr;
};


FS.constructMathTensorProduct = function(dataObj, indicePositionsArr, placeHolderTensorData) {
	// dataObj = {term: MathProduct, mapIndicesIDn: [indiceIDn, …]}
	// indicePositionsArr = [INDICE_UP || INDICE_DOWN, …] in the order of mapIndicesIDn
	// placeHolderTensorData = {tensorUp: MathElement tensor with up indice,
	// 													tensorDown: MathElement tensor with down indice}
	// the function will return a MathTensorProduct (product of 'placeHolderTensor') otimes 'term'
	// where all the indices of 'term' are contracted with the unique indice of 'placeHolderTensor'
	// which appear in sufficiently quantity

	let term = dataObj.term;
	let mapIndicesIDn = dataObj.mapIndicesIDn;
	// console.log('');
	// term.showConsole(OPT.SHOW, '(construct TP) term = ');
	// console.log('(construct TP) mapIndicesIDn: ' + JSON.stringify(mapIndicesIDn));

	let placeHolderMathProduct = new MathProduct();
	let contractionsArr = [];
	if (mapIndicesIDn.length !== indicePositionsArr.length) {
		throw "mapIndicesIDn and indicePositionsArr should have same length";
	};
	// loop over indices: a place holder tensor by indice
	for(let j=0; j<indicePositionsArr.length; j++) {
		let currentPlaceHolder = null;
		let indicePosition = indicePositionsArr[j];
		let currentIndiceIDn = mapIndicesIDn[j];
		// console.log('currentIndiceIDn: ' + currentIndiceIDn);
		// tensor down when indice up and vice-versa
		switch(indicePosition) {
			case CONST.INDICE_UP:
				currentPlaceHolder = placeHolderTensorData.tensorDown.makeCopy();
				break;
			case CONST.INDICE_DOWN:
				currentPlaceHolder = placeHolderTensorData.tensorUp.makeCopy();
				break;
		};
		let placeHolderIndiceIDn = currentPlaceHolder.getTextIndiceAt([0]);
		// multiply this copy on right, no more copy
		placeHolderMathProduct.multiplyOnRightWith(currentPlaceHolder, false);
		// contraction
		contractionsArr.push([placeHolderIndiceIDn, currentIndiceIDn].sort(FH.alphaSort));
	}; // loop over indicePositionsArr
	// placeHolderMathProduct.showConsole(OPT.SHOW, 'placeHolderMathProduct = ');
	// construction of tensor product:
	let returnedTensorProduct = MathExpression.convertTo(placeHolderMathProduct, 'MathTensorProduct');
	// returnedTensorProduct.showConsole(OPT.SHOW, 'returnedTensorProduct = ');
	// insert 'term' as the second factor in the tensor product
	// insert as a MathProduct to accept MathElement as input
	// term.showConsole(OPT.SHOW, 'term = ');
	returnedTensorProduct.list.push(MathExpression.convertTo(term,'MathProduct'));
	// returnedTensorProduct.showConsole(OPT.SHOW, 'returnedTensorProduct = ');
	// console.log('contractionsArr: ' + JSON.stringify(contractionsArr));
	// console.log('currentMathProduct.contractions: ' + JSON.stringify(currentMathProduct.contractions));
	returnedTensorProduct.contractions = FH.joinContractions(contractionsArr, term.contractions);
	// returnedTensorProduct.showConsole(OPT.SHOW, 'returnedTensorProduct = ');
	returnedTensorProduct.updateExpression();

	return returnedTensorProduct;
};


// function used to apply differential operator on a MathProduct
// SUPPOSE MATHOPERATORS HAVE ONE AND ONLY ONE INDICE
FS.constructMathSum = function(dataArr, indicePositionsArr, placeHolderTensorData) {
	// dataArr = [{term: MathProduct computed, mapIndicesIDn: [indiceIDn, …]}, …]
	// indicePositionsArr = [INDICE_UP || INDICE_DOWN, …] in the order of mapIndicesIDn
	// placeHolderTensorData = {tensorUp: MathElement tensor with up indice,
	// 													tensorDown: MathElement tensor with down indice}
	// the function will return a MathSum of MathArgument on which simplifications can be applied

	let returnedMathSum = new MathSum();
	// loop over data elements => one term of the sum
	for (let i=0; i<dataArr.length; i++) {
		// create tensor product with place holder tensors for full contraction of indices
		// console.log('create sum i=' + i);
		// dataArr[i].term.showConsole(OPT.SHOW, 'term = ');
		// console.log('mapIndicesIDn: ' + JSON.stringify(dataArr[i].mapIndicesIDn));
		let currentTerm = FS.constructMathTensorProduct(dataArr[i], indicePositionsArr, placeHolderTensorData);
		// make it a MathArgument
		currentTerm = MathExpression.convertTo(currentTerm, 'MathArgument');
		// add to the sum
		returnedMathSum.addTo(currentTerm, false);
	}; // loop over dataArr

	return returnedMathSum;
};


// from a sum of MathArgument, with "placeHolderTensor" otimes "term"
// get the list of terms, including preElements
// and get list of mapIndicesIDn
FS.constructDataArrFromSum = function(sum, placeHolderTensorData) {
	// placeHolderTensorData = {tensorUp: MathElement tensor with up indice,
	// 													tensorDown: MathElement tensor with down indice}
	// the function returns
	// dataArr = [{term: MathProduct, mapIndicesIDn: [indiceIDn, …]}, …]
	// return empty array if sum is zero

	let dataArr = [];
	// test if sum is a MathSum
	if (sum.constructor.name !== 'MathSum') {
		throw 'MathSum required';
	};

	for (let term of sum.list) {
		// console.log('');
		// term.showConsole(OPT.SHOW, 'term = ');
		// test if term is a MathArgument
		if (term.constructor.name !== 'MathArgument') {
			throw 'MathArgument required';
		};
		// test main list is of length = 2
		if (term.list.length !== 2) {
			throw 'Misformatted term';
		};
		// collect the 3 parts used after
		// post element is ignored
		let preElement = term.preElement;
		let placeHolderMathProduct = term.list[0];
		let currentMathProduct = term.list[1];

		// preElement.showConsole(OPT.SHOW, 'preElement = ');
		// placeHolderMathProduct.showConsole(OPT.SHOW, 'placeHolderMathProduct = ');
		// currentMathProduct.showConsole(OPT.SHOW, 'currentMathProduct = ');

		// multiply preElement and currentMathProduct
		currentMathProduct.multiplyOnLeftWith(preElement, false);

		// construct contractions for this product
		// and construct mapIndicesIDn
		let contractionsArr = term.contractions;
		// console.log('contractionsArr: ' + JSON.stringify(contractionsArr));
		let newContractionsArr = [];
		let mapIndicesIDn = [];
		for (let cont of contractionsArr) {
			let found = false;
			for (let factor of placeHolderMathProduct.list) {
				let factorIndiceIDn = factor.indices.id + '-0';
				// console.log('factorIndiceIDn: ' + factorIndiceIDn);
				let idx = cont.indexOf(factorIndiceIDn);
				// console.log('idx: ' + idx);
				if (idx>-1) {
					found = true;
					let idx2 = idx === 0 ? 1 : 0;
					// console.log('idx2: ' + idx2);
					mapIndicesIDn.push(cont[idx2]);
				};
			};
			if (!found) {
				newContractionsArr.push(cont);
			};
		};
		// console.log('newContractionsArr: ' + JSON.stringify(newContractionsArr));
		// console.log('mapIndicesIDn: ' + JSON.stringify(mapIndicesIDn));
		currentMathProduct.contractions = newContractionsArr;
		// currentMathProduct.showConsole(OPT.SHOW, 'currentMathProduct = ');

		// collect data
		dataArr.push({term: currentMathProduct, mapIndicesIDn: mapIndicesIDn})
	};

	return dataArr;
};


// create a array of obj containing the data at level i of derivation
// original data are in substitutionDataArr
// indicePositionsArr is the ordered list of position (up or down) of indices
// initialFromArr is the (ordered) list of indices in orginal expression
FS.createItemsByDerivativesData = function(substitutionDataArr, i, indicePositionsArr, initialFromArr) {

	// console.log('substitutionDataArr (before): ' + JSON.stringify(substitutionDataArr));

	// the returned data used in substitution
	// at this level of derivation
	let substitutionData = [];

	// permutation
	// the code generates only one permutation (identity)
	// since all the terms are explicitly listed in an array
	let permutationsSum = [];
	for (let k=0; k<indicePositionsArr.length; k++) {
		permutationsSum.push(k);
	};
	if (permutationsSum.length>0) {
		permutationsSum = [permutationsSum];
	};

	// construct each new term for each obj in substitutionDataArr
	if (substitutionDataArr.length === 0) {
		// in this situation, the result is 0
		// fill with null, here encoded as the string '0'
		// will be decoded as 'null' when reading the saved file
		let substitutionTerm = {};
		substitutionTerm.itemJSON = '0';
		// add this object
		substitutionData.push(substitutionTerm);
	} else {
		// loop on obj (new terms)
		for (let obj of substitutionDataArr) {
			// the two elements of obj
			let term = obj.term;
			let mapIndicesIDn = obj.mapIndicesIDn;

			// the object where to collect data in the format of a substitution
			let substitutionTerm = {};

			// from: first the derivations (number of them given by i)
			let fromIdxArr = [];
			for (let k=1; k<=i; k++) {
				fromIdxArr.push([-k,0]);
			};
			// from: then the "initial" data
			if (i>0) {
				for (let kArr of initialFromArr) {
					fromIdxArr.push([-(i+1),kArr[0]]);
				};
			} else {
				fromIdxArr = initialFromArr;
			};
			// to: computed using mapIndicesIDn
			let toIdxArr = FH.getIdxArrFromIndicesIDnArr(mapIndicesIDn, term);

			substitutionTerm.mapIndices = [];
			for (let k=0; k<fromIdxArr.length; k++) {
				substitutionTerm.mapIndices.push({from: fromIdxArr[k], to: toIdxArr[k]});
			};

			// makeCopy before change Indice to preserve original term
			// obj.term.showConsole(OPT.SHOW, 'obj.term (before): ');
			substitutionTerm.item = term.makeCopy().changeIndicesIDToNumber();
			substitutionTerm.itemJSON = substitutionTerm.item.toJSON();
			// obj.term.showConsole(OPT.SHOW, 'obj.term (after): ');
			substitutionTerm.permutationsSum = permutationsSum;

			// add this object
			substitutionData.push(substitutionTerm);
		}; // loop substitutionDataArr
	}; // test result is 0

	// console.log('substitutionDataArr (after): ' + JSON.stringify(substitutionDataArr));

	return substitutionData;
};


// one of the main functions in this file
// create a file containing the data for the substitution
// in normal coordinates of an expression obtained
// as the repeated application of a derivation on a MathElement
// return a latex string containing equations
// => save to display result
// initial data are given in initialData
// file path is given in initialData
// operatorData = {AppliedOperator: MathOperator to apply,
// 								NonCommutative: MathOperator instance,
// 								Commutative: MathOperator instance,
// 								Christoffel: MathElement Christoffel symbol}
// where
// AppliedOperator is the one applied: used for the rule of derivation
// and appears in display
// NonCommutative is the one applied on noncommutative elements
// Commutative is the one applied on commutative elements
// Christoffel is the used for Riemann covariant derivatives
// TODO: if a Riemann covariant derivative is used as AppliedOperator
// then substitute Christoffel in normal coordinates
// (which can be computed by this code and save before)
// placeHolderTensorData = {
// 	tensorUp: instance of MathElement with one and only one indice up,
// 	tensorDown: instance of MathElement with one and only one indice down
// }
// => tensors used to maintain contractions
// => appear in returned latex expression
FS.createSubstitutionsDataNormCoord = function(initialData, operatorData, placeHolderTensorData, rel_path) {
	// initial parameters
	let latexToReturn = '';
	let latexLetter = initialData.latexLetter;
	let saveSumFileRoot = initialData.sumFileRoot;
	let optLaTeX = initialData.optLaTeX;

	// create initial arrays
	let indicePositionsArr = [];
	let initialFromArr = [];
	for (let map of initialData.newItems[0].mapIndices) {
		let pos = initialData.orgItem.indices.list[map.from[0]].pos;
		indicePositionsArr.push(pos);
		initialFromArr.push(map.from);
	};
	// console.log('indicePositionsArr: ' + JSON.stringify(indicePositionsArr));

	// initial data used in FS.applyOperatorToMathProduct
	let dataArr = [];
	for (let obj of initialData.newItems) {
		let mapIndicesIDn = [];
		for (let map of obj.mapIndices) {
			mapIndicesIDn.push(obj.item.getTextIndiceAt(map.to));
		};
		dataArr.push({term: obj.item, mapIndicesIDn: mapIndicesIDn});
	};

	// the final substitution data saved in a file
	// to be read at runtime and used in computation
	let finalSubstitution = {};
	finalSubstitution.termElement = initialData.termElement;
	finalSubstitution.minLevel = initialData.minLevel;
	finalSubstitution.maxLevel = initialData.maxLevel;
	finalSubstitution.orgItemJSON = initialData.orgItem.toJSON();
	finalSubstitution.derivationOperatorJSON = initialData.derivationOperator.toJSON();
	// items are stored in their JSON format
	// => construct them at runtime

	// array to collect substitution data
	finalSubstitution.newItemsByDerivatives = [];

	// loop to generate derivatives (i>0)
	for (let i=0; i<=finalSubstitution.maxLevel; i++) {
		console.log('');
		console.log('i=' + i);
		let newDataArr = [];
		if (i>0) {
			// apply derivation on all terms and collect results
			for (data of dataArr) {
				let appliedOperator = FS.applyOperatorToMathProduct(data.term, operatorData, data.mapIndicesIDn);
				for (let obj of appliedOperator) {
					newDataArr.push({term: obj.term, mapIndicesIDn: obj.mapIndicesIDn});
				};
			};
			indicePositionsArr = [CONST.INDICE_DOWN].concat(indicePositionsArr);
		} else {
			// just use the data, no need to apply operator
			newDataArr = dataArr;
		};
		// transform to sum
		let sum = FS.constructMathSum(newDataArr, indicePositionsArr, placeHolderTensorData);
		// substitution of first derivation of metric inverse in "main"
		sum.substituteItemAndDerivativesWithData(SUBST.gi);
		// sum.showConsole(OPT.SHOW, 'sum = ');

		// create sum in normal coordinates
		let NCSum = sum.makeCopy();
		// substitute metric to normal coordinates
		NCSum.substituteItemAndDerivativesWithData(SUBST.gNC);
		// collect commutative objects in "pre"
		NCSum.collectCommutativeElementsInPreElement();
		// update expression to suppress "1" in pre element
		NCSum.updateExpression();
		// sort pre element
		NCSum.sortPreElementsByProductWeight();
		// simplify expression using rules
		for (let j=0; j<NCSum.list.length; j++) {
			FH.printProgressMessage('term number ' + j + '/' + NCSum.list.length);
			let term = NCSum.list[j];
			// TODO: use FR functions, with FR.RULES_ARRAY_1 and FR.RULES_ARRAY_2, or FR.RULES_ARRAY_1_2
			// FC.applyRulesOnTerm(term, RULE_DATA);
			FR.applyArrayOfRulesOnTerm(FR.RULES_ARRAY_1_2, term, RULE_DATA);
		};
		FH.printProgressMessage('');
		NCSum.updateExpression();
		NCSum.sortBySumWeight();
		NCSum.fillNormalizedIndicesArray();
		NCSum.simplifyExpression();
		NCSum.showNormalized(OPT.SHOW, 'NCSum = ');

		// export as latex
		latexToReturn += NCSum.getLaTeXEquation(optLaTeX, latexLetter + '('+ i +') = ') + CONST.CR + CONST.CR;

		// save sum as json
		let sumFilePath = saveSumFileRoot + '-' + i + '.json';
		FH.saveItemToFile(NCSum, rel_path + sumFilePath);

		// collect data for the substitutions file
		let substitutionDataArr = FS.constructDataArrFromSum(NCSum, placeHolderTensorData);
		// create substitution data for this number of derivations
		let substitutionData = FS.createItemsByDerivativesData(substitutionDataArr, i, indicePositionsArr, initialFromArr);
		// remove 'term' since we collect only JSON version
		for (let substitutionTerm of substitutionData) {
			substitutionTerm.item = undefined;
		};
		// collect these data
		finalSubstitution.newItemsByDerivatives.push(substitutionData);

		// prepare the sum for the next iteration on derivations
		console.log('');
		// collect commutative objects in "pre"
		sum.collectCommutativeElementsInPreElement();
		// update expression to suppress "1" in pre element
		sum.updateExpression();
		// sort pre element
		sum.sortPreElementsByProductWeight();
		sum.fillNormalizedIndicesArray();
		sum.simplifyExpression();
		sum.showNormalized(OPT.SHOW, 'sum = ');
		// data to be used at beginning of the loop
		dataArr = FS.constructDataArrFromSum(sum, placeHolderTensorData);
	}; // loop on derivations

	// save substitution data in JSON file
	FH.saveToFileWithSureFolder(rel_path + initialData.filePath, JSON.stringify(finalSubstitution, null, "\t"));

	return latexToReturn;
};


// one of the main functions in this file
// create a file containing the data for the substitution
// in normal coordinates of an expression obtained
// as the repeated application of gauge derivative (nabla) on a MathElement
// into repeated applications of a Riemann covariant derivative (hnabla)
// return a latex string containing equations
// => save to display result
// initial data are given in initialData
// file path is given in initialData
// operatorData = {AppliedOperator: MathOperator to apply,
// 								NonCommutative: MathOperator instance,
// 								Commutative: MathOperator instance,
// 								Christoffel: MathElement Christoffel symbol}
// where
// AppliedOperator is the one applied: used for the rule of derivation
// and appears in display
// NonCommutative is the one applied on noncommutative elements
// Commutative is the one applied on commutative elements
// Christoffel is the one used for Riemann covariant derivatives
// placeHolderTensorData = {
// 	tensorUp: instance of MathElement with one and only one indice up,
// 	tensorDown: instance of MathElement with one and only one indice down
// }
// => tensors used to maintain contractions
// => appear in returned latex expression
// the stategy is to compute repeated applications of hnabla
// in terms of repeated applications of nabla
// then express repeated applications of nabla in terms
// of repeated applications of hnabla
// then extract a rule at each level and apply recursively this rule for next level…
// so here we have to maintain two substitutions data:
// the one to be saved
// the one to be used recursively
FS.createSubstitutionsDataGaugeToRiemann = function(initialData, operatorData, placeHolderTensorData, rel_path) {
	// initial parameters
	let latexToReturn = '';
	let latexLetter = initialData.latexLetter;
	let saveSumFileRoot = initialData.sumFileRoot;
	let optLaTeX = initialData.optLaTeX;

	let toNC = initialData.toNormalCoordinates;
	let substitutionsGamma = null;
	// convert to normal coodinates?
	if (toNC) {
		console.log('load Gamma substitutions');
		substitutionsGamma = FC.getSubstitutionsFromFile(rel_path + initialData.fileSubstitutionsGammaToNC);
	};

	// initial element item on which to apply the derivations
	let orgItem = initialData.orgItem.makeCopy();
	let initialExpression = MathProduct.newFrom(orgItem);
	orgItem.showConsole(OPT.SHOW, 'orgItem = ');
	initialExpression.showConsole(OPT.SHOW, 'initialExpression = ');

	// initialize initialFromArr from orgItem
	let initialFromArr = orgItem.getIndicesArrIdx();
	console.log('initialFromArr: ' + JSON.stringify(initialFromArr));

	let indicesArrIdx = initialExpression.getIndicesArrIdx();
	console.log('indicesArrIdx: ' + JSON.stringify(indicesArrIdx));
	// initialize mapIndicesIDn with outerIndicesArrIdx
	// initialize indicePositionsArr with outerIndicesArrIdx
	let mapIndicesIDn = [];
	let indicePositionsArr = [];
	for (let arrIdx of indicesArrIdx) {
		mapIndicesIDn.push(initialExpression.getTextIndiceAt(arrIdx));
		indicePositionsArr.push(initialExpression.getIndicePositionAt(arrIdx));
	};
	// console.log('mapIndicesIDn: ' + JSON.stringify(mapIndicesIDn));
	// console.log('indicePositionsArr: ' + JSON.stringify(indicePositionsArr));

	// initial data
	// here just the initial expression
	let dataArr = [];
	dataArr.push({term: initialExpression, mapIndicesIDn: mapIndicesIDn});
	// console.log('dataArr: ' + JSON.stringify(dataArr));

	// the final substitution data saved in a file
	// to be read at runtime and used in computation
	let finalSubstitution = {};
	finalSubstitution.termElement = initialData.termElement;
	finalSubstitution.minLevel = initialData.minLevel;
	finalSubstitution.maxLevel = initialData.maxLevel;
	finalSubstitution.orgItemJSON = initialData.orgItem.toJSON();
	finalSubstitution.derivationOperatorJSON = initialData.derivationOperator.toJSON();
	// items are stored in their JSON format
	// => construct them at runtime

	// array to collect substitution data
	finalSubstitution.newItemsByDerivatives = [];

	// the substitution constructed step by step and used at each level
	let activeSubstitution = {};
	activeSubstitution.termElement = initialData.termElement;
	activeSubstitution.minLevel = initialData.minLevel;
	activeSubstitution.maxLevel = 1; // start with 1 and then increase at each level
	activeSubstitution.orgItem = initialData.orgItem;
	// activeSubstitution.derivationOperator = initialData.derivationOperator;
	activeSubstitution.derivationOperators = [initialData.derivationOperator];
	// array to collect substitution data
	activeSubstitution.newItemsByDerivatives = [];

	// collect consecutive applications of the two derivations
	let orgItemMapIndicesIDn = [];
	for (let arrIdx of initialFromArr) {
		orgItemMapIndicesIDn.push(orgItem.getTextIndiceAt(arrIdx));
	};
	let appliedOpDataArr = [];
	appliedOpDataArr.push({term: orgItem, mapIndicesIDn: orgItemMapIndicesIDn});
	let toRemoveOpDataArr = [];
	toRemoveOpDataArr.push({term: orgItem, mapIndicesIDn: orgItemMapIndicesIDn});

	// no derivation: data are just trivial and will not be used
	let substitutionData = FS.createItemsByDerivativesData(dataArr, 0, indicePositionsArr, initialFromArr);
	// console.log('i=0 substitutionData: ' + JSON.stringify(substitutionData));
	// collect these data
	activeSubstitution.newItemsByDerivatives.push(substitutionData);

	// console.log('dataArr: ' + JSON.stringify(dataArr));
	// loop to generate derivatives (i>0)
	for (let i=1; i<=finalSubstitution.maxLevel; i++) {
		console.log('');
		console.log('i=' + i);

		let newDataArr = [];

		for (data of dataArr) {
			// apply operator on elements and collect the result
			// data.term.showConsole(OPT.SHOW, 'data.term = ');
			// console.log('data.mapIndicesIDn: ' + JSON.stringify(data.mapIndicesIDn));
			let appliedOperator = FS.applyOperatorToMathProduct(data.term, operatorData, data.mapIndicesIDn);
			for (let obj of appliedOperator) {
				newDataArr.push({term: obj.term, mapIndicesIDn: obj.mapIndicesIDn});
			};
		};
		indicePositionsArr = [CONST.INDICE_DOWN].concat(indicePositionsArr);

		// for (let obj of newDataArr) {
		// 	obj.term.showConsole(OPT.SHOW, 'obj term = ');
		// 	console.log('obj mapIndicesIDn: ' + JSON.stringify(obj.mapIndicesIDn));
		// };

		// transform to sum
		console.log('create sum');
		let sum = FS.constructMathSum(newDataArr, indicePositionsArr, placeHolderTensorData);
		// sum.showConsole(OPT.SHOW, 'sum = ');

		// make a copy of the sum now before collectCommutativeElementsInPreElement()
		// this copy is used later to compute subsitutions
		let newSum = sum.makeCopy();

		// prepare the sum for the next iteration on derivations
		console.log('simplify sum');
		// collect commutative objects in "pre"
		sum.collectCommutativeElementsInPreElement();
		// update expression to suppress "1" in pre element
		sum.updateExpression();
		// sort pre element
		sum.sortPreElementsByProductWeight();
		sum.fillNormalizedIndicesArray();
		sum.simplifyExpression();
		sum.showNormalized(OPT.SHOW, 'sum = ');

		// export as latex
		latexToReturn += sum.getLaTeXEquation(optLaTeX, 'u' + '('+ i +') = ') + CONST.CR + CONST.CR;

		// create structure to compute substitutions
		console.log('');
		console.log('create structure to compute substitutions');
		let currentIndiceIDn = '';
		// derivatives applied (left hand side) on original item
		let currentAppliedOpTerm = operatorData.AppliedOperator.makeCopy();
		currentIndiceIDn = currentAppliedOpTerm.getTextIndiceAt([0]);
		// console.log('currentIndiceIDn: ' + currentIndiceIDn);
		// console.log('data at i-1:');
		// appliedOpDataArr[i-1].term.showConsole(OPT.SHOW, 'term = ');
		// console.log('mapIndicesIDn: ' +JSON.stringify(appliedOpDataArr[i-1].mapIndicesIDn));
		currentAppliedOpTerm.applyTo(appliedOpDataArr[i-1].term);
		currentAppliedOpMapIndicesIDn = [currentIndiceIDn].concat(appliedOpDataArr[i-1].mapIndicesIDn);
		// currentAppliedOpTerm.showConsole(OPT.SHOW, 'curr. appl. op. = ');
		// console.log('currentAppliedOpMapIndicesIDn: ' + JSON.stringify(currentAppliedOpMapIndicesIDn));
		appliedOpDataArr.push({term: currentAppliedOpTerm,
														mapIndicesIDn: currentAppliedOpMapIndicesIDn});

		// derivatives to remove/replace (in the right hans side) on original item
		let currentToRemoveOpTerm = operatorData.NonCommutative.makeCopy();
		currentIndiceIDn = currentToRemoveOpTerm.getTextIndiceAt([0]);
		// console.log('currentIndiceIDn: ' + currentIndiceIDn);
		currentToRemoveOpTerm.applyTo(toRemoveOpDataArr[i-1].term);
		currentToRemoveOpMapIndicesIDn = [currentIndiceIDn].concat(toRemoveOpDataArr[i-1].mapIndicesIDn);
		// currentToRemoveOpTerm.showConsole(OPT.SHOW, 'curr. to rem. op. = ');
		toRemoveOpDataArr.push({term: currentToRemoveOpTerm,
														mapIndicesIDn: currentToRemoveOpMapIndicesIDn});

		// here 'term' are MathElement but FS.constructMathTensorProduct
		// converts to MathProduct
		let currentAppliedOpTensorProduct = FS.constructMathTensorProduct(appliedOpDataArr[i], indicePositionsArr, placeHolderTensorData);
		// currentAppliedOpTensorProduct.showConsole(OPT.SHOW, 'curr. appl. op. = ');

		let currentToRemoveOpTensorProduct = FS.constructMathTensorProduct(toRemoveOpDataArr[i], indicePositionsArr, placeHolderTensorData);
		// currentToRemoveOpTensorProduct.showConsole(OPT.SHOW, 'curr. to rem. op. = ');

		// intertwin hnabla and nabla in the equation
		// presently: LHS = hnabla^n, RHS = nabla^n + remainder
		// change to: LHS = nabla^n, RHS = hnabla^n - remainder
		// first change sign of the sum
		newSum.multiplyByInteger(-1);
		// add hnabla^n which is currentAppliedOpTensorProduct converted to MathArgument
		newSum.addTo(MathExpression.convertTo(currentAppliedOpTensorProduct, 'MathArgument'));
		// add nabla^n (to cancel it) which is currentToRemoveOpTensorProduct converted to MathArgument
		newSum.addTo(MathExpression.convertTo(currentToRemoveOpTensorProduct, 'MathArgument'));
		// newSum.showConsole(OPT.SHOW, 'newSum = ');
		newSum.updateExpression();
		newSum.sortBySumWeight();
		newSum.fillNormalizedIndicesArray();
		newSum.simplifyExpression();
		newSum.showNormalized(OPT.SHOW, 'new sum = ');
		// console.log(newSum.toJSON());

		// convert to normal coodinates?
		// don't apply collectCommutativeElementsInPreElement() before
		if (toNC) {
			console.log('substitute Christoffel')
			newSum.substituteItemAndDerivativesWithData(substitutionsGamma);
			// collect commutative objects in "pre"
			// console.log('collect in pre element');
			newSum.collectCommutativeElementsInPreElement();
			// sort pre element
			// console.log('sort pre element');
			newSum.sortPreElementsByProductWeight();
			// sort sum
			// console.log('sort sum');
			newSum.sortBySumWeight();
			// console.log('update expression');
			newSum.updateExpression();
			// console.log('fill normalized indices');
			newSum.fillNormalizedIndicesArray();
			// console.log('simplify');
			newSum.simplifyExpression();
			newSum.showNormalized(OPT.SHOW, 'new sum = ');
		};


		if (i>1) {
			console.log('substitute with previous level');
			activeSubstitution.maxLevel = i-1;
			// TODO: apply substitutions at level i-1
			newSum.substituteItemAndDerivativesWithData(activeSubstitution);

			console.log('new sum after substitutions');
			newSum.updateExpression();
			// collect commutative objects in "pre"
			newSum.collectCommutativeElementsInPreElement();
			// sort pre element
			newSum.sortPreElementsByProductWeight();
			newSum.sortBySumWeight();
			newSum.fillNormalizedIndicesArray();
			newSum.simplifyExpression();
			newSum.showNormalized(OPT.SHOW, 'new sum = ');
		};

		// collect data for substitutions
		let substitutionDataArr = FS.constructDataArrFromSum(newSum, placeHolderTensorData);
		// create substitution data for this number of derivations
		let substitutionData = FS.createItemsByDerivativesData(substitutionDataArr, i, indicePositionsArr, initialFromArr);
		activeSubstitution.newItemsByDerivatives.push(substitutionData);

		// data to be used at beginning of the loop
		dataArr = FS.constructDataArrFromSum(sum, placeHolderTensorData);

	}; // loop on derivation level

	// show substitutions
	activeSubstitution.maxLevel = initialData.maxLevel;
	latexSubstitution = FC.showSubstitution(activeSubstitution);


	// prepare for the file version of substitutions: remove 'term'
	for (let i=0; i<activeSubstitution.newItemsByDerivatives.length; i++) {
		let substitutionData = activeSubstitution.newItemsByDerivatives[i];
		// remove 'term' for JSON version
		for (let substitutionTerm of substitutionData) {
			substitutionTerm.item = undefined;
		};
		finalSubstitution.newItemsByDerivatives.push(substitutionData);
	};
	console.log(finalSubstitution.newItemsByDerivatives.length);
	// save substitution data in JSON file
	FH.saveToFileWithSureFolder(rel_path + initialData.filePath, JSON.stringify(finalSubstitution, null, "\t"));



	return latexToReturn;
};


module.exports = {
	FS: FS,
};
