/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');

const {
				Rational,
				Integer
			} = require(REL_PATH + 'MATH-OBJECTS/MathRational');

const {
				Polynomials,
				RationalExpression,
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument,
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');


const {PROTO, SUBST, RULE_DATA, SELECTION} = require(REL_PATH + 'INIT-FILES/initialization-data');

// MathElement and MathProduct from which to construct the others
// for noncommutative torus
const PROTO_NCT = {};

PROTO_NCT.delta = new MathOperator();
PROTO_NCT.delta
	.setID('delta')
	.setLetter('δ')
	.setMathLetter(CONST.MATH_DELTA)
	.setLaTeXLetter('\\delta')
	.addIndiceDown()
	.setSumWeight(16)
	.setProductWeight(16);

// for Noncommutative torus
PROTO_NCT.k = new MathElement();
PROTO_NCT.k
	.setID('k')
	.setLetter('k')
	.setMathLetter('k')
	.setLaTeXLetter('k')
	.setProductKindToNonCommutative()
	.setSumWeight(22)
	.setProductWeight(22);


// h tensor (see paper)
PROTO_NCT.hi = new MathElement();
PROTO_NCT.hi
	.setID('hi')
	.setLetter('h')
	.setMathLetter('h')
	.setLaTeXLetter('h')
	.setProductKindToCommutative()
	.addIndiceUp()
	.addIndiceUp()
	.setSumWeight(40)
	.setProductWeight(40);

// f tensor (see paper)
PROTO_NCT.fi = new MathElement();
PROTO_NCT.fi
	.setID('fi')
	.setLetter('f')
	.setMathLetter('f')
	.setLaTeXLetter('f')
	.setProductKindToCommutative()
	.addIndiceUp()
	.addIndiceUp()
	.setAntisymmetryIndices([0,1])
	.setSumWeight(41)
	.setProductWeight(41);



// powered versions of delta
PROTO_NCT.delta1 = new MathOperator();
PROTO_NCT.delta1
	.setBehavior(CONST.OPERATOR_NO_RULE)
	.setID('delta1')
	.setLetter('δ1')
	.setMathLetter(CONST.MATH_DELTA)
	.setLaTeXLetter('\\delta')
	.addIndiceDown()
	.setSumWeight(16)
	.setProductWeight(16);

PROTO_NCT.delta2 = new MathOperator();
PROTO_NCT.delta2
	.setBehavior(CONST.OPERATOR_NO_RULE)
	.setID('delta2')
	.setLetter('δ2')
	.setMathLetter(CONST.MATH_DELTA + '2')
	.setLaTeXLetter('\\delta^2')
	.addIndiceDown()
	.addIndiceDown()
	.setSymmetryIndices([0,1])
	.setSumWeight(17)
	.setProductWeight(17);

PROTO_NCT.delta3 = new MathOperator();
PROTO_NCT.delta3
	.setBehavior(CONST.OPERATOR_NO_RULE)
	.setID('delta3')
	.setLetter('δ3')
	.setMathLetter(CONST.MATH_DELTA + '3')
	.setLaTeXLetter('\\delta^3')
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceDown()
	.setSymmetryIndices([0,1,2])
	.setSumWeight(18)
	.setProductWeight(18);

PROTO_NCT.delta4 = new MathOperator();
PROTO_NCT.delta4
	.setBehavior(CONST.OPERATOR_NO_RULE)
	.setID('delta4')
	.setLetter('δ4')
	.setMathLetter(CONST.MATH_DELTA + '4')
	.setLaTeXLetter('\\delta^4')
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceDown()
	.setSymmetryIndices([0,1,2,3])
	.setSumWeight(19)
	.setProductWeight(19);


// repeated applications of nabla on k => powered of delta
PROTO_NCT.delta0k = MathProduct.newFrom(PROTO_NCT.k);

PROTO_NCT.delta1k = PROTO_NCT.delta1.makeCopy()
	.applyTo(PROTO_NCT.k.makeCopy());

PROTO_NCT.delta2k = PROTO_NCT.delta2.makeCopy()
	.applyTo(PROTO_NCT.k.makeCopy());

PROTO_NCT.delta3k = PROTO_NCT.delta3.makeCopy()
	.applyTo(PROTO_NCT.k.makeCopy());

PROTO_NCT.delta4k = PROTO_NCT.delta4.makeCopy()
	.applyTo(PROTO_NCT.k.makeCopy());



// initial substitution of H^{\mu\nu} => g^{\mu\nu} k^2
PROTO_NCT.H_NCT_0 = MathProduct.newFrom(PROTO.gi)
	.multiplyOnRightWith(PROTO_NCT.k)
	.multiplyOnRightWith(PROTO_NCT.k);

// initial substitution of u => k^2
PROTO_NCT.u_NCT_0 = MathProduct.newFrom(PROTO_NCT.k)
	.multiplyOnRightWith(PROTO_NCT.k);

// initial substitution of L^\mu => 2 g^{\mu\nu} k (\delta_\nu k)
// (for P_1 in paper)
PROTO_NCT.L_NCT_0 = PROTO.nabla.makeCopy()
		.applyTo(PROTO_NCT.k.makeCopy());
PROTO_NCT.L_NCT_0 = MathProduct.newFrom(PROTO_NCT.L_NCT_0)
		.multiplyOnLeftWith(PROTO_NCT.k)
		.multiplyOnLeftWith(PROTO.gi)
		.multiplyByInteger(2)
		.setContraction([0,1],[2,-1,0]);


// initial substitution of q => g^{\mu\nu} k (\delta_\mu \delta_\nu k)
// (for P_1 in paper)
PROTO_NCT.q_NCT_0 = PROTO.nabla.makeCopy()
		.applyTo(PROTO_NCT.k.makeCopy());
PROTO_NCT.q_NCT_0 = PROTO.nabla.makeCopy()
		.applyTo(PROTO_NCT.q_NCT_0);
PROTO_NCT.q_NCT_0 = MathProduct.newFrom(PROTO_NCT.q_NCT_0)
	.multiplyOnLeftWith(PROTO_NCT.k)
	.multiplyOnLeftWith(PROTO.gi)
	.setContraction([0,0],[2,-1,0])
	.setContraction([0,1],[2,-2,0]);


// initial substitution of L^\mu => h^{\mu\nu} ( k (\delta_\nu k) + (\delta_\nu k) k )
// (for P_2 in paper)
PROTO_NCT.L_P2_NCT_0 = PROTO.nabla.makeCopy()
		.applyTo(PROTO_NCT.k.makeCopy());
PROTO_NCT.L_P2_NCT_0 = MathProduct.newFrom(PROTO_NCT.L_P2_NCT_0)
		.multiplyOnLeftWith(PROTO_NCT.k)
		.multiplyOnLeftWith(PROTO_NCT.hi)
		.setContraction([0,1],[2,-1,0]);

PROTO_NCT.L_P2_NCT_1 = PROTO.nabla.makeCopy()
		.applyTo(PROTO_NCT.k.makeCopy());
PROTO_NCT.L_P2_NCT_1 = MathProduct.newFrom(PROTO_NCT.L_P2_NCT_1)
		.multiplyOnRightWith(PROTO_NCT.k)
		.multiplyOnLeftWith(PROTO_NCT.hi)
		.setContraction([0,1],[1,-1,0]);



// specific subsitutions for Noncommutative Torus

let SUBST_NCT = {};

// substitute repeated nabla derivations on k by powered derivations delta
const SUBST_NABLAk0 = [
	{
		item: MathProduct.newFrom(PROTO_NCT.k),
		mapIndices: [
		],
		permutationsSum: [],
	}
];

const SUBST_NABLAk1 = [
	{
		item: MathProduct.newFrom(PROTO_NCT.delta1k),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
		],
		permutationsSum: [ [0] ],
	}
];

const SUBST_NABLAk2 = [
	{
		item: MathProduct.newFrom(PROTO_NCT.delta2k),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
		],
		permutationsSum: [ [0,1] ],
	}
];

const SUBST_NABLAk3 = [
	{
		item: MathProduct.newFrom(PROTO_NCT.delta3k),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-3,0], to:[0,-1,2]},
		],
		permutationsSum: [ [0,1,2] ],
	}
];

const SUBST_NABLAk4 = [
	{
		item: MathProduct.newFrom(PROTO_NCT.delta4k),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-3,0], to:[0,-1,2]},
			{from:[-4,0], to:[0,-1,3]},
		],
		permutationsSum: [ [0,1,2,3] ],
	}
];

SUBST_NCT.NABLAk = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 4,
	orgItem: PROTO_NCT.k,
	derivationOperators: [PROTO.nabla],
	newItemsByDerivatives: [SUBST_NABLAk0, SUBST_NABLAk1, SUBST_NABLAk2, SUBST_NABLAk3, SUBST_NABLAk4]
};


// substitute hi to constant element (all derivatives are null)
// 0 is not used
const SUBST_NCT_hiToConstant0 = [
	{
		item: null,
	}
];

const SUBST_NCT_hiToConstant1 = [
	{
		item: null,
	}
];

// hi constant, for partial
SUBST_NCT.hiToConstant = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 1,
	orgItem: PROTO_NCT.hi,
	derivationOperators: [PROTO.partial],
	newItemsByDerivatives: [SUBST_NCT_hiToConstant0, SUBST_NCT_hiToConstant1]
};


// subsitution h^{\mu\nu} = g^{\mu\nu} + f^{\mu\nu}
const SUBST_hi_to_gi_fi = [
	{
		item: MathProduct.newFrom(PROTO.gi),
		mapIndices: [
			{from:[0], to:[0,0]},
			{from:[1], to:[0,1]}
		],
		permutationsSum: [ [0,1] ],
	},
	{
		item: MathProduct.newFrom(PROTO_NCT.fi),
		mapIndices: [
			{from:[0], to:[0,0]},
			{from:[1], to:[0,1]}
		],
		permutationsSum: [ [0,1] ],
	}
]

// hi to gi + fi
SUBST_NCT.hi_to_gi_fi = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.PRE_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 0,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 0,
	orgItem: PROTO_NCT.hi,
	derivationOperators: [PROTO.partial],
	newItemsByDerivatives: [SUBST_hi_to_gi_fi]
};


// substitute H^{\mu\nu} to g^{\mu\nu} u
const SUBST_H_to_gi_u = [
	{
		item: PROTO.H_0,
		mapIndices: [
			{from:[0], to:[0,0]},
			{from:[1], to:[0,1]}
		],
		permutationsSum: [ [0,1] ],
	}
];

SUBST_NCT.H_to_gi_u = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 0,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 0,
	orgItem: PROTO.H,
	derivationOperators: [PROTO.nabla],
	newItemsByDerivatives: [SUBST_H_to_gi_u]
};



// substitute q to 0 (P2 operator)
const SUBST_NCT_q_to_0_0 = [
	{
		item: null,
	}
];

const SUBST_NCT_q_to_0_1 = [
	{
		item: null,
	}
];

const SUBST_NCT_q_to_0_2 = [
	{
		item: null,
	}
];

// q to 0 for nabla
SUBST_NCT.qTo0 = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 0,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 2,
	orgItem: PROTO.q,
	derivationOperators: [PROTO.nabla],
	newItemsByDerivatives: [SUBST_NCT_q_to_0_0, SUBST_NCT_q_to_0_1, SUBST_NCT_q_to_0_2]
};

RULE_DATA.id.fiID = PROTO_NCT.fi.getID();
RULE_DATA.id.delta2ID = PROTO_NCT.delta2.getOperatorID();

RULE_DATA.spectralData.spectralItems.push(
	{itemID: PROTO_NCT.k.getID(), power: new Rational(1,2)}
);


module.exports = {
	PROTO_NCT: PROTO_NCT,
	SUBST_NCT: SUBST_NCT
};
