/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');


const {
				Rational,
				Integer
			} = require(REL_PATH + 'MATH-OBJECTS/MathRational');

const {
				Polynomials,
				RationalExpression,
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');

const {
				MathExpression,
				MathProduct,
				MathTensorProduct,
				MathSum,
				MathArgument,
			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');



// MathElement and MathProduct from which to construct the others
const PROTO = {};

// dimension letter as polynomials
PROTO.dim = RationalExpression.newFromPolynomials(
							Polynomials.newFromLetterCoefficientAndPower('d', 1, 1) );


/*
* DEFINITIONS OF MATHOPERATORS
*/

PROTO.nabla = new MathOperator();
PROTO.nabla
	.setID('nabla')
	.setLetter('∇')
	.setMathLetter(CONST.MATH_NABLA)
	.setLaTeXLetter(CONST.LATEX_OPERATOR_DEFAULT_LETTER)
	.addIndiceDown()
	.setProductKindToNonCommutative()
	.setSumWeight(4)
	.setProductWeight(4);

PROTO.partial = new MathOperator();
PROTO.partial
	.setID('partial')
	.setLetter('𝛛')
	.setMathLetter(CONST.MATH_PARTIAL)
	.setLaTeXLetter('\\partial')
	.addIndiceDown()
	.setProductKindToCommutative()
	.setSumWeight(10)
	.setProductWeight(10);

PROTO.hnabla = new MathOperator();
PROTO.hnabla
	.setBehavior(CONST.OPERATOR_RIEMANN_RULE)
	.setID('hnabla')
	.setLetter('D')
	.setMathLetter(CONST.MATH_HAT_NABLA)
	.setLaTeXLetter('\\hnabla')
	.addIndiceDown()
	.setSumWeight(15)
	.setProductWeight(15);

// powered versions of nabla
PROTO.nabla1 = new MathOperator();
PROTO.nabla1
	.setBehavior(CONST.OPERATOR_NO_RULE)
	.setID('nabla1')
	.setLetter('∇1')
	.setMathLetter(CONST.MATH_NABLA)
	.setLaTeXLetter(CONST.LATEX_OPERATOR_DEFAULT_LETTER)
	.addIndiceDown()
	.setSumWeight(5)
	.setProductWeight(5);

PROTO.nabla2 = new MathOperator();
PROTO.nabla2
	.setBehavior(CONST.OPERATOR_NO_RULE)
	.setID('nabla2')
	.setLetter('∇2')
	.setMathLetter(CONST.MATH_NABLA + '2')
	.setLaTeXLetter(CONST.LATEX_OPERATOR_DEFAULT_LETTER + '^2')
	.addIndiceDown()
	.addIndiceDown()
	.setSumWeight(6)
	.setProductWeight(6);

PROTO.nabla3 = new MathOperator();
PROTO.nabla3
	.setBehavior(CONST.OPERATOR_NO_RULE)
	.setID('nabla3')
	.setLetter('∇3')
	.setMathLetter(CONST.MATH_NABLA + '3')
	.setLaTeXLetter(CONST.LATEX_OPERATOR_DEFAULT_LETTER + '^3')
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceDown()
	.setSumWeight(7)
	.setProductWeight(7);

PROTO.nabla4 = new MathOperator();
PROTO.nabla4
	.setBehavior(CONST.OPERATOR_NO_RULE)
	.setID('nabla4')
	.setLetter('∇4')
	.setMathLetter(CONST.MATH_NABLA + '4')
	.setLaTeXLetter(CONST.LATEX_OPERATOR_DEFAULT_LETTER + '^4')
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceDown()
	.setSumWeight(8)
	.setProductWeight(8);

// powered versions of hnabla
PROTO.hnabla1 = new MathOperator();
PROTO.hnabla1
	.setBehavior(CONST.OPERATOR_NO_RULE)
	.setID('hnabla1')
	.setLetter('D1')
	.setMathLetter(CONST.MATH_HAT_NABLA)
	.setLaTeXLetter('\\hnabla')
	.addIndiceDown()
	.setSumWeight(15)
	.setProductWeight(5);

PROTO.hnabla2 = new MathOperator();
PROTO.hnabla2
	.setBehavior(CONST.OPERATOR_NO_RULE)
	.setID('hnabla2')
	.setLetter('D2')
	.setMathLetter(CONST.MATH_HAT_NABLA + '2')
	.setLaTeXLetter('\\hnabla^2')
	.addIndiceDown()
	.addIndiceDown()
	.setSumWeight(16)
	.setProductWeight(6);

PROTO.hnabla3 = new MathOperator();
PROTO.hnabla3
	.setBehavior(CONST.OPERATOR_NO_RULE)
	.setID('hnabla3')
	.setLetter('D3')
	.setMathLetter(CONST.MATH_HAT_NABLA + '3')
	.setLaTeXLetter('\\hnabla^3')
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceDown()
	.setSumWeight(17)
	.setProductWeight(7);

PROTO.hnabla4 = new MathOperator();
PROTO.hnabla4
	.setBehavior(CONST.OPERATOR_NO_RULE)
	.setID('hnabla4')
	.setLetter('D4')
	.setMathLetter(CONST.MATH_HAT_NABLA + '4')
	.setLaTeXLetter('\\hnabla^4')
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceDown()
	.setSumWeight(18)
	.setProductWeight(8);

// hlaplacian
PROTO.hlaplacian = new MathOperator();
PROTO.hlaplacian
	.setBehavior(CONST.OPERATOR_NO_RULE)
	.setID('hlaplacian')
	.setLetter('𝚫')
	.setMathLetter(CONST.MATH_LAPLACIAN)
	.setLaTeXLetter('\\hDelta')
	.setSumWeight(19)
	.setProductWeight(9);

// symmetric part of PROTO.hnabla2
PROTO.Shnabla2 = new MathOperator();
PROTO.Shnabla2
	.setID('Shnabla2')
	.setLetter('𝚫')
	.setMathLetter(CONST.MATH_LAPLACIAN)
	.setLaTeXLetter('\\hDelta')
	.addIndiceDown()
	.addIndiceDown()
	.setSymmetryIndices([0,1])
	.setSumWeight(6)
	.setProductWeight(6);

// symmetric part of PROTO.hnabla2
PROTO.Shnabla2U = new MathOperator();
PROTO.Shnabla2U
	.setID('Shnabla2U')
	.setLetter('𝚫')
	.setMathLetter(CONST.MATH_LAPLACIAN)
	.setLaTeXLetter('\\hDelta')
	.addIndiceUp()
	.addIndiceUp()
	.setSymmetryIndices([0,1])
	.setSumWeight(5)
	.setProductWeight(5);

// anticommutator of laplacian with hnabla
PROTO.anticomHlaplacianHnabla = new MathOperator();
PROTO.anticomHlaplacianHnabla
	.setID('anticomHLHN')
	.setLetter('{𝚫,D}')
	.setMathLetter('AntiComLD')
	.setLaTeXLetter('\\anticomHLHN')
	.addIndiceDown()
	.setSumWeight(23)
	.setProductWeight(23);



/*
* DEFINITIONS OF MATHELEMENTS
*/

PROTO.H = new MathElement();
PROTO.H
	.setID('H')
	.setLetter('H')
	.setMathLetter('H')
	.setLaTeXLetter('H')
	.setProductKindToNonCommutative()
	.addIndiceUp()
	.addIndiceUp()
	.setSymmetryIndices([0,1])
	.setSumWeight(20)
	.setProductWeight(20);

PROTO.L = new MathElement();
PROTO.L
	.setID('L')
	.setLetter('L')
	.setMathLetter('L')
	.setLaTeXLetter('L')
	.setProductKindToNonCommutative()
	.addIndiceUp()
	.setSumWeight(15)
	.setProductWeight(15);

PROTO.N = new MathElement();
PROTO.N
	.setID('N')
	.setLetter('N')
	.setMathLetter('N')
	.setLaTeXLetter('N')
	.setProductKindToNonCommutative()
	.addIndiceUp()
	.setSumWeight(9)
	.setProductWeight(9);

PROTO.q = new MathElement();
PROTO.q
	.setID('q')
	.setLetter('q')
	.setMathLetter('q')
	.setLaTeXLetter('q')
	.setProductKindToNonCommutative()
	.setSumWeight(10)
	.setProductWeight(10);

PROTO.u = new MathElement();
PROTO.u
	.setID('u')
	.setLetter('u')
	.setMathLetter('u')
	.setLaTeXLetter('u')
	.setProductKindToNonCommutative()
	.setSumWeight(21)
	.setProductWeight(21);

PROTO.p = new MathElement();
PROTO.p
	.setID('p')
	.setLetter('p')
	.setMathLetter('p')
	.setLaTeXLetter('p')
	.setProductKindToNonCommutative()
	.addIndiceUp()
	.setSumWeight(16)
	.setProductWeight(16);

// gauge curvature
PROTO.F = new MathElement();
PROTO.F
	.setID('F')
	.setLetter('F')
	.setMathLetter('F')
	.setLaTeXLetter('F')
	.setProductKindToNonCommutative()
	.addIndiceDown()
	.addIndiceDown()
	.setAntisymmetryIndices([0,1])
	.setSumWeight(8)
	.setProductWeight(8);

// gauge tensor with indices up
PROTO.FU = new MathElement();
PROTO.FU
	.setID('FU')
	.setLetter('F')
	.setMathLetter('F')
	.setLaTeXLetter('F')
	.setProductKindToNonCommutative()
	.addIndiceUp()
	.addIndiceUp()
	.setAntisymmetryIndices([0,1])
	.setSumWeight(7)
	.setProductWeight(7);

// commutator of F with u
PROTO.comFu = new MathElement();
PROTO.comFu
	.setID('comFu')
	.setLetter('[F,u]')
	.setMathLetter('ComFu')
	.setLaTeXLetter('\\comFu')
	.setProductKindToNonCommutative()
	.addIndiceDown()
	.addIndiceDown()
	.setAntisymmetryIndices([0,1])
	.setSumWeight(6)
	.setProductWeight(6);

// commutator of FU with u
PROTO.comFUu = new MathElement();
PROTO.comFUu
	.setID('comFUu')
	.setLetter('[F,u]')
	.setMathLetter('ComFUu')
	.setLaTeXLetter('\\comFUu')
	.setProductKindToNonCommutative()
	.addIndiceUp()
	.addIndiceUp()
	.setAntisymmetryIndices([0,1])
	.setSumWeight(5)
	.setProductWeight(5);

// commutator of F with N
PROTO.comFN = new MathElement();
PROTO.comFN
	.setID('comFN')
	.setLetter('[F,N]')
	.setMathLetter('ComFN')
	.setLaTeXLetter('\\comFN')
	.setProductKindToNonCommutative()
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceUp()
	.setAntisymmetryIndices([0,1])
	.setSumWeight(4)
	.setProductWeight(4);


// metric with lower indices
PROTO.g = new MathElement();
PROTO.g
	.setID('g')
	.setLetter('g')
	.setMathLetter('g')
	.setLaTeXLetter('g')
	.setProductKindToCommutative()
	.addIndiceDown()
	.addIndiceDown()
	.setSymmetryIndices([0,1])
	.setSumWeight(40)
	.setProductWeight(40);

// metric with upper indices
PROTO.gi = new MathElement();
PROTO.gi
	.setID('gi')
	.setLetter('g')
	.setMathLetter('g')
	.setLaTeXLetter('g')
	.setProductKindToCommutative()
	.addIndiceUp()
	.addIndiceUp()
	.setSymmetryIndices([0,1])
	.setSumWeight(39)
	.setProductWeight(39);

// Riemannian tensor
PROTO.RT = new MathElement();
PROTO.RT
	.setID('RT')
	.setLetter('R')
	.setMathLetter('R')
	.setLaTeXLetter('R')
	.setProductKindToCommutative()
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceDown()
	.setSumWeight(30)
	.setProductWeight(30);

// Riemannian tensor
PROTO.RTU = new MathElement();
PROTO.RTU
	.setID('RTU')
	.setLetter('R')
	.setMathLetter('R')
	.setLaTeXLetter('R')
	.setProductKindToCommutative()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setSumWeight(29)
	.setProductWeight(29);

// square of Riemannian tensor
PROTO.RTSQUARE = new MathElement();
PROTO.RTSQUARE
	.setID('RTS')
	.setLetter('|R|^2')
	.setMathLetter('RR')
	.setLaTeXLetter('\\abs{R}^2')
	.setProductKindToCommutative()
	.setSumWeight(28)
	.setProductWeight(28);

// Ricci tensor
PROTO.Ric = new MathElement();
PROTO.Ric
	.setID('Ric')
	.setLetter('Ric')
	.setMathLetter('Ric')
	.setLaTeXLetter('\\Ric')
	.setProductKindToCommutative()
	.addIndiceDown()
	.addIndiceDown()
	.setSymmetryIndices([0,1])
	.setSumWeight(25)
	.setProductWeight(25);

// Ricci tensor with indices up
PROTO.RicU = new MathElement();
PROTO.RicU
	.setID('RicU')
	.setLetter('Ric')
	.setMathLetter('Ric')
	.setLaTeXLetter('\\Ric')
	.setProductKindToCommutative()
	.addIndiceUp()
	.addIndiceUp()
	.setSymmetryIndices([0,1])
	.setSumWeight(24)
	.setProductWeight(24);

// square of Ricci tensor
PROTO.RicSQUARE = new MathElement();
PROTO.RicSQUARE
	.setID('RiccS')
	.setLetter('|Ric|^2')
	.setMathLetter('RicRic')
	.setLaTeXLetter('\\abs{\\Ric}^2')
	.setProductKindToCommutative()
	.setSumWeight(23)
	.setProductWeight(23);

// Scalar curvature
PROTO.SC = new MathElement();
PROTO.SC
	.setID('SC')
	.setLetter('S')
	.setMathLetter('S')
	.setLaTeXLetter('\\SC')
	.setProductKindToCommutative()
	.setSumWeight(20)
	.setProductWeight(20);

// square of Scalar curvature
PROTO.SCSQUARE = new MathElement();
PROTO.SCSQUARE
	.setID('SCS')
	.setLetter('|S|^2')
	.setMathLetter('SS')
	.setLaTeXLetter('\\SC^2')
	.setProductKindToCommutative()
	.setSumWeight(19)
	.setProductWeight(19);

// placeholder for sum of products of gi tensors with change of indices
// see substitution below
PROTO.sumgigi = new MathElement();
PROTO.sumgigi
	.setID('gigi')
	.setLetter('Sumg')
	.setMathLetter('Sumg')
	.setLaTeXLetter('\\Sumgigi')
	.setProductKindToCommutative()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.addIndiceUp()
	.setSymmetryIndices([0,1])
	.setSymmetryIndices([2,3])
	.setSumWeight(38)
	.setProductWeight(38);

// placeholder for sum of products of g tensors with change of indices
// see substitution below
PROTO.sumgg = new MathElement();
PROTO.sumgg
	.setID('gg')
	.setLetter('Sumg')
	.setMathLetter('Sumg')
	.setLaTeXLetter('\\Sumgg')
	.setProductKindToCommutative()
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceDown()
	.addIndiceDown()
	.setSymmetryIndices([0,1])
	.setSymmetryIndices([2,3])
	.setSumWeight(37)
	.setProductWeight(37);

/*
* DEFINITIONS OF COMBINATIONS OF MATHOPERATORS AND MATHELEMENTS
*/

// OPERATORS ON 1
// powered nabla applied on MathElement_one
PROTO.nabla11 = PROTO.nabla1.makeCopy()
	.applyTo(MathElement_one.makeCopy());

PROTO.nabla21 = PROTO.nabla2.makeCopy()
	.applyTo(MathElement_one.makeCopy());

PROTO.nabla31 = PROTO.nabla3.makeCopy()
	.applyTo(MathElement_one.makeCopy());

PROTO.nabla41 = PROTO.nabla4.makeCopy()
	.applyTo(MathElement_one.makeCopy());



// OPERATORS ON H
// nabla repeateadly applied on H
PROTO.H0 = PROTO.H;

PROTO.H1 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.H0.makeCopy());

PROTO.H2 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.H1.makeCopy());

PROTO.H3 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.H2.makeCopy());

PROTO.H4 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.H3.makeCopy());



// OPERATORS ON L
// repeated nabla applied on L
PROTO.L0 = PROTO.L;

PROTO.L1 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.L0.makeCopy());

PROTO.L2 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.L1.makeCopy());

PROTO.L3 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.L2.makeCopy());



// OPERATORS ON N
// powered nabla applied on N
PROTO.nabla1N = PROTO.nabla1.makeCopy()
	.applyTo(PROTO.N.makeCopy());

PROTO.nabla2N = PROTO.nabla2.makeCopy()
	.applyTo(PROTO.N.makeCopy());

PROTO.nabla3N = PROTO.nabla3.makeCopy()
	.applyTo(PROTO.N.makeCopy());

// hnabla applied to N
PROTO.hnablaN = PROTO.hnabla.makeCopy()
	.applyTo(PROTO.N.makeCopy());

// powered hnabla applied on N
PROTO.hnabla1N = PROTO.hnabla1.makeCopy()
	.applyTo(PROTO.N.makeCopy());

PROTO.hnabla2N = PROTO.hnabla2.makeCopy()
	.applyTo(PROTO.N.makeCopy());

PROTO.hnabla3N = PROTO.hnabla3.makeCopy()
	.applyTo(PROTO.N.makeCopy());

// repeated nabla applied on N
PROTO.N0 = PROTO.N;

PROTO.N1 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.N0.makeCopy());

PROTO.N2 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.N1.makeCopy());

PROTO.N3 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.N2.makeCopy());

// one hlaplacian applied to N
PROTO.hlaplacian1N = PROTO.hlaplacian.makeCopy()
	.applyTo(PROTO.N.makeCopy());

// symmetric part of hnabla2 applied to N
PROTO.Shnabla2N = PROTO.Shnabla2.makeCopy()
	.applyTo(PROTO.N.makeCopy());

// hnabla, hlaplacian applied to N
PROTO.hnablahlaplacianN = PROTO.hnabla.makeCopy()
	.applyTo(PROTO.hlaplacian1N.makeCopy());

// hlaplacian, hnabla applied to N
PROTO.hlaplacianhnablaN = PROTO.hlaplacian.makeCopy()
	.applyTo(PROTO.hnablaN.makeCopy());

// anticommutator {hlaplacian, hnabla} applied to N
PROTO.anticomhlaplacianhnablaN = PROTO.anticomHlaplacianHnabla.makeCopy()
	.applyTo(PROTO.N.makeCopy());


// OPERATORS ON q
// powered nabla applied on q
PROTO.nabla1q = PROTO.nabla1.makeCopy()
	.applyTo(PROTO.q.makeCopy());

PROTO.nabla2q = PROTO.nabla2.makeCopy()
	.applyTo(PROTO.q.makeCopy());

// powered hnabla applied on q
PROTO.hnabla1q = PROTO.hnabla1.makeCopy()
	.applyTo(PROTO.q.makeCopy());

PROTO.hnabla2q = PROTO.hnabla2.makeCopy()
	.applyTo(PROTO.q.makeCopy());

// one hlaplacian applied to u
PROTO.hlaplacian1q = PROTO.hlaplacian.makeCopy()
	.applyTo(PROTO.q.makeCopy());

// repeated nabla applied on q
PROTO.q0 = PROTO.q;

PROTO.q1 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.q0.makeCopy());

PROTO.q2 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.q1.makeCopy());

PROTO.q3 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.q2.makeCopy());

PROTO.q4 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.q3.makeCopy());



// OPERATORS ON u
// powered nabla applied on u
PROTO.nabla1u = PROTO.nabla1.makeCopy()
	.applyTo(PROTO.u.makeCopy());

PROTO.nabla2u = PROTO.nabla2.makeCopy()
	.applyTo(PROTO.u.makeCopy());

PROTO.nabla3u = PROTO.nabla3.makeCopy()
	.applyTo(PROTO.u.makeCopy());

PROTO.nabla4u = PROTO.nabla4.makeCopy()
	.applyTo(PROTO.u.makeCopy());

// hnabla applied to u
PROTO.hnablau = PROTO.hnabla.makeCopy()
	.applyTo(PROTO.u.makeCopy());

// powered hnabla applied on u
PROTO.hnabla1u = PROTO.hnabla1.makeCopy()
	.applyTo(PROTO.u.makeCopy());

PROTO.hnabla2u = PROTO.hnabla2.makeCopy()
	.applyTo(PROTO.u.makeCopy());

PROTO.hnabla3u = PROTO.hnabla3.makeCopy()
	.applyTo(PROTO.u.makeCopy());

PROTO.hnabla4u = PROTO.hnabla4.makeCopy()
	.applyTo(PROTO.u.makeCopy());

// one hlaplacian applied to u
PROTO.hlaplacian1u = PROTO.hlaplacian.makeCopy()
	.applyTo(PROTO.u.makeCopy());

// two hlaplacian applied to u
PROTO.hlaplacian2u = PROTO.hlaplacian.makeCopy()
	.applyTo(PROTO.hlaplacian1u.makeCopy());

// symmetric part of hnabla2 applied to u
PROTO.Shnabla2u = PROTO.Shnabla2.makeCopy()
	.applyTo(PROTO.u.makeCopy());

// symmetric part of hnabla2 (up indices) applied to u
PROTO.Shnabla2Uu = PROTO.Shnabla2U.makeCopy()
	.applyTo(PROTO.u.makeCopy());

// hnabla, hlaplacian applied to u
PROTO.hnablahlaplacianu = PROTO.hnabla.makeCopy()
	.applyTo(PROTO.hlaplacian1u.makeCopy());

// hlaplacian, hnabla applied to u
PROTO.hlaplacianhnablau = PROTO.hlaplacian.makeCopy()
	.applyTo(PROTO.hnablau.makeCopy());

// anticommutator {hlaplacian, hnabla} applied to u
PROTO.anticomhlaplacianhnablau = PROTO.anticomHlaplacianHnabla.makeCopy()
	.applyTo(PROTO.u.makeCopy());

// hnabla, hlaplacian, hnabla applied to u
PROTO.hnablahlaplacianhnablau = PROTO.hnabla.makeCopy()
	.applyTo(PROTO.hlaplacianhnablau);

// nabla repeateadly applied on u
PROTO.u0 = PROTO.u;

PROTO.u1 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.u0.makeCopy());

PROTO.u2 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.u1.makeCopy());

PROTO.u3 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.u2.makeCopy());

PROTO.u4 = PROTO.nabla.makeCopy()
	.applyTo(PROTO.u3.makeCopy());

// hnabla repeateadly applied on u
PROTO.uh0 = PROTO.u;

PROTO.uh1 = PROTO.hnabla.makeCopy()
	.applyTo(PROTO.uh0.makeCopy());

PROTO.uh2 = PROTO.hnabla.makeCopy()
	.applyTo(PROTO.uh1.makeCopy());

PROTO.uh3 = PROTO.hnabla.makeCopy()
	.applyTo(PROTO.uh2.makeCopy());

PROTO.uh4 = PROTO.hnabla.makeCopy()
	.applyTo(PROTO.uh3.makeCopy());



// OPERATORS ON p
// powered nabla applied on p
PROTO.nabla1p = PROTO.nabla1.makeCopy()
	.applyTo(PROTO.p.makeCopy());

PROTO.nabla2p = PROTO.nabla2.makeCopy()
	.applyTo(PROTO.p.makeCopy());

PROTO.nabla3p = PROTO.nabla3.makeCopy()
	.applyTo(PROTO.p.makeCopy());



// OPERATORS ON GAUGE CURVATURE
// nabla applied on F
PROTO.NABLA_F = PROTO.nabla.makeCopy()
	.applyTo(PROTO.F.makeCopy());

// powered 1 nabla applied on F
PROTO.nabla1F = PROTO.nabla1.makeCopy()
	.applyTo(PROTO.F.makeCopy());

// hnabla applied on F
PROTO.hnablaF = PROTO.hnabla.makeCopy()
	.applyTo(PROTO.F.makeCopy());


// OPERATORS ON RIEMANN TENSOR
// powered full covariant derivatives applied to RT
PROTO.hnabla1RT = PROTO.hnabla1.makeCopy()
	.applyTo(PROTO.RT.makeCopy());

PROTO.hnabla2RT = PROTO.hnabla2.makeCopy()
	.applyTo(PROTO.RT.makeCopy());

// full covariant derivatives applied to RT
// uses repeated applications of hnabla
PROTO.hnablaRT_1 = PROTO.hnabla.makeCopy()
	.applyTo(PROTO.RT.makeCopy());

PROTO.hnablaRT_2 = PROTO.hnabla.makeCopy()
	.applyTo(PROTO.hnablaRT_1);



// OPERATORS ON SCALAR CURVATURE
// powered full covariant derivatives applied to SC
PROTO.hnablaSC = PROTO.hnabla.makeCopy()
	.applyTo(PROTO.SC.makeCopy());

PROTO.hlaplacianSC = PROTO.hlaplacian.makeCopy()
	.applyTo(PROTO.SC.makeCopy());

// // nabla applied on FS
// PROTO.nabla1FS = PROTO.nabla1.makeCopy()
// 	.applyTo(PROTO.FS.makeCopy());





/*
* DEFINITIONS OF ELEMENTS FOR SUBSTITUTIONS
*/

// normal coordinates for g up to 4th derivation
// use powers of hnabla
// do not use this first rule since nothing happens
PROTO.g_pow0nc_0 = MathProduct.newFrom(PROTO.g);

// maps to 0
PROTO.g_pow1nc_0 = null;

PROTO.g_pow2nc_0 = MathProduct.newFrom(PROTO.RT)
	.multiplyByRational(1,3);

PROTO.g_pow3nc_0 = MathProduct.newFrom(PROTO.hnabla1RT)
	.multiplyByRational(1,6);

PROTO.g_pow4nc_0 = MathProduct.newFrom(PROTO.hnabla2RT)
	.multiplyByRational(1,20);

PROTO.g_pow4nc_1 = MathProduct.newFrom(PROTO.gi)
	.multiplyOnRightWith(PROTO.RT)
	.multiplyOnRightWith(PROTO.RT)
	.setContraction([0,0], [1,3])
	.setContraction([0,1], [2,3])
	.multiplyByRational(2,45);

// normal coordinates for g up to 4th derivation
// uses repeated applications of hnabla
// do not use this rule since nothing happens
PROTO.g0nc_0 = MathProduct.newFrom(PROTO.g);

// maps to 0
PROTO.g1nc_0 = null;

PROTO.g2nc_0 = MathProduct.newFrom(PROTO.RT)
	.multiplyByRational(1,3);

PROTO.g3nc_0 = MathProduct.newFrom(PROTO.hnablaRT_1)
	.multiplyByRational(1,6);

PROTO.g4nc_0 = MathProduct.newFrom(PROTO.hnablaRT_2)
	.multiplyByRational(1,20);

PROTO.g4nc_1 = MathProduct.newFrom(PROTO.gi)
	.multiplyOnRightWith(PROTO.RT)
	.multiplyOnRightWith(PROTO.RT)
	.setContraction([0,0], [1,3])
	.setContraction([0,1], [2,3])
	.multiplyByRational(2,45);

// first derivative of inverse metric in terms of first derivative of metric
PROTO.Pgi1_0 = PROTO.partial.makeCopy().applyTo(PROTO.g.makeCopy());
PROTO.Pgi1_0 = MathProduct.newFrom(PROTO.Pgi1_0);
PROTO.Pgi1_0.multiplyOnLeftWith(PROTO.gi, true)
	.multiplyOnLeftWith(PROTO.gi, true)
	.setContraction([0,1],[2,-2,0])
	.setContraction([1,1],[2,-2,1])
	.multiplyByInteger(-1);


// initial substitution of H^{\mu\nu} => g^{\mu\nu} u
PROTO.H_0 = MathProduct.newFrom(PROTO.gi)
	.multiplyOnRightWith(PROTO.u);

// initial substitution of L^\mu => N^\mu + Gamma^\mu u
// N^mu
PROTO.L_0 = MathProduct.newFrom(PROTO.N);

// 1st element in Gamma^\mu u
PROTO.L_1 = PROTO.partial.makeCopy().applyTo(PROTO.gi.makeCopy());
PROTO.L_1 = MathProduct.newFrom(PROTO.L_1);
PROTO.L_1.multiplyOnLeftWith(PROTO.gi)
	.multiplyOnLeftWith(PROTO.g)
	.multiplyOnRightWith(PROTO.u)
	.setContraction([0,0], [2,-2,0])
	.setContraction([0,1], [2,-2,1])
	.setContraction([1,1], [2,-1,0])
	.multiplyByRational(-1,2);

// 2nd element in Gamma^\mu u
PROTO.L_2 = PROTO.partial.makeCopy().applyTo(PROTO.gi.makeCopy());
PROTO.L_2 = MathProduct.newFrom(PROTO.L_2);
PROTO.L_2	.multiplyOnRightWith(PROTO.u)
	.setContraction([0,-1,0], [0,-2,1]);


// initial substitution of \Gamma^\mu_{\nu\rho}
PROTO.Gamma_00 = MathProduct.newFrom(MathElement_one);
PROTO.Gamma_00.multiplyOnLeftWith(PROTO.partial.makeCopy().applyTo(PROTO.g.makeCopy()));

PROTO.Gamma_0 = PROTO.partial.makeCopy().applyTo(PROTO.g.makeCopy());
PROTO.Gamma_0 = MathProduct.newFrom(PROTO.Gamma_0);
PROTO.Gamma_0.multiplyOnLeftWith(PROTO.gi)
	.multiplyOnRightWith(MathElement_one)
	.setContraction([0,1], [1,-2,1])
	.multiplyByRational(1,2);

PROTO.Gamma_1 = PROTO.partial.makeCopy().applyTo(PROTO.g.makeCopy());
PROTO.Gamma_1 = MathProduct.newFrom(PROTO.Gamma_1);
PROTO.Gamma_1.multiplyOnLeftWith(PROTO.gi)
	.multiplyOnRightWith(MathElement_one)
	.setContraction([0,1], [1,-2,1])
	.multiplyByRational(1,2);

PROTO.Gamma_2 = PROTO.partial.makeCopy().applyTo(PROTO.g.makeCopy());
PROTO.Gamma_2 = MathProduct.newFrom(PROTO.Gamma_2);
PROTO.Gamma_2.multiplyOnLeftWith(PROTO.gi)
	.multiplyOnRightWith(MathElement_one)
	.setContraction([0,1], [1,-1,0])
	.multiplyByRational(-1,2);


// normal coordinates for nabla applied to H
// _i => i-th term in the sum
PROTO.H0nc_0 = MathProduct.newFrom(PROTO.gi)
	.multiplyOnRightWith(PROTO.u);

PROTO.H1nc_0 = MathProduct.newFrom(PROTO.gi)
	.multiplyOnRightWith(PROTO.nabla1u);

PROTO.H2nc_0 = MathProduct.newFrom(PROTO.gi)
	.multiplyOnRightWith(PROTO.nabla2u);

PROTO.H2nc_1 = MathProduct.newFrom(PROTO.gi)
	.multiplyOnRightWith(PROTO.gi)
	.multiplyOnRightWith(PROTO.RT)
	.multiplyOnRightWith(PROTO.u)
	.setContraction([0,0], [2,0])
	.setContraction([1,0], [2,2])
	.multiplyByRational(-1,3);


// normal coordinates for nabla applied on L
// _i => i-th term in the sum
PROTO.L0nc_0 = MathProduct.newFrom(PROTO.N);

PROTO.L1nc_0 = MathProduct.newFrom(PROTO.nabla1N);

PROTO.L1nc_1 = MathProduct.newFrom(PROTO.gi)
	.multiplyOnRightWith(PROTO.Ric)
	.multiplyOnRightWith(PROTO.u)
	.setContraction([0,1], [1,1])
	.multiplyByRational(-2,3);


// elements to replace nabla^n 1 in post element
// _i => i-th term in the sum
PROTO.NABLA2_0 = MathProduct.newFrom(PROTO.F)
	.multiplyByRational(1,2);

PROTO.NABLA2_1 = MathProduct.newFrom(PROTO.nabla21)
	.multiplyByRational(1,2);

PROTO.NABLA2_2 = MathProduct.newFrom(PROTO.nabla21)
	.multiplyByRational(1,2);

// nabla3 1
PROTO.NABLA3_0 = MathProduct.newFrom(PROTO.nabla31);

// nabla3 1 with minus sign
PROTO.NABLA3_1 = MathProduct.newFrom(PROTO.nabla31)
	.multiplyByInteger(-1);

// FF and nabla4 1
PROTO.NABLA4_0 = MathProduct.newFrom(PROTO.F)
	.multiplyOnRightWith(PROTO.F)
	.multiplyByRational(1,4);

PROTO.NABLA4_1 = MathProduct.newFrom(PROTO.nabla41)
	.multiplyByRational(3,4);

PROTO.NABLA4_2 = MathProduct.newFrom(PROTO.nabla41)
	.multiplyByRational(1,4);

PROTO.NABLA4_3 = MathProduct.newFrom(PROTO.nabla41)
	.multiplyByRational(1,4);

PROTO.NABLA4_4 = MathProduct.newFrom(PROTO.nabla41)
	.multiplyByRational(-1,4);

// nabla on F as MathProduct
PROTO.NABLA_F_PROD = MathProduct.newFrom(PROTO.NABLA_F);



// 1/4 of product of two PROTO.gi
PROTO.gigi_0 = MathProduct.newFrom(PROTO.gi);
PROTO.gigi_0.multiplyOnRightWith(PROTO.gi.makeCopy())
	.multiplyByRational(1,4);

// 1/4 of product of two PROTO.g
PROTO.gg_0 = MathProduct.newFrom(PROTO.g);
PROTO.gg_0.multiplyOnRightWith(PROTO.g.makeCopy())
	.multiplyByRational(1,4);


/*
* STARTING OBJECTS FOR THE COMPUTATION
*/

// Construction of P = P1 + P2 + q
let P1 = MathProduct.newFrom(PROTO.H)
					.multiplyOnRightWith(PROTO.nabla)
					.multiplyOnRightWith(PROTO.nabla)
					.setContraction([0,0],[1,0])
					.setContraction([0,1],[2,0]);

let P2 = MathProduct.newFrom(PROTO.L)
					.multiplyOnRightWith(PROTO.nabla)
					.setContraction([0,0],[1,0]);

PROTO.P = new MathSum()
					.addTo(P1)
					.addTo(P2)
					.addTo(PROTO.q)
					.multiplyByInteger(-1);

// Construction of iK = L + iK1
let iK1 = MathProduct.newFrom(PROTO.H)
					.multiplyOnRightWith(PROTO.nabla)
					.setContraction([0,0],[1,0])
					.multiplyByInteger(2);

PROTO.iK = new MathSum()
					.addTo(PROTO.L)
					.addTo(iK1);




/*
* SUBSTITUTIONS GIVEN BY HAND
*/

// substitution rules
const SUBST = {};


// Construction of substitution rules:
// {from: fromArrIdx, to: toArrIdx}
// fromArrIdx is the position of an indice in the item to replace
// toArrIdx is the position on the item which replaces the previous one
// the correct way to think is:
// the letter 'mu' at fromArrIdx will be at position toArrIdx after subsitution
// in other words:
// if one has 'mu1' at fromArrIdx
// then 'mu1' will be at toArrIdx after substitution
// permutationSum: array of permutations of 'to' arrays
// if toArrIdx0 and toArrIdx1 are the 0-th and 1-th 'to' arrIdx in mapIndices,
// with mapIndices rule:
// 'mu1' -> toArrIdx0
// 'mu2' -> toArrIdx1
// then a permutation [1,0] will produce
// 'mu1' -> toArrIdx1
// 'mu2' -> toArrIdx0
// i.e. it will permute the toArrIdx's according to their position in mapIndices


// substitutions for derivatives of metric in normal coordinates
// version with powered applications of hnabla
// start at level 1
const SUBST_g_pow0 = [
	{
		item: PROTO.g_pow0nc_0,
		mapIndices: [
			{from:[0], to:[0,1]}, // alpha
			{from:[1], to:[0,1]}  // beta
		],
		// no permutation
		permutationsSum:[ [0,1] ]
	}
];

const SUBST_g_pow1 = [
	{
		item: PROTO.g_pow1nc_0
	}
];

const SUBST_g_pow2 = [
	{
		item: PROTO.g_pow2nc_0,
		mapIndices: [
			{from:[-1,0], to:[0,0]}, // nu1
			{from:[-2,0], to:[0,2]}, // nu2
			{from:[-3,0], to:[0,1]}, // alpha
			{from:[-3,1], to:[0,3]}  // beta
		],
		// full permutation of nu1, nu2
		permutationsSum:[ [0,1,2,3], [1,0,2,3] ]
	}
];

const SUBST_g_pow3 = [
	{
		item: PROTO.g_pow3nc_0,
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]}, // nu1
			{from:[-2,0], to:[0,-2,0]}, // nu2
			{from:[-3,0], to:[0,-2,2]}, // nu3
			{from:[-4,0], to:[0,-2,1]}, // alpha
			{from:[-4,1], to:[0,-2,3]}  // beta
		],
		// full permutation of nu1, nu2, nu3
		permutationsSum:[ [0,1,2,3,4], [0,2,1,3,4],
											[1,0,2,3,4], [1,2,0,3,4],
										 	[2,0,1,3,4], [2,1,0,3,4]
										],
	}
];

const SUBST_g_pow4 = [
	{
		item: PROTO.g_pow4nc_0,
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]}, // nu1
			{from:[-2,0], to:[0,-1,1]}, // nu2
			{from:[-3,0], to:[0,-2,0]}, // nu3
			{from:[-4,0], to:[0,-2,2]}, // nu4
			{from:[-5,0], to:[0,-2,1]}, // alpha
			{from:[-5,1], to:[0,-2,3]}  // beta
		],
		// full permutation of nu1, nu2, nu3, nu4
		permutationsSum:[ [0,1,2,3,4,5], [0,1,3,2,4,5],
											[0,2,1,3,4,5], [0,2,3,1,4,5],
											[0,3,1,2,4,5], [0,3,2,1,4,5],
											[1,0,2,3,4,5], [1,0,3,2,4,5],
											[1,2,0,3,4,5], [1,2,3,0,4,5],
											[1,3,0,2,4,5], [1,3,2,0,4,5],
										 	[2,0,1,3,4,5], [2,0,3,1,4,5],
											[2,1,0,3,4,5], [2,1,3,0,4,5],
											[2,3,0,1,4,5], [2,3,1,0,4,5],
											[3,0,1,2,4,5], [3,0,2,1,4,5],
											[3,1,0,2,4,5], [3,1,2,0,4,5],
											[3,2,0,1,4,5], [3,2,1,0,4,5]
										 ],
	},
	{
		item: PROTO.g_pow4nc_1,
		mapIndices: [
			{from:[-1,0], to:[1,0]}, // nu1
			{from:[-2,0], to:[1,2]}, // nu2
			{from:[-3,0], to:[2,0]}, // nu3
			{from:[-4,0], to:[2,2]}, // nu4
			{from:[-5,0], to:[1,1]}, // alpha
			{from:[-5,1], to:[2,1]}  // beta
		],
		// full permutation of nu1, nu2, nu3, nu4
		permutationsSum:[ [0,1,2,3,4,5], [0,1,3,2,4,5],
											[0,2,1,3,4,5], [0,2,3,1,4,5],
											[0,3,1,2,4,5], [0,3,2,1,4,5],
											[1,0,2,3,4,5], [1,0,3,2,4,5],
											[1,2,0,3,4,5], [1,2,3,0,4,5],
											[1,3,0,2,4,5], [1,3,2,0,4,5],
										 	[2,0,1,3,4,5], [2,0,3,1,4,5],
											[2,1,0,3,4,5], [2,1,3,0,4,5],
											[2,3,0,1,4,5], [2,3,1,0,4,5],
											[3,0,1,2,4,5], [3,0,2,1,4,5],
											[3,1,0,2,4,5], [3,1,2,0,4,5],
											[3,2,0,1,4,5], [3,2,1,0,4,5]
										 ],
	}
];

// normal coordinates for g up to 4th derivation
// use powers of hnabla
SUBST.g_powNC = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 4,
	orgItem: PROTO.g,
	derivationOperators: [PROTO.partial],
	newItemsByDerivatives: [SUBST_g_pow0,
													SUBST_g_pow1,
													SUBST_g_pow2,
													SUBST_g_pow3,
													SUBST_g_pow4]
};






// substitutions for derivatives of metric in normal coordinates
// version with repeated applications of hnabla
// start at level 1
const SUBST_g0 = [
	{
		item: PROTO.g0nc_0,
		mapIndices: [
			{from:[0], to:[0,1]}, // alpha
			{from:[1], to:[0,1]}  // beta
		],
		// no permutation
		permutationsSum:[ [0,1] ]
	}
];

const SUBST_g1 = [
	{
		item: null
	}
];

const SUBST_g2 = [
	{
		item: PROTO.g2nc_0,
		mapIndices: [
			{from:[-1,0], to:[0,0]}, // nu1
			{from:[-2,0], to:[0,2]}, // nu2
			{from:[-3,0], to:[0,1]}, // alpha
			{from:[-3,1], to:[0,3]}  // beta
		],
		// full permutation of nu1, nu2
		permutationsSum:[ [0,1,2,3], [1,0,2,3] ]
	}
];

const SUBST_g3 = [
	{
		item: PROTO.g3nc_0,
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]}, // nu1
			{from:[-2,0], to:[0,-2,0]}, // nu2
			{from:[-3,0], to:[0,-2,2]}, // nu3
			{from:[-4,0], to:[0,-2,1]}, // alpha
			{from:[-4,1], to:[0,-2,3]}  // beta
		],
		// full permutation of nu1, nu2, nu3
		permutationsSum:[ [0,1,2,3,4], [0,2,1,3,4],
											[1,0,2,3,4], [1,2,0,3,4],
										 	[2,0,1,3,4], [2,1,0,3,4]
										],
	}
];

const SUBST_g4 = [
	{
		item: PROTO.g4nc_0,
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]}, // nu1
			{from:[-2,0], to:[0,-2,0]}, // nu2
			{from:[-3,0], to:[0,-3,0]}, // nu3
			{from:[-4,0], to:[0,-3,2]}, // nu4
			{from:[-5,0], to:[0,-3,1]}, // alpha
			{from:[-5,1], to:[0,-3,3]}  // beta
		],
		// full permutation of nu1, nu2, nu3, nu4
		permutationsSum:[ [0,1,2,3,4,5], [0,1,3,2,4,5],
											[0,2,1,3,4,5], [0,2,3,1,4,5],
											[0,3,1,2,4,5], [0,3,2,1,4,5],
											[1,0,2,3,4,5], [1,0,3,2,4,5],
											[1,2,0,3,4,5], [1,2,3,0,4,5],
											[1,3,0,2,4,5], [1,3,2,0,4,5],
										 	[2,0,1,3,4,5], [2,0,3,1,4,5],
											[2,1,0,3,4,5], [2,1,3,0,4,5],
											[2,3,0,1,4,5], [2,3,1,0,4,5],
											[3,0,1,2,4,5], [3,0,2,1,4,5],
											[3,1,0,2,4,5], [3,1,2,0,4,5],
											[3,2,0,1,4,5], [3,2,1,0,4,5]
										 ],
	},
	{
		item: PROTO.g4nc_1,
		mapIndices: [
			{from:[-1,0], to:[1,0]}, // nu1
			{from:[-2,0], to:[1,2]}, // nu2
			{from:[-3,0], to:[2,0]}, // nu3
			{from:[-4,0], to:[2,2]}, // nu4
			{from:[-5,0], to:[1,1]}, // alpha
			{from:[-5,1], to:[2,1]}  // beta
		],
		// full permutation of nu1, nu2, nu3, nu4
		permutationsSum:[ [0,1,2,3,4,5], [0,1,3,2,4,5],
											[0,2,1,3,4,5], [0,2,3,1,4,5],
											[0,3,1,2,4,5], [0,3,2,1,4,5],
											[1,0,2,3,4,5], [1,0,3,2,4,5],
											[1,2,0,3,4,5], [1,2,3,0,4,5],
											[1,3,0,2,4,5], [1,3,2,0,4,5],
										 	[2,0,1,3,4,5], [2,0,3,1,4,5],
											[2,1,0,3,4,5], [2,1,3,0,4,5],
											[2,3,0,1,4,5], [2,3,1,0,4,5],
											[3,0,1,2,4,5], [3,0,2,1,4,5],
											[3,1,0,2,4,5], [3,1,2,0,4,5],
											[3,2,0,1,4,5], [3,2,1,0,4,5]
										 ],
	}
];


// normal coordinates for g up to 4th derivation
// uses repeated applications of hnabla
SUBST.gNC = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 4,
	orgItem: PROTO.g,
	derivationOperators: [PROTO.partial],
	newItemsByDerivatives: [SUBST_g0,
													SUBST_g1,
													SUBST_g2,
													SUBST_g3,
													SUBST_g4]
};





// substitutions for inverse metric into metric
const SUBST_gi0 = [
	{
		item: null,
	}
];

const SUBST_gi1 = [
	{
		item: PROTO.Pgi1_0,
		mapIndices: [
			{from:[-1,0], to:[2,-1,0]},
			{from:[-2,0], to:[0,0]},
			{from:[-2,1], to:[1,0]}
		],
		permutationsSum: [ [0,1,2] ]
	}
];

// minimal substitution of inverse metric
// first derivative only:
// all other substitutions are "computed" from that one
// see for instance FS.createSubstitutionsDataNormCoord
SUBST.gi = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 1,
	orgItem: PROTO.gi,
	derivationOperators: [PROTO.partial],
	newItemsByDerivatives: [SUBST_gi0, SUBST_gi1]
};

// substitute gi to constant element (all derivatives are null)
// 0 is not used
const SUBST_giToConstant0 = [
	{
		item: null,
	}
];

const SUBST_giToConstant1 = [
	{
		item: null,
	}
];

// gi constant, for partial
SUBST.giToConstant = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 1,
	orgItem: PROTO.gi,
	derivationOperators: [PROTO.partial],
	newItemsByDerivatives: [SUBST_giToConstant0, SUBST_giToConstant1]
};


// substitution N to 0 under hnabla and hlaplacian
// 0 derivative
const SUBST_N0 = [
	{
		item: null,
	}
];

// 1 derivative
const SUBST_N1 = [
	{
		item: null,
	}
];

// 2 derivatives
const SUBST_N2 = [
	{
		item: null,
	}
];

// 3 derivatives
const SUBST_N3 = [
	{
		item: null,
	}
];

// substitution N to 0 under derivations
SUBST.N_TO_ZERO = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 0,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 3,
	orgItem: PROTO.N,
	derivationOperators: [PROTO.nabla, PROTO.hnabla, PROTO.hlaplacian],
	newItemsByDerivatives: [SUBST_N0, SUBST_N1, SUBST_N2, SUBST_N3]
};


// substitute repeated nabla derivations on u by powered derivations
const SUBST_NABLAu0 = [
	{
		item: MathProduct.newFrom(PROTO.u),
		mapIndices: [
		],
		permutationsSum: [],
	}
];

const SUBST_NABLAu1 = [
	{
		item: MathProduct.newFrom(PROTO.nabla1u),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
		],
		permutationsSum: [ [0] ],
	}
];

const SUBST_NABLAu2 = [
	{
		item: MathProduct.newFrom(PROTO.nabla2u),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
		],
		permutationsSum: [ [0,1] ],
	}
];

const SUBST_NABLAu3 = [
	{
		item: MathProduct.newFrom(PROTO.nabla3u),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-3,0], to:[0,-1,2]},
		],
		permutationsSum: [ [0,1,2] ],
	}
];

const SUBST_NABLAu4 = [
	{
		item: MathProduct.newFrom(PROTO.nabla4u),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-3,0], to:[0,-1,2]},
			{from:[-4,0], to:[0,-1,3]},
		],
		permutationsSum: [ [0,1,2,3] ],
	}
];

SUBST.NABLAu = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 2,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 4,
	orgItem: PROTO.u,
	derivationOperators: [PROTO.nabla],
	newItemsByDerivatives: [SUBST_NABLAu0, SUBST_NABLAu1, SUBST_NABLAu2, SUBST_NABLAu3, SUBST_NABLAu4]
};




// substitute repeated hnabla derivations on u by powered derivations
const SUBST_HNABLAu0 = [
	{
		item: MathProduct.newFrom(PROTO.u),
		mapIndices: [
		],
		permutationsSum: [],
	}
];

const SUBST_HNABLAu1 = [
	{
		item: MathProduct.newFrom(PROTO.hnabla1u),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
		],
		permutationsSum: [ [0] ],
	}
];

const SUBST_HNABLAu2 = [
	{
		item: MathProduct.newFrom(PROTO.hnabla2u),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
		],
		permutationsSum: [ [0,1] ],
	}
];

const SUBST_HNABLAu3 = [
	{
		item: MathProduct.newFrom(PROTO.hnabla3u),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-3,0], to:[0,-1,2]},
		],
		permutationsSum: [ [0,1,2] ],
	}
];

const SUBST_HNABLAu4 = [
	{
		item: MathProduct.newFrom(PROTO.hnabla4u),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-3,0], to:[0,-1,2]},
			{from:[-4,0], to:[0,-1,3]},
		],
		permutationsSum: [ [0,1,2,3] ],
	}
];

SUBST.HNABLAu = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 4,
	orgItem: PROTO.u,
	derivationOperators: [PROTO.hnabla],
	newItemsByDerivatives: [SUBST_HNABLAu0, SUBST_HNABLAu1, SUBST_HNABLAu2, SUBST_HNABLAu3, SUBST_HNABLAu4]
};



// substitute repeated hnabla derivations on N by powered derivations
const SUBST_HNABLAN0 = [
	{
		item: MathProduct.newFrom(PROTO.N),
		mapIndices: [
			{from:[0], to:[0,0]}
		],
		permutationsSum:[ [0] ],
	}
];

const SUBST_HNABLAN1 = [
	{
		item: MathProduct.newFrom(PROTO.hnabla1N),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-2,0]},
		],
		permutationsSum: [ [0,1] ],
	}
];

const SUBST_HNABLAN2 = [
	{
		item: MathProduct.newFrom(PROTO.hnabla2N),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-3,0], to:[0,-2,0]},
		],
		permutationsSum: [ [0,1,2] ],
	}
];

const SUBST_HNABLAN3 = [
	{
		item: MathProduct.newFrom(PROTO.hnabla3N),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-3,0], to:[0,-1,2]},
			{from:[-4,0], to:[0,-2,0]},
		],
		permutationsSum: [ [0,1,2,3] ],
	}
];


SUBST.HNABLAN = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 3,
	orgItem: PROTO.N,
	derivationOperators: [PROTO.hnabla],
	newItemsByDerivatives: [SUBST_HNABLAN0, SUBST_HNABLAN1, SUBST_HNABLAN2, SUBST_HNABLAN3]
};



// substitute repeated hnabla derivations on q by powered derivations
const SUBST_HNABLAq0 = [
	{
		item: MathProduct.newFrom(PROTO.q),
		mapIndices: [
		],
		permutationsSum: [],
	}
];

const SUBST_HNABLAq1 = [
	{
		item: MathProduct.newFrom(PROTO.hnabla1q),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
		],
		permutationsSum: [ [0] ],
	}
];

const SUBST_HNABLAq2 = [
	{
		item: MathProduct.newFrom(PROTO.hnabla2q),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
		],
		permutationsSum: [ [0,1] ],
	}
];

SUBST.HNABLAq = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 2,
	orgItem: PROTO.q,
	derivationOperators: [PROTO.hnabla],
	newItemsByDerivatives: [SUBST_HNABLAq0, SUBST_HNABLAq1, SUBST_HNABLAq2]
};





// substitute u to constant element (all derivatives are null)
const SUBST_uToConstant0 = [
	{
		item: MathProduct.newFrom(MathElement_one),
		mapIndices: [
		],
		permutationsSum: [ ],
	}
];

const SUBST_uToConstant1 = [
	{
		item: null,
	}
];

const SUBST_uToConstant2 = [
	{
		item: null,
	}
];

const SUBST_uToConstant3 = [
	{
		item: null,
	}
];

const SUBST_uToConstant4 = [
	{
		item: null,
	}
];

// u constant, for hnabla
SUBST.uToConstant = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 4,
	orgItem: PROTO.u,
	derivationOperators: [PROTO.hnabla, PROTO.hlaplacian],
	newItemsByDerivatives: [SUBST_uToConstant0,
													SUBST_uToConstant1,
													SUBST_uToConstant2,
													SUBST_uToConstant3,
													SUBST_uToConstant4
												]
};

// // u constant, for hlaplacian
// SUBST.uToConstant_hlaplacian = {
// 	// indicate in which part of MathArgument to look at
// 	termElement: CONST.MAIN_ELEMENT,
// 	// min level of derivations to consider => to reduce the computation
// 	minLevel: 1,
// 	// max level of derivations to consider => to reduce the computation
// 	maxLevel: 4,
// 	orgItem: PROTO.u,
// 	derivationOperators: [PROTO.hlaplacian],
// 	newItemsByDerivatives: [SUBST_uToConstant0,
// 													SUBST_uToConstant1,
// 													SUBST_uToConstant2,
// 													SUBST_uToConstant3,
// 													SUBST_uToConstant4
// 												]
// };


// substitute nabla's in post element
// 2 derivatives
// nabla_12 => 1/2 F + 1/2 nabla_12 + 1/2 nabla_21
const SUBST_NABLA2 = [
	{
		item: PROTO.NABLA2_0, // F
		mapIndices: [
			{from:[-1,0], to:[0,0]},
			{from:[-2,0], to:[0,1]}
		],
		permutationsSum:[ [0,1] ],
	},
	{
		item: PROTO.NABLA2_1, // nabla_12
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]}
		],
		permutationsSum:[ [0,1] ],
	},
	{
		item: PROTO.NABLA2_1, // nabla_21
		mapIndices: [
			{from:[-1,0], to:[0,-1,1]},
			{from:[-2,0], to:[0,-1,0]}
		],
		permutationsSum:[ [0,1] ],
	}
];

// 3 derivatives
const SUBST_NABLA3 = [
	{
		item: PROTO.NABLA3_0, // nabla3 1
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-3,0], to:[0,-1,2]}
		],
		permutationsSum:[ [0,1,2] ],
	}
];

// 4 derivatives
// nabla_1234 => 1/4 FF + 3/4 nabla_1234 + 1/4 nabla_1243 + 1/4 nabla_2134 - 1/4 nabla_2143
// FF and nabla4 1
const SUBST_NABLA4 = [
	{
		item: PROTO.NABLA4_0, // F F
		mapIndices: [
			{from:[-1,0], to:[0,0]},
			{from:[-2,0], to:[0,1]},
			{from:[-3,0], to:[1,0]},
			{from:[-4,0], to:[1,1]}
		],
		permutationsSum:[ [0,1,2,3] ],
	},
	{
		item: PROTO.NABLA4_1, // nabla_1234
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-3,0], to:[0,-1,2]},
			{from:[-4,0], to:[0,-1,3]}
		],
		permutationsSum:[ [0,1,2,3] ],
	},
	{
		item: PROTO.NABLA4_2, // nabla_1243
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-3,0], to:[0,-1,3]},
			{from:[-4,0], to:[0,-1,2]}
		],
		permutationsSum:[ [0,1,2,3] ],
	},
	{
		item: PROTO.NABLA4_3, // nabla_2134
		mapIndices: [
			{from:[-1,0], to:[0,-1,1]},
			{from:[-2,0], to:[0,-1,0]},
			{from:[-3,0], to:[0,-1,2]},
			{from:[-4,0], to:[0,-1,3]}
		],
		permutationsSum:[ [0,1,2,3] ],
	},
	{
		item: PROTO.NABLA4_4, // nabla_2143
		mapIndices: [
			{from:[-1,0], to:[0,-1,1]},
			{from:[-2,0], to:[0,-1,0]},
			{from:[-3,0], to:[0,-1,3]},
			{from:[-4,0], to:[0,-1,2]}
		],
		permutationsSum:[ [0,1,2,3] ],
	}
];


SUBST.NABLA_POST_ELEMENT = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.POST_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 2,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 4,
	orgItem: MathElement_one,
	derivationOperators: [PROTO.nabla],
	newItemsByDerivatives: [ [], [], SUBST_NABLA2, SUBST_NABLA3, SUBST_NABLA4]
};

// nabla^3 on 1
// this substituton does not give a satisfactory result
const SUBST_NABLA3_ONE = [
	{
		item: PROTO.NABLA_F_PROD, // nabla_1 F_23
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-2,0]},
			{from:[-3,0], to:[0,-2,1]}
		],
		permutationsSum:[ [0,1,2] ],
	},
	{
		item: PROTO.NABLA3_0, // nabla_132 1
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,2]},
			{from:[-3,0], to:[0,-1,1]}
		],
		permutationsSum:[ [0,1,2] ],
	},
	{
		item: PROTO.NABLA3_1, // - nabla_321 1
		mapIndices: [
			{from:[-1,0], to:[0,-1,2]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-3,0], to:[0,-1,0]}
		],
		permutationsSum:[ [0,1,2] ],
	},
	{
		item: PROTO.NABLA3_0, // nabla_231 1
		mapIndices: [
			{from:[-1,0], to:[0,-1,2]},
			{from:[-2,0], to:[0,-1,0]},
			{from:[-3,0], to:[0,-1,1]}
		],
		permutationsSum:[ [0,1,2] ],
	}
];

SUBST.NABLA3_ONE_TO_NABLA_F_POST_ELEMENT = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.POST_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 3,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 3,
	orgItem: MathElement_one,
	derivationOperators: [PROTO.nabla],
	newItemsByDerivatives: [ [], [], [], SUBST_NABLA3_ONE]
};



// substitution used to test simplified expressions
// hnabla F => sum of nabla3 on one
const SUBST_HNABLA_F = [
	{
		item: PROTO.NABLA3_0, // + nabla3_123 1
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-2,1], to:[0,-1,2]}
		],
		permutationsSum:[ [0,1,2] ],
	},
	{
		item: PROTO.NABLA3_1, // - nabla3_132 1
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-1,2]},
			{from:[-2,1], to:[0,-1,1]}
		],
		permutationsSum:[ [0,1,2] ],
	},
	{
		item: PROTO.NABLA3_0, // + nabla3_321 1
		mapIndices: [
			{from:[-1,0], to:[0,-1,2]},
			{from:[-2,0], to:[0,-1,1]},
			{from:[-2,1], to:[0,-1,0]}
		],
		permutationsSum:[ [0,1,2] ],
	},
	{
		item: PROTO.NABLA3_1, // - nabla3_231 1
		mapIndices: [
			{from:[-1,0], to:[0,-1,2]},
			{from:[-2,0], to:[0,-1,0]},
			{from:[-2,1], to:[0,-1,1]}
		],
		permutationsSum:[ [0,1,2] ],
	}
];

SUBST.HNABLA_F_TO_NABLA3_ONE_POST_ELEMENT = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.POST_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 1,
	orgItem: PROTO.F,
	derivationOperators: [PROTO.hnabla],
	newItemsByDerivatives: [ [], SUBST_HNABLA_F]
};


// substitute hnabla^2 on u
// by its symmetric and antisymmetric parts
// PROTO.hnabla2 = Shnabla2 + 1/2 PROTO.F
// PROTO.hnabla2u => PROTO.Shnabla2u + 1/2 (PROTO.F PROTO.u - PROTO.u PROTO.F)
// PROTO.hnabla2u => PROTO.Shnabla2u + 1/2 PROTO.comFu

PROTO.hnabla2u_1 = MathProduct.newFrom(PROTO.Shnabla2u);

PROTO.hnabla2u_2 = MathProduct.newFrom(PROTO.comFu);
PROTO.hnabla2u_2.multiplyByRational(1,2);

PROTO.hnabla2u_0 = null;
const SUBST_hnabla2u0 = [
	{
		item: PROTO.hnabla2u_0
	}
]

const SUBST_hnabla2u1 = [
	{
		item: PROTO.hnabla2u_1,
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-1,1], to:[0,-1,1]}
		],
		permutationsSum:[ [0,1] ],
	},
	{
		item: PROTO.hnabla2u_2,
		mapIndices: [
			{from:[-1,0], to:[0,0]},
			{from:[-1,1], to:[0,1]}
		],
		permutationsSum:[ [0,1] ],
	}
];

SUBST.hnabla2u = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 1,
	orgItem: PROTO.u,
	derivationOperators: [PROTO.hnabla2],
	newItemsByDerivatives: [SUBST_hnabla2u0, SUBST_hnabla2u1]
};


// substitution of anticommutator {laplacian, nabla} on u to corresponding sum
const SUBST_anticomHLHNu1 = [
	{
		item: MathProduct.newFrom(PROTO.hlaplacianhnablau),
		mapIndices: [
			{from:[-1,0], to:[0,-2,0]}
		],
		permutationsSum:[ [0] ],
	},
	{
		item: MathProduct.newFrom(PROTO.hnablahlaplacianu),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]}
		],
		permutationsSum:[ [0] ],
	}
];

SUBST.anticomHLHNu = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 1,
	orgItem: PROTO.u,
	derivationOperators: [PROTO.anticomHlaplacianHnabla],
	newItemsByDerivatives: [[], SUBST_anticomHLHNu1]
};

// substitute hnabla^2 on N
// by its symmetric and antisymmetric parts
// PROTO.hnabla2 = Shnabla2 + 1/2 PROTO.F
// PROTO.hnabla2N => PROTO.Shnabla2N + 1/2 (PROTO.F PROTO.N - PROTO.N PROTO.F)
// PROTO.hnabla2N => PROTO.Shnabla2N + 1/2 PROTO.comFN

PROTO.hnabla2N_1 = MathProduct.newFrom(PROTO.Shnabla2N);

PROTO.hnabla2N_2 = MathProduct.newFrom(PROTO.comFN);
PROTO.hnabla2N_2.multiplyByRational(1,2);

PROTO.hnabla2N_0 = null;
const SUBST_hnabla2N0 = [
	{
		item: PROTO.hnabla2N_0
	}
]

const SUBST_hnabla2N1 = [
	{
		item: PROTO.hnabla2N_1,
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-1,1], to:[0,-1,1]},
			{from:[-2,0], to:[0,-2,0]}
		],
		permutationsSum:[ [0,1,2] ],
	},
	{
		item: PROTO.hnabla2N_2,
		mapIndices: [
			{from:[-1,0], to:[0,0]},
			{from:[-1,1], to:[0,1]},
			{from:[-2,0], to:[0,2]}
		],
		permutationsSum:[ [0,1,2] ],
	}
];

SUBST.hnabla2N = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 1,
	orgItem: PROTO.N,
	derivationOperators: [PROTO.hnabla2],
	newItemsByDerivatives: [SUBST_hnabla2N0, SUBST_hnabla2N1]
};


// substitution of anticommutator {laplacian, nabla} on N to corresponding sum
const SUBST_anticomHLHNN1 = [
	{
		item: MathProduct.newFrom(PROTO.hlaplacianhnablaN),
		mapIndices: [
			{from:[-1,0], to:[0,-2,0]},
			{from:[-2,0], to:[0,-3,0]}
		],
		permutationsSum:[ [0,1] ],
	},
	{
		item: MathProduct.newFrom(PROTO.hnablahlaplacianN),
		mapIndices: [
			{from:[-1,0], to:[0,-1,0]},
			{from:[-2,0], to:[0,-3,0]}
		],
		permutationsSum:[ [0,1] ],
	}
];

SUBST.anticomHLHNN = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.MAIN_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 1,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 1,
	orgItem: PROTO.N,
	derivationOperators: [PROTO.anticomHlaplacianHnabla],
	newItemsByDerivatives: [[], SUBST_anticomHLHNN1]
};



// substitution of sumgigi by its sum of 3 terms
const SUBST_gigi0 = [
	{
		item: PROTO.gigi_0, // gi^12 gi^34
		mapIndices: [
			{from:[0], to:[0,0]},
			{from:[1], to:[0,1]},
			{from:[2], to:[1,0]},
			{from:[3], to:[1,1]}
		],
		permutationsSum:[ [0,1,2,3] ],
	},
	{
		item: PROTO.gigi_0, // gi^13 gi^24
		mapIndices: [
			{from:[0], to:[0,0]},
			{from:[1], to:[1,0]},
			{from:[2], to:[0,1]},
			{from:[3], to:[1,1]}
		],
		permutationsSum:[ [0,1,2,3] ],
	},
	{
		item: PROTO.gigi_0, // gi^14 gi^23
		mapIndices: [
			{from:[0], to:[0,0]},
			{from:[1], to:[1,0]},
			{from:[2], to:[1,1]},
			{from:[3], to:[0,1]}
		],
		permutationsSum:[ [0,1,2,3] ],
	}
];


SUBST.sumgigi = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.PRE_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 0,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 0,
	orgItem: PROTO.sumgigi,
	derivationOperators: [PROTO.nabla],
	newItemsByDerivatives: [SUBST_gigi0]
};


// substitution of sumgg by its sum of 3 terms
const SUBST_gg0 = [
	{
		item: PROTO.gg_0, // g_12 g_34
		mapIndices: [
			{from:[0], to:[0,0]},
			{from:[1], to:[0,1]},
			{from:[2], to:[1,0]},
			{from:[3], to:[1,1]}
		],
		permutationsSum:[ [0,1,2,3] ],
	},
	{
		item: PROTO.gg_0, // g_13 g_24
		mapIndices: [
			{from:[0], to:[0,0]},
			{from:[1], to:[1,0]},
			{from:[2], to:[0,1]},
			{from:[3], to:[1,1]}
		],
		permutationsSum:[ [0,1,2,3] ],
	},
	{
		item: PROTO.gg_0, // g_14 g_23
		mapIndices: [
			{from:[0], to:[0,0]},
			{from:[1], to:[1,0]},
			{from:[2], to:[1,1]},
			{from:[3], to:[0,1]}
		],
		permutationsSum:[ [0,1,2,3] ],
	}
];


SUBST.sumgg = {
	// indicate in which part of MathArgument to look at
	termElement: CONST.PRE_ELEMENT,
	// min level of derivations to consider => to reduce the computation
	minLevel: 0,
	// max level of derivations to consider => to reduce the computation
	maxLevel: 0,
	orgItem: PROTO.sumgg,
	derivationOperators: [PROTO.nabla],
	newItemsByDerivatives: [SUBST_gg0]
};








/*
* RULES DATA
*/

// rules data used by simplification functions
const RULE_DATA = {};

RULE_DATA.elements = {};
RULE_DATA.elements.u = PROTO.u;
RULE_DATA.elements.g = PROTO.g;
RULE_DATA.elements.gi = PROTO.gi;
RULE_DATA.elements.dim = PROTO.dim;
RULE_DATA.elements.Ric = PROTO.Ric;
RULE_DATA.elements.SC = PROTO.SC;
RULE_DATA.elements.hnabla = PROTO.hnabla;
RULE_DATA.elements.hlaplacian = PROTO.hlaplacian;

RULE_DATA.id = {};
RULE_DATA.id.uID = PROTO.u.getID();
RULE_DATA.id.gID = PROTO.g.getID();
RULE_DATA.id.giID = PROTO.gi.getID();
RULE_DATA.id.RTID = PROTO.RT.getID();
RULE_DATA.id.RicID = PROTO.Ric.getID();
RULE_DATA.id.RicUID = PROTO.RicU.getID();
RULE_DATA.id.FID = PROTO.F.getID();
RULE_DATA.id.FUID = PROTO.FU.getID();
RULE_DATA.id.hnablaID = PROTO.hnabla.getID();
RULE_DATA.id.Shnabla2ID = PROTO.Shnabla2.getID();
RULE_DATA.id.comFuID = PROTO.comFu.getID();

RULE_DATA.tensorData = {};
// raise indices of tensors
RULE_DATA.tensorData.F = {
			tensorDown: PROTO.F,
			tensorUp: PROTO.FU,
			tensorDerivationLevel: 0, // tensor is not under derivation
		};
RULE_DATA.tensorData.RT = {
			tensorDown: PROTO.RT,
			tensorUp: PROTO.RTU,
			tensorDerivationLevel: 0,  // tensor is not under derivation
		};
RULE_DATA.tensorData.Ric = {
			tensorDown: PROTO.Ric,
			tensorUp: PROTO.RicU,
			tensorDerivationLevel: 0,  // tensor is not under derivation
		};
RULE_DATA.tensorData.comFu = {
			tensorDown: PROTO.comFu,
			tensorUp: PROTO.comFUu,
			tensorDerivationLevel: 0, // tensor is not under derivation
		};

RULE_DATA.tensorData.Shnabla2 = {
			tensorDown: PROTO.Shnabla2,
			tensorUp: PROTO.Shnabla2U,
			tensorDerivationLevel: -1, // tensor is a derivation, up to 1
		};

// full self contractions of Riemann tensor
let RTSQUARE1 = MathProduct.newFrom(PROTO.RTSQUARE);
let RTSQUARE2 = MathProduct.newFrom(PROTO.RTSQUARE);
RTSQUARE2.multiplyByInteger(-1);
let RTSQUARE3 = MathProduct.newFrom(PROTO.RTSQUARE);
RTSQUARE3.multiplyByRational(1,2);
let RTSQUARE4 = MathProduct.newFrom(PROTO.RTSQUARE);
RTSQUARE4.multiplyByRational(-1,2);
// indiceConfigurations: permutation of indices of second occurence
// contractedItem: the item produced by this configuration
let contractionsListRTS = [
	{
		indiceConfigurations: ['0123', '1032', '2301', '3210'],
		contractedItem: RTSQUARE1
	},
	{
		indiceConfigurations: ['1023', '0132', '3201', '2310'],
		contractedItem: RTSQUARE2
	},
	{
		indiceConfigurations: ['0213', '2031', '1302', '3120', '0321', '1230', '2103', '3012'],
		contractedItem: RTSQUARE3
	},
	{
		indiceConfigurations: ['2013', '0231', '3102', '1320', '0312', '1203', '2130', '3021'],
		contractedItem: RTSQUARE4
	}
];
RULE_DATA.tensorData.RTS = {
	tensorFirst: PROTO.RT,
	tensorSecond: PROTO.RTU,
	contractionsList: contractionsListRTS
};

// full self contractions of Ricci tensor
let RICCISQUARE = MathProduct.newFrom(PROTO.RicSQUARE);
let contractionListRicS = [
	{
		indiceConfigurations: ['01', '10'],
		contractedItem: RICCISQUARE
	}
];
RULE_DATA.tensorData.RicS = {
	tensorFirst: PROTO.Ric,
	tensorSecond: PROTO.RicU,
	contractionsList: contractionListRicS
};


RULE_DATA.powerData = {};
RULE_DATA.powerData.SCS = {
	item: PROTO.SC,
	power: 2,
	newItem: PROTO.SCSQUARE
};

RULE_DATA.spectralData = {
	// elements that commute with 'u' and their power as fraction of spectral value 'r'
	// TODO: consider elements with spectral function in 'u' more complicated (not just powers of 'r')
	spectralItems: [
		{itemID: PROTO.u.getID(), power: new Rational(1,1)}
	]
};

/*
* SELECTIONS
*/

// some predefined selections of terms in MathSum
const SELECTION = {}

SELECTION.PENabla0 = {
	main: [
	],
	post: [
		{
			element: MathElement_one, min: 1, max: 1,
			operators: [
				{op: PROTO.nabla, min: 0, max: 0}
			]
		}
	]
};

SELECTION.PENabla1 = {
	main: [
	],
	post: [
		{
			element: MathElement_one, min: 1, max: 1,
			operators: [
				{op: PROTO.nabla, min: 1, max: 1}
			]
		}
	]
};

SELECTION.PENabla2 = {
	main: [
	],
	post: [
		{
			element: MathElement_one, min: 1, max: 1,
			operators: [
				{op: PROTO.nabla, min: 2, max: 2}
			]
		}
	]
};

SELECTION.PENabla3 = {
	main: [
	],
	post: [
		{
			element: MathElement_one, min: 1, max: 1,
			operators: [
				{op: PROTO.nabla, min: 3, max: 3}
			]
		}
	]
};

SELECTION.PENabla4 = {
	main: [
	],
	post: [
		{
			element: MathElement_one, min: 1, max: 1,
			operators: [
				{op: PROTO.nabla, min: 4, max: 4}
			]
		}
	]
};

SELECTION.PENabla2_on_One = {
	main: [
	],
	post: [
		{
			element: MathElement_one, min: 1, max: 1,
			operators: [
				{op: PROTO.nabla2, min: 1, max: 1}
			]
		}
	]
};



module.exports = {
	PROTO: PROTO,
	SUBST: SUBST,
	RULE_DATA: RULE_DATA,
	SELECTION: SELECTION,
};
