/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const chalk = require('chalk');

// global data shared by all the code
const GLOBAL = {
	listElementID: [],
	defaultID: '1234', // something compatible with regexp
	currentIndiceID: [0,0,0,0,-1], // next will be [0,0,0,0]
	productWeightData: {},
	productWeightOperatorData: {},
	sumWeightData: {},
	sumWeightOperatorData: {},
};

// some 'constants' used as default values
const CONST = {
	// commutative or non commtutative behavior
	ELEMENT_COMMUTATIVE: 0,
	ELEMENT_NON_COMMUTATIVE: 1,
	// operator obeys Leibniz rule
	OPERATOR_LEIBNIZ_RULE: 0,
	// operator obeys Riemann derivation rule
	OPERATOR_RIEMANN_RULE: 1,
	// operator with no rule: of order greater that 2 for instance
	OPERATOR_NO_RULE: -1,
	// operator status
	OPERATOR_STATUS_LATENT: 0,
	OPERATOR_STATUS_APPLIED: 1,
	// indices positions
	INDICE_UP: 1,
	INDICE_DOWN: -1,
	// indices behavior
	INDICE_INNER: 0,
	INDICE_OUTER: 1,
	// flags for MathArgument parts
	PRE_ELEMENT: -2,
	MAIN_ELEMENT: 0,
	POST_ELEMENT: -1,
	// some letters by default
	DIMENSION_LETTER: 'd',
	ELEMENT_DEFAULT_LETTER: '𝟙',
	OPERATOR_DEFAULT_LETTER: '∇',
	SYMBOL_ONE: '1',
	SYMBOL_BOLD_ONE: '𝟙',
	SYMBOL_ZERO: '0',
	SYMBOL_EXPRESSION: ', ',
	SYMBOL_PRODUCT: '×',
	SYMBOL_POLYNOMIALS_PRODUCT: '·',
	SYMBOL_POLYNOMIALS_SUM: ' + ',
	SYMBOL_POLYNOMIALS_DIFF: ' - ',
	SYMBOL_RATIONAL_OVER: '/',
	SYMBOL_POLYNOMIAL_OPEN: '(',
	SYMBOL_POLYNOMIAL_CLOSE: ')',
	SYMBOL_TENSOR_PRODUCT: '•',
	SYMBOL_EXPONENT: '^',
	SYMBOL_INDICE_UP: '^',
	SYMBOL_INDICE_DOWN: '_',
	SYMBOL_OPERATOR_APPLY: '·',
	SYMBOL_SUM: ' + ',
	SYMBOL_MAP_INDICES: '→',
	SYMBOL_POST_ELEMENT: ' ◁ ',
	SYMBOL_PRE_ELEMENT: ' * ',
	// open and close indices list
	SYMBOL_OPEN_INDICES: '(',
	SYMBOL_CLOSE_INDICES: ')',
	// open and close class className
	SYMBOL_OPEN_CLASS: '«',
	SYMBOL_CLOSE_CLASS: '» ',
	// closures of lists and items
	SYMBOL_DEFAULT_EMPTY_LIST: '{}',
	SYMBOL_DEFAULT_OPEN_LIST: '{ ',
	SYMBOL_DEFAULT_END_LIST: ' }',
	SYMBOL_DEFAULT_OPEN_ITEM: '[',
	SYMBOL_DEFAULT_END_ITEM: ']',
	SYMBOL_TENSOR_PRODUCT_EMPTY_LIST: '()',
	SYMBOL_TENSOR_PRODUCT_OPEN_ITEM: '(',
	SYMBOL_TENSOR_PRODUCT_END_ITEM: ')',
	SYMBOL_OPEN_ID: '❮',
	SYMBOL_CLOSE_ID: '❯',
	SYMBOL_OUTER_INDEX: 'mu',
	SYMBOL_INNER_INDEX: 'nu',
	ARRAY_LETTERS_SHORT: 'abcdefghijklmnopqrstuvwxyz'.split(''),
	ARRAY_LETTERS_LONG: 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'.split(''),
	TIMER: {},
	LATEX_DEFAULT_LETTER: '\\bbbone',
	LATEX_OPERATOR_DEFAULT_LETTER: '\\nabla',
	LATEX_EXPONENT: '^',
	LATEX_SUB: '_',
	LATEX_POLYNOMIALS_SUM: ' + ',
	LATEX_POLYNOMIAL_OPEN: '(',
	LATEX_POLYNOMIAL_CLOSE: ')',
	LATEX_CMD_FRAC: '\\tfrac',
	LATEX_OUTER_INDEX: '\\mu',
	LATEX_INNER_INDEX: '\\nu',
	LATEX_OPERATOR_APPLY: ' ',
	LATEX_PRODUCT: ' ',
	LATEX_TENSOR_PRODUCT: ' \\otimes ',
	LATEX_SUM: ' + ',
	LATEX_RIGHTARROW: ' \\rightarrow ',
	LATEX_THIN_SPACE: ' \\, ',
	LATEX_SPECTRAL_OPERATOR_LETTER: '\\SO',
	LATEX_SPECTRAL_OPERATOR_OPEN: '[ ',
	LATEX_SPECTRAL_OPERATOR_CLOSE: ' ] ',
	LATEX_BEGIN_EQUATION: '\\begin{align*}',
	LATEX_END_EQUATION: '\\end{align*}',
	LATEX_ALIGN_SEP: ' & ',
	LATEX_SUM_NEWLINE: '\\\\',
	LATEX_LINE_NUMBER: ' \\EqLine',
	MATH_PLUS: 'Plus',
	MATH_TIMES: 'Times',
	// MATH_TIMES: function (txt) {return 'Times[' + txt + ']';},
	MATH_TENSOR_PRODUCT: 'TensorProduct',
	MATH_POWER: 'Power',
	MATH_SUBSCRIPT: 'Subscript',
	MATH_SUPERSCRIPT: 'Superscript',
	MATH_SUBSUPERSCRIPT: 'Subsuperscript',
	MATH_BRACKET_OPEN: '[',
	MATH_BRACKET_CLOSE: ']',
	MATH_DEFAULT_LETTER: '\\[NumberSign]',
	MATH_ONE: '1',
	MATH_ZERO: '0',
	MATH_ONE_MATRIX: '𝟙',
	MATH_OPERATOR_DEFAULT_LETTER: '\\[Del]',
	MATH_OUTER_INDEX: '\\[Mu]',
	MATH_INNER_INDEX: '\\[Nu]',
	MATH_NABLA: '\\[Del]',
	MATH_HAT_NABLA: 'OverHat["\\[Del]"]',
	MATH_PARTIAL: '\\[PartialD]',
	MATH_DELTA: '\\[Delta]',
	MATH_LAPLACIAN: 'OverHat["\\[CapitalDelta]"]',
	CR: `
`,
};

// various symbols used to construct display
// they change according to subclass of MathExpression
const DISPLAY_SYMBOLS = {};

DISPLAY_SYMBOLS.MATHEXPRESSION = {
		symbol_prefactor_one: CONST.SYMBOL_ONE,
		symbol_prefactor_zero: CONST.SYMBOL_ZERO,
		b_prefactor_non_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_one_non_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_zero_non_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_one_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_zero_empty_list: CONST.SYMBOL_PRODUCT,
		symbol_open_list: CONST.SYMBOL_DEFAULT_OPEN_LIST,
		symbol_end_list: CONST.SYMBOL_DEFAULT_END_LIST,
		symbol_empty_list: CONST.SYMBOL_DEFAULT_EMPTY_LIST,
		symbol_open_item: CONST.SYMBOL_DEFAULT_OPEN_ITEM,
		symbol_end_item: CONST.SYMBOL_DEFAULT_END_ITEM,
		symbol_operation: CONST.SYMBOL_EXPRESSION,
		symbol_open_item_number: '',
		symbol_end_item_number: ''
};

DISPLAY_SYMBOLS.MATHPRODUCT = {
		symbol_prefactor_one: '',
		symbol_prefactor_zero: CONST.SYMBOL_ZERO,
		b_prefactor_non_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_one_non_empty_list: '',
		b_prefactor_zero_non_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_one_empty_list: '',
		b_prefactor_zero_empty_list: CONST.SYMBOL_PRODUCT,
		symbol_open_list: '',
		symbol_end_list: '',
		symbol_empty_list: CONST.SYMBOL_BOLD_ONE,
		symbol_open_item: CONST.SYMBOL_DEFAULT_OPEN_ITEM,
		symbol_end_item: CONST.SYMBOL_DEFAULT_END_ITEM,
		symbol_operation: CONST.SYMBOL_PRODUCT,
		symbol_open_item_number: '',
		symbol_end_item_number: ''
};

DISPLAY_SYMBOLS.MATHTENSORPRODUCT = {
		symbol_prefactor_one: '',
		symbol_prefactor_zero: CONST.SYMBOL_ZERO,
		b_prefactor_non_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_one_non_empty_list: '',
		b_prefactor_zero_non_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_one_empty_list: '',
		b_prefactor_zero_empty_list: CONST.SYMBOL_PRODUCT,
		symbol_open_list: '',
		symbol_end_list: '',
		symbol_empty_list: CONST.SYMBOL_TENSOR_PRODUCT_EMPTY_LIST,
		symbol_open_item: CONST.SYMBOL_TENSOR_PRODUCT_OPEN_ITEM,
		symbol_end_item: CONST.SYMBOL_TENSOR_PRODUCT_END_ITEM,
		symbol_operation: CONST.SYMBOL_TENSOR_PRODUCT,
		symbol_open_item_number: '',
		symbol_end_item_number: ''
};

DISPLAY_SYMBOLS.MATHARGUMENT = {
		symbol_prefactor_one: '',
		symbol_prefactor_zero: CONST.SYMBOL_ZERO,
		b_prefactor_non_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_one_non_empty_list: '',
		b_prefactor_zero_non_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_one_empty_list: '',
		b_prefactor_zero_empty_list: CONST.SYMBOL_PRODUCT,
		symbol_open_list: '',
		symbol_end_list: '',
		symbol_empty_list: CONST.SYMBOL_TENSOR_PRODUCT_EMPTY_LIST,
		symbol_open_item: CONST.SYMBOL_TENSOR_PRODUCT_OPEN_ITEM,
		symbol_end_item: CONST.SYMBOL_TENSOR_PRODUCT_END_ITEM,
		symbol_operation: CONST.SYMBOL_TENSOR_PRODUCT,
		symbol_open_item_number: '',
		symbol_end_item_number: ''
};

DISPLAY_SYMBOLS.MATHSUM = {
		symbol_prefactor_one: '',
		symbol_prefactor_zero: '',
		b_prefactor_non_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_one_non_empty_list: '',
		b_prefactor_zero_non_empty_list: CONST.SYMBOL_PRODUCT,
		b_prefactor_empty_list: CONST.SYMBOL_ZERO,
		b_prefactor_one_empty_list: CONST.SYMBOL_ZERO,
		b_prefactor_zero_empty_list: CONST.SYMBOL_ZERO,
		symbol_open_list: '',
		symbol_end_list: '',
		symbol_empty_list: '',
		symbol_open_item: '',
		symbol_end_item: '',
		symbol_operation: '%B_SYMBOL_SUM%' + CONST.SYMBOL_SUM,
		symbol_open_item_number: ' {#',
		symbol_end_item_number: '}'
};


const OPT = {};
// global options to show in console
// can be used anywhere since imported in all files
// colors: https://www.w3.org/wiki/CSS/Properties/color/keywords
OPT.SHOW = {
	type: false,
	contractions: true,
	mapIndices: false,
	position: true,
	sumContractions: false,
	showItemNumber: true,
	display: {
		formatText: function (txt) { return chalk.rgb(120, 120, 120)(txt) },
		formatRational: function (txt) { return chalk.black(txt) },
		formatPolynomials: function (txt) { return chalk.black(txt) },
		formatRationalExpression: function (txt) { return chalk.black(txt) },
		formatPrefactor: function (txt) { return chalk.black(txt) },
		formatList: function (txt) { return chalk.black(txt) },
		formatLetter: function (txt, kind = CONST.ELEMENT_NON_COMMUTATIVE) {
			switch(kind) {
				case CONST.ELEMENT_NON_COMMUTATIVE:
					return chalk.bold.black(txt);
					break;
				case CONST.ELEMENT_COMMUTATIVE:
					return chalk.bold.blue(txt);
					break;
			};
		},
		formatPosition: function (txt) { return chalk.keyword('magenta')(txt) },
		formatIndices: function (txt) { return chalk.keyword('red')(txt) },
		formatClass: function (txt) { return chalk.keyword('olive')(txt) },
		formatPreElement: function(txt) { return txt },
		formatPostElement: function(txt) { return chalk.keyword('purple')(txt) },
		formatContractions: function (txt) { return chalk.keyword('green')(txt) },
	},
};

// global options to construct LaTeX strings
OPT.LTX = {
	showSpectralOperators: true,
	showLineNumbers: true,
	showOneElement: true,
	replacePolynomials: true,
	factorizePolynomials: true,
};

OPT.LTX.polynomialsStringDictionary = {
	};

// global options to construct Mathematica FullForm strings
OPT.MATH = {
	showSpectralOperators: false,
	showPreElementListInMathArgument: false,
	showRationalExpressionInMathArgument: false,
	showPostElementInMathArgument: false,
};

// global options to spectral function export to Mathematica
OPT.SPEC = {
	applyDimensionValue: false,
	dimension: 2,
	spectralFunctionString: 'I\\[Alpha]',
	letterSpectralArgument: 'r',
	letterHalfDimensionSpectral: 'm',
	shiftHalfDimension: 0
};

// global options for titles
OPT.TITLE = {
	widthTitle: 70,
	widthSubtitle: 70,
	widthSubsubtitle: 70,
	widthZone: 70,
};

// object to store file paths
const FILES = {
	SUBST: {
		Gamma_to_nc: 'RESULTS/SUBSTITUTIONS/SubstitutionsGamma.json',
		H_to_nc: 'RESULTS/SUBSTITUTIONS/SubstitutionsH.json',
		L_to_nc: 'RESULTS/SUBSTITUTIONS/SubstitutionsL.json',
		hnabla_u_to_nc: 'RESULTS/SUBSTITUTIONS/Substitutions-hnabla-u-NC.json',
		hnabla_u_to_Gamma: 'RESULTS/SUBSTITUTIONS/Substitutions-hnabla-u-Gamma.json',
		hnabla_q_to_nc: 'RESULTS/SUBSTITUTIONS/Substitutions-hnabla-q-NC.json',
		hnabla_q_to_Gamma: 'RESULTS/SUBSTITUTIONS/Substitutions-hnabla-q-Gamma.json',
		hnabla_N_to_nc: 'RESULTS/SUBSTITUTIONS/Substitutions-hnabla-N-NC.json',
		hnabla_N_to_Gamma: 'RESULTS/SUBSTITUTIONS/Substitutions-hnabla-N-Gamma.json',
		nabla_H_to_delta: 'RESULTS/SUBSTITUTIONS-NCT/Substitutions-NCT-H.json',
		hnabla_u_to_delta: 'RESULTS/SUBSTITUTIONS-NCT/Substitutions-NCT-u.json',
		nabla_L_to_delta: 'RESULTS/SUBSTITUTIONS-NCT/Substitutions-NCT-L.json',
		nabla_L_P2_to_delta: 'RESULTS/SUBSTITUTIONS-NCT/Substitutions-NCT-L-P2.json',
		nabla_q_to_delta: 'RESULTS/SUBSTITUTIONS-NCT/Substitutions-NCT-q.json'
	},
	JSON: {
		R2: 'RESULTS/MAIN-R2-R4/R2.json',
		R2_N0: 'RESULTS/MAIN-R2-R4/R2-N0.json',
		R4_save: 'RESULTS/MAIN-R2-R4/R4-after-nabla.json',
		R4_split: [
				'RESULTS/MAIN-R2-R4/R4-after-nabla-0.json',
				'RESULTS/MAIN-R2-R4/R4-after-nabla-1.json',
				'RESULTS/MAIN-R2-R4/R4-after-nabla-2.json',
				'RESULTS/MAIN-R2-R4/R4-after-nabla-3.json',
				'RESULTS/MAIN-R2-R4/R4-after-nabla-4.json'
			],
		R4: [
				'RESULTS/MAIN-R2-R4/R4-0.json',
				'RESULTS/MAIN-R2-R4/R4-1.json',
				'RESULTS/MAIN-R2-R4/R4-2.json',
				'RESULTS/MAIN-R2-R4/R4-3.json',
				'RESULTS/MAIN-R2-R4/R4-4.json'
			],
		R4_N0: [
				'RESULTS/MAIN-R2-R4/R4-0-N0.json',
				'RESULTS/MAIN-R2-R4/R4-1-N0.json',
				'RESULTS/MAIN-R2-R4/R4-2-N0.json',
				'RESULTS/MAIN-R2-R4/R4-3-N0.json',
				'RESULTS/MAIN-R2-R4/R4-4-N0.json'
			],
		R4_ns: [
				'RESULTS/MAIN-R2-R4/R4-0-ns.json',
				'RESULTS/MAIN-R2-R4/R4-1-ns.json',
				'RESULTS/MAIN-R2-R4/R4-2-ns.json',
				'RESULTS/MAIN-R2-R4/R4-3-ns.json',
				'RESULTS/MAIN-R2-R4/R4-4-ns.json'
			],
		R4_ns_N0: [
				'RESULTS/MAIN-R2-R4/R4-0-ns-N0.json',
				'RESULTS/MAIN-R2-R4/R4-1-ns-N0.json',
				'RESULTS/MAIN-R2-R4/R4-2-ns-N0.json',
				'RESULTS/MAIN-R2-R4/R4-3-ns-N0.json',
				'RESULTS/MAIN-R2-R4/R4-4-ns-N0.json'
			],
		R2_NCT_P1_ns: 'RESULTS/NCT-R2-R4/R2-NCT-P1-ns.json',
		R2_NCT_P2_ns: 'RESULTS/NCT-R2-R4/R2-NCT-P2-ns.json',
		R2_NCT_P1: 'RESULTS/NCT-R2-R4/R2-NCT-P1.json',
		R2_NCT_P2: 'RESULTS/NCT-R2-R4/R2-NCT-P2.json',
		R2_NCT: 'RESULTS/NCT-R2-R4/R2-NCT.json',
		R4_NCT_ns: 'RESULTS/NCT-R2-R4/R4-NCT-ns.json',
	},
	TEX: {
		R2: 'LATEX/MAIN-R2-R4/R2.tex',
		R2_N0: 'LATEX/MAIN-R2-R4/R2-N0.tex',
		R4: [
				'LATEX/MAIN-R2-R4/R4-0.tex',
				'LATEX/MAIN-R2-R4/R4-1.tex',
				'LATEX/MAIN-R2-R4/R4-2.tex',
				'LATEX/MAIN-R2-R4/R4-3.tex',
				'LATEX/MAIN-R2-R4/R4-4.tex'
			],
		R4_N0: [
				'LATEX/MAIN-R2-R4/R4-0-N0.tex',
				'LATEX/MAIN-R2-R4/R4-1-N0.tex',
				'LATEX/MAIN-R2-R4/R4-2-N0.tex',
				'LATEX/MAIN-R2-R4/R4-3-N0.tex',
				'LATEX/MAIN-R2-R4/R4-4-N0.tex'
			],
		R4_ns: [
				'LATEX/MAIN-R2-R4/R4-0-ns.tex',
				'LATEX/MAIN-R2-R4/R4-1-ns.tex',
				'LATEX/MAIN-R2-R4/R4-2-ns.tex',
				'LATEX/MAIN-R2-R4/R4-3-ns.tex',
				'LATEX/MAIN-R2-R4/R4-4-ns.tex'
			],
		R4_ns_N0: [
				'LATEX/MAIN-R2-R4/R4-0-ns-N0.tex',
				'LATEX/MAIN-R2-R4/R4-1-ns-N0.tex',
				'LATEX/MAIN-R2-R4/R4-2-ns-N0.tex',
				'LATEX/MAIN-R2-R4/R4-3-ns-N0.tex',
				'LATEX/MAIN-R2-R4/R4-4-ns-N0.tex'
			],
		// R4_N0_SEL_Nabla_u_4_times : [
		// 		'LATEX/MAIN-R2-R4/R4-0-N0-Nu-4x-12.tex',
		// 		'LATEX/MAIN-R2-R4/R4-0-N0-Nu-4x-13.tex',
		// 		'LATEX/MAIN-R2-R4/R4-0-N0-Nu-4x-14.tex',
		// 	],
		R4_0_u_parallel_N0: 'LATEX/u-PARALLEL/R4-0-u-parallel-N0.tex',
		R4_2_u_parallel_N0: 'LATEX/u-PARALLEL/R4-2-u-parallel-N0.tex',
		R4_3_u_parallel_N0: 'LATEX/u-PARALLEL/R4-3-u-parallel-N0.tex',
		R4_4_u_parallel_N0: 'LATEX/u-PARALLEL/R4-4-u-parallel-N0.tex',
		R4_0_u_parallel: 'LATEX/u-PARALLEL/R4-0-u-parallel.tex',
		R4_2_u_parallel: 'LATEX/u-PARALLEL/R4-2-u-parallel.tex',
		R4_3_u_parallel: 'LATEX/u-PARALLEL/R4-3-u-parallel.tex',
		R4_4_u_parallel: 'LATEX/u-PARALLEL/R4-4-u-parallel.tex',
		R2_NCT_P1_ns: 'LATEX/NCT-R2-R4/R2-NCT-P1-ns.tex',
		R2_NCT_P2_ns: 'LATEX/NCT-R2-R4/R2-NCT-P2-ns.tex',
		R2_NCT_P1: 'LATEX/NCT-R2-R4/R2-NCT-P1.tex',
		R2_NCT_P2: 'LATEX/NCT-R2-R4/R2-NCT-P2.tex',
		R2_NCT: 'LATEX/NCT-R2-R4/R2-NCT.tex',
		R2_NCT_d2: 'LATEX/NCT-R2-R4/R2-NCT-d2.tex',
	},
	MATH: {
		R2: 'MATHEMATICA/MAIN-R2-R4/R2.txt',
		R2_N0: 'MATHEMATICA/MAIN-R2-R4/R2-N0.txt',
		R4: [
				'MATHEMATICA/MAIN-R2-R4/R4-0.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-1.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-2.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-3.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-4.txt'
			],
		R4_N0: [
				'MATHEMATICA/MAIN-R2-R4/R4-0-N0.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-1-N0.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-2-N0.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-3-N0.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-4-N0.txt'
			],
		R4_ns: [
				'MATHEMATICA/MAIN-R2-R4/R4-0-ns.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-1-ns.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-2-ns.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-3-ns.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-4-ns.txt'
			],
		R4_ns_N0: [
				'MATHEMATICA/MAIN-R2-R4/R4-0-ns-N0.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-1-ns-N0.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-2-ns-N0.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-3-ns-N0.txt',
				'MATHEMATICA/MAIN-R2-R4/R4-4-ns-N0.txt'
			],
		R4_0_u_parallel_N0: 'MATHEMATICA/u-PARALLEL/R4-0-u-parallel-N0.txt',
		R4_2_u_parallel_N0: 'MATHEMATICA/u-PARALLEL/R4-2-u-parallel-N0.txt',
		R4_3_u_parallel_N0: 'MATHEMATICA/u-PARALLEL/R4-3-u-parallel-N0.txt',
		R4_4_u_parallel_N0: 'MATHEMATICA/u-PARALLEL/R4-4-u-parallel-N0.txt',
	}
};



// not used
const display_default = {
	formatText: function (txt) { return txt },
	formatRational: function (txt) { return txt },
	formatPolynomials: function (txt) { return txt },
	formatRationalExpression: function (txt) { return txt },
	formatPrefactor: function (txt) { return txt },
	formatList: function (txt) { return txt },
	formatLetter: function (txt) { return txt },
	formatPosition: function (txt) { return txt },
	formatIndices: function (txt) { return txt },
	formatClass: function (txt) { return txt },
	formatPreElement: function(txt) { return txt },
	formatPostElement: function(txt) { return txt },
	formatContractions: function (txt) { return txt },
};


module.exports = {
	GLOBAL: GLOBAL,
	CONST: CONST,
	DISPLAY_SYMBOLS: DISPLAY_SYMBOLS,
	OPT: OPT,
	FILES: FILES,
};
