/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');



class MathElement {

	constructor() { // MathElement
		// productKind of element, for product and interaction with neigbor
		this.productKind = CONST.ELEMENT_COMMUTATIVE;
		// letter to use to display in terminal
		this.letter = CONST.ELEMENT_DEFAULT_LETTER;
		// letter to use in Mathematica exportation
		this.math_letter = CONST.MATH_DEFAULT_LETTER;
		// latex command (letter) to use in LaTeX
		this.latex_letter = CONST.LATEX_DEFAULT_LETTER;
		// unique ID to compare
		this.id = FH.createUniqueID(2, '@', GLOBAL.listElementID);
		// weight to sort sums
		this.sum_weight = 0;
		// weight to sort products
		this.product_weight = 0;
		// list and properties of indices
		// initialize id to empty string
		// list: [{pos: CONST.INDICE_DOWN or CONST.INDICE_UP,
		// 				contraction: CONST.INDICE_OUTER or CONST.INDICE_INNER}, …]
		// REMARK: contraction is not updated currently when indice is contracted…
		this.indices = {id: GLOBAL.defaultID, list:[], symmetric: [], antisymmetric: []};
		// list of pairs of indices contractions
		this.contractions = [];
		// informations stored during copies
		this.copiedInfo = {mapIndicesID: []};
		// string to sort in product for commutative elements
		this.productStringWeight = '';
		// top down data about indices: position, kind, etc…
		this.normalizedIndices = [];

	};

	static showSumWeight() { // MathElement
		console.log('Sum Weights: ' + JSON.stringify(GLOBAL.sumWeightData));
	};

	static showProductWeight() { // MathElement
		console.log('Product Weights: ' + JSON.stringify(GLOBAL.productWeightData));
	};

	static showCurrentIndiceID() { // MathElement
		console.log('Current Indice ID: ' + JSON.stringify(GLOBAL.currentIndiceID));
	};

	updateExpression() {
		return this;
	}

	getID() { // MathElement
		return this.id;
	};

	getChainedID() { // MathElement
		return this.getID();
	};

	setID(id) { // MathElement
		let currentID = this.id;
		if (GLOBAL.listElementID.indexOf(id) === -1) {
			this.id = id;
		};
		// add in list of used ID
		GLOBAL.listElementID.push(id);
		// remove old one
		let i = GLOBAL.listElementID.indexOf(currentID);
		if (i > -1) {
			GLOBAL.listElementID.splice(i,1);
		};
		return this;
	};

	// change id:
	// add newID to element
	// does not check if newID exists
	// remove oldID from global array
	// does not add newID to global list
	flipID(oldID, newID) { // MathElement
		this.id = newID;
		// remove old one
		let i = GLOBAL.listElementID.indexOf(oldID);
		if (i > -1) {
			GLOBAL.listElementID.splice(i,1);
		};
		return this;
	};

	getArrID() { // MathElement
		 return this.id;
	};

	getIndicesID() { // MathElement
		// return indice id even if no explicit indices
		return [this.indices.id];
	};

	// return [arrIdx of every indices]
	getIndicesArrIdx() { // MathElement
		let arr = [];
		for (let i = 0; i < this.indices.list.length; i++) {
			arr.push([i]);
		};
		return arr;
	};

	// return the idx-th indice position (CONST.INDICE_DOWN or CONST.INDICE_UP) in MathElement
	getIndicePositionAtIndex(idx) { // MathElement
		return this.indices.list[idx].pos;
	}

	// return the indice position (CONST.INDICE_DOWN or CONST.INDICE_UP) at position arrIdx
	getIndicePositionAt(arrIdx) { // MathElement
 		if (arrIdx.length === 1) {
			// arrIdx = [idx]
			let idx = arrIdx[0];
			return this.getIndicePositionAtIndex(idx)
		} else if (arrIdx.length === 2) {
			// arrIdx = [-n, idx]
			let idx = arrIdx[1];
			let itemArrIdx = arrIdx.slice(0,-1);
			return this.getItemAt(itemArrIdx).getIndicePositionAtIndex(idx);
		} else {
			throw "error in array"
		};
	};


	getTextIndiceAt(arrIdx) { // MathElement
		// console.log('arrIdx: ' + JSON.stringify(arrIdx));
		// arrIdx is an array of indices in the hierarchy of objets
		// here, it should be reduced to length <= 2
		// if first element is <0 and MathOperator => search in its argument
		if (arrIdx.length === 1) {
			// pure MathElement
			let l = this.indices.list.length;
			let idx = arrIdx[0];
			if (idx < l && 0 <= idx) {
				return this.indices.id + '-' + idx;
			};
		} else if (arrIdx.length === 2) {
			// MathOperator => look inside if first idx <-1
			let idx = arrIdx[0];
			if (idx === -1) {
				// can be MathOperator or last MathElement in argument
				return this.indices.id + '-' + arrIdx[1];
			} else if (idx < -1
					&& this.constructor.name === 'MathOperator'
					&& this.status === CONST.OPERATOR_STATUS_APPLIED) {
				let nextArrIdx = FH.copyData(arrIdx);
				nextArrIdx[0]++;
				// console.log('call argument');
				return this.argument.getTextIndiceAt(nextArrIdx);
			};
		};
		return '';
	};


	getPositionsByID(itemID, maxLevel, termElement) { // MathElement
		if (this.id === itemID) {
			return [[]];
		};
		return null;
	};

	getFirstPositionByID(itemID, maxLevel, termElement) { // MathElement
		if (this.id === itemID) {
			return [];
		};
		return null;
	};

	getLatentOperator() { // MathElement
		if (this.constructor.name === 'MathOperator'
							&& this.status === CONST.OPERATOR_STATUS_LATENT) {
			return [[]];
		};
		return null;
	};

	getLastLatentOperator() { // MathElement
		if (this.constructor.name === 'MathOperator'
							&& this.status === CONST.OPERATOR_STATUS_LATENT) {
			return [];
		};
		return null;
	};


	getCommutativeElements() { // MathElement
		if (this.constructor.name === 'MathOperator') {
			if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
				let argArrIdx = this.argument.getCommutativeElements();
				if (argArrIdx === null) {
					// nothing in argument
					return null;
				} else if (argArrIdx[0].length == 0) {
					// argument is the MathElement
					return [[-2]];
				} else {
					// MathElement found deeper
					return [[argArrIdx[0][0]--]];
				};
			} else {
				// not applied operator
				return null;
			};
		} else {
			// MathElement, direct test
			// don't consider MathElement_one as a commutative element
			if (this.productKind === CONST.ELEMENT_COMMUTATIVE
						&& this.id !== MathElement_one.id) {
				return [[]];
			} else {
				return null;
			};
		};
	};

	getOperatorAppliedToCommutative(operatorID) { // MathElement
		if (this.constructor.name === 'MathOperator') {
			// this.showConsole(OPT.SHOW, 'this: ');
			if (this.id === operatorID
							&& this.status === CONST.OPERATOR_STATUS_APPLIED) {
				if (this.argument.constructor.name === 'MathElement') {
					if (this.argument.productKind === CONST.ELEMENT_COMMUTATIVE) {
						// found commutative element
						return [[-2]];
					} else {
						// not commutative
						return null;
					};
				} else {
					// argument is MathOperator => look inside
					let argArrIdx = this.argument.getOperatorAppliedToCommutative(operatorID);
					if (argArrIdx !== null) {
						// found something
						argArrIdx[0]--;
					}
					return argArrIdx;
				};
			} else {
				// not operatorID or not applied
				return null;
			}
		} else {
			// not a MathOperator
			return null;
		};
	};


	getProductWeight() { // MathElement
		let arrPW = [];
		let idx = GLOBAL.productWeightData[this.id]['idx'];
		// arrPW: [… , [#ME, #applied MO1, #applied MO1, …], …]
		arrPW[idx] = [1];
		return arrPW;
	};

	getProductStringWeight() { // MathElement
		// this.showConsole(OPT.SHOW, 'this: ');
		// console.log('getProductWeight: ' + JSON.stringify(this.getProductWeight()));
		return FH.constructStringFromProductWeight(
								this.getProductWeight(),
								GLOBAL.productWeightData,
								GLOBAL.productWeightOperatorData);
	};

	getSumWeight() { // MathElement
		let arrSW = [];
		let idx = GLOBAL.sumWeightData[this.id]['idx'];
		// arrPW: [… , [#ME, #applied MO1, #applied MO1, …], …]
		arrSW[idx] = [1];
		return arrSW;
	};

	getSumStringWeight() { // MathElement
		return FH.constructStringFromSumWeight(this.getSumWeight(),
																	GLOBAL.sumWeightData, GLOBAL.sumWeightOperatorData);
	};

	getItemAtIndex(idx) { // MathElement
		// can only be used with negative idx
		return this.getItemAt([idx]);
	};

	getItemAt(arrIdx) {
		// console.log('arrIdx: ' + JSON.stringify(arrIdx));
		// should be fired only with 1 negative element
		if (arrIdx.length === 1) {
			let idx = arrIdx[0];
			if (idx === -1) {
				return this;
			} else if (idx < -1
					&& this.constructor.name === 'MathOperator'
					&& this.status === CONST.OPERATOR_STATUS_APPLIED) {
				let nextArrIdx = FH.copyData(arrIdx);
				nextArrIdx[0]++;
				return this.argument.getItemAt(nextArrIdx);
			};
		};
		return null;
	};

	// return a object {MathElementArr}
	// array with references to all MathElements in 'this'
	getContent() { // MathElement
		return {MathElementArr: [this], MathExpressionArr: []};
	};

	// return obj data of this:
	// 	obj = {
	// 		id: indiceID,
	// 		elementID: id of the MathElement,
	// 		arrIdx: [i] with i # of indice,
	// 		symmetric: arrSym,
	// 		antisymmetric: arrAntiSym
	// 	}
	collectIndicesArray() { // MathElement
		let arr = [];
		for (let i=0; i<this.indices.list.length; i++) {
			let arrSym = [];
			if (this.indices.symmetric.includes(i)) {
				arrSym = this.indices.symmetric;
			};
			let arrAntiSym = [];
			if (this.indices.antisymmetric.includes(i)) {
				arrAntiSym = this.indices.antisymmetric;
			};
			let obj = {id: this.indices.id,
								elementID: this.id,
								arrIdx: [i],
								symmetric: arrSym,
								antisymmetric: arrAntiSym};
			arr.push(obj);
		};
		return arr;
	};

	makeCopy() { // MathElement
		// console.log('\tCopy of ' + this.displayText(OPT.SHOW, []));
		// console.log('\tthis mapIndicesID = ' + JSON.stringify(this.copiedInfo.mapIndicesID));

		let copy = FH.duplicateInstance(this);
		// copy data properties
		copy.letter = FH.copyString(this.letter);
		copy.math_letter = FH.copyString(this.math_letter);
		copy.latex_letter = FH.copyString(this.latex_letter);
		copy.id = FH.copyString(this.id);
		copy.indices = FH.copyData(this.indices);
		copy.copiedInfo = FH.copyData(this.copiedInfo);
		// copy.productKind, copy.sum_weight and copy.product_weight ignored here…
		// copy.contractions => below

		// new indices ID
		let oldIndID = copy.indices.id;
		// console.log('oldIndID: ' + oldIndID);
		let newIndID = FH.nextIndiceID(GLOBAL.currentIndiceID);
		copy.indices.id = newIndID;

		// update copiedInfo
		copy.copiedInfo.mapIndicesID = [];
		// store data about this copy for containers elements
		copy.copiedInfo.mapIndicesID = copy.copiedInfo.mapIndicesID.concat([ [oldIndID, newIndID] ]);

		// special case: applied Operator
		if (this.constructor.name === 'MathOperator'
						&& this.status === CONST.OPERATOR_STATUS_APPLIED) {
			copy.argument = this.argument.makeCopy();
			let newObjCopyInfo = FH.copyData(copy.argument.copiedInfo);
			copy.copiedInfo.mapIndicesID = copy.copiedInfo.mapIndicesID.concat(newObjCopyInfo.mapIndicesID);
		};

		// update contractions
		let strCont = JSON.stringify(this.contractions);
		// console.log(strCont);
		let re = new RegExp(oldIndID, "g");
		strCont = strCont.replace(re, newIndID);
		// console.log(strCont);
		let arrCont = JSON.parse(strCont);
		copy.contractions = arrCont;

		// console.log('\tcopy mapIndicesID = ' + JSON.stringify(copy.copiedInfo.mapIndicesID));

		return copy;
	};

	setProductKindToCommutative() { // MathElement
		this.productKind = CONST.ELEMENT_COMMUTATIVE;
		return this;
	};

	setProductKindToNonCommutative() { // MathElement
		this.productKind = CONST.ELEMENT_NON_COMMUTATIVE;
		return this;
	};

	setLetter(str) { // MathElement
		this.letter = str;
		return this;
	};

	setMathLetter(str) { // MathElement
		this.math_letter = str;
		return this;
	};

	setLaTeXLetter(str) { // MathElement
		this.latex_letter = str;
		return this;
	};

	setProductWeight(w) { // MathElement
		if (Number.isInteger(w) && w > 0) {
			this.product_weight = w;
			// console.log('');
			// console.log('CALL setProductWeight');
			FH.addWeight({w: w, id: this.id, letter: this.letter}, GLOBAL.productWeightData);
		} else {
			throw "Weight should be positive integer";
		};
		return this;
	};

	setSumWeight(w) { // MathElement
		if (Number.isInteger(w) && w > 0) {
			this.sum_weight = w;
			// console.log('');
			// console.log('CALL setSumWeight');
			FH.addWeight({w: w, id: this.id, letter: this.letter}, GLOBAL.sumWeightData);
		} else {
			throw "Weight should be positive integer";
		};
		return this;
	};

	addIndiceUp() { // MathElement
		this.indices.list.push({pos: CONST.INDICE_UP, contraction: CONST.INDICE_OUTER});
		return this;
	};

	addIndiceDown() { // MathElement
		this.indices.list.push({pos: CONST.INDICE_DOWN, contraction: CONST.INDICE_OUTER});
		return this;
	};


	setSymmetryIndices(arr) { // MathElement
		// arr = [index of symmetric indices]
		let addSymmetry = true;
		if (arr.length < 2) {
			throw "You should provide at least 2 indices";
		} else if (this.indices.list.length >= arr.length) {
			let thePos = this.indices.list[arr[0]]['pos'];
			for (let i=1; i<arr.length; i++) {
				if (this.indices.list[arr[i]] === undefined || this.indices.list[arr[i]]['pos'] !== thePos) {
					throw "You should provide existing indices and at same position";
					addSymmetry = false;
				}
			}
			if (addSymmetry) {
				this.indices.symmetric = arr;
			}
		} else {
			throw "There is not enough indices in this element";
		};
		return this;
	};

	setAntisymmetryIndices(arr) { // MathElement
		// arr = [index of antisymmetric indices]
		let addSymmetry = true;
		if (arr.length < 2) {
			throw "You should provide at least 2 indices";
		} else if (this.indices.list.length >= arr.length) {
			let thePos = this.indices.list[arr[0]]['pos'];
			for (let i=1; i < arr.length; i++) {
				if (this.indices.list[arr[i]] === undefined || this.indices.list[arr[i]]['pos'] !== thePos) {
					throw "You should provide existing indices and at same position";
					addSymmetry = false;
				}
			}
			if (addSymmetry) {
				this.indices.antisymmetric = arr;
			}
		} else {
			throw "There is not enough indices in this element";
		};
		return this;
	};

	setContraction(arrIdx1, arrIdx2) { // MathElement
		// TODO: take into account negative index
		// here arrIdx1 and arrIdx2 should contain only 1 element each
		let srtIdx1 = this.getTextIndiceAt(arrIdx1);
		let srtIdx2 = this.getTextIndiceAt(arrIdx2);
		let arrCont = this.contractions;
		let srtContraction = JSON.stringify(arrCont);
		// test if one of the indices is already in list
		// this test is stronger than only testing if the couple is in the set
		if ( srtIdx1 !== '' && srtIdx2 !== ''
					&& !srtContraction.includes(srtIdx1)
					&& !srtContraction.includes(srtIdx2) ) {
			arrCont.push([srtIdx1, srtIdx2].sort(FH.alphaSort));
			this.contractions = FH.copyData(arrCont); // necessary to copy?
		} else {
			throw "Error while adding contraction indices";
		};
		return this;
	};

	setContractionByIndices(idx1, idx2) { // MathElement
		// TODO: take into account negative index
		this.setContraction([idx1], [idx2]);
		return this;
	};


	isOne() { // MathElement
		return ( this.productKind === CONST.ELEMENT_COMMUTATIVE
						&& this.letter === CONST.ELEMENT_DEFAULT_LETTER );
	};

	propagateNormalizedIndicesArray(arrNormalizedIndices) { // MathElement
		this.normalizedIndices = arrNormalizedIndices;
		return this;
	};

	getStringRepresentation(displaySymbols) { // MathElement
		let symbolOpenID = CONST.SYMBOL_OPEN_ID;
		let symbolCloseID = CONST.SYMBOL_CLOSE_ID;
		if (displaySymbols !== undefined
						&& displaySymbols.useInternalDisplaySymbols === false) {
			symbolOpenID = '';
			symbolCloseID = '';
		};

		let str = '';
		str += symbolOpenID + this.id + symbolCloseID;
		if (this.indices.list.length > 0) {
			let id = this.indices.id;
			str += CONST.SYMBOL_OPEN_INDICES;
			let indicesArrTxt = [];
			for (let i=0; i<this.indices.list.length; i++) {
				let theIndice = this.indices.list[i];
				let indicesPosTxt = '';
				if (theIndice.pos === CONST.INDICE_UP) {
					indicesPosTxt += '^';
				} else {
					indicesPosTxt += '_';
				};
				let indiceID = id + '-' + i
				let e = this.normalizedIndices[indiceID];
				if (e.kind === CONST.INDICE_OUTER) {
					indicesArrTxt.push(indicesPosTxt + CONST.SYMBOL_OUTER_INDEX + e.number);
				} else {
					indicesArrTxt.push(indicesPosTxt + CONST.SYMBOL_INNER_INDEX + e.number);
				};
			};
			str += indicesArrTxt.join('');
			str += CONST.SYMBOL_CLOSE_INDICES;
		};

		return str;
	};

	getReducedStringRepresentationCountingElementID(elID) { // MathElement
		// this.showConsole(OPT.SHOW, 'this: ');
		// console.log('this.normalizedIndices: ' + JSON.stringify(this.normalizedIndices));
		let str = '';
		let counter = 0;
		if (this.id === elID) {
			counter++;
		} else {
			str += this.letter;
			if (this.indices.list.length > 0) {
				let id = this.indices.id;
				str += CONST.SYMBOL_OPEN_INDICES;
				let indicesArrTxt = [];
				for (let i=0; i<this.indices.list.length; i++) {
					let theIndice = this.indices.list[i];
					let indicesPosTxt = '';
					if (theIndice.pos === CONST.INDICE_UP) {
						indicesPosTxt += '^';
					} else {
						indicesPosTxt += '_';
					};
					let indiceID = id + '-' + i;
					let e = this.normalizedIndices[indiceID];
					if (e.kind === CONST.INDICE_OUTER) {
						indicesArrTxt.push(indicesPosTxt + CONST.SYMBOL_OUTER_INDEX + e.number);
					} else {
						indicesArrTxt.push(indicesPosTxt + CONST.SYMBOL_INNER_INDEX + e.number);
					};
				};
				str += indicesArrTxt.join('');
				str += CONST.SYMBOL_CLOSE_INDICES;
			}; // if indices length

		};
		// no indices in this string representation

		return {counter: counter, string: str};
	};

	getMathematica(opt) { // MathElement
		// letter
		let letterTxt = this.math_letter;
		let returnedTxt = letterTxt;

		// indices
		if (this.indices.list.length > 0) {
			let id = this.indices.id;
			let indicesArrTxtUp = [];
			let indicesArrTxtDown = [];
			let collectorArr = null;
			for (let i=0; i<this.indices.list.length; i++) {
				let theIndice = this.indices.list[i];
				if (theIndice.pos === CONST.INDICE_UP) {
					collectorArr = indicesArrTxtUp;
				} else {
					collectorArr = indicesArrTxtDown;
				};
				let indiceID = id + '-' + i
				let e = this.normalizedIndices[indiceID];
				if (e.kind === CONST.INDICE_OUTER) {
					collectorArr.push(CONST.MATH_SUBSCRIPT
												+ CONST.MATH_BRACKET_OPEN
													+ CONST.MATH_OUTER_INDEX
												+ ', '
													+ e.number
												+ CONST.MATH_BRACKET_CLOSE);
				} else {
					collectorArr.push(CONST.MATH_SUBSCRIPT
												+ CONST.MATH_BRACKET_OPEN
													+ CONST.MATH_INNER_INDEX
												+ ', '
													+ e.number
												+ CONST.MATH_BRACKET_CLOSE);
				};
			};

			let indicesTxtUp = '{' + indicesArrTxtUp.join(', ') + '}';
			let indicesTxtDown = '{' + indicesArrTxtDown.join(', ') + '}';

			if (indicesArrTxtUp.length > 0 && indicesArrTxtDown.length > 0) {
				returnedTxt = CONST.MATH_SUBSUPERSCRIPT
											+ CONST.MATH_BRACKET_OPEN
												+ letterTxt
											+ ', '
												+ indicesTxtDown
											+ ', '
												+ indicesTxtUp
											+ CONST.MATH_BRACKET_CLOSE;
			} else if (indicesArrTxtUp.length === 0 && indicesArrTxtDown.length > 0) {
				returnedTxt = CONST.MATH_SUBSCRIPT
											+ CONST.MATH_BRACKET_OPEN
												+ letterTxt
											+ ', '
												+ indicesTxtDown
											+ CONST.MATH_BRACKET_CLOSE;
			} else if (indicesArrTxtUp.length > 0 && indicesArrTxtDown.length === 0) {
				returnedTxt = CONST.MATH_SUPERSCRIPT
											+ CONST.MATH_BRACKET_OPEN
												+ letterTxt
											+ ', '
												+ indicesTxtUp
											+ CONST.MATH_BRACKET_CLOSE;
			};
		};

		// console.log('MathElement');
		// this.showConsole(OPT.SHOW, 'this: ');
		// console.log(returnedTxt);
		// console.log('');

		return returnedTxt;
	};

	getLaTeX(opt) { // MathElement
		// letter
		let letterTxt = this.latex_letter;

		// indices
		// TODO: use package tensor
		// TODO: add ';' if MathElement has different kinds of indices
		let indicesTxt = '';
		if (this.indices.list.length > 0) {
			let id = this.indices.id;
			let indicesArrTxtUp = [];
			let indicesArrTxtDown = [];
			let collectorArr = null;
			for (let i=0; i<this.indices.list.length; i++) {
				let theIndice = this.indices.list[i];
				if (theIndice.pos === CONST.INDICE_UP) {
					collectorArr = indicesArrTxtUp;
				} else {
					collectorArr = indicesArrTxtDown;
				};
				let indiceID = id + '-' + i
				let e = this.normalizedIndices[indiceID];
				if (e.kind === CONST.INDICE_OUTER) {
					collectorArr.push(CONST.LATEX_OUTER_INDEX
												+ CONST.LATEX_SUB
												+ '{'
												+ e.number
												+ '}');
				} else {
					collectorArr.push(CONST.LATEX_INNER_INDEX
												+ CONST.LATEX_SUB
												+ '{'
												+ e.number
												+ '}');
				};
			};
			let indicesTxtUp = indicesArrTxtUp.join('');
			let indicesTxtDown = indicesArrTxtDown.join('');
			if (indicesTxtUp !== '') {
				indicesTxt += CONST.LATEX_EXPONENT + '{' + indicesTxtUp + '}';
			};
			if (indicesTxtDown !== '') {
				indicesTxt += CONST.LATEX_SUB + '{' + indicesTxtDown + '}';
			};
		};

		let txt = letterTxt + indicesTxt;
		// console.log('(MathElement) ltx: ' + txt);
		// console.log('(MathElement) sign: ' + '1');
		return {ltx: txt, sign: 1};
	};

	displayText(opt, arrIdx) { // MathElement
//		console.log("MathElement : " + JSON.stringify(opt));
		let od = opt.display;
		// letter
		let letterTxt = this.letter;
		letterTxt = od.formatLetter(letterTxt, this.productKind);
		// position inside hierarchy
		let positionTxt = '';
		if (opt.position) {
			if (arrIdx.length > 0) {
				positionTxt += '(' + arrIdx.join(',') + ')';
			};
		};
		positionTxt = od.formatPosition(positionTxt);
		// indices
		let indicesTxt = '';
		if (opt.normalized) {
			if (this.indices.list.length > 0) {
				let id = this.indices.id;
				indicesTxt += CONST.SYMBOL_OPEN_INDICES;
				let indicesArrTxt = [];
				for (let i=0; i<this.indices.list.length; i++) {
					let theIndice = this.indices.list[i];
					let indicesPosTxt = '';
					if (theIndice.pos === CONST.INDICE_UP) {
						indicesPosTxt += CONST.SYMBOL_INDICE_UP;
					} else {
						indicesPosTxt += CONST.SYMBOL_INDICE_DOWN;
					};
					let indiceID = id + '-' + i
					let e = this.normalizedIndices[indiceID];
					if (e.kind === CONST.INDICE_OUTER) {
						indicesArrTxt.push(indicesPosTxt + CONST.SYMBOL_OUTER_INDEX + e.number);
					} else {
						indicesArrTxt.push(indicesPosTxt + CONST.SYMBOL_INNER_INDEX + e.number);
					};
				};
				indicesTxt += indicesArrTxt.join('');
				indicesTxt += CONST.SYMBOL_CLOSE_INDICES;
			};
		} else {
			if (this.indices.list.length > 0) {
				indicesTxt += CONST.SYMBOL_OPEN_INDICES;
				if (opt.mapIndices) {
					if (this.copiedInfo.mapIndicesID.length > 0) {
						indicesTxt += this.copiedInfo.mapIndicesID[0][0];
						indicesTxt += CONST.SYMBOL_MAP_INDICES;
					};
				};
				indicesTxt += this.indices.id;
				for (let i=0; i<this.indices.list.length; i++) {
					let theIndice = this.indices.list[i];
					if (theIndice.pos === CONST.INDICE_UP) {
						indicesTxt += CONST.SYMBOL_INDICE_UP + i;
					} else {
						indicesTxt += CONST.SYMBOL_INDICE_DOWN + i;
					};
				};
				indicesTxt += CONST.SYMBOL_CLOSE_INDICES;
			};
		};
		indicesTxt = od.formatIndices(indicesTxt);
		// class name
		let classTxt = '';
		if (opt.type) {
			classTxt = CONST.SYMBOL_OPEN_CLASS
								+ this.constructor.name
								+ CONST.SYMBOL_CLOSE_CLASS;
		};
		classTxt = od.formatClass(classTxt);
		// contractions of indices
		let contractionsTxt = ''
		if (opt.contractions) {
			contractionsTxt = FH.formatContractionDisplay(this.contractions);
		};
		contractionsTxt = od.formatContractions(contractionsTxt);
		return classTxt + letterTxt + positionTxt + indicesTxt + contractionsTxt;
	};

	showConsole(opt, text = '') { // MathElement
		console.log(opt.display.formatText(text) + this.displayText(opt, []));
		return this;
	};

	toObject() { // MathElement
		return {
			// class constructor
			class: this.constructor.name,
			// product kind
			pK: this.productKind,
			// letter
			l: this.letter,
			// Mathematica
			math: this.math_letter,
			// LaTeX
			ltx: this.latex_letter,
			// id
			id: this.id,
			// sum weight
			sW: this.sum_weight,
			// product weight
			pW: this.product_weight,
			// indices data
			ind: FH.duplicateInstance(this.indices),
			// contractions array
			cont: FH.copyData(this.contractions),
//			copiedInfo: FH.duplicateInstance(this.copiedInfo)
		};
	};

	toJSON(...args) { // MathElement
		return JSON.stringify(this.toObject(), ...args);
	};


};

class MathOperator extends MathElement {
	constructor() { // MathOperator
		super();
		// default is non commutative
		this.productKind = CONST.ELEMENT_NON_COMMUTATIVE;
		// the behavior of the operator concerning products
		this.behavior = CONST.OPERATOR_LEIBNIZ_RULE;
		this.letter = CONST.OPERATOR_DEFAULT_LETTER;
		this.math_letter = CONST.MATH_OPERATOR_DEFAULT_LETTER;
		this.latex_letter = CONST.LATEX_OPERATOR_DEFAULT_LETTER;
		// status of the operator : latent or applied
		this.status = CONST.OPERATOR_STATUS_LATENT;
		// the argument of the operator when applied
		// MathElement
		// TODO: MathProduct, but then propagate its prefactor, for instance by returning a MathProduct
		this.argument = null;
	};

	static showProductWeightOperators() { // MathOperator
		console.log('Product Weights Operators: ' + JSON.stringify(GLOBAL.productWeightOperatorData));
	};

	static showSumWeightOperators() { // MathOperator
		console.log('Sum Weights Operators: ' + JSON.stringify(GLOBAL.sumWeightOperatorData));
	};

	getPositionsByID(itemID, maxLevel, termElement) { // MathOperator
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			let arrArg = this.argument.getPositionsByID(itemID, maxLevel, termElement);
			if (arrArg !== null) {
				let idx = null;
				let arrIdx = arrArg[0]; // empty array or contains negative index
				if (arrIdx.length > 0) {
					// contains negative index
					idx = arrIdx[0];
					idx--;
					// return [[idx]];
				} else {
					// element found as last argument
					idx = -2;
					// return [[-2]];
				};
				// test max level and null
				if (-idx - 1 > maxLevel || idx === null) {
					return null;
				} else {
					return [[idx]];
				};
			} else {
				// no element found
				return null;
			};
		} else {
			// not applied
			if (this.id === itemID) {
				return [[-1]];
			};
			return null;
		};
	};

	getFirstPositionByID(itemID, maxLevel, termElement) { // MathOperator
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			let arrArg = this.argument.getFirstPositionByID(itemID, maxLevel);
			if (arrArg !== null) {
				let idx = null;
				if (arrArg.length > 0) {
					// contains negative index
					idx = arrArg[0];
					idx--;
				} else {
					// element found as last argument
					idx = -2;
				};
				if (-idx - 1 > maxLevel) {
					return null;
				} else {
					return [idx];
				};
			} else {
				// no element found
				return null;
			};
		} else {
			// not applied
			if (this.id === itemID) {
				return [-1];
			};
			return null;
		};
	};


	getID() { // MathOperator
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			return this.argument.getID();
		} else {
			return this.id;
		};
	};

	getOperatorID() { // MathOperator
		return this.id;
	};

	// return a string containing the chained list of id until applied MathElement
	getChainedID() { // MathOperator
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			return this.id + this.argument.getChainedID();
		} else {
			return this.id;
		};
	};


	getArrID() { // MathOperator
		let arr = [this.id];
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			let argID = this.argument.getArrID();
			arr.push(argID);
		};
		return arr;
	};

	getIndicesID() { // MathOperator
		let arr = [this.indices.id];
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			let argID = this.argument.getIndicesID();
			arr.push(argID);
		};
		return arr;
	};

	// return [arrIdx of every indices]
	getIndicesArrIdx() { // MathOperator
		let arr = [];
		let arrSub = null;
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			// MathOperator itself
			for (let a of super.getIndicesArrIdx()) {
				arr.push( [-1].concat(a) );
			};
			//  argument
			arrSub = this.argument.getIndicesArrIdx();
			for (let a of arrSub) {
				if (a.length === 1) {
					// comes from a MathElement
					arr.push( [-2].concat(a) );
				} else {
					// increment depth
					a[0]--;
					arr.push(a);
				};
			};
		} else {
			// not applied
			arr = super.getIndicesArrIdx();
		}
		return arr;
	};

	getProductWeight() { // MathOperator
		let arrPW = [];
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			// array of [#ME, #applied MO1, #applied MO2, …]
			let argPW = this.argument.getProductWeight();
			let argID = this.argument.getID();
			let opID = this.id;
			let argIdx = GLOBAL.productWeightData[argID]['idx'];
			let opIdx = GLOBAL.productWeightOperatorData[opID]['idx'];
			// console.log('opIdx: ' + opIdx);
			// get data for this ME
			// [#ME, #applied MO1, #applied MO2, …] for this ME
			let arrArg = argPW[argIdx];
			if (arrArg[opIdx+1] !== undefined) {
				// increment number of application of this operator
				arrArg[opIdx+1] = arrArg[opIdx+1] + 1;
			} else {
				// first time this operator is applied
				arrArg[opIdx+1] = 1;
			};
			arrPW[argIdx] = arrArg;
		};
		return arrPW;
	};

	getSumWeight() { // MathOperator
		let arrSW = [];
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			// array of [#ME, #applied MO1, #applied MO2, …]
			let argSW = this.argument.getSumWeight();
			let argID = this.argument.getID();
			let opID = this.id;
			let argIdx = GLOBAL.sumWeightData[argID]['idx'];
			let opIdx = GLOBAL.sumWeightOperatorData[opID]['idx'];
			// console.log('opIdx: ' + opIdx);
			// get data for this ME
			// [#ME, #applied MO1, #applied MO2, …] for this ME
			let arrArg = argSW[argIdx];
			if (arrArg[opIdx+1] !== undefined) {
				// increment number of application of this operator
				arrArg[opIdx+1] = arrArg[opIdx+1] + 1;
			} else {
				// first time this operator is applied
				arrArg[opIdx+1] = 1;
			};
			arrSW[argIdx] = arrArg;
		};
		return arrSW;
	};

	getStringRepresentation(displaySymbols) { // MathOperator
		let str = super.getStringRepresentation(displaySymbols);
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			str += '·' + this.argument.getStringRepresentation(displaySymbols);
		};
		return str;
	};

	getReducedStringRepresentationCountingElementID(elID) { // MathOperator
		// elID is not an operator ID
		// this.showConsole(OPT.SHOW, 'this: ');
		let {string: str}
							= super.getReducedStringRepresentationCountingElementID(elID);
		let counter = 0;
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			let {counter: c, string: strArg} = this.argument.getReducedStringRepresentationCountingElementID(elID);
			str += '·' + strArg;
			counter = c;
		};
		return {counter: counter, string: str};
	};

	// return a object {MathElementArr}
	// array with references to all MathElements in 'this'
	getContent() { // MathElement
		let MathElementArr = [this];
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			let argContent = this.argument.getContent();
			MathElementArr = MathElementArr.concat(argContent.MathElementArr);
		};
		return {MathElementArr: MathElementArr, MathExpressionArr: []};
	};

	setBehavior(behavior) {
		this.behavior = behavior;
		return this;
	};

	setOperatorToLatent() { // MathOperator
		this.status = CONST.OPERATOR_STATUS_LATENT;
		this.argument = null;
		return this;
	};

	setProductWeight(w) { // MathOperator
		if (Number.isInteger(w) && w > 0) {
			this.product_weight = w;
			// console.log('');
			// console.log('CALL setProductWeight');
			FH.addWeight({w: w, id: this.id, letter: this.letter}, GLOBAL.productWeightOperatorData);
		} else {
			throw "Weight should be positive integer";
		};
		return this;
	};

	setSumWeight(w) { // MathOperator
		if (Number.isInteger(w) && w > 0) {
			this.sum_weight = w;
			// console.log('');
			// console.log('CALL setSumWeight');
			FH.addWeight({w: w, id: this.id, letter: this.letter}, GLOBAL.sumWeightOperatorData);
		} else {
			throw "Weight should be positive integer";
		};
		return this;
	};

	// return obj data of this and element on which it is applied:
	// 	obj = {
	// 		id: indiceID,
	// 		elementID: id of the MathElement,
	// 		arrIdx: [i] with i # of indice,
	// 		symmetric: arrSym,
	// 		antisymmetric: arrAntiSym
	// 	}
	collectIndicesArray() { // MathOperator
		let arr = [];
		let currArr = super.collectIndicesArray();
		for (let i=0; i<currArr.length; i++) {
			let cObj = currArr[i];
			cObj.arrIdx.unshift(-1);
			arr.push(cObj);
		};
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			let arrArg = this.argument.collectIndicesArray();
			for (let i=0; i<arrArg.length; i++) {
				let cObj = arrArg[i];
				if (cObj.arrIdx[0] < 0) {
					cObj.arrIdx[0]--;
				} else {
					cObj.arrIdx.unshift(-2);
				};
				arr.push(cObj);
			};
		};
		return arr;
	};

	// apply operator to item without copying
	applyTo(item) { // MathOperator
		switch (item.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
				if (this.status === CONST.OPERATOR_STATUS_LATENT) {
					this.argument = item;
					this.status = CONST.OPERATOR_STATUS_APPLIED;
					return this;
				} else {
					throw "Operator already applied";
				};
				break;
			default:
				throw "MathElement or MathOperator required";
				break;
		};
	};

	propagateNormalizedIndicesArray(arrNormalizedIndices) { // MathElement
		this.normalizedIndices = arrNormalizedIndices;
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			this.argument.propagateNormalizedIndicesArray(arrNormalizedIndices);
		};
		return this;
	};

	getMathematica(opt) { // MathOperator
		let returnedTxt = super.getMathematica(opt);
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			let argTxt = this.argument.getMathematica(opt);
			returnedTxt += CONST.MATH_BRACKET_OPEN
											+ argTxt
											+ CONST.MATH_BRACKET_CLOSE;
		};
		// console.log('MathOperator');
		// this.showConsole(OPT.SHOW, 'this: ');
		// console.log(returnedTxt);
		// console.log('');

		return returnedTxt;
	};

	getLaTeX(opt) { // MathOperator
		let txtObj = super.getLaTeX(opt);
		let txt = txtObj.ltx;
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			let argTxtObj = this.argument.getLaTeX(opt);
			txt += CONST.LATEX_OPERATOR_APPLY + argTxtObj.ltx;
			txt = '(' + txt + ')'
		};

		// console.log('(MathOperator) ltx: ' + txt);
		// console.log('(MathOperator) sign: ' + '1');
		return {ltx: txt, sign: 1};
	};

	displayText(opt, arrIdx) { // MathOperator
//		console.log("MathOperator : " + JSON.stringify(opt));
		let od = opt.display;
		// for MathOperator, start negative index for arguments
		let newArrIdx = null;
		if (arrIdx[arrIdx.length-1] < 0) {
			// last element negative
			// => already passed through a MathOperator before
			// => inside argument
			newArrIdx = FH.copyData(arrIdx);
		} else {
			// first time we meet a MathOperator
			// => create the first negative index
			newArrIdx = arrIdx.concat([-1]);
		};
		// call super to display this element
		let txt = super.displayText(opt, newArrIdx);
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			newArrIdx[newArrIdx.length-1]--;
			let newOpt = FH.duplicateInstance(opt);
			newOpt.type = false;
			txt += CONST.SYMBOL_OPERATOR_APPLY + this.argument.displayText(newOpt, newArrIdx);
		};

		return txt;
	};

	toObject() {
		let obj = super.toObject();
		obj.be = this.behavior;
		obj.st = this.status;
		if (this.status === CONST.OPERATOR_STATUS_APPLIED) {
			obj.arg = this.argument.toObject();
		};
		return obj;
	};

};


const MathElement_one = new MathElement();
MathElement_one
	.setID('one')
	.setLetter(CONST.ELEMENT_DEFAULT_LETTER)
	.setMathLetter(CONST.MATH_ONE_MATRIX)
	.setLaTeXLetter(CONST.LATEX_DEFAULT_LETTER)
	.setSumWeight(1)
	.setProductWeight(1);



module.exports = {
	MathElement: MathElement,
	MathOperator: MathOperator,
	MathElement_one: MathElement_one,
};


// const {
// 				MathExpression,
// 				MathProduct,
// 				MathTensorProduct,
// 				MathSum,
// 				MathArgument,
// 			} = require(REL_PATH + 'MATH-OBJECTS/MathExpression');
