/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');


const {
				Polynomials,
				RationalExpression
			} = require(REL_PATH + 'MATH-OBJECTS/MathRationalExpression');

const {
				MathElement,
				MathOperator,
				MathElement_one,
			} = require(REL_PATH + 'MATH-OBJECTS/MathElement');


class MathExpression {
	constructor() { // MathExpression
		// multiplicative prefactor
		// this.prefactor = new Integer(1);
		this.prefactor = RationalExpression.newFromLetterAndRational(CONST.DIMENSION_LETTER,1,1);
		// array of elements: MathElement, MathProduct, MathSum, MathArgument…
		this.list = [];
		// list of pairs of indices contractions
		this.contractions = [];
		// informations stored during copies
		this.copiedInfo = {mapIndicesID: []};
		this.sumStringWeight = '';
		// string representation to compare MathExpression (without prefactor)
		this.stringRepresentation = '';
		// top down data about indices: position, kind, etc…
		this.normalizedIndices = [];

		// the symbols to use for display
		this.display_symbols = DISPLAY_SYMBOLS.MATHEXPRESSION;

	};

	// MathExpression
	static newFrom(item) { // MathExpression
		let newME = new MathExpression();
		switch (item.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
				newME.list.push(item.makeCopy());
				break;
			default:
				throw "MathElement, or MathOperator required";
				breaks;
		};
		newME.updateExpression();
		return newME;
	};

	// transform an item into a type (one of MathExpression subclasses)
	// this does not copy the original item
	// so, best use is myItem = MathExpression.convertTo(myItem, newType)
	// this avoid to let the original item in life
	static convertTo(item, type) { // MathExpression
		// type: MathProduct, MathTensorProduct, MathArgument
		// item: MathElement, MathOperator, MathProduct, MathTensorProduct
		switch (type) {
			case 'MathProduct':
				switch (item.constructor.name) {
					case 'MathElement':
					case 'MathOperator':
						let newME = new MathProduct();
						newME.list.push(item);
						newME.contractions = FH.joinContractions(newME.contractions, item.contractions);
						newME.updateExpression();
						return newME;
						break;
					case 'MathProduct':
						return item;
						break;
					default:
						throw "MathElement, MathOperator, or MathProduct required";
						break;
				};
				break;

			case 'MathTensorProduct':
				switch (item.constructor.name) {
					case 'MathElement':
					case 'MathOperator':
					case 'MathProduct':
						let newItem = MathExpression.convertTo(item, 'MathProduct');
						let newME = new MathTensorProduct();
						newME.list.push(newItem);
						newME.contractions = FH.joinContractions(newME.contractions, newItem.contractions);
						newME.updateExpression();
						return newME;
						break;
					case 'MathTensorProduct':
						return item;
						break;
					default:
						throw "MathElement, MathOperator, MathProduct, or MathTensorProduct required";
						break;
				};
				break;

			case 'MathArgument':
				switch (item.constructor.name) {
					case 'MathElement':
					case 'MathOperator':
					case 'MathProduct':
					case 'MathTensorProduct':
						let newItem = MathExpression.convertTo(item, 'MathTensorProduct');
						let newME = new MathArgument();
						for (let i=0; i<newItem.list.length; i++) {
							newME.list.push(newItem.list[i]);
						};
						// newME.prefactor = newME.prefactor.multiplyBy(newItem.prefactor);
						newME.prefactor = newME.prefactor.multiply(newItem.prefactor);
						newME.contractions = FH.joinContractions(newME.contractions, newItem.contractions);
						newME.updateExpression();
						return newME;
						break;
					case 'MathArgument':
						return item;
						break;
					default:
						throw "MathElement, MathOperator, MathProduct, MathTensorProduct, or MathArgument required";
						break;
				};
				break;

			default:
				throw "MathProduct, MathTensorProduct or MathArgument required for type";
				break;
		};
	};

	// return array of arrIdx of all elements with itemID
	// maxLevel is the max number of derivations applied
	getPositionsByID(itemID, maxLevel, termElement) { // MathExpression
		let arrIdx = [];
		for (let i=0; i<this.list.length; i++) {
			let currentArrIdx = this.list[i].getPositionsByID(itemID, maxLevel, termElement);
			if (currentArrIdx !== null) {
				for (let e of currentArrIdx) {
					arrIdx.push([i].concat(e));
				}
			};
		};
		return arrIdx;
	};

	// return arrIdx of the first element with itemID
	getFirstPositionByID(itemID, maxLevel, termElement) { // MathExpression
		for (let i=0; i<this.list.length; i++) {
			let currentArrIdx = this.list[i].getFirstPositionByID(itemID, maxLevel, termElement);
			if (currentArrIdx !== null) {
					return [i].concat(currentArrIdx);
			};
		};
		return null;
	};

	// collect the operators in MathExpression that are not applied
	// => used in computation
	getLatentOperator() { // MathExpression
		let arrIdx = [];
		for (let i=0; i<this.list.length; i++) {
			let currentArrIdx = this.list[i].getLatentOperator();
			if (currentArrIdx !== null) {
				for (let e of currentArrIdx) {
					arrIdx.push([i].concat(e));
				}
			};
		};
		return arrIdx;
	};

	// find thearrIdx of the last operator in MathExpression
	// which is not applied, or return null is no latent operator
	// => used in computation
	getLastLatentOperator() { // MathExpression
		for (let i=this.list.length-1; i>=0; i--) {
			let currentArrIdx = this.list[i].getLastLatentOperator();
			if (currentArrIdx !== null) {
					return [i].concat(currentArrIdx);
			};
		};
		return null;
	};

	// return array of arrIdx of commutative elements
	getCommutativeElements() { // MathExpression
		let arrIdx = [];
		for (let i=0; i<this.list.length; i++) {
			let currentArrIdx = this.list[i].getCommutativeElements();
			if (currentArrIdx !== null) {
				for (let e of currentArrIdx) {
					arrIdx.push([i].concat(e));
				}
			};
		};
		return arrIdx;
	};

	// return array of arrIdx of commutative elements on which an operator applies
	getOperatorAppliedToCommutative(operatorID) { // MathExpression
		let arrIdx = [];
		for (let i=0; i<this.list.length; i++) {
			let currentArrIdx = this.list[i].getOperatorAppliedToCommutative(operatorID);
			if (currentArrIdx !== null) {
				for (let e of currentArrIdx) {
					arrIdx.push([i].concat(e));
				}
			};
		};
		return arrIdx;
	};


	// return [ids] of all elements
	getID() { // MathExpression
		let arr = [];
		for (let i =0; i < this.list.length; i++) {
			arr.push(this.list[i].getID());
		};
		return arr;
	};

	// return pyramids of ids of elements and operators
	getArrID() { // MathExpression
		let arr = [];
		for (let i =0; i < this.list.length; i++) {
			arr.push(this.list[i].getArrID());
		};
		return arr;
	};

	// return [indice ids] of all elements
	getIndicesID() { // MathExpression
		let arr = [];
		for (let i = 0; i < this.list.length; i++) {
			// let arrIndID = this.list[i].getIndicesID();
			// add array even if empty
			arr.push(this.list[i].getIndicesID());
			// if (arrIndID.length > 0) {
			// 	arr.push(this.list[i].getIndicesID());
			// };
		};
		return arr;
	};

	// return [arrIdx of every indices]
	getIndicesArrIdx() { // MathExpression
		let arr = [];
		let arrSub = null;
		for (let i = 0; i < this.list.length; i++) {
			arrSub = this.list[i].getIndicesArrIdx();
			// add idx = i in front
			for (let a of arrSub) {
				arr.push( [i].concat(a) );
			};
		};
		return arr;
	};

	getOuterIndices() { // MathExpression
		let arr = [];
		let arrCont = this.contractions;
		let srtCont = JSON.stringify(arrCont);
		// console.log(srtCont);
		let arrIndicesArrIdx = this.getIndicesArrIdx();
		for (let arrIdx of arrIndicesArrIdx) {
			let indiceID = this.getTextIndiceAt(arrIdx);
			// console.log(indiceID);
			if (!srtCont.includes(indiceID)) {
				arr.push(arrIdx);
			};
		};
		return arr;
	};

	// return the indice id at position arrIdx
	getTextIndiceAt(arrIdx) { // MathExpression
 		// arrIdx is an array of indices in the hierarchy of objets
 		// last item is the index of the indice in MathElement
		if (arrIdx.length > 1) {
			let idx = arrIdx[0];
			let nextArrIdx = arrIdx.slice(1);
			if (Number.isInteger(idx)) {
				if (0 <= idx && idx < this.list.length) {
					// main list
					return this.list[idx].getTextIndiceAt(nextArrIdx);
				} else if (idx === -1 && this.constructor.name === 'MathArgument') {
					// post element
					return this.postElement.getTextIndiceAt(nextArrIdx);
				} else if (idx === -2 && this.constructor.name === 'MathArgument') {
					// pre element
					return this.preElement.getTextIndiceAt(nextArrIdx);
				};
			};
		};
		return '';
	};

	// return the indice position (CONST.INDICE_DOWN or CONST.INDICE_UP) at position arrIdx
	getIndicePositionAt(arrIdx) { // MathExpression
 		// arrIdx is an array of indices in the hierarchy of objets
 		// last item is the index of the indice in MathElement
		if (arrIdx.length > 1) {
			// console.log('arrIdx (in get): ' + JSON.stringify(arrIdx));
			let idx = arrIdx[arrIdx.length-1];
			let itemArrIdx = arrIdx.slice(0,-1);
			// this.showConsole(OPT.SHOW, 'this = ');
			// console.log('itemArrIdx: ' + JSON.stringify(itemArrIdx));
			// this.getItemAt(itemArrIdx).showConsole(OPT.SHOW, 'item inside = ');
			return this.getItemAt(itemArrIdx).getIndicePositionAtIndex(idx)
		} else {
			throw "error in array"
		};
	};

	getListLength() { // MathExpression
		return this.list.length;
	};

	getSumWeight() { // MathExpression
		if (this.constructor.name === 'MathSum') {
			throw "getSumWeight does not apply to a MathSum object";
		};
		let arrSW = [];
		for (let item of this.list) {
			let arrItemSW = item.getSumWeight();
			for (let idx = 0; idx < arrItemSW.length; idx++) {
				if (arrSW[idx] === undefined) {
					// no previous data at this idx => add the returned data
					arrSW[idx] = arrItemSW[idx];
				} else if (arrItemSW[idx] !== undefined) {
					// combine arrSW[idx] with arrItemSW[idx]
					// index O is #ME, always present
					arrSW[idx][0] = arrSW[idx][0] + arrItemSW[idx][0];
					// loop over #applied MO
					let max = Math.max(arrSW[idx].length, arrItemSW[idx].length);
					for (let opIdx=0; opIdx<max; opIdx++) {
						if (arrSW[idx][opIdx+1] === undefined
									&& arrItemSW[idx][opIdx+1] !== undefined) {
							arrSW[idx][opIdx+1] = arrItemSW[idx][opIdx+1];
						} else if (arrSW[idx][opIdx+1] !== undefined
									&& arrItemSW[idx][opIdx+1] !== undefined) {
							arrSW[idx][opIdx+1] = arrSW[idx][opIdx+1] + arrItemSW[idx][opIdx+1];
						};
						// other configurations:
						// arrSW[idx][opIdx+1] is defined but not arrItemSW[idx][opIdx+1]
						// => arrSW[idx][opIdx+1] remains as it is
						// arrSW[idx][opIdx+1] and arrItemSW[idx][opIdx+1] are not defined
						// => arrSW[idx][opIdx+1] remains undefined
					}; // for opIdx
				}; // if arrItemSW[idx] !== undefined
			}; // for idx
		}; // for item
		return arrSW;
	};

	getSumStringWeight() { // MathExpression
		return FH.constructStringFromWeight(this.getSumWeight(), GLOBAL.sumWeightData, GLOBAL.sumWeightOperatorData);
	};



	makeCopy() { // MathExpression
		// console.log('\tCopy of ' + this.displayText(OPT.SHOW, []).replace(/%B_SYMBOL_SUM%/g, ""));
		// console.log('\tthis mapIndicesID = ' + JSON.stringify(this.copiedInfo.mapIndicesID));

		let copy = FH.duplicateInstance(this);
		// copy all inner objects: prefactor, elements in list, copiedInfo
		copy.prefactor = this.prefactor.makeCopy();
		copy.list = [];
		// copy information to propagate up
		copy.copiedInfo = FH.copyData(this.copiedInfo);
		// copy.contractions => below

		// update copiedInfo
		copy.copiedInfo.mapIndicesID = [];
		for (let i=0; i < this.list.length; i++){
			let newObj = this.list[i].makeCopy();
			let newObjCopyInfo = FH.copyData(newObj.copiedInfo);
			copy.list.push(newObj);
			// store data about this copy for following container elements
			copy.copiedInfo.mapIndicesID = copy.copiedInfo.mapIndicesID
																		.concat(newObjCopyInfo.mapIndicesID);
		};

		// special cases of subclasses with extra properties
		if (this.constructor.name === 'MathArgument') {
			// copy preElement
			copy.preElement = this.preElement.makeCopy();
			// copy postElement
			copy.postElement = this.postElement.makeCopy();
			// collect copy infos from postElement
			// here only mapIndicesID…
			copy.copiedInfo.mapIndicesID = copy.copiedInfo.mapIndicesID
																		.concat(copy.preElement.copiedInfo.mapIndicesID)
																		.concat(copy.postElement.copiedInfo.mapIndicesID);
		};

		// console.log('');
		// this.showConsole(OPT.SHOW, '(makeCopy) term: ');
		// copy.showConsole(OPT.SHOW, '(makeCopy) copy (avant cont): ');


		if (this.constructor.name !== 'MathSum') {
			// update contraction indices with data in copy.copiedInfo
			// no contractions in MathSum
			let mapIndicesID = copy.copiedInfo.mapIndicesID
			// console.log( '(makeCopy) mapIndicesID = ' + JSON.stringify(mapIndicesID) );
			let strCont = JSON.stringify(this.contractions);
			// console.log('(makeCopy) strCont (avant) = ' + strCont);
			for (let i=0; i < mapIndicesID.length; i++) {
				let oldIndID = mapIndicesID[i][0];
				let newIndID = mapIndicesID[i][1];
				let re = new RegExp(oldIndID, "g");
				// console.log('(makeCopy) replace ' + oldIndID + ' by ' + newIndID);
				strCont = strCont.replace(re, newIndID);
			};
			// console.log('(makeCopy) strCont (après) = ' + strCont);
			let arrCont = JSON.parse(strCont);
			// sort each couples
			for (let e of arrCont) {e.sort(FH.alphaSort);};
			// add created new array
			copy.contractions = arrCont;
			// console.log('\tcopy mapIndicesID = ' + JSON.stringify(copy.copiedInfo.mapIndicesID));
		};
		// copy.showConsole(OPT.SHOW, '(makeCopy) copy (après cont): ');

		return copy;
	};

	setContraction(arrIdx1, arrIdx2) { // MathExpression
		// TODO negative idx for MathOperators and MathArguments
		// TODO check up + down
		// arrIdx1 and arrIdx2 are arrays of successive indices in the hierarchy
		// should be correct to obtain good result
		// prevent contraction in different terms of a sum
		if (this.constructor.name === 'MathSum') {
			throw "Contraction not possible on MathSum";
		};
		let srtIdx1 = this.getTextIndiceAt(arrIdx1);
		let srtIdx2 = this.getTextIndiceAt(arrIdx2);
		let arrCont = this.contractions;
		let srtContraction = JSON.stringify(arrCont);
		// test if one of the indices is already in list
		// this test is stronger than only testing if the couple is in the array
		if ( srtIdx1 !== '' && srtIdx2 !== ''
					&& !srtContraction.includes(srtIdx1)
					&& !srtContraction.includes(srtIdx2) ) {
			arrCont.push([srtIdx1, srtIdx2].sort(FH.alphaSort));
			this.contractions = FH.copyData(arrCont); // necessary to copy?
		} else {
			throw "Error while adding contraction indices";
		};
		return this;
	};

	checkContractionArray() { // MathExpression
		let arrCont = this.contractions;
		let collected = [];
		for (let c of arrCont) {
			for (let str of c) {
				if (collected.indexOf(str) === -1 ) {
					collected.push(str);
				} else {
					console.log('');
					console.log('');
					this.showConsole(OPT.SHOW, 'item: ');
					console.log('CHECK CONTRATION ARRAY BAD: ' + JSON.stringify(arrCont));
					console.log('');
				};
			};
		};
	}


	// get reference to item at idx if it exists, otherwise return null
	// can get item in post (idx=-1) and pre (idx=-2) element for MathArgument
	getItemAtIndex(idx) { // MathExpression
		if (Number.isInteger(idx)) {
			if (0 <= idx && idx < this.list.length) {
				return this.list[idx];
			} else if (idx === -1 && this.constructor.name === 'MathArgument') {
				// post element
				return this.postElement;
			} else if (idx === -2 && this.constructor.name === 'MathArgument') {
				// pre element
				return this.preElement;
			};
		};
		return null;
	};

	// return reference to item in the chain of indices arrIdx1
	// null if no object there
	getItemAt(arrIdx) { // MathExpression
		// iteratively use getItemAtIndex(idx)
		if (arrIdx.length > 1) {
			let nextArrIdx = arrIdx.slice(1);
			let currentItem = this.getItemAtIndex(arrIdx[0]);
			return currentItem.getItemAt(nextArrIdx);
		} else if (arrIdx.length == 1) {
			return this.getItemAtIndex(arrIdx[0]);
		} else {
			return null;
		};
	};

	// return a object {MathElementArr, MathExpressionArr}
	// two arrays with references to all MathElements and all MathExpression in 'this'
	getContent() { // MathExpression
		let MathElementArr = [];
		let MathExpressionArr = [this]; // add itself
		for (let item of this.list) {
			let content = item.getContent();
			MathElementArr = MathElementArr.concat(content.MathElementArr);
			MathExpressionArr = MathExpressionArr.concat(content.MathExpressionArr);
		};
		// console.log('MathExpression: ');
		// console.log('MathElementArr: ' + MathElementArr.length);
		// console.log('MathExpressionArr: ' + MathExpressionArr.length);
		return {MathElementArr: MathElementArr, MathExpressionArr: MathExpressionArr};
	};

	// change all the IndiceID in MathExpression
	// starting from "firstAsText" => first number + number of digits
	// default is "0001"
	changeIndicesIDToNumber(firstAsText) { // MathExpression
		if (firstAsText === undefined) {
			firstAsText = '0001';
		};
		let nDigits = firstAsText.length;
		let start = Number(firstAsText);
		if (start === NaN) {
			throw "firstAsText should be the string of a valid number"
		};
		// console.log('nDigits = ' + nDigits);
		// console.log('start = ' + start);

		if (this.constructor.name === 'MathSum') {
			for (let term of this.list) {
				term.changeIndicesIDToNumber(firstAsText);
			};
			return this;
		};

		// Not a MathSum
		let currentNumber = start;
		let {MathElementArr, MathExpressionArr} = this.getContent();
		for (let el of MathElementArr) {
			let currentIndiceID = el.indices.id;
			// console.log('currentIndiceID: ' + currentIndiceID);
			let newIndiceID = FH.convertToString(currentNumber, nDigits);
			// console.log('newIndiceID: ' + newIndiceID);
			let mapIndiceIDArr = [[currentIndiceID, newIndiceID]];
			for (let exp of MathExpressionArr) {
				let currentContArr = exp.contractions;
				// console.log('currentContArr: ' + JSON.stringify(currentContArr));
				let newContArr = FH.updateContractionArray(currentContArr, mapIndiceIDArr);
				// console.log('newContArr: ' + JSON.stringify(newContArr));
				exp.contractions = newContArr;
			};
			el.indices.id = newIndiceID;
			currentNumber++;
		};

		return this;
	};

	// change all the IndiceID in MathExpression
	// using the global position GLOBAL.currentIndiceID
	changeIndicesIDToGlobal() {
		if (this.constructor.name === 'MathSum') {
			for (let term of this.list) {
				term.changeIndicesIDToGlobal();
			};
			return this;
		};

		// Not a MathSum
		let {MathElementArr, MathExpressionArr} = this.getContent();
		for (let el of MathElementArr) {
			let currentIndiceID = el.indices.id;
			// console.log('currentIndiceID: ' + currentIndiceID);
			let newIndiceID = FH.nextIndiceID(GLOBAL.currentIndiceID);
			// console.log('newIndiceID: ' + newIndiceID);
			let mapIndiceIDArr = [[currentIndiceID, newIndiceID]];
			for (let exp of MathExpressionArr) {
				let currentContArr = exp.contractions;
				// console.log('currentContArr: ' + JSON.stringify(currentContArr));
				let newContArr = FH.updateContractionArray(currentContArr, mapIndiceIDArr);
				// console.log('newContArr: ' + JSON.stringify(newContArr));
				exp.contractions = newContArr;
			};
			el.indices.id = newIndiceID;
		};

		return this;
	};

	// insert item as the new index idx, without copying item
	// arrIdx ??? => could be useful
	// PRIVATE METHOD
	insertItemAtIndex(item, idx, updateContractions) { // MathExpression
		// TODO: take into account negative index
		// if i=0, item is the new first
		// if i=this.list.length, item is the new last
		// if updateContractions is true, update list of contractions of 'this' with the one in item
		if (!Number.isInteger(idx) || idx < 0 || this.list.length < idx) {
			throw "index not an integer or not in the good range";
		};
		let newItem = null;
		switch (this.constructor.name) {
			case 'MathProduct':
				switch (item.constructor.name) {
					case 'MathElement':
					case 'MathOperator':
						this.list.splice(idx, 0, item);
						break;
					case 'MathProduct':
						// collect item.prefactor
						// this.prefactor.multiplyBy(item.prefactor);
						this.prefactor = this.prefactor.multiply(item.prefactor);
						// https://stackoverflow.com/questions/7032550/javascript-insert-an-array-inside-another-array
						// collect item.list
						this.list = this.list.slice(0, idx).concat(item.list).concat(this.list.slice(idx));
						break;
					default:
						throw "MathElement, MathOperator, or MathProduct required";
						break;
				};
				break;

			case 'MathTensorProduct':
			case 'MathArgument': // insert in the main list, not in postElement
				switch (item.constructor.name) {
					case 'MathElement':
					case 'MathOperator':
					case 'MathProduct':
						newItem = MathExpression.convertTo(item, 'MathProduct');
						this.list.splice(idx, 0, newItem);
						// collect and normalisation of prefactor is done in updateExpression()
						break;
					case 'MathTensorProduct':
						// collect item.prefactor
						// this.prefactor.multiplyBy(item.prefactor);
						this.prefactor = this.prefactor.multiply(item.prefactor);
						// collect item.list
						this.list = this.list.slice(0, idx).concat(item.list).concat(this.list.slice(idx));
						break;
					default:
						throw "MathElement, MathOperator, MathProduct, or MathTensorProduct required";
						break;
				};
				break;

			case 'MathSum':
			// this case is not very useful since sum is commutative
			// so the same result can be obtained with addTo(item)
			switch (item.constructor.name) {
				case 'MathElement':
				case 'MathOperator':
					// promote to MathProduct
					newItem = MathExpression.convertTo(item, 'MathProduct');
					this.list.splice(idx, 0, newItem);
					break;
				case 'MathProduct':
				case 'MathTensorProduct':
					// insert directly
					this.list.splice(idx, 0, item);
					break;
				case 'MathSum':
					// item.prefactor = 1 by convention, no need to worry…
					// insert directly all the elements of item.list
					this.list = this.list.slice(0, idx).concat(item.list).concat(this.list.slice(idx));
					break;
				default:
					throw "MathElement, MathOperator, MathProduct, MathTensorProduct, or MathSum required";
					break;
			};
				break;

			default:
				throw "MathProduct, MathTensorProduct, or MathSum required";
				break;
		};
		if (updateContractions && this.constructor.name !== 'MathSum') {
			// merge contractions of this with contractions of item
			this.contractions = FH.joinContractions(this.contractions, item.contractions);
		};
		this.updateExpression();
		return this;
	};



	removeItemAtIndex(idx, returnedInfo, updateContractions = true) { // MathExpression
		// TODO: take into account negative index
		// returnedInfo is an object which collects infos about the removed object
		let oldItem = this.getItemAtIndex(idx);
		let arrIndicesID = oldItem.getIndicesID();
		returnedInfo.indicesID = FH.copyData(arrIndicesID);
		returnedInfo.type = oldItem.constructor.name;
		if (oldItem !== null) {
			this.list.splice(idx, 1);
		} else {
			throw "no item at this index";
		};
		if (updateContractions
						&& arrIndicesID.length > 0
						&& this.constructor.name !== 'MathSum') {
			// remove all occurences of contractions containing indices in the removed item
			// flatten arrIndicesID
			let flatArrIndicesID = FH.flattenArray(arrIndicesID);
			let arrCont = this.contractions;
			let newArrCont = [];
			for (e of arrCont) {
				let stre = JSON.stringify(e);
				let found = false;
				for (let i=0; i<flatArrIndicesID.length; i++) {
					let indID = flatArrIndicesID[i];
					if (stre.includes(indID)) {
						found = true;
					};
				};
				if (!found) { newArrCont.push(e); };
			};
			this.contractions = newArrCont;
		};
		this.updateExpression();
		return this;
	};

	// replace the obj at idx by item, without copying item
	// arrIdx ??? => could be useful
	// PRIVATE METHOD
	replaceByItemAtIndex(item, idx, returnedInfo, updateContractions = true) { // MathExpression
		// TODO: take into account negative index
		// remove old item
		this.removeItemAtIndex(idx, returnedInfo, updateContractions);
		// insert new item
		this.insertItemAtIndex(item, idx);
		// updateExpression
		this.updateExpression();
		return this;
	};

	// setPrefactor(h) { // MathExpression
	// 	CONST.DIMENSION_LETTER
	// 	this.prefactor = Rational.newFrom(h);
	// 	this.updateExpression();
	// 	return this;
	// };

	// multiplyBy(h) { // MathExpression
	// 	this.prefactor.multiplyBy(h);
	// 	this.updateExpression();
	// 	return this;
	// };

	multiplyByRational(n,d) { // MathExpression
		// this.prefactor.showConsole(OPT.SHOW, 'multiplyByRational avant = ');
		this.prefactor.multiplyByRational(n,d);
		// this.prefactor.showConsole(OPT.SHOW, 'multiplyByRational après = ');
		this.updateExpression();
		// this.prefactor.showConsole(OPT.SHOW, 'multiplyByRational après update = ');
		return this;
	};

	multiplyByInteger(n) { // MathExpression
		// this.prefactor.showConsole(OPT.SHOW, 'multiplyByInteger avant = ');
		this.multiplyByRational(n,1);
		// this.prefactor.showConsole(OPT.SHOW, 'multiplyByInteger après = ');
		this.updateExpression();
		return this;
	};

	addToPrefactor(ratExp) { // MathExpression
		this.prefactor = this.prefactor.add(ratExp);
		this.updateExpression();
		return this;
	};

	updateExpression() { // MathExpression
		// empty list if prefactor is zero
		if (this.prefactor.isZero() && this.list.length > 0) {
			this.list = [];
		};
		// nothing here
		// MathSum: distribute on terms such that global prefactor = 1
		// MathTensorProduct: collect all prefactor of the factors, expand if contains MathSum
		// MathSumArguments: distribute on terms such that global prefactor = 1
	};

	// return obj data of this.
	// go deep into MathElement and MathOperator
	// 	obj = {
	// 		id: indiceID,
	// 		elementID: id of the MathElement,
	// 		arrIdx: array to get position of indice in this,
	// 		symmetric: arrSym,
	// 		antisymmetric: arrAntiSym
	// 	}
	// notice that arrIdx contains as last element the 1 of the indice
	// all previous data in arrIdx give position of the MathElement/MathOperator
	collectIndicesArray() { // MathExpression
		// to get position, symmetries, kind…
		if (this.constructor.name === 'MathSum') {
			throw "collectIndicesArray does not apply to a MathSum object";
		};
		let arr = [];
		for(let i=0; i<this.list.length; i++) {
			let item = this.list[i];
			let arrItem = item.collectIndicesArray();
			for (let a of arrItem) {
				// add i at start to create array of position
				a.arrIdx.unshift(i);
				arr.push(a);
			};
		};
		if (this.constructor.name === 'MathArgument') {
			let arrPostItem = this.postElement.collectIndicesArray();
			for (let a of arrPostItem) {
				a.arrIdx.unshift(-1);
				arr.push(a);
			};
			let arrPreItem = this.preElement.collectIndicesArray();
			for (let a of arrPreItem) {
				a.arrIdx.unshift(-2);
				arr.push(a);
			};
		};
		return arr;
	};

	// return object with keys 'indiceID' and values 'obj data':
	// 	obj = {
	// 		id: indiceID,
	// 		elementID: id of the MathElement,
	// 		arrIdx: array to get position of element in this,
	// 		n: indice number,
	// 		symmetric: arrSym,
	// 		antisymmetric: arrAntiSym
	// 	}
	collectIndiceIDtoIndiceObj() { // MathExpression
		if (this.constructor.name === 'MathSum') {
			throw "collectIndiceIDtoIndiceObj does not apply to a MathSum object";
		};
		let arrIndices = this.collectIndicesArray();
		let indiceIDtoIndiceObj = {};
		for (let e of arrIndices) {
			e.n = e.arrIdx[e.arrIdx.length - 1];
			e.arrIdx.splice(-1,1);
			let indiceID = e.id + '-' + e.n;
			indiceIDtoIndiceObj[indiceID] = e;
		};
		return indiceIDtoIndiceObj;
	};


	// arrange all the indices in a predefined way
	// the algorithm is not simple…
	fillNormalizedIndicesArray() { // MathExpression
		// for a MathSum, apply to its terms
		if (this.constructor.name === 'MathSum') {
			for (let i=0; i<this.list.length; i++) {
				FH.printProgressMessage('Normalization: term ' + i + '/' + this.list.length);
				this.list[i].fillNormalizedIndicesArray();
			};
			FH.printProgressMessage('');
			return this;
		};

		// sort function for indice objects:
		// increasing pos, then increasing n
		function sortObjIndices(e1, e2) {
			if (e1.pos === e2.pos) {
				return e1.n - e2.n;
			} else {
				return e1.pos - e2.pos;
			};
		};

		// function to sort contraction objects:
		// increasing source pos
		// and then increasing target pos
		// => unique presentation for non commuting MathElement
		// => not a unique presentation for commuting MathElement (in preElement for instance)
		function sortObjCont(objCont1, objCont2) {
			// get sources obj of contractions
			let es1 = objCont1[0];
			let es2 = objCont2[0];
			if (es1.pos === es2.pos) {
				// same source position
				// test if both n in symmetric or antisymmetric
				if ( (es1.symmetric.includes(es1.n) && es2.symmetric.includes(es2.n))
							||
							(es1.antisymmetric.includes(es1.n) && es2.antisymmetric.includes(es2.n)) ) {
					// => increasing target pos
					let et1 = objCont1[1];
					let et2 = objCont2[1];
					return et1.pos - et2.pos;
				} else {
					// not in symmetric or antisymmetric
					// => increasing n
					return es1.n - es2.n;
				};
			} else {
				// not same pos
				// => increasing source pos
				return es1.pos - es2.pos;
			};
		};

		function normalizeSymmetricIndices(arrObjCont) {
			// console.log('(sym before) arrObjCont: ' + JSON.stringify(arrObjCont));
			let normalizationData = {};
			let objByIndiceID = {};
			// collect all the data from symmetric indices by indiceID
			for (let objCont of arrObjCont) {
				// console.log('(sym) objCont: ' + JSON.stringify(objCont));
				for (let e of objCont) {
					// console.log('(sym) e: ' + JSON.stringify(e));
					if (e.symmetric.length > 0) {
						// collect
						if (objByIndiceID[e.id] !== undefined) {
							objByIndiceID[e.id].push(e);
						} else {
							objByIndiceID[e.id] = [e];
						};
					}; // if length
				}; // for e
			}; // for objCont
			// for each indiceID, collect the e.n
			// sort the array of e.n and insert in objects in increasing order
			let keysID = Object.keys(objByIndiceID);
			for (let k of keysID) {
				let arrIndiceNumber = [];
				for (let e of objByIndiceID[k]) {
					arrIndiceNumber.push(e.n);
				};
				// console.log('arrIndiceNumber: ' + JSON.stringify(arrIndiceNumber));
				let arrIndiceNumberSorted = FH.copyData(arrIndiceNumber);
				arrIndiceNumberSorted.sort();
				// console.log('arrIndiceNumber (sorted): ' + JSON.stringify(arrIndiceNumber));
				normalizationData[k] = {};
				normalizationData[k].arrIndiceNumber = arrIndiceNumber;
				normalizationData[k].arrIndiceNumberSorted = arrIndiceNumberSorted;
				// normalizationData[k].sign = 1;
			};
			// console.log('(sym after) arrObjCont: ' + JSON.stringify(arrObjCont));
			return {sign: 1, data: normalizationData};
		};

		function normalizeAntisymmetricIndices(arrObjCont) {
			let normalizationData = {};
			let objByIndiceID = {};
			// collect all the data from antisymmetric indices by indiceID
			for (let objCont of arrObjCont) {
				for (let e of objCont) {
					// console.log('(norm antisym) e: ' + JSON.stringify(e));
					if (e.antisymmetric.length > 0) {
						// console.log('(norm antisym) collected')
						// collect
						if (objByIndiceID[e.id] !== undefined) {
							objByIndiceID[e.id].push(e);
						} else {
							objByIndiceID[e.id] = [e];
						};
					}; // if length
				}; // for e
			}; // for objCont
			// console.log('(norm antisym) objByIndiceID: ' + JSON.stringify(objByIndiceID));
			// for each indiceID, collect the e.n
			// => get the signature of the permutation which sorts this array
			// => collect all the signs to put in front of 'this'
			let keysID = Object.keys(objByIndiceID);
			// cumulated sign to apply in from of 'this'
			let sign = 1;
			for (let k of keysID) {
				let arrIndiceNumber = [];
				for (let e of objByIndiceID[k]) {
					arrIndiceNumber.push(e.n);
				};
				// get signature and sort arrIndiceNumber => arrIndiceNumberSorted
				let arrIndiceNumberSorted = FH.copyData(arrIndiceNumber);
				let s = FH.signaturePermutation(arrIndiceNumberSorted);
				// console.log('norm antisym sign: ' + s);
				sign = sign*s;
				normalizationData[k] = {};
				normalizationData[k].arrIndiceNumber = arrIndiceNumber;
				normalizationData[k].arrIndiceNumberSorted = arrIndiceNumberSorted;
				// normalizationData[k].sign = s;
			};
			return {sign: sign, data: normalizationData};
		};

		function updateArrObjContWithData(arrObjCont, normalizationData, indiceIDtoIndiceObj) {
			// console.log('(update) arrObjCont')
			// for (let arrCont of arrObjCont) {
			// 	console.log('arrCont: ' + JSON.stringify(arrCont));
			// }
			// let sign = 1;
			for (let objCont of arrObjCont) {
				for (let i=0; i<2; i++) {
					let e = objCont[i];
					let indiceIDRoot = e.id;
					if (normalizationData[indiceIDRoot] !== undefined) {
						// need treatement: change objCont[i] according to the permutation
						// arrIndiceNumber => arrIndiceNumberSorted
						let n = e.n;
						let arrIndiceNumber = normalizationData[indiceIDRoot].arrIndiceNumber;
						// console.log('(update) arrIndiceNumber: ' + JSON.stringify(arrIndiceNumber));
						// get current index of n in arrIndiceNumber
						let idx = arrIndiceNumber.indexOf(n);
						// test if this n is involved
						// (some 'n' of an element can be out of the symmetry/antisymmetry arrays)
						if (idx > -1) {
							// let s = normalizationData[indiceIDRoot].sign;
							// sign = sign*s;
							// console.log('permut sign: ' + s);
							let arrIndiceNumberSorted = normalizationData[indiceIDRoot].arrIndiceNumberSorted;
							// console.log('(update) arrIndiceNumberSorted: ' + JSON.stringify(arrIndiceNumberSorted));
							// console.log('(update) n: ' + n + ' => idx: ' + idx);
							// change indice object: use new n at idx in arrIndiceNumberSorted
							let newIndiceID = indiceIDRoot + '-' + arrIndiceNumberSorted[idx];
							// console.log('(update) newIndiceID: ' + newIndiceID);
							// update reference to this indice object
							if (indiceIDtoIndiceObj[newIndiceID] !== undefined) {
								objCont[i] = indiceIDtoIndiceObj[newIndiceID];
							} else {
								// console.log('newIndiceID: ' + newIndiceID + ' => no indice object');
								// console.log('indiceIDtoIndiceObj: ' + JSON.stringify(indiceIDtoIndiceObj));
								// console.log('arrObjCont: ' + JSON.stringify(arrObjCont));
								throw "Missing indice object";
							}; // test problem
						}; // if idx > -1
					};
				};
			};
			// console.log('sign: ' + sign);
			// return sign;
		};


		function insertPosInArrayIndices(arrIndices) {
			let maxIdx = 0;
			for (let e of arrIndices) {
				// find max index in all e.arrIdx
				for (let i of e.arrIdx) {
					if (i > maxIdx) {maxIdx = i;}
				};
			};
			// compute 'base'
			let base = Math.pow(10, Math.ceil(Math.log10(maxIdx + 1)));
			if (base === 1) {base = 10;};
			// console.log('base = ' + base);
			// insert pos parameter in objects of arrIndices
			for (let e of arrIndices) {
				e.pos = FH.createRealNumberFromIndicesArray(e.arrIdx, base);
			};
		}

		function getMinimumIndiceNumber(item, maxStart, indiceIDtoIndiceObj) {
			// start with given max
			// this value is returned if there is no assigned number to the indices of item
			let minNumber = maxStart;
			let indiceIDRoot = item.indices.id;
			// loop over all indices in this item
			for (let idx=0; idx<item.indices.list.length; idx++) {
				let indiceID = indiceIDRoot + '-' + idx;
				// console.log('indiceID: ' + indiceID);
				if (indiceIDtoIndiceObj[indiceID].kind !== undefined) {
					// a number has been assigned => compare to current minNumber
					let number = indiceIDtoIndiceObj[indiceID].number;
					if (number < minNumber) {
						minNumber = number;
					};
				};
			}; // for idx on item.indices.list
			// console.log('minNumber: ' + minNumber);
			return minNumber;
		};

		function sortPreElement(preElement, maxStart, indiceIDtoIndiceObj) {
			let arrChars = CONST.ARRAY_LETTERS_LONG;
			// append a letter to productStringWeight using minNumber
			for (let item of preElement.list) {
				item.productStringWeight = item.getProductStringWeight();
				let minNumber = getMinimumIndiceNumber(item, maxStart, indiceIDtoIndiceObj);
				item.productStringWeight += arrChars[minNumber];
			};
			preElement.list.sort(FH.sortProductStringWeight);
		};

		// DATA MODEL FOR THIS METHOD
		// the indices in MathExpression become object indices,
		// collected in arrIndices: [e1, e2, ...]
		// e1: {
		// 			arrIdx: [idx,...], (position of MathElement in MathExpression)
		// 			n: indice number in its MathElement,
		// 			id: id of indice,
		// 			elementID: id of MathElement,
		// 			symmetric: [...], (array of symmetric indices)
		// 			antisymmetric: [...], (array of antisymmetric indices)
		// 			pos: real number, (position in MathExpression)
		// 			kind: defined later, (how this indice is contracted)
		// 			number: defined later, (apparent indice number in MathExpression)
		// 		}
		// pos: used to create unique numerotation of indices
		// it is a real number contructed with 'base', a power of 10.
		// this defines the position of the MathElement from left to right in 'this',
		// the non integer part goes inside operators…
		// 'base' is a power of 10 computed with max value of idx in the arrIdx
		// it's done in function insertPosInArrayIndices
		// kind: used to determine if inner or outer

		// model for contractions:
		// arrObjCont: [..., [e1,e2], ...]
		// e1, e2: indice objects of contracted indices
		// [e1,e2] sorted according to position
		// arrObjCont sorted according to source pos and then target pos

		// indiceIDtoIndiceObj: it is a correspondance indiceID => its indice object
		// where indiceID is 'aaaa-0', contained in arrCont = this.contractions
		// this is used to get a fast access to the indice object

		// list of contrations of 'this': arrays of [indiceID1, indiceID2]
		let arrCont = FH.copyData(this.contractions);

		// array of indice objects
		let arrIndices = null;
		// object of indice objects
		let indiceIDtoIndiceObj = null;
		// array of pairs of contraction objects
		let arrObjCont = null;
		// the while loop should converge to a normalisation of position of indices
		// and elements in preElement
		// => return newArrObjCont and indiceIDtoIndiceObj
		let somethingFound = true;
		while (somethingFound) {
			// flag to false, change later if something found
			somethingFound = false;

			// collect informations about indices in object indices e1,...
			arrIndices = this.collectIndicesArray();
			// console.log('arrIndices: ' + JSON.stringify(arrIndices));
			// remove last index of arrIdx: it is the index of the indice in the MathElement
			// save this index in an extra property 'n'
			// construct indiceIDtoIndiceObj
			indiceIDtoIndiceObj = {};
			for (let e of arrIndices) {
				e.n = e.arrIdx[e.arrIdx.length - 1]; // collect last element
				e.arrIdx.splice(-1,1); // remove last element
				let indiceID = e.id + '-' + e.n;
				indiceIDtoIndiceObj[indiceID] = e;
			};

			// fill object indices with 'pos'
			insertPosInArrayIndices(arrIndices);
			// sort array of indice objects
			arrIndices.sort(sortObjIndices);

			// console.log('');
			// console.log('');
			// console.log('');
			// console.log('');
			// this.showConsole(OPT.SHOW, 'this: ');
			// console.log('indiceIDtoIndiceObj: ' + JSON.stringify(indiceIDtoIndiceObj));
			// console.log('');
			// console.log('arrCont: ' + JSON.stringify(arrCont));
			// console.log('');

			// model for contractions:
			// arrObjCont: [..., [e1,e2], ...]
			// e1, e2: indice objects of contracted indices
			// [e1,e2] sorted according to position
			// arrObjCont sorted according to source pos and then target pos
			arrObjCont = [];
			for (let c of arrCont) {
				let e1 = indiceIDtoIndiceObj[c[0]];
				let e2 = indiceIDtoIndiceObj[c[1]];
				// console.log('c: ' + JSON.stringify(c));
				// console.log('e1: ' + JSON.stringify(e1));
				// console.log('e2: ' + JSON.stringify(e2));
				// console.log('');
				// check consistency
				if (e1 === undefined) {
					console.log('(arrCont) problem in c: ' + c[0]);
					throw "Problem in contraction";
				};
				if (e2 === undefined) {
					console.log('(arrCont) problem in c: ' + c[1]);
					throw "Problem in contraction";
				};
				let objCont = [e1, e2].sort( (e1,e2) => {return e1.pos - e2.pos;} );
				arrObjCont.push(objCont);
			};
			arrObjCont.sort(sortObjCont);

			// console.log('');
			// console.log('arrObjCont: ' + JSON.stringify(arrObjCont));
			// console.log('');

			let normalizationData = null;
			let sign = null;
			let returnedObj = null;
			// take care of symmetric indices
			returnedObj = normalizeSymmetricIndices(arrObjCont);
			normalizationData = returnedObj.data;
			// console.log('(sym) normalizationData: ' + JSON.stringify(normalizationData));
			updateArrObjContWithData(arrObjCont, normalizationData, indiceIDtoIndiceObj);
			// if (sign !== 1) {
			// 	throw "sign should be 1";
			// };

			// take care of antisymmetric indices
			returnedObj = normalizeAntisymmetricIndices(arrObjCont);
			normalizationData = returnedObj.data;
			sign = returnedObj.sign;
			// console.log('(antisym) normalizationData: ' + JSON.stringify(normalizationData));
			updateArrObjContWithData(arrObjCont, normalizationData, indiceIDtoIndiceObj);
			this.multiplyByInteger(sign);
			this.updateExpression();

			// console.log('');
			// console.log('(after sym and antisym) arrObjCont: ' + JSON.stringify(arrObjCont));
			// console.log('');

			// fill contracted elements with data: kind, number
			// loop to construct inner contractions
			let innerIdx = 1;
			for (let objCont of arrObjCont) {
				// console.log('(loop inner) objCont: ' + JSON.stringify(objCont));
				// from left to right…
				// fill objects to normalize contractions
				for (let e of objCont) {
					// console.log('(loop inner) e: ' + JSON.stringify(e));
					if (e.kind === undefined) {
						e.kind = CONST.INDICE_INNER;
						e.number = innerIdx;
						// console.log('(loop inner has changed) e: ' + JSON.stringify(e));
					} else {
						throw "indice number already assigned";
					};
				};
				// increase indice number
				innerIdx++;
			};

			// test if the order of MathElement in preElement is normalized
			// normalized means:
			// the elements in preElement are grouped by their id (metric, inverse metric, tensors…)
			// one can commute elements inside these groups, but one can't mix elements from different groups
			// to order in each group: from left to right, the elements should be ordered along
			// the increasing of the minimum of their assigned indice numbers
			// one can get this minimum by the function getMinimumIndiceNumber
			if (this.constructor.name === 'MathArgument' && this.preElement.list.length > 1) {
				let arrPreElement = this.preElement.list;
				// console.log(arrPreElement[0].getChainedID());
				let currentElementID = arrPreElement[0].getChainedID();
				// let currentElementID = arrPreElement[0].id;
				// get minimal assigned indice number in arrPreElement[0]
				// start with max assigned number + 1
				let currentNumber = getMinimumIndiceNumber(arrPreElement[0], innerIdx + 1, indiceIDtoIndiceObj);
				// arrPreElement[0].showConsole(OPT.SHOW, '[0] (norm.) = ')
				// search in elements in the list if currentNumber is the smallest one
				for (let i=1; i<arrPreElement.length; i++) {
					// arrPreElement[i].showConsole(OPT.SHOW, '['+i+'] (norm.) = ');
					// console.log(arrPreElement[i].getChainedID());
					let testElementID = arrPreElement[i].getChainedID();
					// let testElementID = arrPreElement[i].id;
					let testNumber = getMinimumIndiceNumber(arrPreElement[i], innerIdx + 1, indiceIDtoIndiceObj);
					if (testElementID === currentElementID) {
						// test order of assigned indice number
						if (testNumber < currentNumber) {
							// not normalized => sort preElement according to sum weight + assigned indice numbers
							sortPreElement(this.preElement, innerIdx + 1, indiceIDtoIndiceObj);
							// this.showConsole(OPT.SHOW, 'this (norm.) = ');
							somethingFound = true;
							// quit the for loop on arrPreElement => return to while loop
							break;
						};
					} else {
						// we are now in another group => update ID of elements in this group
						currentElementID = testElementID;
						// initialize minimum number to search in this group
						currentNumber = testNumber;
					};
				}; // for on arrPreElement
			}; // if MathArgument and preElement contains more than one element
		}; // while loop

		// create the array of contraction for 'this'
		let newArrCont = [];
		for (let objCont of arrObjCont) {
			let e0 = objCont[0];
			let e1 = objCont[1];
			newArrCont.push([e0.id + '-' + e0.n, e1.id + '-' + e1.n]);
		};
		// console.log('newArrCont: ' + JSON.stringify(newArrCont));
		// update the array of contractions in 'this'
		this.contractions = newArrCont;

		// loop on arrIndices to fill with outer data indices
		// for all indices without kind attributed
		let outerIdx = 0;
		for (let e of arrIndices) {
			if (e.kind === undefined) {
				outerIdx++;
				e.kind = CONST.INDICE_OUTER;
				e.number = outerIdx;
			};
		};
		// console.log('outerIdx: ' + outerIdx);
		// console.log('end of normalisation, arrIndices: ' + JSON.stringify(arrIndices));
		// if (outerIdx%2 !== 0) {
		// 	throw 'Inconsistancy in outer indices';
		// };

		// TODO: loop over all MathElement/MathOperator in this and check order of indices
		// change according to an array of data obj:
		// - MathElement
		// - possibles permutations and associated sign
		//  => order the indices to get min on left according to possibles permutations

		// insert these data inside all the pyramid of objects
		this.propagateNormalizedIndicesArray(indiceIDtoIndiceObj);

		return this;
	};


	// add arrNormalizedIndices to every item down the pyramid from 'this'
	// so that MathElement and MathOperator know how to display their indices.
	propagateNormalizedIndicesArray(arrNormalizedIndices) { // MathExpression
		this.normalizedIndices = arrNormalizedIndices;
		for (let item of this.list) {
			item.propagateNormalizedIndicesArray(arrNormalizedIndices);
		};
		if (this.constructor.name === 'MathArgument') {
			this.preElement.propagateNormalizedIndicesArray(arrNormalizedIndices);
			this.postElement.propagateNormalizedIndicesArray(arrNormalizedIndices);
		};
		return this;
	}


	fillSumStringWeight() { // MathExpression
		switch (this.constructor.name) {
			case 'MathProduct':
			case 'MathTensorProduct':
			case 'MathArgument':
				this.sumStringWeight = this.getSumStringWeight();
				// console.log('this.sumStringWeight: ' + this.sumStringWeight);
				break;
			case 'MathSum':
				for (let i=0; i<this.list.length; i++) {
 					this.list[i].fillSumStringWeight();
				};
				break;
			default:
				break;
		};
		return this;
	};

	fillStringRepresentation() { // MathExpression
		switch (this.constructor.name) {
			case 'MathProduct':
			case 'MathTensorProduct':
				this.stringRepresentation = this.getStringRepresentation();
				break;
			case 'MathArgument':
				this.stringRepresentation = this.getStringRepresentation();
				this.preElement.stringRepresentation = this.preElement.getStringRepresentation();
				break;
			case 'MathSum':
				for (let i=0; i<this.list.length; i++) {
 					this.list[i].fillStringRepresentation();
				};
				break;
			default:
				break;
		};
		return this;
	};

	getStringRepresentation(displaySymbols) { // MathExpression
		let ds = null;
		let useInternalDisplaySymbols = true;
		if (displaySymbols !== undefined
						&& displaySymbols.useInternalDisplaySymbols === false) {
			ds = displaySymbols;
			useInternalDisplaySymbols = false;
		} else {
			ds = this.display_symbols;
		};

		if (this.constructor.name === 'MathSum') {
			throw "getStringRepresentation does not applied to a MathSum object";
		};

		// list
		let listTxt = '';
		if (this.list.length > 0) {
			listTxt += ds.symbol_open_list;
			let txtArray = [];
			for (let i=0; i<this.list.length; i++) {
				txtArray.push(ds.symbol_open_item
												+ this.list[i].getStringRepresentation(ds)
												+ ds.symbol_end_item);
			};
			listTxt += txtArray.join(ds.symbol_operation);
			listTxt += ds.symbol_end_list;
		};

		// post element for MathArgument
		// pre element is ignored to compare MathArgument
		let postTxt = '';
		if (this.constructor.name === 'MathArgument') {
			postTxt += CONST.SYMBOL_POST_ELEMENT;
			postTxt += this.postElement.getStringRepresentation(ds);
		};
		let str = listTxt + postTxt;

		return str;
	};

	getStringArrayRepresentation() { // MathExpression

		let ds = {
				useInternalDisplaySymbols: false,
				symbol_prefactor_one: CONST.SYMBOL_ONE,
				symbol_prefactor_zero: CONST.SYMBOL_ZERO,
				b_prefactor_non_empty_list: CONST.SYMBOL_PRODUCT,
				b_prefactor_one_non_empty_list: CONST.SYMBOL_PRODUCT,
				b_prefactor_zero_non_empty_list: CONST.SYMBOL_PRODUCT,
				b_prefactor_empty_list: CONST.SYMBOL_PRODUCT,
				b_prefactor_one_empty_list: CONST.SYMBOL_PRODUCT,
				b_prefactor_zero_empty_list: CONST.SYMBOL_PRODUCT,
				symbol_open_list: '',
				symbol_end_list: '',
				symbol_empty_list: '',
				symbol_open_item: '',
				symbol_end_item: '',
				symbol_operation: '',
				symbol_open_item_number: '',
				symbol_end_item_number: ''
		};


		if (this.constructor.name === 'MathSum') {
			throw "getStringArrayRepresentation() does not applied to a MathSum object";
		};

		let txtArray = [];
		for (let i=0; i<this.list.length; i++) {
				txtArray.push(this.list[i].getStringRepresentation(ds));
		};
		return txtArray;
	};

	getReducedStringRepresentationCountingElementID(elID) { // MathExpression
		let ds = this.display_symbols;

		if (this.constructor.name === 'MathSum') {
			throw "getReducedStringRepresentationCountingElementID does not applied to a MathSum object";
		};

		// this.showConsole(OPT.SHOW, 'this: ');
		let counter = 0;
		// list
		let listTxt = '';
		if (this.list.length > 0) {
			listTxt += ds.symbol_open_list;
			let txtArray = [];
			for (let i=0; i<this.list.length; i++) {
				let {counter: c, string: str} = this.list[i].getReducedStringRepresentationCountingElementID(elID);
				if (str !== '') {
					txtArray.push(ds.symbol_open_item + str + ds.symbol_end_item);
				};
				counter = counter + c;
			};
			listTxt += txtArray.join(ds.symbol_operation);
			listTxt += ds.symbol_end_list;
		};

		// post element for MathArgument
		// pre element for MathArgument
		let postTxt = '';
		let preTxt = '';
		if (this.constructor.name === 'MathArgument') {
			let {counter: postc, string: postStr} = this.postElement.getReducedStringRepresentationCountingElementID(elID);
			postTxt += postStr;
			let {counter: prec, string: preStr} = this.preElement.getReducedStringRepresentationCountingElementID(elID);
			preTxt += preStr;
			// counter is not used in postElement and preElement
			return {counter: counter, string: listTxt, preString: preTxt, postString: postTxt};
		};

		return {counter: counter, string: listTxt};
	};

	// create a MathSum based on 'this' in which
	// the original item at arrIdx is replaced by items in arrItemsData
	// if replacement is not available, return null
	// used in substituteItemAndDerivativesWithData
	// the level of derivatives is obtained by the idx of the item to find
	// => no check of what is this derivation
	constructNewSumWithSubstitutions(arrIdx, newItemsByDerivatives) { // MathExpression
		// find idx of orgItem in its MathProduct in 'this'
		// it is the last idx in arrIdx which is >= 0
		// negative idx at the end corresponds to number of applied derivations
		let orgItemIdx = arrIdx[arrIdx.length - 1];
		// console.log('orgItemIdx = ' + orgItemIdx);
		let nDerivations = 0;
		if (orgItemIdx < 0) {
			// inside operator
			// keep number of applied derivations
			nDerivations = -(orgItemIdx + 1);
			 // take previous idx in arrIdx to get container
			orgItemIdx = arrIdx[arrIdx.length - 2];
		};

		// if no given substitution for this level of derivations,
		// return null
		if (newItemsByDerivatives[nDerivations] === undefined) {
			console.log('newItemsByDerivatives[nDerivations] is undefined');
			return null;
		};

		// the new MathSum that will be returned
		let newSum = new MathSum();

		// console.log('nDerivations: ' + nDerivations);
		// console.log('newItemsByDerivatives[nDerivations]: ' + JSON.stringify(newItemsByDerivatives[nDerivations]));

		// loop over MathProduct in newItemsByDerivatives[nDerivations] to replace orgItem
		for (let newItemData of newItemsByDerivatives[nDerivations]) {
			let newItem = newItemData.item;
			// console.log('');
			// console.log('newItem: ' + JSON.stringify(newItem));
			// no contribution to the sum if new item is 'null'
			if (newItem === null) {
				// console.log('null new item');
				continue;
			};
			let mapIndices = newItemData.mapIndices;
			let permutationsSum = newItemData.permutationsSum;
			// console.log('');
			// console.log('mapIndices: ' + JSON.stringify(mapIndices));
			// console.log('permutationsSum: ' + JSON.stringify(permutationsSum));

			// don't modify permutationsSum, make a copy
			let permutationsSumCopy = FH.copyData(permutationsSum);
			if (mapIndices.length === 0) {
				if (permutationsSum.length > 0) {
					throw "permutationsSum should be empty";
				} else {
					// add a "perm" to get into the next loop
					// this perm will not be used since mapIndices is empty
					// important to work with a copy of permutationsSum here
					permutationsSumCopy.push([0]);
				};
			};
			// loop over permutations
			for (let perm of permutationsSumCopy) {
				// console.log('');
				// console.log('perm: ' + JSON.stringify(perm));
				// copy newItem
				let newItemCopy = newItem.makeCopy();
				// copy this
				let thisCopy = this.makeCopy();

				// copy arrIdx since it will be manipulated
				let arrIdxCopy = FH.copyData(arrIdx);

				// find item in thisCopy in which to replace at orgItemIdx
				// this is a MathProduct
				// remove last idx from arrIdxCopy until found
				let replaceItemContainer = thisCopy.getItemAt(arrIdxCopy);
				// console.log('arrIdxCopy: ' + JSON.stringify(arrIdxCopy));
				while (replaceItemContainer.constructor.name !== 'MathProduct') {
					arrIdxCopy.splice(-1,1);
					if (arrIdxCopy.length === 0) {
						replaceItemContainer = thisCopy;
					} else {
						replaceItemContainer = thisCopy.getItemAt(arrIdxCopy);
					};
				};

				// get contractions of 'thisCopy'
				let currentArrCont = thisCopy.contractions;
				let srtCurrentArrCont = JSON.stringify(currentArrCont);
				// console.log('currentArrCont: ' + JSON.stringify(currentArrCont));
				// update srtCurrentArrCont
				let replaceItem = replaceItemContainer.getItemAtIndex(orgItemIdx);
				// replaceItemContainer.showConsole(OPT.SHOW, 'before = ');
				for (let i=0; i<mapIndices.length; i++) {
					// when mapIndices is empty, this loop is bypassed
					// this is where perm is used
					// console.log('in mapIndices loop, i=' + i);
					// from
					let arrIndiceIdx = mapIndices[i].from;
					let srtIndiceIDFrom = replaceItem.getTextIndiceAt(arrIndiceIdx);
					// to (using perm)
					arrIndiceIdx = mapIndices[perm[i]].to;
					let srtIndiceIDTo = newItemCopy.getTextIndiceAt(arrIndiceIdx);
					let re = new RegExp(srtIndiceIDFrom, "g");
					srtCurrentArrCont = srtCurrentArrCont.replace(re, srtIndiceIDTo);
				};
				// replace in 'thisCopy'
				let retInfo = {};
				replaceItemContainer.replaceByItemAtIndex(newItemCopy, orgItemIdx, retInfo, false);

				// replaceItemContainer.showConsole(OPT.SHOW, 'after = ');
				// newItemCopy.showConsole(OPT.SHOW, 'after = ');
				// console.log('currentArrCont: ' + JSON.stringify(currentArrCont));

				// add contraction array
				let arrCont = FH.joinContractions(JSON.parse(srtCurrentArrCont), newItemCopy.contractions);
				// console.log('newItemCopy.contractions: ' + JSON.stringify(newItemCopy.contractions));
				// console.log('arrCont: ' + JSON.stringify(arrCont));
				for (let e of arrCont) {e.sort(FH.alphaSort);};
				thisCopy.contractions = arrCont;

				// thisCopy.showConsole(OPT.SHOW, 'new term: ');
				thisCopy.checkContractionArray();

				// add to the collecting sum
				newSum.addTo(thisCopy, false);
			}; // loop on permutationsSum
		}; // loop on newItems

		return newSum;
	};

	getMathematica(opt) { // MathExpression

		let returnedTxt = '';
		let listTxt = '';
		// collect arguments of 'Times' for non MathSum
		let timesArray = [];

		// main prefactor
		if (!this.prefactor.isOne()) {
			timesArray.push(this.prefactor.getMathematica(opt));
		};

		// main list => collect in arrays
		let listTxtArray = [];
		for (let i=0; i<this.list.length; i++) {
			listTxtArray.push(this.list[i].getMathematica(opt));
		};

		switch (this.constructor.name) {
			case 'MathProduct':
				// everything will be in 'Times'
				if (!this.isOne()) {
					timesArray = timesArray.concat(listTxtArray);
					returnedTxt = CONST.MATH_TIMES
												+ CONST.MATH_BRACKET_OPEN
													+ timesArray.join(', ')
												+ CONST.MATH_BRACKET_CLOSE;
					// returnedTxt = CONST.MATH_TIMES(timesArray.join(', '));
				} else {
					// test now what is the prefactor
					if (timesArray.length>0) {
						// prefactor contains info
						returnedTxt = timesArray[0];
					} else {
						// prefactor is 1 (and product is 1 also)
						returnedTxt = CONST.MATH_ONE;
					};
				};
				// console.log('MathProduct');
				// this.showConsole(OPT.SHOW, 'this: ');
				// console.log(returnedTxt);
				// console.log('');
				break;

			case 'MathTensorProduct':
				listTxt = CONST.MATH_TENSOR_PRODUCT
										+ CONST.MATH_BRACKET_OPEN
											+ listTxtArray.join(', ')
										+ CONST.MATH_BRACKET_CLOSE;
				if (timesArray.length>0) {
					// only 1 element, the prefactor
					returnedTxt = CONST.MATH_TIMES
												+ CONST.MATH_BRACKET_OPEN
													+ timesArray[0]
												+ ', '
													+ listTxt
												+ CONST.MATH_BRACKET_CLOSE;
					// returnedTxt = CONST.MATH_TIMES(timesArray[0] + ', ' + listTxt);
				} else {
					returnedTxt = listTxt;
				};
				// console.log('MathTensorProduct');
				// this.showConsole(OPT.SHOW, 'this: ');
				// console.log(returnedTxt);
				// console.log('');
				break;

			case 'MathArgument':
				// prefactor is prefactor of preElement
				let preElementTxt = this.preElement.getMathematica(opt);
				// list: (as in MathTensorProduct)
				listTxt = CONST.MATH_TENSOR_PRODUCT
										+ CONST.MATH_BRACKET_OPEN
											+ listTxtArray.join(', ')
										+ CONST.MATH_BRACKET_CLOSE;

				// encapsulate in spectral operator
				// TODO…

				// postElement
				// TODO…

				if (preElementTxt !== CONST.MATH_ONE) {
					returnedTxt = CONST.MATH_TIMES
												+ CONST.MATH_BRACKET_OPEN
													+ preElementTxt
												+ ', '
													+ listTxt
												+ CONST.MATH_BRACKET_CLOSE;
						// returnedTxt = CONST.MATH_TIMES(preElementTxt + ', ' + listTxt);
				} else {
					returnedTxt = listTxt;
				};
				// console.log('MathArgument');
				// this.showConsole(OPT.SHOW, 'this: ');
				// console.log('preElementTxt: ' + preElementTxt);
				// console.log(returnedTxt);
				// console.log('');
				break;

			case 'MathSum':
				if (this.list.length === 0) {
					returnedTxt = CONST.MATH_ZERO;
				} else {
					let joinList = ',' + CONST.CR;
					returnedTxt = CONST.MATH_PLUS
													+ CONST.MATH_BRACKET_OPEN
													+ CONST.CR
														+ listTxtArray.join(joinList)
													+ CONST.CR
													+ CONST.MATH_BRACKET_CLOSE;
				}; // sum = 0 or not
				break;

			default:
				throw "MathProduct, MathTensorProduct, MathArgument, or MathSum required";
				break;
		};

		return returnedTxt;
	};

	getLaTeX(opt) { // MathExpression

		// main prefactor
		let prefactorObj = this.prefactor.getLaTeX(opt);
		let sign = prefactorObj.sign;
		let prefactorTxt = prefactorObj.ltx;

		// main list => collect in arrays
		let txtArray = [];
		let signArray = [];
		for (let i=0; i<this.list.length; i++) {
			let itemObj = this.list[i].getLaTeX(opt);
			txtArray.push(itemObj.ltx);
			signArray.push(itemObj.sign)
		};

		let returnedTxt = '';
		switch (this.constructor.name) {
			case 'MathProduct':
				// prefactor is main prefactor => nothing to do
				// console.log('(MathProduct) prefactorTxt: ' + prefactorTxt);
				// list:
				// if (this.isOne()) {
				// 	returnedTxt = ''
				// } else {
				// 	returnedTxt = txtArray.join(CONST.LATEX_PRODUCT);
				// };
				returnedTxt = txtArray.join(CONST.LATEX_PRODUCT);
				// check: MathElement/MathOperator return sign=1
				for (let s of signArray) {
					if (s<0) {
						throw "sign should be 1";
					};
				};
				if (prefactorTxt === '1') {
					prefactorTxt = '';
				};
				if (prefactorTxt !== '' && returnedTxt !== '') {
					prefactorTxt += CONST.LATEX_THIN_SPACE;
				};
				returnedTxt = prefactorTxt + returnedTxt;
				break;

			case 'MathTensorProduct':
				// prefactor is main prefactor => nothing to do
				// console.log('(MathTensorProduct) prefactorTxt: ' + prefactorTxt);
				// list:
				returnedTxt = txtArray.join(CONST.LATEX_TENSOR_PRODUCT);
				// check: normalization of MathTensorProduct
				for (let s of signArray) {
					if (s<0) {
						throw "sign should be 1";
					};
				};
				if (prefactorTxt !== '' && returnedTxt !== '') {
					prefactorTxt += CONST.LATEX_THIN_SPACE;
				};
				returnedTxt = prefactorTxt + returnedTxt;
				break;

			case 'MathArgument':
				// check: main prefactor is 1 by normalization
				// console.log('(MathArgument) prefactorTxt: ' + prefactorTxt);
				if (sign<0) {
					throw "sign should be 1";
				}
				// prefactor is prefactor of preElement
				// => get preElement
				let preElementObj = this.preElement.getLaTeX(opt);
				sign = preElementObj.sign;
				let preElementTxt = preElementObj.ltx;
				// trick to remove '\bbbone' in pre element: regexp replacement
				// better solution?
				// console.log("preElementTxt (before): " + preElementTxt);
				let regbbbone = new RegExp('\\' + CONST.LATEX_DEFAULT_LETTER, "g");
				// console.log("regbbbone: " + regbbbone);
				preElementTxt = preElementTxt.replace(regbbbone, '');
				// console.log("preElementTxt (after): " + preElementTxt);
				// list:
				let listTxt = txtArray.join(CONST.LATEX_TENSOR_PRODUCT);
				// encapsulate in spectral operator
				if (listTxt !== '') {
					let k = this.list.length;
					listTxt = CONST.LATEX_SPECTRAL_OPERATOR_LETTER
										+ CONST.LATEX_SUB + '{' + k + '}'
										+ CONST.LATEX_SPECTRAL_OPERATOR_OPEN
										+ listTxt
										+ CONST.LATEX_SPECTRAL_OPERATOR_CLOSE;
				};
				// check: normalization as in MathTensorProduct
				for (let s of signArray) {
					if (s<0) {
						throw "sign should be 1";
					};
				};
				// postElement
				let postElementTxt = '';
				if (!this.postElement.isOne()) {
					let postElementObj = this.postElement.getLaTeX(opt);
					// check
					if (postElementObj.sign<0) {
						throw "sign should be 1";
					};
					postElementTxt = postElementObj.ltx;
					if (postElementTxt === '1') {
						postElementTxt = '';
					};
					if (postElementTxt !== '') {
						postElementTxt = CONST.LATEX_THIN_SPACE + postElementTxt;
					};
				};
				// space between pre and main
				if (preElementTxt !== '' && listTxt !== '') {
					preElementTxt += CONST.LATEX_THIN_SPACE;
				};
				returnedTxt = preElementTxt + listTxt + postElementTxt;
				break;

			case 'MathSum':
				if (this.list.length === 0) {
					returnedTxt = '0';
				} else {
					// console.log('(MathSum) prefactorTxt: ' + prefactorTxt);
					if (sign<0) {
						throw "sign should be 1";
					};
					// end of line in align '\\'
					// + newline in text
					// + '&' for align
					let latexJoinList = CONST.LATEX_SUM_NEWLINE
														+ CONST.CR
														+ CONST.LATEX_ALIGN_SEP;
					for (let i=0; i<txtArray.length; i++) {
						let signTxt = '';
						if (signArray[i]<0) {
							signTxt = '-';
						} else if(i>0) {
							// no sign '+' for first item
							signTxt = '+';
						};
						returnedTxt += signTxt + txtArray[i];
						if (opt.showLineNumbers) {
							let nLine = i+1;
							returnedTxt += CONST.LATEX_LINE_NUMBER + '{' + nLine + '} ';
						};
						if (i !== txtArray.length-1) {
							returnedTxt += latexJoinList;
						};
					};
				}; // sum = 0 or not
				// '{} &' before expression
				returnedTxt = '{}' + CONST.LATEX_ALIGN_SEP + returnedTxt;
				break;

			default:
				throw "MathProduct, MathTensorProduct, MathArgument, or MathSum required";
				break;
		};

		// console.log('(MathExpression) ltx: ' + returnedTxt);
		// console.log('(MathExpression) sign: ' + '1');
		return {ltx: returnedTxt, sign: sign};
	};

	getLaTeXEquation(opt, text = '') { // MathExpression
		let latexObj = this.getLaTeX(opt);
		let latex = latexObj.ltx;
		let sign = latexObj.sign;
		let signTxt = '';
		if (sign<0) {
			signTxt = '-';
		};
		return CONST.LATEX_BEGIN_EQUATION
						+ CONST.CR
						+ text
						+ signTxt
						+ latex
						+ CONST.CR
						+ CONST.LATEX_END_EQUATION;
	};

	getLaTeXInEquation(opt, text = '') { // MathExpression
		let latexObj = this.getLaTeX(opt);
		let latex = latexObj.ltx;
		let sign = latexObj.sign;
		let signTxt = '';
		if (sign<0) {
			signTxt = '-';
		};
		return text
						+ signTxt
						+ latex;
	};

	displayText(opt, arrIdx) { // MathExpression
		// console.log(this.constructor.name + ' => showItemNumber:' + opt.showItemNumber);
		let ds = this.display_symbols;
		let od = opt.display;
		let newOpt = FH.duplicateInstance(opt); // functions inside, no FH.copyData
		if (this.constructor.name !== 'MathSum') {
			// console.log('define new optShow');
			newOpt.type = false;
			newOpt.contractions = false;
		};
		if (this.constructor.name === 'MathSum') {
			newOpt.showItemNumber = false;
		};
		// after prefactor: afterPrefactorLeftTxt + afterPrefactorRightTxt
		let afterPrefactorLeftTxt = ''; // part affected by prefactor
		let afterPrefactorRightTxt = ''; // part affected by list
		// prefactor
		let prefactorTxt = '';
		if (this.prefactor.isNegative()) {
			prefactorTxt += '(' + this.prefactor.displayText(newOpt) + ')';
			afterPrefactorLeftTxt = '%B_PREFACTOR_';
		} else if (this.prefactor.isOne()) {
			prefactorTxt += '%SYMBOL_PREFACTOR_ONE%';
			afterPrefactorLeftTxt = '%B_PREFACTOR_ONE_';
		} else if (this.prefactor.isZero()) {
			prefactorTxt += '%SYMBOL_PREFACTOR_ZERO%';
			afterPrefactorLeftTxt = '%B_PREFACTOR_ZERO_';
		} else {
			prefactorTxt += this.prefactor.displayText(newOpt);
			afterPrefactorLeftTxt = '%B_PREFACTOR_';
		};
		prefactorTxt = od.formatPrefactor(prefactorTxt);
		// list
		let listTxt = '';
		if (this.list.length > 0) {
			let txtArray = [];
			for (let i=0; i<this.list.length; i++) {
				// send info on current index i in displayText()
				let newArrIdx = arrIdx.concat([i]);
				// opt.type = false in sub elements
				let itemTxt = '%SYMBOL_BEGIN_ITEM%'
													+ this.list[i].displayText(newOpt, newArrIdx)
													+ '%SYMBOL_END_ITEM%';
				if (opt.showItemNumber) {
					let nLine = i+1;
					itemTxt += '%SYMBOL_BEGIN_ITEM_NUMBER%' + nLine + '%SYMBOL_END_ITEM_NUMBER%';
				};
				txtArray.push(itemTxt);
				// txtArray.push('%SYMBOL_BEGIN_ITEM%'
				// 												+ this.list[i].displayText(newOpt, newArrIdx)
				// 												+ '%SYMBOL_END_ITEM%');
			};
			listTxt += '%SYMBOL_BEGIN_LIST%';
			listTxt += txtArray.join('%SYMBOL_JOIN_LIST%');
			listTxt += '%SYMBOL_END_LIST%';
			afterPrefactorRightTxt = 'NON_EMPTY_LIST%';
		} else {
			listTxt += '%SYMBOL_EMPTY_LIST%';
			afterPrefactorRightTxt = 'EMPTY_LIST%';
		};
		listTxt = od.formatList(listTxt);
		// class name
		let classTxt = '';
		if (opt.type) {
			classTxt = CONST.SYMBOL_OPEN_CLASS
								+ this.constructor.name
								+ CONST.SYMBOL_CLOSE_CLASS;
		};
		classTxt = od.formatClass(classTxt);
		// contractions of indices
		// not shown for MathSum
		let contractionsTxt = '';
		if (opt.contractions) {
			if (this.constructor.name !== 'MathSum' || opt.sumContractions) {
				contractionsTxt = '%B_CONTRACTIONS%' + FH.formatContractionDisplay(this.contractions);
			};
		};
		contractionsTxt = od.formatContractions(contractionsTxt);
		// pre element for MathArgument
		let preTxt ='';
		if (this.constructor.name === 'MathArgument') {
			let newArrIdx = arrIdx.concat([-2]);
			preTxt += this.preElement.displayText(newOpt, newArrIdx);
			preTxt += CONST.SYMBOL_PRE_ELEMENT;
			// preTxt = preTxt.replace(/%B_SYMBOL_SUM%/g, "");
		};
		preTxt = od.formatPreElement(preTxt);
		// globalTxt = globalTxt.replace(/%B_SYMBOL_SUM%/g, "\n" + " ".repeat(l));
		// post element for MathArgument
		let postTxt = '';
		if (this.constructor.name === 'MathArgument') {
//			let newArrIdx = [-1].concat(arrIdx);
			let newArrIdx = arrIdx.concat([-1]);
			postTxt += CONST.SYMBOL_POST_ELEMENT;
			postTxt += this.postElement.displayText(newOpt, newArrIdx);
		};
		postTxt = od.formatPostElement(postTxt);

		let globalTxt = classTxt
									+ prefactorTxt
									+ afterPrefactorLeftTxt
									+ afterPrefactorRightTxt
									+ preTxt
									+ listTxt
									+ postTxt
									+ contractionsTxt;
		globalTxt = globalTxt
					.replace(/%SYMBOL_PREFACTOR_ONE%/g, ds.symbol_prefactor_one)
					.replace(/%SYMBOL_PREFACTOR_ZERO%/g, ds.symbol_prefactor_zero)
					.replace(/%B_PREFACTOR_NON_EMPTY_LIST%/g, ds.b_prefactor_non_empty_list)
					.replace(/%B_PREFACTOR_ONE_NON_EMPTY_LIST%/g, ds.b_prefactor_one_non_empty_list)
					.replace(/%B_PREFACTOR_ZERO_NON_EMPTY_LIST%/g, ds.b_prefactor_zero_non_empty_list)
					.replace(/%B_PREFACTOR_EMPTY_LIST%/g, ds.b_prefactor_empty_list)
					.replace(/%B_PREFACTOR_ONE_EMPTY_LIST%/g, ds.b_prefactor_one_empty_list)
					.replace(/%B_PREFACTOR_ZERO_EMPTY_LIST%/g, ds.b_prefactor_zero_empty_list)
					.replace(/%SYMBOL_BEGIN_LIST%/g, ds.symbol_open_list)
					.replace(/%SYMBOL_END_LIST%/g, ds.symbol_end_list)
					.replace(/%SYMBOL_EMPTY_LIST%/g, ds.symbol_empty_list)
					.replace(/%SYMBOL_BEGIN_ITEM%/g, ds.symbol_open_item)
					.replace(/%SYMBOL_END_ITEM%/g, ds.symbol_end_item)
					.replace(/%SYMBOL_BEGIN_ITEM_NUMBER%/g, ds.symbol_open_item_number)
					.replace(/%SYMBOL_END_ITEM_NUMBER%/g, ds.symbol_end_item_number)
					.replace(/%SYMBOL_JOIN_LIST%/g, ds.symbol_operation);
		if (this.constructor.name === 'MathSum') {
			globalTxt = globalTxt.replace(/%B_CONTRACTIONS%/g, "%B_SYMBOL_SUM%");
		} else {
			globalTxt = globalTxt.replace(/%B_CONTRACTIONS%/g, '');
		}

		return globalTxt;
	};

	showConsole(opt, text = '') { // MathExpression
		let newOpt = FH.duplicateInstance(opt); // functions inside, no FH.copyData
		newOpt.normalized = false;

		let l = text.length;
		let globalTxt = newOpt.display.formatText(text) + this.displayText(newOpt, []);
		globalTxt = globalTxt.replace(/%B_SYMBOL_SUM%/g, "\n" + " ".repeat(l));

		console.log(globalTxt);

		return this;
	};

	showNormalized(opt, text = '') { // MathExpression
		// console.log('normalized: ' + this.constructor.name + ' => showItemNumber:' + opt.showItemNumber);
		let newOpt = FH.duplicateInstance(opt); // functions inside, do not use FH.copyData
		newOpt.type = false;
		newOpt.contractions = false;
		newOpt.mapIndices = false;
		newOpt.position = false;
		newOpt.sumContractions = false;

		newOpt.normalized = true;

		let l = text.length;
		let globalTxt = newOpt.display.formatText(text) + this.displayText(newOpt, []);
		globalTxt = globalTxt.replace(/%B_SYMBOL_SUM%/g, "\n" + " ".repeat(l));

		console.log(globalTxt);

		return this;
	};

	toObject() { // MathExpression
		let listObjects = [];
		for (let i=0; i<this.list.length; i++) {
			listObjects.push(this.list[i].toObject());
		};
		return {
			// class constructor
			class: this.constructor.name,
			// prefactor
			pf: this.prefactor.toObject(),
			// main list
			list: listObjects,
			// contractions array
			cont: FH.copyData(this.contractions),
		};
	};

	toJSON(...args) { // MathExpression
		return JSON.stringify(this.toObject(), ...args);
	};

	isZero() { // MathExpression
		return this.prefactor.isZero();
	};


};



// product of MathElement with common rational prefactor
class MathProduct extends MathExpression {
	constructor() { // MathProduct
		super();

		// the symbols to use for display
		this.display_symbols = DISPLAY_SYMBOLS.MATHPRODUCT;

	};

	// return a MathProduct from
	// MathElement, MathOperator, or MathProduct
	// make a copy of original item
	static newFrom(item) { // MathProduct
		switch (item.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
				let newME = new MathProduct();
				let newItem = item.makeCopy();
				newME.list.push(newItem);
				newME.contractions = FH.joinContractions(newME.contractions, newItem.contractions);
				newME.updateExpression();
				return newME;
				break;
			case 'MathProduct':
				return item.makeCopy();
				break;
			default:
				throw "MathElement, MathOperator, or MathProduct required";
				breaks;
		};
	};


	updateExpression() { // MathProduct
		super.updateExpression();

		// should contain at least the MathElement_one MathElement
		if (this.list.length === 0) {
			this.list.push(MathElement_one.makeCopy());
		};

		// remove duplicated MathElement_one
		let arrOne = [];
		for (let i=0; i<this.list.length; i++) {
			if (this.list[i].id === MathElement_one.id) {
				arrOne.push(i);
			};
		};
		// console.log('updateExpression MathProduct, arrOne: ' + JSON.stringify(arrOne));
		if (arrOne.length > 0) {
			// remove all MathElement_one except the first one
			for (let i=arrOne.length-1; i>0; i--) {
				this.list.splice(arrOne[i], 1);
			};
			if (this.list.length > 1) {
				// we can remove the first MathElement_one
				this.list.splice(arrOne[0], 1);
			};
		};
		// order this.list according to commutative and product_weight
		// do not cross MathOperator from right to left
		// TODO
		return this;
	};

	sortByProductWeight() { // MathProduct
		this.fillProductStringWeight();
		let arrComm = this.getCommutativeElements();
		// remove negative idx, since consider all the 'block'
		// here, since in MathProduct, arrays are of length 2
		// so element at 1 is to consider
		for (let a of arrComm) {
			if (a[1] < 0) {
				a.splice(1,1);
			};
		};
		// console.log('arrComm: ' + JSON.stringify(arrComm));
		// move commutative elements in front
		// collect them in array to sort
		let arrItemComm = [];
		for (let a of arrComm) {
			let idx = a[0];
			let item = this.getItemAtIndex(idx);
			this.list.splice(idx, 1);
			this.list.unshift(item);
			arrItemComm.push(item);
		};
		// for(let item of arrItemComm) {
		// 	item.showConsole(OPT.SHOW, 'item: ');
		// }
		arrItemComm.sort(FH.sortProductStringWeight);
		// console.log('');
		// for(let item of arrItemComm) {
		// 	item.showConsole(OPT.SHOW, 'item: ');
		// }
		for (let i=0; i<arrItemComm.length; i++) {
			this.list[i] = arrItemComm[i];
		};
		return this;
	};

	fillProductStringWeight() { // MathProduct
		for (let item of this.list) {
			item.productStringWeight = item.getProductStringWeight();
		}
		return this;
	};


	multiplyOnRightWith(item, copy=true) { // MathProduct
		// item: MathElement, MathOperator, MathProduct
		let newItem = null;
		if (copy) {
			newItem = item.makeCopy();
		} else {
			newItem = item;
		};
		switch (newItem.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
				this.list.push(newItem);
				break;
			case 'MathProduct':
				// this.prefactor = this.prefactor.multiplyBy(newItem.prefactor);
				this.prefactor = this.prefactor.multiply(newItem.prefactor);
				for (let i=0; i < newItem.list.length; i++) {
					this.list.push(newItem.list[i]);
				};
				break;
			default:
				throw "MathElement, MathOperator, or MathProduct required";
				break;
		};
		this.contractions = FH.joinContractions(this.contractions, newItem.contractions);
		this.updateExpression();
		return this;
	};

	multiplyOnLeftWith(item, copy=true) { // MathProduct
		// item: MathElement, MathOperator, MathProduct
		let newItem = null;
		if (copy) {
			newItem = item.makeCopy();
		} else {
			newItem = item;
		};
		switch (newItem.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
				this.list.unshift(newItem);
				break;
			case 'MathProduct':
				// this.prefactor.multiplyBy(newItem.prefactor);
				this.prefactor = this.prefactor.multiply(newItem.prefactor);
				for (let i=newItem.list.length - 1; i >= 0; i--) {
					this.list.unshift(newItem.list[i]);
				};
				break;
			default:
				throw "MathElement, MathOperator, or MathProduct required";
				break;
		};
		this.contractions = FH.joinContractions(this.contractions, newItem.contractions);
		this.updateExpression();
		return this;
	};

	innerMultiplyAtIndex(item, idx) { // MathProduct
		// TODO: take into account negative index
		// item: MathElement, MathOperator, MathProduct
		// insert item as the (new) idx-th factor as a product
		let newItem = item.makeCopy();
		this.insertItemAtIndex(newItem, idx);
		this.contractions = FH.joinContractions(this.contractions, newItem.contractions);
		this.updateExpression();
		return this;
	};

	removeFactorAtIndex(idx, updateContraction = true) { // MathProduct
		// TODO: take into account negative index
		// remove item at idx, with or without update of contraction
		// set boolean to false if the item will be insert after in same element.
		// get reference to item before removing using getItemAtIndex(i)
		// TODO
	};

	isOne() { // MathProduct
		if (this.list.length === 1
						&& this.list[0].id === MathElement_one.id) {
			return true;
		};
		return false;
	};

};



class MathTensorProduct extends MathExpression {
	constructor() { // MathTensorProduct
		super();

		// the symbols to use for display
		this.display_symbols = DISPLAY_SYMBOLS.MATHTENSORPRODUCT;

	};

	// return a MathTensorProduct from
	// MathElement, MathOperator, MathProduct, or MathTensorProduct
	// make a copy of original item
	static newFrom(item) { // MathTensorProduct
		switch (item.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
			case 'MathProduct':
				let newME = new MathTensorProduct();
				let newItem = MathProduct.newFrom(item);
				newME.list.push(newItem);
				newME.contractions = FH.joinContractions(newME.contractions, newItem.contractions);
				newME.copiedInfo = newItem.copiedInfo;
				newME.updateExpression();
				return newME;
				break;
			case 'MathTensorProduct':
				return item.makeCopy();
				break;
			default:
				throw "MathElement, MathOperator, MathProduct, or MathTensorProduct required";
				breaks;
		};
	};

	updateExpression() { // MathTensorProduct
		super.updateExpression();
		// MathTensorProduct: collect all prefactors in the global one
		for (let i=0; i < this.list.length; i++) {
			let obj = this.list[i];
			// this.prefactor.multiplyBy(obj.prefactor);
			this.prefactor = this.prefactor.multiply(obj.prefactor);
			obj.prefactor = RationalExpression.newFromLetterAndRational(CONST.DIMENSION_LETTER,1,1);
		};
		return this;
	};

	tensorOnRightWith(item) { // MathTensorProduct
		//item: MathElement, MathOperator, MathProduct, MathTensorProduct
		let newItem = null;
		switch (item.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
			case 'MathProduct':
				// convert to new MathProduct
				newItem = MathProduct.newFrom(item);
				this.list.push(newItem);
				break;
			case 'MathTensorProduct':
				newItem = item.makeCopy();
				// this.prefactor.multiplyBy(newItem.prefactor);
				this.prefactor = this.prefactor.multiply(newItem.prefactor);
				for (let i=0; i < newItem.list.length; i++) {
					this.list.push(newItem.list[i]);
				};
				break;
			default:
				throw "MathElement, MathOperator, MathProduct, or MathTensorProduct required";
				break;
		};
		this.contractions = FH.joinContractions(this.contractions, newItem.contractions);
		this.updateExpression();
		return this;
	};

	tensorOnLeftWith(item) { // MathTensorProduct
//		 item: MathElement, MathOperator, MathProduct, MathTensorProduct
		let newItem = null;
		switch (item.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
			case 'MathProduct':
				// convert to MathProduct
				newItem = MathProduct.newFrom(item);
				this.list.unshift(newItem);
				break;
			case 'MathTensorProduct':
				newItem = item.makeCopy();
				// this.prefactor.multiplyBy(newItem.prefactor);
				this.prefactor = this.prefactor.multiply(newItem.prefactor);
				for (let i=newItem.list.length - 1; i >= 0; i--) {
					this.list.unshift(newItem.list[i]);
				};
				break;
			default:
				throw "MathElement, MathOperator, MathProduct, or MathTensorProduct required";
				break;
		};
		this.contractions = FH.joinContractions(this.contractions, newItem.contractions);
		this.updateExpression();
		return this;
	};

	multiplyOnRightAtIndex(item, idx) { // MathTensorProduct
		// item: MathElement, MathOperator, MathProduct
		// multiply item on the right at the idx-th factor
		let targetItem = this.getItemAtIndex(idx);
		targetItem.multiplyOnRightWith(item);
		this.contractions = FH.joinContractions(this.contractions, targetItem.contractions);
		this.updateExpression();
		return this;
	};

	multiplyOnLeftAtIndex(item, idx) { // MathTensorProduct
		// item: MathElement, MathOperator, MathProduct
		// multiply item on the left at the idx-th factor
		let targetItem = this.getItemAtIndex(idx);
		targetItem.multiplyOnLeftWith(item);
		this.contractions = FH.joinContractions(this.contractions, targetItem.contractions);
		this.updateExpression();
		return this;
	};

	innerTensorAtIndex(item, idx) { // MathTensorProduct
		// item: MathElement, MathOperator, MathProduct, MathTensorProduct
		// insert item as the (new) idx-th factor as a tensor product
//		let newItem = item.makeCopy();
		let newItem = MathExpression.convertTo(item.makeCopy(), 'MathTensorProduct');
		this.insertItemAtIndex(newItem, idx);
		this.contractions = FH.joinContractions(this.contractions, newItem.contractions);
		this.updateExpression();
		return this;
	}

	removeTensorFactorAtIndex(idx, updateContraction = true) { // MathTensorProduct
		// remove item at idx, with or without update of contraction
		// set boolean to false if the item will be insert after in same element.
		// get reference to item before removing using getItemAtIndex(i)
	};

};



class MathArgument extends MathTensorProduct {
	constructor() { // MathArgument
		super();
		// element before the list
		// MathProduct of commutative MathElement
		this.preElement = MathProduct.newFrom(MathElement_one);

		// element after the list, MathElement
		// In the paper, we use "v" as a constant vector
		// here we use "1" as the MathElement MathElement_one
		// this substitution is harmless for the computation
		// since we only perform symbolic computations on this MathElement
		// TODO: change this MathElement to a "v" for notational purpose
		this.postElement = MathProduct.newFrom(MathElement_one);

		// the symbols to use for display
		this.display_symbols = DISPLAY_SYMBOLS.MATHARGUMENT;

	};

	static newFrom(item) { // MathArgument
		let newME = new MathArgument();
		switch (item.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
			case 'MathProduct':
			case 'MathTensorProduct':
				let newItem = MathTensorProduct.newFrom(item);
				// newME.prefactor.multiplyBy(newItem.prefactor);
				newME.prefactor = newME.prefactor.multiply(newItem.prefactor);
				newME.list = newItem.list;
				newME.contractions = FH.joinContractions(newME.contractions, newItem.contractions);
				newME.copiedInfo = newItem.copiedInfo;
				newME.updateExpression();
				return newME;
				break;
			case 'MathArgument':
				return item.makeCopy();
			default:
				throw "MathElement, MathOperator, MathProduct, or MathTensorProduct required";
				break;
		};
	};

	getID() { // MathArgument
		return {"main": super.getID(),
						"pre": this.preElement.getID(),
						"post": this.postElement.getID()};
	};

	// return pyramids of ids of elements and operators
	getArrID() { // MathArgument
		return {"main": super.getArrID(),
						"pre": this.preElement.getArrID(),
						"post": this.postElement.getArrID()};
	};

	getIndicesID() { // MathArgument
		return {"main": super.getIndicesID(),
						"pre": this.preElement.getIndicesID(),
						"post": this.postElement.getIndicesID()};
	};

	// return [arrIdx of every indices]
	getIndicesArrIdx() { // MathArgument
		let arr = [];
		let arrSub = null;
		// preElement => add idx = -2 in front
		arrSub = this.preElement.getIndicesArrIdx();
		for (let a of arrSub) {
			arr.push( [-2].concat(a) );
		};
		// main
		arr = arr.concat(super.getIndicesArrIdx());
		// postElement => add idx = -1 in front
		arrSub = this.postElement.getIndicesArrIdx();
		for (let a of arrSub) {
			arr.push( [-1].concat(a) );
		}
		return arr;
	};

	// return a object {MathElementArr, MathExpressionArr}
	// two arrays with references to all MathElements and all MathExpression in 'this'
	getContent() { // MathArgument
		let content = super.getContent();
		let preContent = this.preElement.getContent();
		let postContent = this.postElement.getContent();
		content.MathElementArr = content.MathElementArr.concat(preContent.MathElementArr);
		content.MathElementArr = content.MathElementArr.concat(postContent.MathElementArr);
		content.MathExpressionArr = content.MathExpressionArr.concat(preContent.MathExpressionArr);
		content.MathExpressionArr = content.MathExpressionArr.concat(postContent.MathExpressionArr);

		// console.log('MathArgument: ');
		// console.log('MathElementArr: ' + content.MathElementArr.length);
		// console.log('MathExpressionArr: ' + content.MathExpressionArr.length);
		return content;
	};


	getSumStringWeight() { // MathArgument
		let mainStr = FH.constructStringFromWeight(this.getSumWeight(), GLOBAL.sumWeightData, GLOBAL.sumWeightOperatorData);
		let postStr = FH.constructStringFromWeight(this.postElement.getSumWeight(), GLOBAL.sumWeightData, GLOBAL.sumWeightOperatorData);
		// return postStr + '-' + mainStr;
		let preStr = FH.constructStringFromWeight(this.preElement.getSumWeight(), GLOBAL.sumWeightData, GLOBAL.sumWeightOperatorData);
		return postStr + '-' + mainStr + '-' + preStr;
	};

	getSelectionData() { // MathArgument
		let mainSumWeight = this.getSumWeight();
		let postSumWeight = this.postElement.getSumWeight();
		let selection = {};
		selection.main = FH.getSelectionFromSumWeight(mainSumWeight, GLOBAL.sumWeightData, GLOBAL.sumWeightOperatorData);
		selection.post = FH.getSelectionFromSumWeight(postSumWeight, GLOBAL.sumWeightData, GLOBAL.sumWeightOperatorData);
		return selection;
	};

	// return array of arrIdx of all elements with itemID
	getPositionsByID(itemID, maxLevel, termElement) { // MathArgument
		let arrIdx = [];
		let currentArrIdx = null;
		switch(termElement) {
			case CONST.MAIN_ELEMENT:
				return super.getPositionsByID(itemID, maxLevel, termElement);
				break;
			case CONST.POST_ELEMENT:
				currentArrIdx = this.postElement.getPositionsByID(itemID, maxLevel, termElement);
				for (let e of currentArrIdx) {
					arrIdx.push([-1].concat(e));
				};
				return arrIdx;
				break;
			case CONST.PRE_ELEMENT:
				currentArrIdx = this.preElement.getPositionsByID(itemID, maxLevel, termElement);
				for (let e of currentArrIdx) {
					arrIdx.push([-2].concat(e));
				};
				return arrIdx;
				break;
			default:
				throw "term element ill defined";
				break;
		};
	};

	// return arrIdx of the first element with itemID
	getFirstPositionByID(itemID, maxLevel, termElement) { // MathArgument
		let arrIdx = null;
		switch(termElement) {
			case CONST.MAIN_ELEMENT:
				return super.getFirstPositionByID(itemID, maxLevel, termElement);
				break;
			case CONST.POST_ELEMENT:
				arrIdx = this.postElement.getFirstPositionByID(itemID, maxLevel, termElement);
				return [-1].concat(arrIdx);
				break;
			case CONST.PRE_ELEMENT:
				arrIdx = this.preElement.getFirstPositionByID(itemID, maxLevel, termElement);
				return [-2].concat(arrIdx);
				break;
			default:
				throw "term element ill defined";
				break;
		};
	};


	setMixedContraction(arrIdx1, arrIdx2) { // MathArgument
		// TODO: take into account negative index
		// TODO: arrIdx1 in main, arrIdx2 in postElement
		return this;
	};

	setPostElementContraction(arrIdx1, arrIdx2) { // MathArgument
		// TODO: take into account negative index
		// TODO: arrIdx1, arrIdx2 in postElement
		return this;
	};

	multiplyPreElementWith(item) { // MathArgument
		// since product of commutative elements, no 'left' and 'right' distinction
		// item.showConsole(OPT.SHOW, 'multiplyPreElementWith:');
		switch (item.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
			case 'MathProduct':
				this.preElement.multiplyOnRightWith(item);
				break;
			default:
				throw "MathElement, MathOperator, or MathProduct required";
				break;
		};
		this.contractions = FH.joinContractions(this.contractions, item.contractions);
		this.updateExpression();
		// this.preElement.showConsole(OPT.SHOW, 'preElement: ');
		return this;
	};

	collectCommutativeElementsInPreElement() { // MathArgument
		// this.showConsole(OPT.SHOW, 'current term before: ');
		let arrComm = this.getCommutativeElements();
		// console.log('arrComm: ' + JSON.stringify(arrComm));
		let c = 0;
		while (arrComm.length > 0) {
			FH.printProgressMessage('Collect comm. el. #' + c);
			let arrIdx = FH.copyData(arrComm[0]);
			arrIdx = arrIdx.slice(0, 2);
			// console.log('arrIdx to get currentItem: ' + JSON.stringify(arrIdx));
			let currentItem = this.getItemAt(arrIdx);
			// currentItem.showConsole(OPT.SHOW, 'currentItem to move: ');
			let itemContainer = this.getItemAtIndex(arrIdx[0]);
			// itemContainer.showConsole(OPT.SHOW, 'itemContainer (before): ');
			if (itemContainer.list.length === 1) {
				// we are going to remove last element in this MathProduct
				// => insert MathElement_one"
				itemContainer.list.push(MathElement_one.makeCopy());
				// itemContainer.showConsole(OPT.SHOW, 'itemContainer (add 1): ');
			};
			let retInfo = {};
			itemContainer.removeItemAtIndex(arrIdx[1], retInfo, false);
			// itemContainer.showConsole(OPT.SHOW, 'itemContainer (after): ');
			// don't use multiplyPreElementWith since then there is a copy
			// => update contractions needed
			// this.multiplyPreElementWith(currentItem);
			// just push currentItem in the list of preElement
			this.preElement.list.push(currentItem);
			// this.showConsole(OPT.SHOW, 'current term after: ');
			arrComm = this.getCommutativeElements();
			// console.log('arrComm: ' + JSON.stringify(arrComm));
			// console.log('end of while');
			// console.log('');
		};
		FH.printProgressMessage('');
		// this.showConsole(OPT.SHOW, 'current term after while: ');
		// console.log('');
		// console.log('');
		return this;
	};

	multiplyPostElementOnRightWith(item) { // MathArgument
		switch (item.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
			case 'MathProduct':
				this.postElement.multiplyOnRightWith(item);
				break;
			default:
				throw "MathElement, MathOperator, or MathProduct required";
				break;
		};
		this.contractions = FH.joinContractions(this.contractions, item.contractions);
		this.updateExpression();
		return this;
	};

	multiplyPostElementOnLeftWith(item) { // MathArgument
		switch (item.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
			case 'MathProduct':
				this.postElement.multiplyOnLeftWith(item);
				break;
			default:
				throw "MathElement, MathOperator, or MathProduct required";
				break;
		};
		this.contractions = FH.joinContractions(this.contractions, item.contractions);
		this.updateExpression();
		return this;
	};

	updateExpression() { // MathArgument
		super.updateExpression();
		this.preElement.updateExpression();
		this.postElement.updateExpression();
		// normalize global prefactor to 1
		// collect prefactor of main and postElement into preElement.prefactor
		// this.preElement.prefactor.multiplyBy(this.prefactor);
		this.preElement.prefactor = this.preElement.prefactor.multiply(this.prefactor);
		this.prefactor = RationalExpression.newFromLetterAndRational(CONST.DIMENSION_LETTER,1,1);
		// this.preElement.prefactor.multiplyBy(this.postElement.prefactor);
		this.preElement.prefactor = this.preElement.prefactor.multiply(this.postElement.prefactor);
		this.postElement.prefactor = RationalExpression.newFromLetterAndRational(CONST.DIMENSION_LETTER,1,1);

		this.preElement.updateExpression();
		this.postElement.updateExpression();

		return this;
	};

	toObject() { // MathArgument
		let obj = super.toObject();
		// preElement
		obj.pre = this.preElement.toObject();
		// postElemet
		obj.post = this.postElement.toObject();
		return obj;
	}

	isZero() { // MathArgument
		return this.preElement.isZero();
	};


};





// sum of MathProduct, MathTensorProduct, MathArguments
// need prefactor in objects
// update such that prefactor = 1
class MathSum extends MathExpression {
	constructor() { // MathSum
		super();
		this.prefactor = RationalExpression.newFromLetterAndRational(CONST.DIMENSION_LETTER,1,1);
		// default sum is zero but prefactor is 1

		// the symbols to use for display
		this.display_symbols = DISPLAY_SYMBOLS.MATHSUM;

	};

	static newFrom(item) { // MathSum
		let newME = new MathSum();
		let newItem = null;
		switch (item.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
			case 'MathProduct':
				// convert to MathProduct to get prefactor
				newItem = MathProduct.newFrom(item);
				newME.list.push(newItem);
				// newME.contractions = FH.joinContractions(newME.contractions, newItem.contractions);
				newME.updateExpression();
				return newME;
				break;
			case 'MathTensorProduct':
				newItem = item.makeCopy();
				newME.list.push(newItem);
				// newME.contractions = FH.joinContractions(newME.contractions, newItem.contractions);
				newME.updateExpression();
				return newME;
				break;
			case 'MathArgument':
				newItem = item.makeCopy();
				newME.list.push(newItem);
				// newME.contractions = FH.joinContractions(newME.contractions, newItem.contractions);
				newME.updateExpression();
				return newME;
				break;
			case 'MathSum':
				return item.makeCopy();
				break;
			default:
				throw "MathElement, MathOperator, MathProduct, MathTensorProduct, or MathSum required";
				break;
		};
	};

	static applyMathOperatorTo(operator, item, returnedInfo) { // MathSum
		// apply MathOperator operator on the MathProduct item
		// and use the corresponding rules to return a MathSum
		// both operator and item are copied

		let newSum = new MathSum();
		returnedInfo.mapIndicesIDByTerms = [];

		switch (item.constructor.name) {
			case 'MathProduct':
				// collect all array of indices
				let itemArrIndicesID = item.getIndicesID();
				let opArrIndicesID = operator.getIndicesID();

				returnedInfo.mapIndicesIDByTerms = [];
				if (operator.behavior === CONST.OPERATOR_LEIBNIZ_RULE) {
					for (let i=0; i<item.list.length; i++) {
						let newItem = item.makeCopy();
						// console.log('newItem.copiedInfo: ' + JSON.stringify(newItem.copiedInfo));

						let newItemArrIndicesID = newItem.getIndicesID();
						let orgIndiceID = operator.indices.id;
						let newOp = operator.makeCopy();
						let newIndiceID = newOp.indices.id;
						newOp.status = CONST.OPERATOR_STATUS_LATENT;
						newOp.argument = null;
						// console.log('newOp.copiedInfo: ' + JSON.stringify(newOp.copiedInfo));
						let newOpArrIndicesID = newOp.getIndicesID();

						let currentItem = newItem.getItemAtIndex(i);
						let retInfo = {}; // no need here
						newItem.removeItemAtIndex(i, retInfo, false);
						newOp.applyTo(currentItem);
						newItem.insertItemAtIndex(newOp,i);

						newSum.addTo(newItem, false);
						// return data on the correspondances between old indices IDs
						// and new ones for each term in the new sum
						// this info can be used to update contractions
						// only return change of indiceID of the operator, not its arguments
						let mapIndicesID = [[orgIndiceID, newIndiceID]]
																		.concat(newItem.copiedInfo.mapIndicesID);
						returnedInfo.mapIndicesIDByTerms.push(mapIndicesID);
					};
				};
				return newSum;
				break;

			case 'MathSum':
				// apply the operator on all the terms of the sum
				// collect all returned information on the indiceIDs
				returnedInfo.mapIndicesIDByTerms = [];
				for (let currentItem of item.list) {
					let retInfo = {};
					let currentSum = MathSum.applyMathOperatorTo(operator, currentItem, retInfo);
					newSum.addTo(currentSum, false);
					returnedInfo.mapIndicesIDByTerms.push(retInfo.mapIndicesIDByTerms);
				};
				return newSum;
				break;

			default:
				throw 'MathProduct or MathSum required';
				break;
		};
	};

	// this method does not apply to MathSum
	getContent() { // MathSum
		throw "getContent does not apply to MathSum";
	};




	// this method substitutes all occurence of an item (MathElement) and its derivatives
	// by some given MathProduct
	// all the necessary data are given in the object substitutionData
	// {
	//	orgItem: MathElement to replace,
	//	termElement: CONST.MAIN_ELEMENT, or CONST.POST_ELEMENT,
	//	minLevel: integer, min value of derivations applied to orgItem,
	//	maxLevel: integer, max value of derivations applied to orgItem,
	//	newItemsByDerivatives: [substitionObjectsAtDerivation-i, …]
	// }
	// newItemsByDerivatives should contains as many items from label 0 to maxLevel
	// => subtitution at a level of derivation i uses the data at index i in newItemsByDerivatives
	// substitionObjectsAtDerivation-i are arrays of data for every level of derivations:
	// substitionObjectsAtDerivation-i = [
	// 	{
	//		item: MathProduct that will substitutes the corresponding derivative of orgItem,
	//		mapIndices: [{from: arrIdx, to: arrIdx},…] are the correspondances between original indices and new ones,
	//		permutationsSum:[ [0,1,2,3], … ] are the list of permutations to apply on 'to' in mapIndices
	// 	},
	// …]
	// if 'item' is 'null' then the replacement is by 0,
	// in that case, no need to supply mapIndices and permutationsSum
	// length of each permutation is exactly the number of elements in mapIndices
	// both mapIndices and permutationsSum can be empty arrays
	substituteItemAndDerivativesWithData(substitutionData) { // MathSum
		// only substitute item in main list of MathExpression
		// if orgItem can't be replaced, replace its id by newID to continue the loop
		let newID = FH.createUniqueID(2, '@', GLOBAL.listElementID);

		let orgItem = substitutionData.orgItem;
		let derivationOperators = [];
		if (substitutionData.derivationOperators !== undefined) {
			derivationOperators = substitutionData.derivationOperators;
		} else {
			throw "derivation operators should be provided";
		};
		let newItemsByDerivatives = substitutionData.newItemsByDerivatives;
		// console.log('newItemsByDerivatives: ' + JSON.stringify(newItemsByDerivatives));

		// determine in which part of MathArgument the substitution takes place
		let termElement = null;
		if (substitutionData.termElement === undefined) {
			termElement = CONST.MAIN_ELEMENT;
		} else {
			termElement = substitutionData.termElement;
		};

		// min level of derivations to consider
		// no value => 0
		let minLevel = null;
		if (substitutionData.minLevel === undefined) {
			minLevel = 0;
		} else {
			minLevel = substitutionData.minLevel;
		};
		// max level of derivations to consider
		let maxLevel = substitutionData.maxLevel;
		// console.log('min = ' + minLevel + ' / max  = ' + maxLevel);

		// general tests:
		// orgItem should be MathElement
		if (orgItem.constructor.name !== 'MathElement') {
			throw "MathElement for original item required";
		};
		// substitute items must be MathProduct or 'null'
		// length of perm in permuntationsSum = length of mapIndices
		for (let newItemDerivativeData of newItemsByDerivatives) {
			for (let newItemData of newItemDerivativeData) {
				// check that each item is MathProduct if not 'null'
				if (newItemData.item !== null) {
					if (newItemData.item.constructor.name !== 'MathProduct') {
						throw "MathProduct for new item required";
					};
					let l = newItemData.mapIndices.length;
					for (let perm of newItemData.permutationsSum) {
						if (perm.length !== l) {
							throw "Inconsistancy in permutationsSum";
						};
					};
				}; // test not 'null'
			};
		};

		// unique identifier to find occurences of orgItem in 'this'
		let orgItemID = orgItem.getID();
		// console.log('orgItemID: ' + orgItemID);
		// unique identifier to test applied derivations
		let derivationIDArr = [];
		for (let derivationOperator of derivationOperators) {
			derivationIDArr.push(derivationOperator.getID());
		};

		// keep time used by getFirstArrayIdxForElementIDAndLevels
		// CONST.TIMER.getPosition = 0;
		// let timerStart = null;
		// let timerEnd = null;
		//
		// timerStart = Date.now();

		// loop over first occurence of orgItem in 'this'
		// until all occurences are replaced
		let arrIdx = this.getFirstArrayIdxForElementIDAndLevels(orgItemID, derivationIDArr, minLevel, maxLevel, termElement);
		// this.showConsole(OPT.SHOW, 'this (to find) = ');
		// console.log('arrIdx: ' + JSON.stringify(arrIdx));
		// console.log('');

		// timerEnd = Date.now();
		// CONST.TIMER.getPosition = CONST.TIMER.getPosition + timerEnd - timerStart;

		let counter = 0;
		while (arrIdx !== null) {
			let N = this.list.length;
			let R = arrIdx[0]/N; // ratio to display progress
			counter++;
			FH.printProgressMessage('item found at '
										+ FH.padStringOnRight(JSON.stringify(arrIdx), ' ', 25)
										+ ' ('
										+ '#found: ' + counter
										+ ' / #terms: ' + N
										+ ' / progress: ' + FH.toStringPercent(R)
										+ ')');

			let termIdx = arrIdx[0];
			let currentTerm = this.getItemAtIndex(termIdx);
			// console.log('');
			// currentTerm.showConsole(OPT.SHOW, 'currentTerm (found) = ');
			// console.log('');

			// remove first element, which is the sum index
			arrIdx.splice(0,1);
			// console.log('arrIdx: ' + JSON.stringify(arrIdx));

			// construct new sum using ad hoc method
			let newSum = currentTerm.constructNewSumWithSubstitutions(arrIdx, newItemsByDerivatives);
			// newSum.showConsole(OPT.SHOW, 'newSum = ');
			// newSum is null if replacement impossible => change id of original item
			if (newSum !== null) {
				// insert the collecting sum and replace previous term
				let retInfo = {};
				this.replaceByItemAtIndex(newSum, termIdx, retInfo, false);
			} else {
				// change id
				console.log('change id of MathElement');
				currentTerm.getItemAt(arrIdx).id = newID;
			}

			// this.showConsole(OPT.SHOW, 'this = ');

			// find next occurence of orgItem
			// timerStart = Date.now();
			arrIdx = this.getFirstArrayIdxForElementIDAndLevels(orgItemID, derivationIDArr, minLevel, maxLevel, termElement);
			// timerEnd = Date.now();
			// CONST.TIMER.getPosition = CONST.TIMER.getPosition + timerEnd - timerStart;

		}; // while arrIdx

		this.updateExpression();
		FH.printProgressMessage('');
		// console.log('getFirstArrayIdxForElementIDAndLevels: ' + CONST.TIMER.getPosition + 'ms');
		// console.log('end of substitution');
		return this;
	};

	getFirstArrayIdxForElementIDAndLevels(orgItemID, derivationIDArr, minLevel, maxLevel, termElement) { // MathSum
		// get all positions and find > min level with corresponding derivations
		let arrArrIdx = this.getPositionsByID(orgItemID, maxLevel, termElement);
		// console.log('');
		// console.log('get first array idx…');
		// this.showConsole(OPT.SHOW, 'search in: ');
		// console.log('arrArrIdx (in get): ' + JSON.stringify(arrArrIdx));
		// console.log('');
		if (arrArrIdx.length > 0) {
			for (let arrIdx of arrArrIdx) {
				// console.log('');
				// console.log('arrIdx: ' + JSON.stringify(arrIdx));
				// number of applied derivations (on a MathElement) = nMax - 1
				let nMax = - arrIdx[arrIdx.length-1];
				// console.log('nMax: ' + nMax);
				if (minLevel === 0 && nMax <= 0) {
					// item not under derivation and min level accepts this => return arrIdx
					// console.log('return arrIdx not in derivation');
					return arrIdx;
				};
				if (nMax > minLevel) {
					// number of derivation is >= minLevel
					// arrIdx to capture the first derivation
					let newArrIdx = FH.copyData(arrIdx);
					newArrIdx[newArrIdx.length-1] = -1;
					let firstDerivation = this.getItemAt(newArrIdx);
					// firstDerivation.showConsole(OPT.SHOW, 'first = ');
					let derivationIDtest = true;
					for (let n=1; n<nMax; n++) {
						let testDerivation = firstDerivation.getItemAt([-n]);
						// testDerivation.showConsole(OPT.SHOW, 'test = ');
						// console.log('ID: ' + testDerivation.id);
						if (!derivationIDArr.includes(testDerivation.id)) {
							derivationIDtest = false;
						};
					}; // loop over derivations to test
					if (derivationIDtest) {
						// console.log('return arrIdx in derivation');
						return arrIdx;
					} else {
						// not a good candidate => next in for on arrArrIdx
						// console.log('continue');
						continue;
					};
				};
			}; // for on arrArrIdx
		};
		// console.log('return null');
		return null;
	};

	// collect in a new MathSum references to terms
	// satisfying the selection criteria
	// fillSumStringWeight() should have been applied on terms before
	collectTermsByCriteria(selection) { // MathSum
		let {mainStr, postStr} = FH.constructRegExpFromSelection(selection, GLOBAL.sumWeightData, GLOBAL.sumWeightOperatorData);
		let mainRE = new RegExp(mainStr, "g");
		let fullRE = new RegExp(postStr + '-' + mainStr, "g");
		let resultSum = new MathSum();
		for (let i=0; i<this.list.length; i++) {
			let item = this.list[i];
			switch (item.constructor.name) {
				case 'MathProduct':
				case 'MathTensorProduct':
					if (item.sumStringWeight.match(mainRE)) {
						resultSum.addTo(item, false);
					};
					break;
				case 'MathArgument':
				if (item.sumStringWeight.match(fullRE)) {
					resultSum.addTo(item, false);
				};
					break;
				default:
					break;
			};
		};
		return resultSum;
	};

	// collect in a new MathSum references to terms
	// at positions given by positionsList = [list of integers as indx of this.list]
	collectTermsByPositions(positionsList) { // MathSum
		let resultSum = new MathSum();
		for (let i of positionsList) {
			if (this.list[i] !== undefined) {
				resultSum.addTo(this.list[i], false);
			};
		};
		return resultSum;
	};

	// collect in a new MathSum references to terms
	// at positions given in the range (inclusive) [begin, end]
	collectTermsInRange(range) { // MathSum
		let positionsList = [];
		for (let i=range[0]; i<=range[1]; i++) {
			positionsList.push(i);
		}
		return this.collectTermsByPositions(positionsList);
	};

	// collect in a new MathSum references to terms
	// at positions given after (inclusive) begin
	collectTermsStartingAt(begin) { // MathSum
		let positionsList = [];
		for (let i=begin; i<this.list.length; i++) {
			positionsList.push(i);
		}
		return this.collectTermsByPositions(positionsList);
	};

	// collect in a new MathSum references to terms
	// at positions given before (inclusive) end
	collectTermsUntil(end) { // MathSum
		let positionsList = [];
		for (let i=0; i<=end; i++) {
			positionsList.push(i);
		}
		return this.collectTermsByPositions(positionsList);
	};

	collectCommutativeElementsInPreElement() { // MathSum
		for (let i=0; i<this.list.length; i++) {
			FH.printProgressMessage('term ' + i + '/' + this.list.length);
			// console.log('term number ' + i);
			let currentTerm = this.list[i];
			currentTerm.collectCommutativeElementsInPreElement();
			// this.showConsole(OPT.SHOW, 'sum: ');
		};
		FH.printProgressMessage('');
		return this;
	};

	sortPreElementsByProductWeight() { // MathSum
		for (let term of this.list) {
			term.preElement.sortByProductWeight();
		};
	};


	displaySelectionCriteria(opt, selection) { // MathSum
		let mainSelection = selection.main;
		let postSelection = selection.post;
		let str = 'Selection:\n';
		let strMain = 'Main: \n';
		for (let e of mainSelection) {
			strMain += opt.display.formatLetter(e.element.letter)
											+ ': from ' + e.min + ' to ' + e.max;
			for (let o of e.operators) {
				strMain += ' / applied ' + opt.display.formatLetter(o.op.letter)
											+ ': from ' + o.min + ' to ' + o.max;
			};
			strMain += '\n';
		};
		str += strMain;

		let strPost = 'Post: \n';
		for (let e of postSelection) {
			strPost += opt.display.formatLetter(e.element.letter)
											+ ': from ' + e.min + ' to ' + e.max;
			for (let o of e.operators) {
				strPost += ' / applied ' + opt.display.formatLetter(o.op.letter)
											+ ': from ' + o.min + ' to ' + o.max;
			};
			strPost += '\n';
		};
		str += strPost;

		console.log(str);
		return this;
	};

	updateExpression() { // MathSum
		super.updateExpression();
		// MathSum: normalize prefactor to 1
		 if (this.list.length > 0) {
			 let r = this.prefactor;
			 if (!r.isZero()) {
				 // don't multiply by 0…
				 for (let i=0; i < this.list.length; i++) {
					// this.list[i].multiplyBy(r);
					this.list[i].prefactor = this.list[i].prefactor.multiply(r);
					this.list[i].updateExpression();
				 };
			 };
		 };
		 this.prefactor = RationalExpression.newFromLetterAndRational(CONST.DIMENSION_LETTER,1,1);

		 // remove terms with zero prefactor
		 for (let i=this.list.length - 1; i >=0; i--) {
			 let term = this.list[i];
			 if (term.isZero()) {
				 let retInfo = {};
				 // don't use removeItemAtIndex since call updateExpression after removing
				 // => problems on elements which are removed
				 // this.removeItemAtIndex(i, retInfo, true);
				 // use native splice method
				 this.list.splice(i, 1);
			 };
		 };
	};

	addTo(item, copy=true) { // MathSum
		// item: MathElement, MathOperator, MathProduct, MathSum
		// copy: true to copy item, false to add without copy
		// item.showConsole(OPT.SHOW, 'item = ');
		let newItem = null;
		switch (item.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
				// convert to MathProduct
					newItem = MathProduct.newFrom(item);
				this.list.push(newItem);
				break;
			case 'MathProduct':
			case 'MathTensorProduct':
			case 'MathArgument':
				if (copy) {
					newItem = item.makeCopy();
				} else {
					newItem = item;
				};
				this.list.push(newItem);
				break;
			case 'MathSum':
				if (copy) {
					newItem = item.makeCopy();
				} else {
					newItem = item;
				};
				for (let i=0; i < newItem.list.length; i++) {
					this.list.push(newItem.list[i]);
				};
				break;
			default:
				throw "MathElement, MathOperator, MathProduct, MathTensorProduct, MathArgument, or MathSum required";
				break;
		};
		// this.contractions = FH.joinContractions(this.contractions, newItem.contractions);
		this.updateExpression();
		return this;
	};

	tensorOnRightWith(item) { // MathSum
		// item: MathElement, MathOperator, MathProduct, MathSum
		let newItem = null;
		switch (item.constructor.name) {
			case 'MathElement':
			case 'MathOperator':
			case 'MathProduct':
			case 'MathTensorProduct':
				// convert to MathProduct
				newItem = MathTensorProduct.newFrom(item);
				for (let i=0; i<this.list.length; i++) {
					if (this.list[i].constructor.name !== 'MathArgument') {
						this.list[i] = MathExpression.convertTo(this.list[i], 'MathTensorProduct');
					}; // else it is ready to receive tensor multiplication
					this.list[i].tensorOnRightWith(newItem);
				};
				// this.contractions = FH.joinContractions(this.contractions, newItem.contractions);
				this.updateExpression();
				return this;
				break;
			case 'MathSum':
				newItem = item.makeCopy();
				// from right to left in order to always replace at index i
				for (let i=this.list.length-1; i>=0; i--) {
					let currentTerm = this.list[i];
					let newSum = new MathSum(); // will replace currentTerm
					if (currentTerm.constructor.name !== 'MathArgument') {
						currentTerm = MathExpression.convertTo(currentTerm, 'MathTensorProduct');
					}; // else it is ready to receive tensor multiplication
					for (let j=0; j<newItem.list.length; j++) {
						// console.log('(i,j) = (' + i + ', ' + j +')');
						let newCurrentTerm = currentTerm.makeCopy();
						newCurrentTerm.tensorOnRightWith(newItem.getItemAtIndex(j));
						// newCurrentTerm.showConsole(OPT.SHOW, 'currentTerm = ');
						newSum.addTo(newCurrentTerm);
					};
					// newSum.showConsole(OPT.SHOW, 'newSum = ');
					// replace this.list[i] with newSum
					let returnedInfo = {};
					this.replaceByItemAtIndex(newSum, i, returnedInfo, true);
					// console.log('returnedInfo = ' + JSON.stringify(returnedInfo));
				};
				this.updateExpression();
				return this;
				break;
			default:
				throw "MathElement, MathOperator, MathProduct, MathTensorProduct, or MathSum required";
				break;
		};
	};

	sortBySumWeight() { // MathSum
		this.fillSumStringWeight();
		this.list.sort(FH.sortSumStringWeight);
		return this;
	};

	normalizeToTensorProducts() { // MathSum
		for (let i=0; i < this.list.length; i++) {
			this.list[i] = MathExpression.convertTo(this.list[i], 'MathTensorProduct');
		};
		return this;
	};

	normalizeToArguments() { // MathSum
		for (let i=0; i < this.list.length; i++) {
			this.list[i] = MathExpression.convertTo(this.list[i], 'MathArgument');
		};
		return this;
	};

	// only adapted to sum of same kind of arguments
	// compare terms in the sum with others via their string representation
	// and add their prefactors
	simplifyExpression() { // MathSum
		this.fillStringRepresentation();
		let itemMathArgument = false;
		if (this.list.length > 0) {
			let item = this.list[0];
			if (item.constructor.name === 'MathArgument') {
				itemMathArgument = true;
			};
		};
		if (itemMathArgument) {
			let i=0;
			while (i < this.list.length) {
				FH.printProgressMessage('Simplification: term ' + i + '/' + this.list.length);
				// main+postElement
				let strRep = this.list[i].stringRepresentation;
				// preElement
				let strRepPreEl = this.list[i].preElement.stringRepresentation;
				for (let j = this.list.length - 1; j > i; j--) {
					let item = this.list[j];
					let strRepToTest = item.stringRepresentation;
					let strRepPreElToTest = item.preElement.stringRepresentation;
					// check main+postElement and preElement
					if (strRep === strRepToTest && strRepPreEl === strRepPreElToTest) {
						let retInfo = {};
						this.list[i].preElement.prefactor = this.list[i].preElement.prefactor.add(item.preElement.prefactor);
						this.list.splice(j, 1);
					};
				};
				i++;
			};
		} else {
			// not MathArgument
			let i=0;
			while (i < this.list.length) {
				FH.printProgressMessage('Simplification: term ' + i + '/' + this.list.length);
				// main+postElement
				let strRep = this.list[i].stringRepresentation;
				for (let j = this.list.length - 1; j > i; j--) {
					let item = this.list[j];
					let strRepToTest = item.stringRepresentation;
					// check main+postElement
					if (strRep === strRepToTest) {
						let retInfo = {};
						this.list[i].prefactor = this.list[i].prefactor.add(item.prefactor);
						this.list.splice(j, 1);
					};
				};
				i++;
			};
		}
		FH.printProgressMessage('');
		// to clean zero terms
		this.updateExpression();
		return this;
	}


};





module.exports = {
	MathExpression: MathExpression,
	MathProduct: MathProduct,
	MathTensorProduct: MathTensorProduct,
	MathSum: MathSum,
	MathArgument: MathArgument,
}
