/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');


// Define classes Rational and Integer, their multiplication and addition

class Rational {
	constructor(numerator,denominator) { // Rational
		if (Number.isInteger(numerator) && Number.isInteger(denominator) && denominator !== 0) {
			[this.numerator, this.denominator] = this.constructor.reduce(numerator,denominator);
		} else {
			[this.numerator, this.denominator] = [NaN, 1];
		}
	};

	static reduce(numerator,denominator) { // Rational
		//https://stackoverflow.com/questions/4652468/is-there-a-javascript-function-that-reduces-a-fraction
		let gcd = function(a,b) {
			return b ? gcd(b, a%b) : a;
		};
		let gcdnd = gcd(numerator,denominator);
		let sd = Math.sign(denominator);
		let sgcd = Math.sign(gcdnd);
		return [sd*sgcd*numerator/gcdnd, sd*sgcd*denominator/gcdnd];
	};

	static isInteger(r) { // Rational
			if (r.denominator === 1) {
				return true;
			} else {
				return false;
			}
			// return Number.isInteger(r);
	};

	static newFrom(r) { // Rational
		if (r instanceof Rational) {
			return r.makeCopy();
		} else if (Number.isInteger(r)) {
			return new Rational(r,1);
		} else {
			throw "Rational or integer required";
		};
	};

	makeCopy() { // Rational
		let copy = FH.duplicateInstance(this);
		[copy.numerator, copy.denominator] = FH.copyData([this.numerator, this.denominator]);
		return copy;
	};

	// r is integer or Rational
	// return new Rational
	multiply(r) { // Rational
		// console.log(chalk.red('use of multiply: replace by multiplyBy?'));
		if (r instanceof Rational) {
			return new Rational(this.numerator * r.numerator, this.denominator * r.denominator);
		} else if (Number.isInteger(r)) { // can multiply integers
			return new Rational(this.numerator * r, this.denominator);
		} else {
			throw "multiply by Rational or integer";
		};
	};

	// r is integer or Rational
	// return this
	multiplyBy(r) { // Rational
		if (r instanceof Rational) {
			[this.numerator, this.denominator] = this.constructor.reduce(this.numerator * r.numerator, this.denominator * r.denominator);
		} else if (Number.isInteger(r)) { // can multiply integers
			[this.numerator, this.denominator] = this.constructor.reduce(this.numerator * r, this.denominator);
		} else {
			throw "multiply by Rational or integer";
		};
		return this;
	};

	// return new Rational
	add(r) { // Rational
		if (r instanceof Rational) {
			return new Rational(this.numerator * r.denominator + this.denominator * r.numerator, this.denominator * r.denominator);
		} else if (Number.isInteger(r)) { // can add integers
			return new Rational(this.numerator + this.denominator * r, this.denominator);
		} else {
			return {};
			throw "add Rational or integer";
		};
	};

	// return this
	addTo(r) { // Rational
		if (r instanceof Rational) {
			[this.numerator, this.denominator] = this.constructor.reduce(this.numerator * r.denominator + this.denominator * r.numerator, this.denominator * r.denominator);
			return this;
		} else if (Number.isInteger(r)) { // can add integers
			[this.numerator, this.denominator] = this.constructor.reduce(this.numerator + this.denominator * r, this.denominator);
			return this;
		} else {
			return {};
			throw "add Rational or integer";
		};
	};

	getStringRepresentation() { // Rational
		if (this.denominator !== 1) {
			return this.numerator + '/' + this.denominator;
		} else {
			return this.numerator + '';
		}

	};

	displayText(opt) { // Rational
		let od = opt.display;
		// Rational
		let rationalTxt = '';
		if (this.denominator !== 1) {
			rationalTxt += this.numerator + "/" + this.denominator;
		} else {
			rationalTxt += this.numerator;
		}
		rationalTxt = od.formatRational(rationalTxt);
		// class name
		let classTxt = '';
		if (opt.type) {
			classTxt = CONST.SYMBOL_OPEN_CLASS
								+ this.constructor.name
								+ CONST.SYMBOL_CLOSE_CLASS;
		};
		classTxt = od.formatClass(classTxt);

		return classTxt + rationalTxt;
	};

	showConsole(opt, text = '') { // Rational
		let globalTxt = opt.display.formatText(text) + this.displayText(opt);
		console.log(globalTxt);
		return this;
	};

	toObject() { // Rational
		return {
			class: this.constructor.name,
			num: this.numerator,
			den: this.denominator
		};
	};

	toJSON(...args) { // Rational
		return JSON.stringify(this.toObject(), ...args);
	};

	isEqual(r) {
		if (r instanceof Rational) {
			return (this.numerator === r.numerator
							&& this.denominator === r.denominator)
		} else {
			throw "compare to Rational";
		};
	};

	isZero() { // Rational
		if (this.numerator === 0) {
			return true;
		} else {
			return false;
		};
	};

	isOne() { // Rational
		if (this.numerator === 1 && this.denominator === 1) {
			return true;
		} else {
			return false;
		};
	};

	isNegative() { // Rational
		if (this.numerator * this.denominator < 0) {
			return true;
		} else {
			return false;
		};
	};

};

class Integer extends Rational {
	constructor(integer) { // Integer
		super(integer, 1);
	};

};


module.exports = {
	Rational: Rational,
	Integer: Integer,
}
