/*
Copyright 2019, Thierry Masson, Bruno Iochum

Contact: thierry.masson@cpt.univ-mrs.fr

This file is part of a computer program whose purpose is to help to compute
asymptotic expansion coefficients of the heat-trace for a nonminimal Laplace type operator
using a method developed and exposed in the paper
"Heat coefficient a_4 for nonminimal Laplace type operators" by Bruno Iochum and Thierry Masson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

let REL_PATH = '../';

const chalk = require('chalk');
const {GLOBAL, CONST, DISPLAY_SYMBOLS, OPT, FILES} = require(REL_PATH + 'INIT-FILES/options');
const {FH} = require(REL_PATH + 'FUNCTIONS-FILES/functions-helpers');
const {FF} = require(REL_PATH + 'FUNCTIONS-FILES/functions-factorize');


// Define classes RationalExpression,
// with multiplication, addition, simplification, display

class Polynomials {
	constructor(letter) { // Polynomials
		// [ ..., integer coefficient,...]
		// array index = power of letter, from 0 to max degree
		// order = increasing power
		// fill with "null" if necessary to avoid empty slots
		// reduce to max degree by removing "null" values after max degree
		this.monomials = [];
		// initialization to 0
		this.monomials[0] = 0;
		this.letter = letter;
	};

	// normalize this.monomials:
	// remove "null" and "0" at the end
	// replace "0" by "null", except 0th power
	updateExpression() { // Polynomials
		let p = this.monomials.length-1;
		let c = this.monomials[p];
		while ( p>0 && (c === null || c == 0) ) {
			this.monomials.splice(p,1);
			p--;
			// console.log('p='+p);
			c = this.monomials[p];
		};
		for (let p=this.monomials.length-1; p>0; p--) {
			if (this.monomials[p] == 0) {
				this.monomials[p] = null;
			};
		};
		// console.log('monomials: ' + JSON.stringify(this.monomials));
		return this;
	};

	// construction with letter, coefficient and power
	static newFromLetterCoefficientAndPower(l,c,p) { // Polynomials
		if (!Number.isInteger(c) || !Number.isInteger(p) || p<0) {
			throw "coefficient should be integer, power should be positive integer";
		};
		let poly = new Polynomials(l);
		poly.monomials[p] = c;
		poly.updateExpression();
		return poly;
	};

	// construction of 0th order polynomials with letter and coefficient
	static newFromLetterAndCoefficient(l,c) { // Polynomials
		return Polynomials.newFromLetterCoefficientAndPower(l,c,0);
	};

	// constrution with letter and array of coefficients (at powers)
	static newFromLetterAndArrayOfCoefficient(l,arrCoeff) { // Polynomials
		// console.log(JSON.stringify(arrCoeff));
		let poly = new Polynomials(l);
		// poly.showConsole(OPT.SHOW, 'poly = ');
		let c = null;
		let addPoly = null;
		for (let i=0; i<arrCoeff.length; i++) {
			c = arrCoeff[i];
			if (c !== undefined && c !== null) {
				addPoly = Polynomials.newFromLetterCoefficientAndPower(l,c,i);
				poly = poly.add(addPoly);
			};
			// poly.showConsole(OPT.SHOW, 'poly = ');
		};
		poly.updateExpression();
		return poly;
	};

	getLetter() { // Polynomials
		return this.letter;
	};

	// return array of non vanishing coefficients
	// this does not represent the polynomials
	// used to compute from these coefficient, for instance GCD
	getArrayCoefficients() { // Polynomials
		let arr = [];
		let c = null;
		for (let i=0; i<this.monomials.length; i++) {
			c = this.monomials[i];
			if (c !== undefined && c !== null) {
				arr.push(c);
			};
		};
		return arr;
	}

	// return the coefficient of the highest degree
	getMaxDegreeCoefficient() { // Polynomials
		let c = this.monomials[this.monomials.length-1];
		if (c !== undefined && c !== null) {
			return c;
		} else {
			throw "Error with max degree coefficient";
		};
	};

	// return the degree
	getDegree() { // Polynomials
		return this.monomials.length - 1;
	};

	// create a copy
	makeCopy() { // Polynomials
		let copy = FH.duplicateInstance(this);
		copy.monomials = FH.copyData(this.monomials);
		copy.letter = FH.copyData(this.letter);
		return copy;
	};

	// add polynomials to target one, return a new polynomials
	add(poly) { // Polynomials
		let newPoly = new Polynomials(this.letter);
		let c = null;
		for (let i=0; i<this.monomials.length; i++) {
			c = this.monomials[i];
			if (c !== undefined && c !== null) {
				newPoly.monomials[i] = c;
			};
		};
		for (let i=0; i<poly.monomials.length; i++) {
			c = poly.monomials[i];
			if (c !== undefined && c !== null) {
				if (newPoly.monomials[i] === undefined) {
					newPoly.monomials[i] = c;
				} else {
					newPoly.monomials[i] += c;
				};
			};
		};
		// simplify if zero coefficients
		newPoly.updateExpression();
		return newPoly;
	};

	// multiply polynomials with target one, return a new polynomials
	multiply(poly) { // Polynomials
		let newPoly = new Polynomials(this.letter);
		let c1 = null;
		let c2 = null;
		for (let i=0; i<this.monomials.length; i++) {
			c1 = this.monomials[i];
			// console.log(i + ' => ' + c1);
			if (c1 !== undefined && c1 !== null) {
				for (let j=0; j<poly.monomials.length; j++) {
					c2 = poly.monomials[j];
					// console.log('   ' + j + ' => ' + c2);
					if (c2 !== undefined && c2 !== null) {
						// console.log('      ' + c1*c2);
						if (newPoly.monomials[i+j] === undefined) {
							newPoly.monomials[i+j] = c1*c2;
						} else {
							newPoly.monomials[i+j] += c1*c2;
						};
					};
				};
			};
		};
		// simplify if zero coefficients
		newPoly.updateExpression();
		return newPoly;
	};

	// multiply integer with target polynomials, return a new polynomials
	multiplyInteger(d) { // Polynomials
		let newPoly = new Polynomials(this.letter);
		let c = null;
		for (let i=0; i<this.monomials.length; i++) {
			c = this.monomials[i];
			if (c !== undefined && c !== null) {
				newPoly.monomials[i] = c*d;
			};
		};
		// simplify if zero coefficients
		newPoly.updateExpression();
		return newPoly;
	}

	// multiply integer with target polynomials, return 'this'
	multiplyByInteger(d) { // Polynomials
		let c = null;
		for (let i=0; i<this.monomials.length; i++) {
			c = this.monomials[i];
			if (c !== undefined && c !== null) {
				this.monomials[i] = c*d;
			};
		};
		// simplify if zero coefficients
		this.updateExpression();
		return this;
	};

	// divide target polynomials with integer, return 'this'
	divideByInteger(d) { // Polynomials
		let c = null;
		for (let i=0; i<this.monomials.length; i++) {
			c = this.monomials[i];
			if (c !== undefined && c !== null) {
				c = c/d;
				if (Number.isInteger(c)) {
					this.monomials[i] = c;
				} else {
					throw "should divide by common denominator";
				};
			};
		};
		// simplify if zero coefficients
		this.updateExpression();
		return this;
	};

	// return value of polynomials at integer x
	evaluateAtInteger(x) { // Polynomials
		// return value a if a is integer, or 0 if a is null
		function getIntegerValue(a) {
			let v = 0;
			if (a !== undefined && a !== null) {
				v = a;
			};
			return v;
		};

		// evaluate using Horner's method
		function evaluateArray(arr, x) {
			let r = null;
			let a = getIntegerValue(arr[0]);
			if (arr.length === 1) {
				r = a;
			} else {
				let arrNext = FH.copyData(arr);
				// remove first element
				arrNext.shift();
				r = a + x*evaluateArray(arrNext, x);
			};
			// console.log('arr: ' + JSON.stringify(arr));
			// console.log('a = ' + a);
			// console.log('r = ' + r);
			return r;
		};

		if (Number.isInteger(x)) {
			let arrCoeff = this.monomials;
			return evaluateArray(arrCoeff, x);
		} else {
			throw "evaluation only on integers";
		};
	};

	// return Mathematica FullForm string representation of target polynomials
	getMathematica(opt) { // Polynomials
		let arrPolynomialsTxt = [];

		for (let i=0; i<this.monomials.length; i++) {
			if (this.monomials[i] !== undefined && this.monomials[i] !== null) {
				arrPolynomialsTxt.push(''
						// 'Times'
						+ CONST.MATH_TIMES
						+ CONST.MATH_BRACKET_OPEN
							+ this.monomials[i]
						+ ', '
							// 'Power'
							+ CONST.MATH_POWER
							+ CONST.MATH_BRACKET_OPEN
								+ this.letter
							+ ', '
								+ i
							+ CONST.MATH_BRACKET_CLOSE
						+ CONST.MATH_BRACKET_CLOSE);
			}; // not undefined and not null
		};

		// collect in 'Plus'
		let polynomialsTxt = CONST.MATH_PLUS
												+ CONST.MATH_BRACKET_OPEN
												+ arrPolynomialsTxt.join(', ')
												+ CONST.MATH_BRACKET_CLOSE;

		// console.log('Polynomials');
		// this.showConsole(OPT.SHOW, 'this: ');
		// console.log(polynomialsTxt);
		// console.log('');

		return polynomialsTxt;
	};

	// return LaTeX representation of target polynomials
	getLaTeX(opt) { // Polynomials

		if (opt.factorizePolynomials) {
			// console.log('use FF functions');
			// this.showConsole(OPT.SHOW, 'P: ');
			// delegate to external functions
			let arr = FF.fillArrayWithZeros(this.monomials);
			// case when arr reduced to 1 element 1 or -1
			if (arr.length === 1 && Math.abs(arr[0]) === 1) {
				return {ltx: '1', sign: Math.sign(arr[0])};
			};
			let letter = this.getLetter();
			// collect data about factors
			let factorsObj = FF.getFactorsFromArray(arr);
			// check factor decomposition
			if (!FF.testFactorDecompositionOfArray(factorsObj, arr)) {
				console.log('wrong factor decomposition of ');
				this.showConsole(OPT.SHOW, 'P = ');
			};
			let {globalFactor: globalFactor, arrFactors: arrFactors} = factorsObj;
			// console.log('factorsObj: ' + JSON.stringify(factorsObj));
			let newOpt = FH.duplicateInstance(opt); // option to give to getLaTeX
			newOpt.factorizePolynomials = false;
			// construct latex representation of these factors
			let overAllSign = Math.sign(globalFactor);
			let arrLtx = [];
			if (Math.abs(globalFactor) !== 1) {
				arrLtx.push('' + Math.abs(globalFactor));
			};
			for (let factorData of arrFactors) {
				let poly = Polynomials.newFromLetterAndArrayOfCoefficient(letter, factorData.arr);
				// poly.showConsole(OPT.SHOW, 'P = ');
				let {ltx, sign} = poly.getLaTeX(newOpt);
				// sign should be +1
				if (sign !== 1) {
					throw "sign should be +1";
				};
				let latexPoly = ltx;
				if (factorData.mult > 1) {
					latexPoly += CONST.LATEX_EXPONENT + '{' + factorData.mult + '}'
				};
				arrLtx.push(latexPoly);
			};
			// TODO: remove Math.abs(globalFactor) from string representation
			// pass it as new item in object to construct fraction in front of final string
			// in MathRationalExpression
			return {ltx: arrLtx.join(' '), sign: overAllSign};
		} else {
			// non factorized expression
			let arrPolynomialsTxt = [];
			let deg = this.getDegree();
			let sign = Math.sign(this.monomials[deg]);
			// console.log('(polynomials) deg: ' + deg);
			// console.log('(polynomials) sign: ' + sign);
			// degree > 0
			for (let i=this.monomials.length-1; i>0; i--) {
				// console.log('this.monomials[i]=' + this.monomials[i]);
				if (this.monomials[i] !== undefined && this.monomials[i] !== null) {
					// console.log('i='+i);
					let c = this.monomials[i]*sign;
					// console.log('(polynomials) c: ' + c)
					let powerTxt = i>1 ? CONST.LATEX_EXPONENT + '{' + i + '}' : '';
					let monomialsTxt = '';
					// case 1 => nothing
					// case -1 = >only sign -
					if (c === -1) {
						// sign without value
						monomialsTxt += '- ';
					} else if (c < -1) {
						// sign with absolute value of c
						monomialsTxt += '- ' + Math.abs(c);
					} else if (c === 1) {
						// no '+' in front of higher degree
						// => nothing for higher degree
						if (i !== deg) {
							// just sign '+', no value '1'
							monomialsTxt += '+ ';
						};
					} else {
						if (i !== deg) {
							// sign '+' and value
							monomialsTxt += '+ ' + c;
						} else {
							// no '+' in front of higher degree
							monomialsTxt += '' + c;
						};
					};
					arrPolynomialsTxt.push(''
							+ monomialsTxt
							+ this.letter
							+ powerTxt);
				}; // not undefined and not null
			};

			// degree = 0
			// since no 'letter' => value is necessary
			if (this.monomials[0] !== 0) {
				let c = this.monomials[0]*sign;
				if (c < 0) {
					// sign '-' is in c
					arrPolynomialsTxt.push('- ' + Math.abs(c));
				} else {
					// c > 0
					if (arrPolynomialsTxt.length == 0) {
						// nothing in the expression before this term, no sign '+'
						arrPolynomialsTxt.push('' + c);
					} else {
						// something in the expression before this term => sign +
						arrPolynomialsTxt.push('+ ' + c);
					};
				};
			} else if (arrPolynomialsTxt.length == 0) {
				// nothing in the expression before this term
				// and the coefficient is zero…
				arrPolynomialsTxt.push('0');
			};

			let polynomialsTxt = arrPolynomialsTxt.join(' ');
			if (arrPolynomialsTxt.length > 1) {
				polynomialsTxt = CONST.LATEX_POLYNOMIAL_OPEN
														+ polynomialsTxt
														+ CONST.LATEX_POLYNOMIAL_CLOSE;
			};

			// if (sign < 0) {polynomialsTxt = '-' + polynomialsTxt;}

			// replace polynomials string with opt.polynomialsStringDictionary
			// console.log('(Polynomials) ltx: ' + polynomialsTxt);
			if (opt.replacePolynomials) {
				for (let k in opt.polynomialsStringDictionary) {
					polynomialsTxt = polynomialsTxt.split(k).join(opt.polynomialsStringDictionary[k]);
				};
			};
			// console.log('(Polynomials) ltx: ' + polynomialsTxt);
			// console.log('(Polynomials) sign: ' + sign);
			// TODO: add global factor = 1 here
			return {ltx: polynomialsTxt, sign: sign};
		};
	};


	// return display text representation of target polynomials
	displayText(opt) { // Polynomials
		this.updateExpression();
		let od = opt.display;
		let arrPolynomialsTxt = [];
		// TODO: consider the sign of the coefficients to use + or -

		// degree > 0
		for (let i=this.monomials.length-1; i>0; i--) {
			// console.log('this.monomials[i]=' + this.monomials[i]);
			if (this.monomials[i] !== undefined && this.monomials[i] !== null) {
				// console.log('i='+i);
				let pTxt = i>1 ? CONST.SYMBOL_EXPONENT + i : '';
				let monomialsTxt = '';
				// case 1 => nothing
				// case -1 = >only sign -
				if (this.monomials[i] === -1) {
					monomialsTxt += '-';
				} else if (this.monomials[i] !== 1) {
					monomialsTxt += this.monomials[i] + CONST.SYMBOL_POLYNOMIALS_PRODUCT;
				};
				arrPolynomialsTxt.push(''
						+ monomialsTxt
						+ this.letter
						+ pTxt);
			};
		};

		// degree = 0
		if (this.monomials[0] !== 0) {
			arrPolynomialsTxt.push('' + this.monomials[0]);
		} else if (arrPolynomialsTxt.length == 0) {
			arrPolynomialsTxt.push('0');
		};

		let polynomialsTxt = arrPolynomialsTxt.join(CONST.SYMBOL_POLYNOMIALS_SUM);
		polynomialsTxt = od.formatPolynomials(polynomialsTxt);
		// class name
		let classTxt = '';
		if (opt.type) {
			classTxt = CONST.SYMBOL_OPEN_CLASS
								+ this.constructor.name
								+ CONST.SYMBOL_CLOSE_CLASS;
		};
		classTxt = od.formatClass(classTxt);

		return classTxt + polynomialsTxt;
	}

	showConsole(opt, text = '') { // Polynomials
		let globalTxt = opt.display.formatText(text) + this.displayText(opt);
		console.log(globalTxt);
		return this;
	};

	toObject() { // Polynomials
		return {
			// class constructor
			class: this.constructor.name,
			// monomials
			mon: FH.copyData(this.monomials),
			// letter
			l: this.letter,
		};
	};

	toJSON(...args) {
		return JSON.stringify(this.toObject(), ...args);
	};

	isZero() { // Polynomials
		this.updateExpression();
		if (this.monomials.length == 1 && this.monomials[0] == 0) {
			return true;
		} else {
			return false;
		};
	};

	isOne() { // Polynomials
		this.updateExpression();
		if (this.monomials.length == 1 && this.monomials[0] == 1) {
			return true;
		} else {
			return false;
		};
	};

	isInteger() { // Polynomials
		this.updateExpression();
		if (this.monomials.length == 1) {
			return true;
		} else {
			return false;
		};
	};

	isPositive() { // Polynomials
		this.updateExpression();
		if (this.monomials.length == 1 && this.monomials[0] >= 0) {
			return true;
		} else {
			return false;
		};
	};

	isNegative() { // Polynomials
		this.updateExpression();
		if (this.monomials.length == 1 && this.monomials[0] < 0) {
			return true;
		} else {
			return false;
		};
	};

}


class RationalExpression {
	constructor(numerator, denominator) { // RationalExpression
		this.numerator = numerator;

		if ( !Number.isInteger(denominator) || denominator === 0) {
			throw "denominator should be a non zero integer";
		};

		this.denominator = denominator;

		this.updateExpression();
	};

	static newFromPolynomials(poly) { // RationalExpression
		let denominator = 1;
		return new RationalExpression(poly, denominator);
	};

	static newFromLetterAndRational(l,n,d) { // RationalExpression
		let num = Polynomials.newFromLetterAndCoefficient(l,n);
		return new RationalExpression(num, d);
	};

	static newFromLetterAndInteger(l,n) { // RationalExpression
		return RationalExpression.newFromLetterAndRational(l,n,1);
	};

	getLetter() { // RationalExpression
		return this.numerator.getLetter();
	};

	makeCopy() { // RationalExpression
		let copy = FH.duplicateInstance(this);
		copy.numerator = this.numerator.makeCopy();
		copy.denominator = this.denominator;
		return copy;
	};

	updateExpression() { // RationalExpression
		if (this.numerator.isZero()) {
			this.denominator = 1;
		};

		let arrNum = this.numerator.getArrayCoefficients();
		let commonDivisor = Math.abs(FH.getGCDArray(arrNum.concat([this.denominator])));
		// let commonDivisor = gcdArr(arrNum.concat([this.denominator]));
		this.numerator.divideByInteger(commonDivisor);
		this.denominator = this.denominator/commonDivisor;

		let sd = Math.sign(this.denominator);
		this.numerator.multiplyByInteger(sd);
		this.denominator = sd*this.denominator;

		return this;
	};

	add(ratExp) { // RationalExpression
		let den = this.denominator*ratExp.denominator;
		let num1 = this.numerator.multiplyInteger(ratExp.denominator);
		let num2 = ratExp.numerator.multiplyInteger(this.denominator);
		let num = num1.add(num2);
		return new RationalExpression(num, den);
	}

	addRational(n,d) { // RationalExpression
		let l = this.numerator.letter;
		let ratExp = RationalExpression.newFromLetterAndRational(l,n,d);
		return this.add(ratExp);
	};

	multiply(ratExp) { // RationalExpression
		let num = this.numerator.multiply(ratExp.numerator);
		let den = this.denominator*ratExp.denominator;
		return new RationalExpression(num, den);
	};

	multiplyByInteger(d) { // RationalExpression
		this.numerator.multiplyByInteger(d);
		// simplify if zero coefficients
		this.updateExpression();
		return this;
	};

	multiplyByRational(n,d) { // RationalExpression
		this.numerator.multiplyByInteger(n);
		this.denominator = d*this.denominator;
		// simplify if zero coefficients
		this.updateExpression();
		return this;
	};

	divideByInteger(d) { // RationalExpression
		this.denominator = d*this.denominator;
		// simplify if zero coefficients
		this.updateExpression();
		return this;
	};

	// return value of polynomials at integer x
	evaluateAtInteger(x) { // RationalExpression
		let n = this.numerator.evaluateAtInteger(x);
		let d = this.denominator;
		let l = this.numerator.letter;
		return RationalExpression.newFromLetterAndRational(l,n,d);
	};


	toObject() { // RationalExpression
		return {
			// class constructor
			class: this.constructor.name,
			// numerator
			num: 	this.numerator.toObject(),
			// denominator
			den: this.denominator,
		};
	};

	toJSON(...args) {
		return JSON.stringify(this.toObject(), ...args);
	};

	getMathematica(opt) { // RationalExpression

		let numeratorTxt = this.numerator.getMathematica(opt);

		// fraction
		let fractionTxt = '';
		// just a test…
		let denominatorSign = Math.sign(this.denominator);
		if (denominatorSign < 0) {
			throw "denominator should be positive";
		}
		if (this.denominator === 1) {
			fractionTxt = numeratorTxt;
		} else {
			fractionTxt = CONST.MATH_TIMES
										+ CONST.MATH_BRACKET_OPEN
											+ numeratorTxt
										+ ', '
											+ CONST.MATH_POWER
											+ CONST.MATH_BRACKET_OPEN
											+ this.denominator
											+ ', -1'
											+ CONST.MATH_BRACKET_CLOSE
										+ CONST.MATH_BRACKET_CLOSE;
		};

		// console.log('RationalExpression');
		// this.showConsole(OPT.SHOW, 'this: ');
		// console.log(fractionTxt);
		// console.log('');

		return fractionTxt;
	};

	getLaTeX(opt) { // RationalExpression
		// numerator
		let numObj = this.numerator.getLaTeX(opt);
		let numeratorTxt = numObj.ltx;
		let sign = numObj.sign;

		// fraction
		let fractionTxt = '';
		// just a test…
		let denominatorSign = Math.sign(this.denominator);
		if (denominatorSign < 0) {
			throw "denominator should be positive";
		}
		if (this.denominator === 1) {
			fractionTxt = numeratorTxt;
		} else {
			fractionTxt += CONST.LATEX_CMD_FRAC;
			// fraction 1/den in front of num
			// TODO: get coeff in front of numeratorTxt to construct fraction
			fractionTxt += '{1}';
			fractionTxt += '{' + this.denominator + '}';
			if (numeratorTxt !== '1') {
				fractionTxt += numeratorTxt;
			};
			// fraction num/den
			// fractionTxt += '{' + numeratorTxt + '}';
			// fractionTxt += '{' + this.denominator + '}';
		};

		// if (sign < 0) {fractionTxt = '-' + fractionTxt;}
		// return fractionTxt;
		// console.log('(RationalExpression) ltx: ' + fractionTxt);
		// console.log('(RationalExpression) sign: ' + sign);
		return {ltx: fractionTxt, sign: sign};
	};


	displayText(opt) { // RationalExpression
		this.updateExpression();
		let od = opt.display;
		let newOpt = FH.duplicateInstance(opt); // functions inside, no class
		newOpt.type = false;

		// numerator
		let numeratorTxt = '';
		if (this.numerator.monomials.length > 1) {
			numeratorTxt += CONST.SYMBOL_POLYNOMIAL_OPEN;
			numeratorTxt += this.numerator.displayText(newOpt);
			numeratorTxt += CONST.SYMBOL_POLYNOMIAL_CLOSE;
		} else {
			numeratorTxt += this.numerator.displayText(newOpt);
		};

		// denominator
		let denominatorTxt = '';
		if (this.denominator === 1) {
			denominatorTxt = '';
		} else {
			denominatorTxt += CONST.SYMBOL_RATIONAL_OVER;
			denominatorTxt += this.denominator;
		};

		// class name
		let classTxt = '';
		if (opt.type) {
			classTxt = CONST.SYMBOL_OPEN_CLASS
								+ this.constructor.name
								+ CONST.SYMBOL_CLOSE_CLASS;
		};
		classTxt = od.formatClass(classTxt);

		return classTxt + od.formatRationalExpression(numeratorTxt + denominatorTxt);
	};

	showConsole(opt, text = '') { // RationalExpression
		let globalTxt = opt.display.formatText(text) + this.displayText(opt);
		console.log(globalTxt);
		return this;
	};

	isZero() { // RationalExpression
		this.updateExpression();
		return this.numerator.isZero();
	};

	isOne() { // RationalExpression
		this.updateExpression();
		return this.numerator.isOne() && this.denominator === 1;
	};

	isRational() { // RationalExpression
		this.updateExpression();
		return this.numerator.isInteger();
	};

	isNegative() { // RationalExpression
		this.updateExpression();
		return this.numerator.isNegative() && this.denominator> 0;
	};

};



module.exports = {
	Polynomials: Polynomials,
	RationalExpression: RationalExpression,
}
